\documentstyle[comments,a4,titlepage]{cweb}

\input{/LEDA/SRC/man/MANUAL.mac}

\title{{\bf\huge bigfloats} \\ High-precision floating point numbers
in LEDA}
\author{Christoph Burnikel \and Jochen K\"onemann\\ \and  Kurt Mehlhorn}
     %  {\footnotesize Max-Planck-Institut f\"ur Informatik,}\\[-0.7ex]
     %  {\footnotesize 66123 Saarbr"ucken, Germany}}

%\repnumber{???}
%\begin{acknowledgements}
%\end{acknowledgements}
%\begin{address}
%Jochen K\"onemann\\
%{\footnotesize Max-Planck-Institut f\"ur Informatik,}\\[-0.7ex]
%{\footnotesize 66123 Saarbr\"ucken, Germany}\\
%E-Mail: {\tt JOE@@MPI-SB.MPG.DE}
%\end{address}

%\begin{keywords}
%CWEB,
%LEDA,
%IEEE,
%Literate Programming
%\end{keywords}

\begin{document}

\begin{abstract}
We describe the implementation of the LEDA \cite{lman,ledaMN} data type
{\bf bigfloat}. {\bf bigfloats} represent floating-point numbers of 
arbitrary precision.
{\bf bigfloats} are closed under the operations addition, subtraction,
multiplication, division, and square root. The main features of this data type
are
\begin{enumerate}
\item {\bf bigfloats} conform to the IEEE-standard \cite{ieee,gold} regarding
      their functionality. 
      In particular, {\bf bigfloats} generalize the build-in data type
      {\bf double}. 
\item The user can choose numerical precision and rounding mode for every 
      single operation.
\item Rounding is possible to every precision and in every rounding mode.
      Furthermore, {\bf bigfloats} can be rounded to {\bf doubles} and 
      {\bf integers}.
\end{enumerate}
{\bf bigfloats} are developed using LEDA-{\bf integers} which are integer 
numbers of arbitrary length.
\end{abstract}

\thispagestyle{empty}
%\makempicover


\setcounter{page}{0}
\pagenumbering{Roman}

\tableofcontents
\newpage

\setcounter{page}{0}
\pagenumbering{arabic}

% /KM/usr/joe/cweb/typnamen

@* Introduction.

KURT ODER JOCHEN, BITTE EINLEITUNG EINFUEGEN!

@* Manual Page.

@i bigfloat.man


@* Representation of {\bf bigfloats}. 

|bigfloat|s that are not special values are layed down as a pair of |integers|,
|significant| and |exponent|. 
Additionally, every |bigfloat| contains the flag |special| that determines
whether a special value is represented or not. 
The precision of the significant is held in a separate element variable of type
|long|.
The precision specifies the current binary length of the significant.

@<data members of class |bigfloat|@>=

  integer significant,exponent;
  special_values special;
  long precision;

@ |special_values| is an enumeration type.

@<special values@>=

enum special_values{NOT, PZERO, NZERO, PINF, NINF, NAN};

@ We maintain the following invariants:
\begin{enumerate}
\item if special $\neq$ |NOT| then special determines the special value of
      the corresponding |bigfloat|
\item if special $\equiv$ |NOT| then the real number represented by the
      |bigfloat| is $significant \cdot 2^{exponent}$.
\item we will also guarantee that |precision| is the number of binary digits
      of the |significant|.
\end{enumerate}
A |bigfloat| number does not necessarily have a unique representation 
because there may be zeros at the end of the |significant|. 
However, some functions require a unique representation.
We call a |bigfloat| {\sf normalized} if its |significant| ends in a one.
To guarantee uniqueness some functions normalize their input before working on
it.  
To do this we have an internal procedure |normalize|.

@<function normalize@>=

void bigfloat::normalize (void)
{
  int signum = ::sign(significant);
  if ((signum == 0)&&(special == NOT))
    special = PZERO;

  long z=significant.zeros();
  // |z| is the number of final zeros in the significant
  if (z>0)
    significant=significant >> z;
  precision = significant.length();
  exponent+=z;
}

  
@* The Header-File.

First we give an overview of the header file.
The refinements are defined in the following sections.

@(bigfloat.h@>=

/*{\Manpage {bigfloat} {} {The data type bigfloat} 
}*/

#ifndef BIGFLOAT_H
#define BIGFLOAT_H

#include <LEDA/integer.h>

/* we use the LEDA data type |integer| for the representation 
of |bigfloat|s */

class bigfloat;

@<special values@>@;
@<rounding constants@>@;
@<global variables and constants@>@;
@<constants and global variables dealing with output@>@;
@<special bigfloat values@>@;
 
class bigfloat @;
{@;

/*{\Mdefinition
In general a |bigfloat| is given by two |integer|s $s$ and $e$ where $s$ is the
significant and $e$ is the exponent.
The tuple $(s,e)$ represents the real number $$s \cdot 2^e.$$
In addition, there are the special |bigfloat| values |NaN| ($=$ not a number),
|pZero|, |nZero| ($= +0, -0$), and |pInf|, |nInf| ($= +\infty, -\infty$).
These special values behave as defined by the IEEE floating point standard.
In particular,
$\frac{5}{+0}=\infty$,$\frac{-5}{+0}=-\infty$,$\infty+1=\infty$, 
$\frac{5}{\infty}=+0$,$+\infty+(-\infty)=NaN$ and $0 \cdot \infty = NaN$.

Arithmetic on |bigfloat|s uses two parameters:
The precision |prec| of the result (in number of binary digits) and the rounding
mode |mode|.
Possible rounding modes are:
\begin{itemize}
\item |TO_NEAREST|: round to the closest representable value
\item |TO_ZERO|: round towards zero
\item |TO_INF|: round towards $\infty$
\item |TO_P_INF|: round towards $+\infty$
\item |TO_N_INF|: round towards $-\infty$
\end{itemize}
Operations $+,-,*$ work as follows. 
First, the exact result |z| is computed.
If the rouning mode is EXACT then |z| is the result of the operation. 
Otherwise, let |s| be the significant of the result; |s| is rounded to |prec|
binary places as dictated by |mode|. 
Operations $/$ and $\sqrt{\hspace{1ex}}$ work accordingly except that
EXACT is treated as TO\_NEAREST.

The parameters |prec| and |mode| are either set directly for a single
operation or else they are set globally for every operation to follow. 
The default values are 53 for |prec| and TO\_NEAREST for |mode|.
}*/

  @<data members of class |bigfloat|@>@;

  @<private functions@>@;
  void normalize (void);

 public:@;

  ~bigfloat (void)@+ {@+}

  @<bigfloat creation@>@;
  @<declaration of operational functions@>@;
  @<declaration of arithmetical operators@>@;
  @<declaration of comparison operators@>@;
  @<declaration of checking functions@>@;
  @<mathematical functions@>@;
  @<converting functions@>@;
  @<input/output operations and string contruction@>@;
  @<informative functions@>@;
  @<rounding functions@>@;
}@+ ;@;

@<global functions@>
#endif

@ Creation of {\bf bigfloat}s.

@<bigfloat creation@>=
/*{\Mcreation x }*/

/*{\Mtext 
|bigfloat|s may be constructed from data types |double|, |long|, |int| and
|integer|, without loss of accuracy.
In addition, |bigfloat|s can be created as follows.
}*/

 bigfloat();
 bigfloat(double);
 bigfloat(long);
 bigfloat(int);
 bigfloat(const integer&);

 bigfloat(const integer&,const integer&);
 /*{\Mcreate bigfloat(s,e) introduces a variable \var\ of type |bigfloat|
 and initializes it to the value given by the tuple $(s,e)$ 
 }*/

 bigfloat(special_values);
   
  /* the default copy constructor and assignment operator are used
     (element-wise copy) and do the right operations */

@ |bigfloat| arithmetic has all rounding modes provided by the $IEEE$ standard.
Additionally we have the rounding mode |TO_INF| that for positive numbers rounds
to $+\infty$ and for negative numbers rounds to $-\infty$.
To compute the operations $+,-,*$ exactly, we further introduce the (formal)
rounding mode |EXACT| who means that for these three operations we do not round
at all.

@<rounding constants@>=
enum rounding_modes{TO_NEAREST,TO_ZERO,TO_P_INF,TO_N_INF,TO_INF,EXACT};

@ The default values for arithmetical precision and rounding mode are 
specified by the global variables |global_prec| and |round_mode|.
Another global variable |dbool| serves as a default argument for the
boolean parameter in our arithmetical operations.

@<global variables and constants@>=

extern long global_prec;
extern rounding_modes round_mode;
extern bool dbool;


@ The basic arithmetic operations are given by the functions |add|, |sub|,
|mul|, |div|, and |sqrt|.
We choose a syntax that differs from the syntax of the standard operators 
$+,-,*,/$.
The reason is that the user should be able to specify precision and rounding
mode for each operation separately, without having to manipulate global 
variables before and after the operation.
The latter procedure would be complicated and could produce unwanted side
effects.
Another useful feature of our syntax is that the user can be informed 
after an operation if the operation was carried out exactly or not
(here by the boolean parameter |is_exact|).
This is important for applications when the same set of computations is
repeated with increasing precision, where it helps to save time.

@<declaration of operational functions@>=

/*{\Moperations 1.5 3.5 
}*/

/*{\Mtext
The arithmetical operations $+, -, *, \, sqrt$, the comparison operators 
$<, \leq, >, \geq, =, \neq$ and the stream operators are all available.

Addition, subtraction, multiplication, division and squareroot are implemented
by the functions |add|, |sub|, |mul|, |div| and |sqrt|, respectively.
For example, the call $$add(x,y,prec,mode,is\_exact)$$
computes the sum of |bigfloat|s |x| and |y| up to |prec| binary digits, in
rounding mode |mode|, and returns it.
The optional last parameter |is_exact| is a boolean variable that is set to
|true| if and only if the returned |bigfloat| exactly equals the sum of 
|x| and |y|.
The parameters |prec| and |mode| are also optional and have the global default
values |global_prec| and |round_mode|.
That is, the three calls |add(x,y,glob_prec,round_mode)|, |add(x,y,glob_prec)|,
|add(x,y)| are all equivalent.
The syntax for functions |sub|, |mul|, |div| and |sqrt| is analoguous.

The standard operators $+, -, *, \ $ are also available and are implemented by
their counterparts among the functions |add|, |sub|, |mul|, |div| and |sqrt|.
For example, the call $x+y$
is equivalent to |add(x,y)|.
}*/


friend bigfloat add(const bigfloat&,const bigfloat&,long=global_prec,
rounding_modes=round_mode,bool &is_exact=dbool);@;

friend bigfloat sub(const bigfloat&,const bigfloat&,long=global_prec,
rounding_modes=round_mode,bool &is_exact=dbool);@;

friend bigfloat mul(const bigfloat&,const bigfloat&,long=global_prec,
                     rounding_modes=round_mode,bool &is_exact=dbool);@;

friend bigfloat div(const bigfloat&,const bigfloat&,long=global_prec,
                     rounding_modes=round_mode,bool &is_exact=dbool);@;

friend bigfloat sqrt(const bigfloat&,long=global_prec,
rounding_modes=round_mode, bool &is_exact=dbool, const bigfloat& start=pZero);@;

@ The standard operators $+, -, *, /$ are immediately derived from the
functions |add|, |mul|, etc. 

@<declaration of arithmetical operators@>=
friend bigfloat operator+(const bigfloat&,const bigfloat&);
friend bigfloat operator-(const bigfloat&,const bigfloat&);
friend bigfloat operator * (const bigfloat&,const bigfloat&);
friend bigfloat operator / (const bigfloat&,const bigfloat&);
friend bigfloat operator - (const bigfloat&);


@ Among the comparison operators we only have to implement |==| and |<|,
because the other operators can be easily obtained from these two.

@<declaration of comparison operators@>=

friend bool operator == (const bigfloat&,const bigfloat&);
friend bool operator > (const bigfloat&,const bigfloat&);
friend bool operator != (const bigfloat &a,const bigfloat &b)@;
{@+ return !(a==b);@+}
friend bool operator >= (const bigfloat &a,const bigfloat &b)@;
{@+ return ((a>b)||(a==b));@+}
friend bool operator < (const bigfloat &a,const bigfloat &b)@;
{@+ return (!(a>=b));@+}
friend bool operator <= (const bigfloat &a,const bigfloat &b)@;
{@+ return (!(a>b));@+}

@ The rounding function.

@<rounding functions@>=

/*{\Mtext
A bigfloat \var\ can be rounded by the call 
$$x.round(prec,mode,is\_exact,bias).$$
Here |bias| is an optional integer variable taking values in $\{-1,0,+1\}$, 
with default value $0$.
If |bias| is zero, the function rounds \var\ up to |prec| binary digits in
rounding mode |mode|.
The optional boolean variable |is_exact| is set to true if and only if the
rounding operation did not change the value of $x$. 
If |bias| is nonzero, the result is the same as above, only the value to be
rounded is not \var\ but $\var + bias \cdot \epsilon$, where $\epsilon$ is a
positive infinitesimal.
}*/

bigfloat round(long=0,rounding_modes=round_mode,bool &is_exact=dbool,long=0);@;

bigfloat round_with_error(long=0,rounding_modes=round_mode,bool
                            &is_exact=dbool,long=0);@;


@ The following set of special values is available to the user.

@<special bigfloat values@>=

extern const bigfloat NaN;
extern const bigfloat pInf;
extern const bigfloat nInf;
extern const bigfloat pZero;
extern const bigfloat nZero;


@ We need functions to check if a |bigfloat| is |NaN| or $\pm 0$.
The equality operator fails to give the wanted information, because |NaN| is 
not equal to any |bigfloat|, including itself.
Also, the equality operator gives |pZero == nZero|.
We use the functions |isNaN|, |ispZero| and |isnZero|.
Further, the following (private) functions are used internally:
|isnInf| checks for $-\infty$, |ispInf| checks for $+\infty$,
|isInf| checks for $\pm\infty$, |isZero| checks for $\pm 0$
and finally |isSpecial| checks whether the flag |special| is true.


@<declaration of chec...@>=

/*{\Mtext
There are several functions to check whether the outcome of an arithmetical
computation on |bigfloat|s is a special-valued |bigfloat|.
These functions are necessary because the equality operator fails.
Further operations include some of the most important mathemtaical
functions, functions to approximate |bigfloat|s by |double|s or |integer|s, 
and functions to set global precision, rounding mode, and output mode. 
}*/

bool isNaN()  const { return (special == NAN); }
/*{\Mop returns true if and only if \var\ is |NaN|
}*/

bool isnInf() const  { return (special == NINF);  } 
/*{\Mop returns true if and only if \var\ is |nInf|
}*/

bool ispInf() const  { return (special == PINF);  }
/*{\Mop returns true if and only if \var\ is |pInf|
}*/


bool isnZero() const { return (special == NZERO); }
/*{\Mop returns true if and only if \var\ is |nZero|
}*/

bool ispZero() const { return (special == PZERO); }
/*{\Mop returns true if and only if \var\ is |pZero|
}*/

bool isZero() const  {return ((special==PZERO)||(special==NZERO));}
/*{\Mop returns true if and only if \var\ is |pZero| or |nZero|
}*/

bool isInf (void) const  { return ((special==PINF)||(special==NINF));}
/*{\Mop returns true if and only if \var\ is |pInf| or |nInf|
}*/

bool isSpecial (void) const  { return (special != NOT);}
/*{\Mop returns true if and only if \var\ is $\pm0$, $\pm\infty$ or |NaN|
}*/


@ We provide the user with some mathematical functions on |bigfloat|s.

@<mathematical functions@>=

long sign(void) const;
/*{\Mfunc the sign of \var
}*/

inline friend bigfloat fabs(const bigfloat &b)
{  if (b < pZero) return -b; else return b; 
}
/*{\Mfunc the absolute value if \var
}*/

inline friend bigfloat pow2(const integer &p)  {  return bigfloat(1,p);  }
/*{\Mfunc gives back the corresponding power of $2$
}*/

inline friend integer log2(const bigfloat &x)  
{  return x.get_precision() + x.get_exponent();  }
/*{\Mfunc $log2(x)$ gives back the binary logarithm of $x$, rounded up to the
    next integer (precondition: $x != 0$)
}*/

inline integer ceil(void)  {  return tointeger(TO_P_INF); } 
/*{\Mop rounds up to the next integer
}*/

inline integer floor(void)  {  return tointeger(TO_N_INF); }
/*{\Mop rounds down to the next integer
}*/



@ The following internal function is used to get the sign of special |bigfloat|
values. 
It returns $+1$ for |pZero| and $-1$ for |nZero|, hence it never returns the
sign $0$. 
In the NaN case the returned sign is meaningless.

@<private...@>+=
long non_zero_sign (void) const;


@ We can convert (i.e., round) |bigfloat|s to the types |double| and |integer|.

@<converting functions@>=

  double todouble() const; 
/*{\Mop gives back the |double| value next to \var\ (in rounding mode |TO_NEAREST|) 
}*/

  integer tointeger(rounding_modes=TO_ZERO) const; 
/*{\Mop gives back the |integer| value next to \var\ (in the corresponding rounding mode) 
}*/



@ There are some functions that help getting information of private data of
|bigfloat|s.

@<informative functions@>=
  long get_precision(void) const @+ { @+ return precision; @+ }

  integer get_exponent() const { return exponent;  } 
/*{\Mop returns the exponent of the |bigfloat|
}*/

  integer get_significant() const  {  return significant;  } 
/*{\Mop returns the significant of the |bigfloat| 
}*/


@ Input and output of |bigfloat|s is possible via streams. 
In addition a |bigfloat| can be read from a string.

@<input/output operations and string contruction@>=

  friend ostream& operator<<(ostream&,const bigfloat&); 
/*{\Mbinopfunc writes the |bigfoat| in decimal or binary mode
}*/
  friend istream& operator>>(istream&,bigfloat&); 
/*{\Mbinopfunc the input format is $dd \ldots d.dd \ldots d E d \ldots d$ where
  $d$ stands for any decimal digit.
}*/

@ The form of the output can be influenced by the user by setting the global
variable output to one of the values |DEC_OUT| or |BIN_OUT|. 
So both binary and decimal output (the default value) are possible. 

@<constants and global variables dealing with output@>=

enum output_modes {BIN_OUT,DEC_OUT};
extern output_modes output_mode; 

@ There are functions to change the global parameters.

@<global functions@>=

inline void set_glob_prec(long p)  {  global_prec=p;  }
/*{\Mfunc sets the global precision to |p| 
}*/

inline void set_round_mode(rounding_modes m=TO_NEAREST) {  round_mode=m; }
/*{\Mfunc sets the global rounding mode 
}*/

inline void set_output_mode(output_modes o_mode=DEC_OUT) { output_mode=o_mode; }
/*{\Mfunc sets the output mode (possible values are |BIN_OUT| and |DEC_OUT|) 
}*/


@ The following functions outputs significant and exponent in binary format
and additionally provides the user with the precision. The function was
designed for debugging purpose only.

@<private functions@>+=
  void print_contents (void) const;
  // soll spaeter raus

@* The Implementation.

In the following section we describe the source, element-functions and elements.
At first the form in general is described and afterwards the details are
provided with the element functions.

@(bigfloat.c@>=

@<includes@>@;
@<global identifiers@>@;
@<auxiliary functions@>@;
@<internal functions@>@;
@<public functions@>@;


@ 
The public part is further subdivided into different subgroups.

@<public functions@>=
  @<function normalize@>@;
  @<simple constructors@>@;
  @<general functions@>@;
  @<operational functions@>@;
  @<operators@>


@ We need the following include files:

@<includes@>=

#include <LEDA/bigfloat.h>
#include <iostream.h>
#include <strstream.h>
#include <stdio.h>
/*
#include <unistd.h>
*/
#include <math.h>
#include <string.h>

@* Simple functions.

@ Constructors.

Simple constructors are all constructors beside the constructor from
data type |double| which is described in a later section.
The bodies of the constructors from data types |int|, |long| and |integer|
are all the same and hence are collected in a special refinement.

@<simple constructors@>=

bigfloat::bigfloat ()
{
  special = PZERO;
  exponent=significant=precision=0;
} 

bigfloat::bigfloat (const integer &s,const integer &e)
{
  significant=s;
  exponent=e;
  special = NOT;
  precision = s.length();
}

bigfloat::bigfloat (special_values spec) @+ { @+ special = spec; @+ }

bigfloat::bigfloat (const integer &a)
{ @+  @<constructor body...@>@; @+
}

bigfloat::bigfloat (long a)
{ @+ @<constructor body...@>@; @+
}

bigfloat::bigfloat (int a)
{ @+ @<constructor body...@>@; @+
} 

@

@<constructor body for integer data type@>=
  special = NOT;
  significant=a;
  exponent=0;
  precision=significant.length();

  if (significant==0)
    special=PZERO;


@  We use the constructor from type |special_values| to create
the |bigfloat|s for |NaN|, $\pm 0$, and $\pm \infty$.

@<global identifiers@>=

const bigfloat NaN(NAN);
const bigfloat pInf(PINF);
const bigfloat nInf(NINF);
const bigfloat pZero(PZERO);
const bigfloat nZero(NZERO);

@ Sign functions.

There are two |bigfloat| member functions for sign computations.
One is the ordinary |sign| function that returns the sign $0$ if the
|bigfloat| is zero.
The other function |non_zero_sign| behaves exactly as function |sign|,
but gives back the sign $+1$ for |pZero| and sign $-1$ for |nZero|.
This function is used to compute the nonzero signs of special |bigfloat|
numbers.

@<general functions...@>+=
long bigfloat::sign (void) const
{
  long s = 0;

  switch(special)
  {
    case NOT: s =  ::sign(significant);  break;
    case PZERO: @+ case NZERO: @+ s = 0;  break;
    case PINF:  @+ s = 1;  break;
    case NINF:  @+ s = -1;  break;
    case NAN: @+error_handler(1,"sign: NaN has no sign");
  }

  return s;
}

@
@<internal functions@>=
long bigfloat::non_zero_sign (void) const
{
  long s = 0;

  switch(special)
  { case NOT:  @+ s = ::sign(significant); break;
    case PINF: @+ case PZERO: @+ s =  1; break;
    case NINF: @+ case NZERO: @+ s = -1; break;
    case NAN:  @+ error_handler(1,"sign: NaN has no sign");
  }
  return s;
}


@ Before we come to the core of the |bigfloat| implementation, we list some 
useful functions to compute the maximum of two |long|s, the square of two
|integer|s and the sign for types |long| and |int|, respectively.

@<auxiliary functions@>+=
long max (long l1,long l2)@+ {@+ return (l1>l2 ? l1 : l2);@+}
integer sq (const integer &op) @+ { @+ return op*op; @+ }
long sign(int i) 
@+ { @+ if (i == 0) @+ return 0; @+ else @+ return (i > 0 ? 1 : -1); @+}
long sign(long i) 
@+ { @+ if (i == 0) @+ return 0; @+ else @+ return (i > 0 ? 1 : -1); @+}


@* Rounding. 

The |round| function is the central tool for the implementation of
|bigfloat| arithmetic.
Recall that |x.round(digits,mode,is_exact,bias)| rounds the number
\begin{displaymath}
(significant+bias \cdot \epsilon) \cdot 2^{exponent}
\end{displaymath}
to |digits| binary digits in rounding mode |mode|, where $\epsilon$ is a
positive infinitesimal and bias is out of $\{-1,0,1\}$.
The parameter |is_exact| that tells us if the returned value equals $x$
is later used to detect exact return values of arithmetical operations.

We have global variables for the rounding mode, the precision and a dummy
variable for the data type |bool|. 
These parameters are the default values of the above parameters
|mode|, |digits| and |is_exact|, respectively.

@<global identifiers@>+=
rounding_modes round_mode=TO_NEAREST;
long global_prec=53; 
bool dbool=true;

@ Now we outline the implementation of function |round|.
At first the |bigfloat| is normalized in order to avoid final zeros of the
significant.
Then we distinguish cases.
It might be that the significant has more places than specified by parameter
|digits|.
In this case the bias can be neglected, with the one exception of the 
|TO_NEAREST| mode (details see below).
In all cases except for |EXACT| the significant is rounded to |digits|
places and the exponent is adapted accordingly.
On the other hand, it might be that the length of the significant is less
than |digits|.
In that case the return value depends on the bias (again with the exception
of mode |TO_NEAREST|). 

@<general functions@>+=
bigfloat bigfloat::round (long digits,rounding_modes mode,bool &is_exact,
                          long bias)
{ @;
  normalize();
  if (isSpecial() || (mode == EXACT))
    return *this;
  int test = 0; 
  long shift;

  if (digits < precision)
  {
    switch(mode)
    {
      case TO_NEAREST: 
        @<round to nearest@>@;
        break;
      case TO_ZERO: 
        @<round to zero@>@;
        break;
      case TO_P_INF: 
        @<round to plus infinity@>@;
        break;
      case TO_N_INF: 
        @<round to minus infinity@>@;
        break;
      case TO_INF: 
        @<round to infinity@>@; 
        break;
      case EXACT:; @;
    }
    if (mode!=EXACT)
      exponent+=(precision-digits);
  } 
  else
    if (bias != 0) 
      @<bias rounding@>@;

  is_exact = (digits >= precision) && (bias == 0);
  normalize();
  return *this;
}

@ We need a function that cuts an integer to an arbitrary amount of digits.

@<internal functions@>=
void cut (integer &b,long prec) @+ { @+ b=(b >> (b.length()-prec)); @+ }


@ First we implement rounding in the case that the significant of the
normalized |bigfloat| has more than |digits| places and hence
the rounded value is not equal to the exact value.

We start with the rounding case |TO_NEAREST|.
Rounding to nearest means that a given |bigfloat| number should be rounded 
to a bigfloat with |n|-bit significant and which is nearest to accurate. 
To get the nearest number we have to check the bit n+1.
If it is zero we only have to shift the significant to n digits.
If it is a $1$ we first check whether the bit $n+1$ is the last $1$.
This is true if and only if |precision == digits + 1|.
If |precision > digits+1| we have to increment the significant by one after
shifting it to n binary places. 
Finally, in the case of |precision==digits+1| we have to consider the bias 
to ensure correct rounding: 
If the bias happens to be zero, we round to even.
Otherwise, we have to round down or up, depending on the sign of the bias.

@<round to nearest@>=
cut(significant,digits+1);@;
test = @<significant is even@>;@;

cut(significant,digits);
if (!test)
{
   if ((precision > digits + 1) 
        || (::sign(bias) == ::sign(significant)) 
        || ( (bias == 0) && !@<significant is...@> )
      ) 
        if (::sign(significant)>0) significant++; @+ else significant--;
}


@ It is simple to test whether the significant is even.

@<significant is even@>=
   ( (significant & integer(1)) == 0 )

@ In the |TO_ZERO| case we cut the significant to the wanted amount of
  digits.

@<round to zero@>=
cut(significant,digits);

@ In the |TO_INF| case we always round up because we already know that
the rounded value is not exact. 

@<round to infinity@>=
cut(significant,digits);
if (::sign(significant)>0) significant++; @+ else significant--;

@ The next case is |TO_P_INF|.
Here the significant is rounded up.
If the number is negative this is done by cutting |significant| to length
|digits|.
Otherwise we increment the significant after the cutting.

@<round to plus infinity@>=
cut(significant,digits);
if (::sign(significant) > 0) significant++;

@ The |TO_N_INF|-mode is the opposite case to |TO_P_INF|.

@<round to minus infinity@>=
cut(significant,digits);
if (::sign(significant) < 0) significant--;

@ Now we come to the cases where the wanted precision is less or equal to the
original precision and bias is crucial.
Again we consider the rounding modes separately.
For mode |TO_NEAREST| there is nothing to do.
In the other cases, it depends on the sign of the bias whether we have 
to do nothing, or else we have to add a $1$ or $-1$ at the |digits|$^{th}$
place.

@<bias rounding@>=

{
shift = digits - precision;
long bf_sign=::sign(significant);

switch(mode)
{
  case TO_ZERO: 
    if (::sign(bias) != bf_sign) 
    {
      @<shift significant@>@;
      if (bf_sign==1)
        significant--; else
        significant++;
    }
    break;
  case TO_INF: 
    if (::sign(bias) == bf_sign) 
    {
      @<shift significant@>@;
      if (bf_sign==1)
        significant++; else
        significant--;
    }
    break;
  case TO_P_INF: 
    if (::sign(bias) == 1) 
    {
      @<shift significant@>@;
      significant++;
    }
    break;
  case TO_N_INF: 
    if (::sign(bias) == -1) 
    {
      @<shift significant@>@;
      significant--;
    }
    break;
  /* in the cases |TO_NEAREST| and |EXACT| we have to do nothing */
  case TO_NEAREST: case EXACT: ;
}
}


@ When shifting the significant left by |shift| digits we have to lower
the |bigfloat|'s exponent by the same amount.

@<shift significant@>=

if (shift>0)
{
  significant=significant << shift;
  exponent-=shift;
}

@ Function |round_with_error| does exactly the same as function |round|
but in addition returns the rounding error which is the difference of 
original and rounded value.

@<general functions@>+=

bigfloat bigfloat::round_with_error (long digits,rounding_modes mode,     
                                     bool &is_exact,long bias)
{
  bigfloat temp=(*this);
  round(digits,mode,is_exact,bias);
  return (sub((*this),temp,temp.precision));
}

@ A |bigfloat| $x$ can also be rounded to integer format by the function
|to_integer|.
If $x$ represents a nonzero special value, we throw out an error
message.
Now let us assume that $x$ is a nonspecial value.  
We round $x$ to |length = max(precision+exponent,1)| places and assign the
rounded value to |result|.
Our choice of |length| guarantees that |result| represents the integer next to 
|x|.
Since |result| has a nonnegative exponent (if it is nonzero), we simply return
its significant, shifted by |result.exponent| digits to the left.

@<general functions@>+=
integer bigfloat::tointeger (rounding_modes rmode) const
{
  if ((isNaN())||(isInf()))
    error_handler(1,"bigfloat::tointeger : special values cannot be converted to integer");
  if (!(exponent+(integer)significant.length()).islong())
    error_handler(1,"bigfloat::tointeger : (exp+sig len) has to be in long range");

   bigfloat result=(*this);
   long pow2 = precision+exponent.tolong();
   long length = max(1,pow2);
   result.round(length,rmode);
   if (result.isZero() || (pow2 <= 0)) @+ return integer(0);
   else return (result.significant << result.exponent.tolong());
}

@* Arithmetical functions.

@ The Add-Function.

In the following sections we explain the operational functions
|add|, |sub|, |mul|, |div|, and |sqrt|.
We start right off with the |add| function. 

Let |a| and |b| be the operands of the addition.
At first we make sure that the binary logarithm of |a| (rounded up to
the next integer) is not less than the binary logarithm of |b|.
Then we compute |bigfloat|s |sum| and |error| such that we have the
exact equality
$$a + b = sum + error.$$ 
Here error is bounded by half the least significant digit of |sum| and
|error==0| for |mode==EXACT|.
Finally we give back the rounded value of |sum|. 
Procedure |round| only needs to know the sign of |error|.

@<operational functions@>=

bigfloat add (const bigfloat &x,const bigfloat &y,long prec,rounding_modes mode,
bool& is_exact)
{
  bigfloat a,b;
  is_exact = true;

  @<check of special cases@>@;

  @<find bigger operand@>@;
  /* now $\lceil \log_2{a} \rceil \geq \lceil \log_2{b} \rceil$  */
  
  bigfloat sum, error;
  @<compute sum and error@>@;

  sum.round(prec,mode,is_exact,error.sign());
  return sum;
}


@ It is helpful to know the "bigger" operand. 
To decide which operand is bigger we compute for |x| and |y| the sums of
exponent and precision, called |log_x| and |log_y|.
The difference |diff| of these quantities must not be confused with the
difference of the exponents, |exp_diff|.

@<find bigger ...@>=
  bigfloat *a_ptr, *b_ptr;
  integer log_x = x.exponent+(integer)x.precision;
  integer log_y = y.exponent+(integer)y.precision;
  integer diff = log_x - log_y;

  if (diff >= 0)
  {@+ 
    a_ptr =(bigfloat*)&x;@+ 
    b_ptr =(bigfloat*)&y;@+
  } else
  { @+
    a_ptr =(bigfloat*)&y;@+
    b_ptr =(bigfloat*)&x;@+
  }

  a = *a_ptr;
  b = *b_ptr;

  diff.absolute();
  integer exp_diff=a.exponent-b.exponent;

@ Often it is unnecessary to compute the exact sum. 
If |b| has no influence on the addition's result it suffices to set |sum| 
to |a| and |error| to |b|. 
This is the case if |b| is less than the least significant digit of |a| and
also less than the $(prec+1)^{th}$ digit of |a|.
(Here we can assume that |a| has at least |prec+1| digits by conceptually
shifting left, if it has less digits.)
These restrictions can be formulated as |diff > a.precision| and
|diff > (prec+1)|.
Furthermore the rounding mode must be different from |EXACT|.
If one of the conditions is violated we compute the addition exactly.

@<compute sum and...@>=
if ((mode!=EXACT)&&(diff>max(prec+1,a.precision)))@;
{
  sum=a; 
  error=b;
} 
else
{
  @<exact addition@>@;
}

@ The exact addition of two |bigfloat| numbers can easily be attributed to 
|integer| addition. 
The precondition for the use of |integer| arithmetic is that the two
|bigfloat| operands have equal exponents. 
In this case the result can be calculated as follows:

@<calculate sum@>=
sum=bigfloat(a.significant+b.significant,a.exponent);
  
@ If |a.exponent > b.exponent| we shift |a.significant| leftwards
by |exp_diff == a.exponent - b.exponent| binary places. 
Similarly, if |b.exponent > a.exponent| we shift |b.significant| leftwards 
by |-exp_diff| places. 
In both cases the exponents have to be set to the smallest exponent of |a|
and |b|.


@<exact addition@>=

if (!exp_diff.islong())
  error_handler(1,"bigfloat::add() : exponential difference has to be in long range");

if (exp_diff>0)
{
  a.significant=a.significant << exp_diff.tolong();
  a.exponent=b.exponent;
} else
{
  b.significant=b.significant << (-exp_diff).tolong();
  b.exponent=a.exponent;
}  
@<calculate sum@>@;


@ The rules to handle special cases in addition are simple.
The result is NaN if one of the operands is NaN or for the sums 
$\infty + (-\infty)$ and $(-\infty) + \infty$.
In the other cases we return the sum of |x| and |y|.

@<check of special cases@>=
if (x.isSpecial() || y.isSpecial())
{
  if (x.isNaN() || y.isNaN())
    return NaN;
  if (x.isZero()) return y;
  if (y.isZero()) return x;

  if (x.isInf() && y.isInf())
  {
    if (x.non_zero_sign() == y.non_zero_sign())
      return x;
    else return NaN;
  }
  if (x.isInf()) return x;

  /* it is obvious that |y| has to be $\infty$ */
  return y;
}

@ The Sub-Function.

This function is quite easy because it can be reduced to the |add|
function.  

@<operational functions@>+=

bigfloat sub (const bigfloat &a,const bigfloat &b,long prec,
	      rounding_modes mode,bool &is_exact)
{
  return add(a,-b,prec,mode,is_exact);
}   

@ The Mul-Function.

The multiplication is always done by first computing the exact result and
rounding afterwards, in contrast to the procedure for addition.
Computing the exact result is done by simply multiplying the significands
of the operands and adding their exponents.

@<operational functions@>+=

bigfloat mul (const bigfloat &a,const bigfloat &b,long prec,
              rounding_modes mode,bool &is_exact)
{
  bigfloat result;
  is_exact = true;

  @<special cases for mul@>@;
  
  result.special = NOT;
  result.significant=a.significant*b.significant;
  result.exponent=a.exponent+b.exponent;
  return result.round(prec,mode,is_exact);
}

@ We come to the special case treatment of |mul|.
The result is |NaN| if one of the operands is |NaN| or else,
if one of the operands is zero and the other one infinity.
Otherwise, if one operand is zero, we return zero, and if one of the operands
is infinity, we return infinity.
Here the sign of the return value is the product of the operand signs. 

@<special cases for mul@>=

  long sign_result;

  if ((a.isSpecial())||(b.isSpecial()))
  { 
    if ((a.isNaN())||(b.isNaN()))
      return NaN;
    if ( (a.isZero() && b.isInf()) || (a.isInf() && b.isZero()) )
      return NaN;

    sign_result = a.non_zero_sign()*b.non_zero_sign();

    if (a.isZero() || b.isZero()) 
    {
      if (sign_result == 1)
        return pZero;
      else
        return nZero;
    }
    if (a.isInf() || b.isInf()) 
    {
      if (sign_result == 1)
        return pInf;
      else
        return nInf;
    }
  }


@ The Div-Function.

One important difference between the division and other operational functions
is that the exact calculation of a division in |bigfloat| format is impossible.
Instead we use inexact |integer| division of the significands to approximate
the result up to $(prec+1)$ digits \footnote{we need one extra digit to
guarantee exact rounding}. 
Here it may be necessary to shift the significant of the dividend
to the left.
We also compute the sign of the division remainder in the variable |bias|.
Then the correctly rounded division result can be obtained by calling
the |round| function with this bias.

@<operational functions@>+=

bigfloat div (const bigfloat &a,const bigfloat &b,long prec,
              rounding_modes mode, bool &is_exact)
{
  is_exact = true;
  bigfloat result;
  long bias;

  @<special cases for div@>@;
  @<shift dividend's significant@>@;
  @<compute approximative result@>@;
  result.round(prec,mode,is_exact,bias);
  
  return result;
}


@ We first try to calculate the number of shifting digits for the
dividend |a|.
Let $S_a$ and $S_b$ be the significants of the operands with binary
lengths $l_a$ and $l_b$.
We suppose that $S_a$ and $S_b$ are nonzero.
Then we have 
$2^{l_a-1} \leq S_a < 2^{l_a}$ 
and  
$2^{l_b-1} \leq S_b < 2^{l_b}$
which implies 
$$2^{l_a-l_b-1} < S_a/S_b < 2^{l_a-l_b+1}.$$
Hence the precision of $S_A / S_b$ is not less than $l_a - l_b$ which 
is also true if we use integer division to round to zero.
That is we have to guarantee $l_a-l_b \geq prec+1$.
If $d = l_b - l_a + prec +1 > 0$ we have to shift the significant of |a| by
|d| digits.

@<shift dividend's...@>=
bigfloat aa=a;
long d=prec+b.significant.length()-a.significant.length()+1;
if (d>0)
{
  aa.significant=aa.significant << d;
  aa.exponent-=d;
}

@ Computing the approximation of the result is now easy.
We simply do a Euclidean division of $S_a$ and $S_b$, that is, 
$S_a = S_b\ cdot S_r + R$ where $R$ is the division remainder.
This is equaivalent to $$S_a/S_b = S_r + R/S_b$$
and hence the bias is given by the sign of $R/S_b$.

@<compute approximative result@>=
result.special = NOT;
result.significant=aa.significant/b.significant;

result.exponent=aa.exponent-b.exponent;
result.precision=result.significant.length();

integer R=aa.significant-b.significant*result.significant;
if (R!=0) is_exact=false;
if (mode==EXACT) mode=TO_NEAREST;

bias = sign(R)*sign(b.significant);


@ The special case handling for division is very similar to that for 
multiplication.
We omit the details.

@<special cases for div@>=

if ((a.isSpecial())||(b.isSpecial()))
{
  long sign_result;

  if ((a.isNaN())||(b.isNaN())) 
    return NaN;
  if (((a.isZero())&&(b.isZero()))||((a.isInf())&&(b.isInf())))
    return NaN;
  sign_result = a.non_zero_sign()*b.non_zero_sign();
  if ((a.isInf())||(b.isZero()))
  { 
    if (sign_result == 1)
      return pInf;
    else 
      return nInf;
  }

  /* it is clear that |isZero(a)||(isInf(b)| */

  if (sign_result == 1)
    return pZero; else 
    return nZero;
}


@* the sqrt-function.

We compute an approximative square root of a bigfloat using the iteration
\begin{equation}
x_{n+1}=\frac{x_n^2+a}{2x_n}
\end{equation}
This is Newton approximation, applied to the function $f(x)=x^2-a$. It is
easy to see that for any start value $x_0>0$, we have $x_1 \geq \sqrt{a}$.
Because f is a convex function, we further have
\begin{math}
x_1 \geq x_2 \geq \ldots \geq x_n \geq \sqrt{a}
\end{math}
provided that Equation (1) is evaluated exactly.
Of course, we cannot compute bigfloat divisions exactly.
However, if we use the rounding modes |EXACT| to exclude numerator
numerator $x_n^2+a$ and denominator $2x_n$ and |TO_P_INF| to evaluate
the division, we compute the approximate value $\tilde{x_{n+1}}$ which is
$x_{n+1}$ rounded up after the $p^{th}$ most significant digit.
Hence for the computed sequence $\tilde{x_1},\tilde{x_2}, \ldots $ of
approximative iterations, we still have
\begin{displaymath}
\tilde{x_1} \geq \tilde{x_2} \geq \ldots \geq \tilde{x_n} \geq \sqrt{a}
\end{displaymath}
Also the quadratic speed of Newton's method is preserved. Our termination
condition is $\tilde{x_{n+1}}=\tilde{x_n}$.\footnote{using rounding mode
|TO_NEAREST| is would be difficult to guarantee that the stop condition
will ever be achieved} We claim that in this case we have 
\begin{math} 
\sqrt{a} \in (\tilde{x_n}-\delta,\tilde{x_n}] 
\end{math}
where $\delta$ is the value of the most
significant bit in representation of $x:=\tilde{x_n}$.
To see this we observe that because of rounding to $+\infty$ there exists
$\delta>\Delta>0$ s.th.
\begin{displaymath}
x=\frac{x^2+a}{2x}\cdot (1+\Delta)
\end{displaymath}
which is equivalent to 
\begin{displaymath}
x*\sqrt{\frac{1-\Delta}{1+\Delta}}=
x*\frac{1-\Delta}{\sqrt{1-\Delta^2}}=
\sqrt{a}
\end{displaymath}
It follows that $x\cdot (1-\Delta)<\sqrt{a}$ which proves the statement.
The square root function now looks as follows.
As you can see below we decided to speed up the sqare root by intoducing
a new variable named |t_prec|. Newton approximation is known to double the
accuracy of the result in every iteration step. So we can save time by
replacing the |EXACT| rounding mode by the default rouding mode and set the 
precision to |t_prec| which is multiplicated by 2 at the end of every loop.

The squareroot function has an additional parameter |start| that allows the user
to give his own start value for the Newton iteration.

@<operational functions@>+=
bigfloat sqrt (const bigfloat &a,long prec,rounding_modes mode, 
  	       bool &is_exact,const bigfloat& start)
{
  bigfloat result;
  long t_prec, p_bound = prec+1;
  bool exact=true;

  @<special cases of sqrt@>@;
  @<calculate start value of iteration@>@;
  @<Newton iteration@>@;
  @<sqrt rounding@>@;
  return result;
}

@ There are several special cases to check in the beginning of our function
|sqrt|.
In particular, if a is zero then we return a.

@<special cases of sqrt@>=

  if (a.isZero()) @+ return a;
  if (a.sign() < 0) @+ { @+ is_exact = false; @+ return NaN; @+ }
  if (a.isSpecial()) @+ { @+ is_exact = false; @+ return a; @+ }

@ A good start value saves a lot of iterations and so it saves time. One
can save time in chosing a value bigger than and near to the expected result.
Therefore we take the |bigfloat|s exponent and divide it by two in order
to get the exponent of the start value. If the former exponent was odd we
finally add 1.
The start value's significant is the |double| square root 
of the rounded significant of the given |bigfloat|.

@<calculate start...@>=
long prec_a = a.get_precision();
double d;
if (start.isSpecial() || start <= 0)
{
  result = bigfloat(a.significant,-prec_a);
  integer log_a = a.exponent+prec_a;
  d =  sqrt(result.todouble());
  if (!(log_a%2==0)) 
    d*=2;
  result=d;
  result.exponent += log_a >> 1;
  t_prec=52;
}
else result = start;

@ The Newton iteration is the core of the sqrt function.
We compute $\sqrt{a}$ rounded up to |prec + 1| digits to make rounding
easier.
The iteration terminates if |ne == it| is the exact value of $\sqrt{a}$ or else
the full precision |prec+1| is achieved and |ne == it|.

@<Newton iteration@>=
  bigfloat it; // iterator
  bigfloat& ne = result; // next iteration value
  
  t_prec = ne.precision;
  bigfloat square, sum;
  while(1==1)
  {
    exact = true;
    t_prec*=2; if(t_prec > p_bound) t_prec = p_bound;
    it=ne;
    square = mul(it,it,t_prec,EXACT);
    sum=add(square,a,t_prec,EXACT); 
    ne=div(sum,it,t_prec,TO_P_INF,exact); 
    ne.exponent--;
    if ((t_prec == p_bound) || exact)
      if (ne == it)
         break;
  }
  if (!exact) is_exact = false;

@ We provide all rounding modes except |EXACT| for function |sqrt|. The returned
value is the \underline{exactly} rounded value of $\sqrt{a}$.

@<sqrt rounding@>=
if (exact) result.round(prec,mode,is_exact);
else
{
  switch(mode)
  {
    case TO_NEAREST: result.round(prec,TO_ZERO,is_exact); @+ break;
    case EXACT:
  error_handler(1,"bigfloat::sqrt : the EXACT-rounding mode is not provided!");
    case TO_P_INF: case TO_N_INF: case TO_INF: case TO_ZERO:
       result.round(prec,mode,is_exact);
  }
  is_exact = false;
}
         

@* Comparison Operators.

We still have to implement the operators |==| and |>|.
Remember that the other comparison operators have been reduced to these 
two cases.
Let us begin with operator |==|. 
If no operand is a special value, we normalize the operands to make their representation unique
\footnote{
Since |normalize| is a not a constant member function we first cast
the operands to type |bigfloat&|.
This is allowed since |normalize| does not change the value of a |bigfloat|
but only its representation.
Here we use the idea of logical constance.
}.
Then we only have to compare significant and exponent.
 
@<operators@>+=
bool operator == (const bigfloat &a,const bigfloat &b)
{
  @<special case checking for operator $\equiv$@>@;
  ((bigfloat &)a).normalize();
  ((bigfloat &)b).normalize();
  return ( (a.significant==b.significant)&&(a.exponent==b.exponent) );
}

@ The rules of special-case checking can be summarized as follows.
Comparisons with a |NaN| not allowed.
If both operands are zero, |true| is returned.
Otherwise, |true| is returned if and only if the operands represent 
the same special value.

@<special case checking for operator $\equiv$@>=
  if (a.isSpecial() || b.isSpecial())
  {
    if (a.isZero() && b.isZero())
      return true;
    if (a.isNaN() || b.isNaN())
      error_handler(1,"bigfloat::operator == : NaN case occurred");
    return (a.special == b.special);
  }



@ We come to |operator >|. 
For non-special |bigfloat|s we first check the signs of the 
operands. 
If the signs are different we can immediately compute the result.
Otherwise we find out the operand with the bigger exponent and shift its
significant by the exponential difference to the left.
Afterwards we return the |integer| comparison of the significants.

@<operators@>+=
bool operator > (const bigfloat &a,const bigfloat &b)
{  
  @<special case checking of operator $>$@>@; 
  int sign_a = ::sign(a.significant);
  int sign_b = ::sign(b.significant);
  if (sign_a != sign_b) return (sign_a > sign_b);

  integer sig,diff=a.exponent-b.exponent;
  if (!diff.islong())
    error_handler(1,"bigfloat operator > : exponential difference not in long range");
  if (::sign(diff) >= 0)
  {
    sig=a.significant << (diff.tolong());
    return (sig>b.significant);
  } 
  else
  {
    sig=b.significant << ((-diff).tolong());
    return (a.significant>sig);
  }
}


@ Finally we come to the special case checking for operator |>|.
As before, |NaN| comparisons are not allowed. 
If the |==| operator returns true, operator |>| returns false.
The remaining cases are straightforward.

@<special case checking of operator $>$@>=
  if (a.isSpecial() || b.isSpecial())
  {
    if (a.isNaN() || b.isNaN())
      error_handler(1,"bigfloat::operator > : NaN case occurred!");
    if (a.isZero() && b.isZero()) return false;
    if (a.ispInf()) return !b.ispInf();
    if (b.isnInf()) return !a.isnInf();
    return (a.sign() > b.sign());
  }





@* Conversion between the data types {\bf bigfloat} and {\bf double}.

It is possible to convert a |double| into a |bigfloat| and vice versa. 
If the significant of the |bigfloat| has more than 53 bits or if its 
exponent is too large there will be a loss of information.
For a |double| |d|  we will always guarantee the consistency property
|bigfloat(d).todouble() == d|.

@ Some facts on |double|s.

A |double| is specified by a sign $s \in \{0,1\}$, an exponent 
$e \in [0,2047]$ and a binary fraction $F=f_1f_2 \ldots f_{52}$ with
$f_i \in {0,1}$ for all $1\leq i \leq 52$. 
For $0 < e < 2047$ the triple ($s$,$f$,$e$) represents the number
$$(-1)^s \cdot 1.f \cdot 2^{e-1023}.$$
Such a number is called {\sf normalized}.
If $e$ is zero the number is called {\sf denormalized}. and the value
represented by ($s$,$f$,$e$) is
$$(-1)^s \cdot 0.f \cdot 2^{-1023}.$$
Denormalized numbers are smaller than the smallest normalized number 
|double_min| $= 2^{-1022}$.
Due to the limited exponent range |double|s can over- and underflow. 
To get more security in arithmetical operations there are the values
$\pm \infty$ signaling overflow and the value $NaN$ signaling invalid
operations like $0\cdot \infty$.
$\pm \infty$ are represented by $s \in \{0,1\}$, $e = 2047$ and $f = 0$,
and $NaN$ is represented by any triple $(s,e,f)$ with $e = 2047$ and $f \neq 0$.


@ The function |compose_parts| composes the parameters $(s,e,f)$ to a |double|
value.
The sign $s$ is delivered in |sign_1|, and |exp_11| is the exponent $e$.
Since we use $32$ bit wide |long|s, the $52$ bit of $f$ are split into the 
lower part |least_sig_32| and the higher part |most_sig_20|.
The resulting |double| is made of two |long|s, the higher $32$ bit 
containing |sign_1|, |exp_11| and |most_sig_20|, and the lower $32$ bit
of |least_sig_32|.

@<internal functions@>+=
double compose_parts (long sign_1,long exp_11,long most_sig_20,long least_sig_32)
{
  double a;
  long high32=0;
 
  @<calculate high32@>@;
  @<put it all together@>@;

  return a;
}

@ First we compute the higher $32$ bit of the |doube|.

@<calculate high32@>=
  if (sign_1)
    high32=high32|0x80000000;
  exp_11=exp_11 << 20;
  high32=high32|exp_11;
  high32=high32|most_sig_20;


@ To compose the two |long| parts to one |double| value we use pointer
arithmetic.
For this we need a pointer to a |long| variable that initially
points to the beginning of the resulting |double| |a|. 
Since we would like to achieve machine independent code we have to care for
different byte ordering mechanisms.
Sun Sparc stations use {\sl BIG ENDIAN} byte ordering which means that always
the higher part is saved before the lower part in memory. 
Intel PCs use {\sl LITTLE ENDIAN} byte ordering in which the lower part is
saved first.

@<put it all together@>=
  long *p;
  p=(long*)&a;
#ifndef LITTLE_ENDIAN
  (*p)=high32; @+ p++; @+ (*p)=least_sig_32;
#else
  (*p)=least_sig_32; @+ p++; @+ (*p)=high32;
#endif


@ The mathematical function |pow2| for |double|s that computes $2^{exp}$
for a given exponent |exp| can be realized efficiently using the new
function |compose_parts|.

@<internal functions@>+=
double pow2(long exp)@+ {@+ return compose_parts(0,exp+1023,0,0);@+ }

@ Now we are able to define the special |double| values that we need in the
 following sections.

@<global identifiers@>=

/* since we need the |pow2| and |compose_parts| functions we have to use
   prototyping */

double compose_parts (long,long,long,long);
double pow2(long);

const double double_min=pow2(-1022);
const double NaN_double=compose_parts(0,2047,0,1);
const double pInf_double=compose_parts(0,2047,0,0);
const double nInf_double=-pInf_double;
const double pZero_double=compose_parts(0,0,0,0);
const double nZero_double=compose_parts(1,0,0,0); 


@ The next constructor transforms a |double| |d| into a |bigfloat|. 
First we check whether |d| is denormalized.
In the case that |d| is denormalized, we correct the exponent of |d| to make it
normalized (if it is nonzero) but set a flag that allows us to take back our
changes later.
After that we split |d| into a high and a low part. 
From these two parts we compute the sign, significant and exponent of |d|.

@<simple constructors@>+=

bigfloat::bigfloat (double d)
{  
  int sign;
  long flag=0;
  unsigned long mh=0, ml=0;
  long *p;

  @<check for denormalized number@>@;
  @<determine high and low part@>@;
  @<get the double's sign@>@;
  @<get the significant's value@>@;
  @<get the exponent's value@>@;
  @<check for special values@>
  normalize();
}

@ If |d| is smaller than |double_min| then it is denormalized and
we multiply |d| with $2^{52}$ to get a normalized number.

@<check for denorm...@>=
    if (fabs(d)<double_min)
    {
      d=d*pow2(52);
      flag=1;
    }

@ We split the 64 bit wide |double| representation in two 32 bit wide
|long|s. 
To do this we use a pointer to a |long| value and assign the casted address of
the |double| value to it. 
The |long| pointer works like an |array|.
If the {\sl BIG ENDIAN} byte ordering is active, the first component holds
the higher 32 bit part |mh| and if the {\sl LITTLE ENDIAN} byte ordering is
active, the first component holds the lower 32 bit part |ml|.

@<determine...@>=
    p =(long*) &d;
#ifndef LITTLE_ENDIAN
    mh= *p; @+ p++; @+ ml= *p;
#else
    ml= *p; @+ p++; @+ mh= *p;
#endif 

@ The |sign| of a |double| is denoted by its highest bit.

@<get the double's sign@>=
  if (mh & 0x80000000) @+ sign = -1;
  else @+ sign = 1;
  

@ The significant of the |double| is composed out of |ml| and the least 20 bits of |mh|. 
As we have ensured normalized representation in the beginning of this function
we have to keep in mind that the significant represents the binary fraction and
that there is an implicit one bit that leads this fraction.

@<get the significant...@>=
    long sig=0;
    /* get the significant bits out of the highword */
    sig = mh & 0x000fffff;
    /* we have ensured that |d| is normalized $\Rightarrow$ add the leading one
       bit */
    sig = sig | 0x00100000;
    /* compose significant */
    significant=integer(sig);   
    significant=significant << 32;
    significant=significant + integer(ml);
    if (sign == -1) significant = -significant;

@ The |double| exponent takes the bits $2$ to $12$ bits of |mh|.
We use bit manipulation to extract this value out of |mh|. 
To transform the |double| exponent into the exponent of the resulting
|bigfloat|, we have to correct its value by
\begin{enumerate}
\item the bias $-1023$,
\item |-(significant.length()-1) == -52|, due to the |bigfloat|'s representation
\item -52, if |d| was denormalized
\end{enumerate}

@<get the exp...@>= 
    long e=0;
    /* get the 11 bits of the exponent */
    e = mh & 0x7ff00000;
    /* shift the result to get the right value */
    e = e >> 20;
    /* subtract bias and 52*/
    e -= 1075;
    /* subtract 52 if |d| was denormalized */
    if (flag) @+ e-=52;
    exponent = e;
    

@ Finally, we have to check whether |d| was in a special state.

@<check for spec...@>=
    special = NOT;
    if (e==972) // Inf or NaN-Case
    {

sig = mh & 0x000fffff;
   
      if ((mh & 0x000fffff) == 0 && ml == 0)
        special = (sign > 0) ? PINF : NINF;
      else 
        special = NAN;
    }
    if (d == 0) 
      special = (sign > 0) ? PZERO : NZERO;


@ In several functions we use |integer| values which have - in their
binary representation - one single bit set.

@<global identifiers@>+=

const integer integer_1=integer(1);
const integer integer_52=(integer_1 << 52)-integer_1;
const integer integer_32=(integer_1 << 32)-integer_1;
const integer integer_20=(integer_1 << 20)-integer_1;

@ the todouble function.

This function takes the given |bigfloat| and converts it into |double|
format. 
If the |bigfloat| is in a special state (NaN,pInf,...), we return the
corresponding special |double| value. 
A bigfloat $x$ is called {\em approximable} if 
$$2^{-1074} \leq \mid x \mid < 2^{1024}.$$
For approximable values of $x$, we return the |double| nearest to $x$,
otherwise pInf,nInf,pZero or nZero are returned. 


In the main part of the function we distinguish the cases that the returned
|double| is normalized or denormalized.
At the end the significant |s|, the exponent |t_exp| and the significant |t_sig|
of the |double| are put together.

@<general functions@>+=
double bigfloat::todouble (void) const
{@#
  long s=::sign(significant); 
  long t_exp=0; 
  integer t_sig=0; 
  
  bigfloat rounded_value = fabs(*this);

  @<special case checking of todouble@>@;
  @<rounding step of todouble@>@;
  @<check for normal or denormal return value@>@; 
 
  if (normal)@;
    @<normal case@>@;
  else
    @<denormal case@>@;

  double a;
  @<set the bits of a@>@;
  return a;
}

@ The special cases are first.

@<special case checking of todouble@>=
if (isSpecial())
{
  if (ispZero()) @+ return pZero_double;
  if (isnZero()) @+ return nZero_double;
  if (isNaN())   @+ return NaN_double;
  if (ispInf())  @+ return pInf_double;
  if (isnInf())  @+ return nInf_double;
}

@ We round the |bigfloat| such that all bits of the signficant just fit into
|double| format.
We distinguish cases according to the quantity |log2 = exponent + precision|.
\begin{itemize}
\item If |log2| is between $-1021$ and $1024$, we have to round to $53$ places.
\item If |log2| is between $-1073$ and $-1022$, we have to rund to $1074+log2$ 
      places.
\end{itemize}
If |log2| is outside of $[-1073,1023]$ the conversion overflows or underflows.
Note that $log2$ might change through the rounding.

@<rounding step of todouble@>=
  integer log2=exponent + precision;
  if (log2>1024) return non_zero_sign()*pInf_double;
  if (log2<-1073) return non_zero_sign()*pZero_double;
  if (log2>-1021)
   rounded_value.round(53,TO_NEAREST);
  else 
   rounded_value.round(1074+log2.tolong(),TO_NEAREST);
   

@ Now we decide whether the |bigfloat| is approximable by a normalized |double|
or not.
This is the case if and only if
$$2^{-1022} \leq \mid |rounded_value| \mid < 2^{1024}.$$
Note that |log2| can be $1025$ because of rounding, in which case we return
infinity.

@<check for normal...@>=
  long normal = 1;
  log2 = rounded_value.exponent + rounded_value.precision;

  if (log2==1025) return non_zero_sign()*pInf_double;
  if (log2<-1021) 
    normal = 0;

@ First we assume that we have a denormalized value to store. 
Remember that a denormalized |double| has an implicitly leading 0 bit
and that its unbiased exponent is $-1022$. 
Let us consider the number $k$ of zeros such that 
$|rounded_value| = sig \cdot 2^e$ 
equals 
$0.\underbrace{00\ldots 0}_k \cdot 2^{-1022}$.
This is equivalent to $k = -1022 - e - l$ where $l$ is the length of the
significant $sig$ of |rounded_value|.
We have to ensure that the |bigfloat| significant has length $52 - k$, that
is, we shift the significant by |l_shift = 52 - k - l| places to the left.
(If |l_shift| is negative rightshifts have to be performed).
Simplifying we see that |l_shift = e + 1074|.

@<denormal case@>=

{
  long l_shift=rounded_value.exponent.tolong()+1074;
  if (l_shift>0) @+ t_sig = rounded_value.significant << l_shift;
  else t_sig = rounded_value.significant >> (-l_shift);
}
    
@ Now we assume that we have a normalized value to store.
As normalized |double|s have an implicit leading one, the biased exponent of the
returned |double| is 
\begin{displaymath}
t\_exp=exponent+(l-1)+1023
\end{displaymath}
where again |l| is the length of the |bigfloat| significant.
Finally we cut off the first bit of |significant| (which is necessarily one)
and obtain the |double| significant |t_sig|.

@<normal case@>=
{
  t_exp=rounded_value.exponent.tolong()+(rounded_value.precision-1)+1023;
  /* if the length of the rounded |significant| is lower than 53 we
     have to shift it to the left */
  t_sig=rounded_value.significant << (53-rounded_value.significant.length());
  t_sig=t_sig&integer_52;
}

@ Finally we set the bits of the returned |double|. 
That means that we have to lay down our specified values for sign, significant
For this we use the function |compose_parts| and so we just need to calculate
the arguments for this function.

@<set the...@>=
unsigned long sign,h_sig,l_sig;
sign=(s==(-1));
h_sig=((t_sig >> 32)&integer_20).tolong();
l_sig=(t_sig&integer_32).tolong();
a=compose_parts(sign,t_exp,h_sig,l_sig);													














@* Functions for input and output.

There are two operators which perform input/output.
\begin{itemize}
\item The first one is the |operator <<|. 
     It outputs a |bigfloat| in either binary or decimal representation.
     The global variable |output_mode| determines the output form.
     The two different modes are
     \begin{itemize}
     \item binary (|BIN_OUT|)
     \item decimal (|DEC_OUT|)
     \end{itemize}
     This operator uses the |binout| function for binary output
     and function |decimal_output| for decimal output.
\item The second one is the operator |>>|. 
     It performs decimal input of a |bigfloat| using function |outofchar|.
     Binary input is not yet implemented.
\end{itemize}
We store the default output mode in a variable |output_mode| and define the 
maximal size of a string field that is allocated during input and output.

@<global identifiers@>+=

output_modes output_mode=DEC_OUT;
const long bin_maxlen=10000;

@ Procedure |binout| takes an output stream and an |integer| |b| as input.
It produces an unsigned, binary output of |b|.

@<internal functions@>+=
void binout (ostream &os,integer b)
{
  char temp[bin_maxlen];
  long flag=0,count=0;

  if (b<0) @+ b*=-1;

  do
  {
    temp[count++]=(char)(b%2).tolong()+'0';
    if (b<=1) @+ flag=1; 
    else @+ b/=2;
  } while (!flag);
  for (long i=count-1;i>=0;i--)
    os << temp[i]; 
}

@ The follwing function computes the $n^{th}$ power of a |bigfloat|
in precision |prec| and rounding mode |mode|.
\footnote{
This does not mean that the result is correct up to |prec| digits.
Only every operation within the procedure is carried out in that precison.
}. 
We use it with $n = 10$ for our decimal output.

@<auxiliary functions@>+=
bigfloat powl(const bigfloat &x, long n, long prec = 1, 
              rounding_modes mode = EXACT) 
{
  bigfloat z = 1,y=x;
  long n_prefix = n;
    
  while (n_prefix > 0)
  {
    if (n_prefix % 2) 
      z = mul(z,y,prec,mode);
    n_prefix = n_prefix / 2;
    y = mul(y,y,prec,mode);
  }
  return z;
}

@ Procedure |decimal_output| displays a |bigfloat| |b| in decimal floating
point notation, with |prec| decimal places in the significant.
It first treats the sign of the given |bigfloat b|.
The further procedure works with $\mid b \mid$ instead of |b|.
Then we compute the decimal exponent of |b| which allows us to get the
wanted |prec| decimal places of the output.
Finally we write the output to |ostream os|.

@<internal functions@>+=
void decimal_output(ostream& os, bigfloat b, long prec)
{
  if (!b.get_exponent().islong())
    error_handler(1,"decimal_output: not implemented for large exponents");
  ostrstream oss; // string stream needed for output
  integer dd = 10;
  @<calculate the sign@>@;
  @<compute decimal logarithm of b@>@;
  @<compute decimal significant of b@>@;
  @<output of the result@>@;
}

@ First we calculate the sign of |b| and take the modulus of |b|.

@<calculate the sign@>=
  if (b.sign() < 0) @+ { @+ b = -b; @+ os << "-"; @+}

@ We need to know the decimal logarithm of $b$, rounded up to the next
integer.
For this we use the function |log10| from the standard math library that 
takes a |double| input.
To avoid the problem of overflow in the |double| calculation we first determine
$k = \lceil log_2(b) \rceil$, write $b = 2^k b_{rem}$ and finally compute
$$\log_{10}(b) = \log_{10}(2^k) b_{rem} 
               = k \cdot \log_{10}(2) + \log_10(b_rem).$$

@<compute decimal logarithm of b@>=
long log2_b  = b.get_precision() + b.get_exponent().tolong();
bigfloat b_rem(b.get_significant(),-b.get_precision());
long log10_b = (long) ceil(log2_b * log10(2) + log10(b_rem.todouble()));

@ Now we compute the |prec| decimal digits for the decimal output of |b|.
The idea is to multiply $b$ with an integer power of $10$ or else to divide $b$
by an integer power of $10$ such that the integer fraction of the resulting
number |b_shift| has exactly |prec| decimal digits.
We round this number to get the wanted decimal significant.
To be precise, we do the following.
If $log = |log10_b|$ is less than |prec|, we multiply $b$ by
$10^{prec-log}$ and round to the next integer value, or else, if $log$ is
bigger than |prec|, we divide $b$ by $10^{log-prec}$ and round to the next
integer. 
To do the latter operation we first estimate the binary length |l| of the
integer result, then compute the division in precision |l| and finally do the 
rounding.

@<compute decimal significant...@>=
long digits = b.get_precision() + 32;
long diff = prec - log10_b;
bigfloat multiplicator, divisor;
bigfloat b_shift;
integer significant;
if (diff >= 0)
{
  multiplicator = powl(dd, diff, digits, TO_NEAREST);
  b_shift = mul(b,multiplicator,digits,TO_NEAREST);
}
else
{
  divisor = powl(dd,-diff, digits, TO_NEAREST);
  b_shift = div(b,divisor,digits,TO_NEAREST);
}
significant = b_shift.tointeger(TO_NEAREST);
oss << significant;

@ At last, we have to output the computed value.
We take a floating point format with one digit in front of the decimal point. 

@<output of the result@>=
char *str=oss.str();
if (prec > 1) os << (*(str++)) << "." << str;
else os << str;
if (log10_b != 1) os << "E" << log10_b - 1;

@ Now we are ready to implement the stream operators.
We start with the output.
In the case of decimal output mode we have to compute the decimal precision
|prec| that we deliver to function |decimal_output|.
The basic idea is that the binary precision and the decimal precison 
correspond to each other.
To approximate the decimal precision corresponding to the length |l|
of the binary significant we multiply |l| by $\log_{10}(2)$.

@<operators@>+=
ostream& operator << (ostream& os, const bigfloat& b)
{
  if (b.isSpecial())
  {
    if (b.isNaN()) @+ return os << "NaN";
    if (b.ispInf()) @+ return os << "+Inf";
    if (b.isnInf()) @+ return os << "-Inf";  
    if (b.ispZero()) @+ return os << "+0";
    if (b.isnZero()) @+ return os << "-0";
  }
  int sign_b = sign(b.significant);
  if (output_mode==BIN_OUT)
  {
    if (sign_b<0) os << "-";
    os << "0.";
    if (sign_b>=0)
      binout(os,b.significant);
    else
      binout(os,-b.significant);
    os << "E";
    if (b.exponent<0) @+ os << "-";
    else @+ os << "+";
    binout(os,b.exponent);
  } 
  if (output_mode == DEC_OUT)
    decimal_output(os,b,(long) ceil(log10(2)*b.precision));
  return os;
}

@ We turn to the input operator.
The hard work is done by function |outofchar|.

@<operators@>+=
istream &operator >> (istream &is,bigfloat &b)
{
  char temp[bin_maxlen];
  is >> temp;
  b=outofchar(temp);
  return is;
}


@ Function |outofchar| provides the conversion of |string| to |bigfloat|.
Currently, only decimal input is possible. 
The expected format is:
\begin{displaymath}
\pm dd \cdots d[.dd \cdots d[E\pm dd \cdots d]]
\end{displaymath}
where $d$ is out of $[0,9]$ and $\cdots$ stands for arbitrarily many d's.
The procedure works as follows.
We read the sign, the integer part, the decimal fraction and the exponent 
of the decimal floating point representation in turn.
Then we concatenate the integer part and the decimal fraction into one 
decimal significant $m$ such that the input |bigfloat| is
$$m \cdot 10^{e-l}$$
where |l = frac_length| is the length of the decimal fraction and $e$ is the 
scanned exponent.
Now we distinguish cases.
If the exponent is larger than |l|, the represented number is an integer that
can be computed by multiplication with the integer $10^{e-l}$.
Otherwise, we divide the significant by the integer $10^{l-e}$ in 
precision |prec|.
If |prec| is less or equal to the default value zero, we set |prec|
to |significant.length()| which corresponds to the precision of the decimal
input.

@<internal functions@>+=
bigfloat outofchar (char *rep, long prec = 0)
{
  integer significant=0, exponent=0;
  bigfloat result, pow;
  integer dd = 10;
  long int_length=0, frac_length=0;

  int s;
  @<scan sign s@>@;
  int sign = s;
  @<scan integer part@>@;
  @<scan fraction@>@;
  @<scan exponent@>@;
  if (prec == 0) 
  {
     pow = powl(dd,int_length+frac_length);
     prec = pow.get_precision()+pow.get_exponent().tolong();
  }
  long exp = exponent.tolong() - frac_length;
  if (exp > 0) 
  {
     pow = powl(dd,exp,prec + 10,TO_NEAREST);
     result = mul(significant,pow,prec,TO_NEAREST);
  }
  else
  {
     pow = powl(dd,-exp,prec + 10,TO_NEAREST);
     result = div(significant,pow,prec,TO_NEAREST);
  }
  if (sign == 1) @+ return result; @+ else @+ return -result; 
}

@ First, we scan the optional sign at the beginning of the input.

@<scan sign s@>=
  s=1;
  if (rep[0]=='-') @+ { @+ s=-1; @+ rep++; @+ } 
  else if (rep[0]=='+') @+ rep++;

@ We need a function |isnum| to test if a scanned character is a digit.

@<auxiliary functions@>+=
bool isnum (char ch)
{
  if ((ch>='0')&&(ch<='9'))
    return true; else
    return false;
}

@ Now we scan the input upto the decimal point if there is one.
We pass every character and convert it to |int|. 
Each number is added to the temporary significant that is decimally shifted by one to the left.

@<scan integer...@>=
  while (isnum(*rep))
  { 
    int_length++;
    significant = significant * dd + (*(rep++)-'0');
  }

@ The fraction scan works quite similar as |@<scan integer...@>|.
We step over the input upto the character 'E' or until the input's end is
reached.

@<scan fraction@>=
 if (*rep == '.')
 { rep++;
   while (isnum(*rep))
   {
     significant = significant * dd + (*(rep++)-'0');
     frac_length++;
   }
 }
 

@ To scan the exponent we first read the optional sign and otherwise proceed 
as before.

@<scan exponent@>=
 if (*rep == 'E')
 { 
   rep ++;
   @<scan sign s@>@;
   while (isnum(*rep))
     exponent = exponent * dd + (*(rep++)-'0');
   if (s == -1) exponent = -exponent;
 }

@* Debugging tools.

@ Arithmetic Operators.

There are five arithmetic operators for the different symbols +,-,*,/ and
unaer minus. Each operator can be reduced to the corresponding
operational function. So for example |operator +| takes |add| to perform a 
|bigfloat|-addition.

@<operators@>+=

bigfloat operator + (const bigfloat &a,const bigfloat &b)
{
  #ifdef DEBUG_BF
    long operation = 0;
    double a_d = a.todouble();
    double b_d = b.todouble();
    @<set double mode@>@;
    double result_d = a_d + b_d;
    @<set default mode@>@;
    bigfloat result_bf = add(a,b);
    @<check result@>@;
  #endif
  return add(a,b);
}

bigfloat operator - (const bigfloat &a,const bigfloat &b)
{
#ifdef DEBUG_BF
    long operation = 1;
    double a_d = a.todouble();
    double b_d = b.todouble();
    @<set double mode@>@;
    double result_d = a_d - b_d;
    @<set default mode@>@;
    bigfloat result_bf = sub(a,b);
    @<check result@>@;
#endif
  return sub(a,b);
}

bigfloat operator * (const bigfloat &a,const bigfloat &b)
{
#ifdef DEBUG_BF
    long operation = 2;
    double a_d = a.todouble();
    double b_d = b.todouble();
    @<set double mode@>@;
    double result_d = a_d * b_d;
    @<set default mode@>@;
    bigfloat result_bf = mul(a,b);
    @<check result@>@;
#endif
  return mul(a,b);
}

bigfloat operator / (const bigfloat &a,const bigfloat &b)
{
#ifdef DEBUG_BF
    long operation = 3;
    double a_d = a.todouble();
    double b_d = b.todouble();
    @<set double mode@>@;
    double result_d = a_d / b_d;
    @<set default mode@>@;
    bigfloat result_bf = div(a,b);
    @<check result@>@;
#endif
  return div(a,b);
}

bigfloat operator - (const bigfloat &a)
{
  if (a.isSpecial())
  {
    if (a.isZero()) return pZero;
    if (a.isInf())  return pInf;
    return NaN;
  }
  return bigfloat(-a.significant,a.exponent);
}

@ To check whether the |bigfloat| operator was correct for the global rounding
mode |TO_NEAREST| and global precision $53$,
we compare the bigfloat result with the |double| result.

@<check result@>=
if ( (a != bigfloat(a_d)) || (b != bigfloat(b_d)) )
{
  cout << "a : \n";
  a.print_contents();
  cout << "Kontrolle : \n";
  bigfloat(a_d).print_contents();
  cout << "b : \n";
  b.print_contents();
  cout << "Kontrolle : \n";
  bigfloat(a_d).print_contents();
  error_handler(1,"error: arithmetic operation: operands not correct");
}
  
if (result_bf != bigfloat(result_d))
{
  cout << "----------------------------------------------------\n";
  cout << "bigfloat result: \n";
  result_bf.print_contents();
  cout << "double result: \n";
  bigfloat(result_d).print_contents();
  switch(operation)
  {
    case 0: result_bf = add(a,b); break;
    case 1: result_bf = sub(a,b); break;
    case 2: result_bf = mul(a,b); break;
    case 3: result_bf = div(a,b);
  }
  error_handler(1,"error: double result != bigfloat result");
}
@

@<set double mode@>=
  char* out;
  switch(round_mode)
  {
    case TO_ZERO: ieee_flags("set","direction","tozero",&out); break;
    case TO_P_INF: ieee_flags("set","direction","positive",&out); break;
    case TO_N_INF: ieee_flags("set","direction","negative",&out); break;
    case TO_NEAREST: case EXACT: case TO_INF: break;
  }

@

@<set default mode@>= 
  ieee_flags("set","direction","nearest",&out);

@ The final |print_contents|-function is a quite simple one that outputs
significant and exponent in binary format for control purposes.

@<general functions@>+=

void bigfloat::print_contents (void) const
{
  integer s=significant,e=exponent;
  cout << "Significant: ";
  if (::sign(significant)==-1) 
    cout << "-";
    

  binout(cout, abs(s));
  cout << "\n";
  cout << "Exponent: ";
  if (e<0)
  {
    cout << "-";
    e=e*integer(-1);
  }
  binout(cout,e);
  cout << "\n";
  cout << "Precision: " << precision << "\n";

  cout << "special flag: ";
  switch(special)
  {
    case NOT:
      cout << "NOT\n";
      break;
    case PZERO:
      cout << "PZERO\n";
      break;
    case NZERO:
      cout << "NZERO\n";
      break;
    case PINF:
      cout << "PINF\n";
      break;
    case NINF:
      cout << "NINF\n";
      break;
    case NAN:
      cout << "NAN\n";
      break;
  }
}



   


@* References.

\begin{thebibliography}{IEE87}
\bibitem[IEE87]{ieee}IEEE standard 754-1985 for binary floating-point
arithmetic, ieee.{\sl reprinted in SIGPLAN 22,2:9-25,1987}
\bibitem[IE\_Go]{gold}David Goldberg. What every computer scientist should know 
about floating-point arithmetic
\bibitem[MN89]{ledaMN}K.Mehlhorn and S.N\"aher.LEDA: A library of 
efficient data types and algorithms.
\bibitem[N\"ah93]{lman}S.N\"aher.LEDA manual.Technical report MPI-I-93-109,
Max-Planck-Institut f\"ur Informatik, 1993
\end{thebibliography}

@

\end{document} 

