\documentclass[a4paper,11pt,twoside,structure=flat]{cweb}

\pagestyle{myheadings}
\markboth{\textsc{Rational Vectors} \textrm{(rat\_vector)}}%
{\textsc{Rational Vectors} \textrm{(rat\_vector)}}

\usepackage{/KM/usr/seel/tex/sty/comments}

\input MANUAL.pagesize
\input MANUAL.mac
\input{/KM/usr/mehlhorn/tex/papermacros}
\input{/KM/usr/seel/litprog/geo/geo_abrev}

@i types.w
@i leda_types.w

@s LEDA_RAT_VECTOR_H ledatypewriter 
@s integer int
@s rational int
@s rat_vector int

\begin{document}
\title{Vectors with Rational Coordinates in d-Space\\
(class \rvec)}
\author{Geokernel}
\date{\today}
\maketitle
\tableofcontents
\newpage


@* The Manual Page of class \rvec.

\input{rat_vector.man}


@* The Header File of class \rvec.
The type \rvec\ is an item class with representation class \grep. It
shares this representation class with hyperplanes, points, and
directions. We derive \rvec\ from handle\_base and derive \grep\ from
handle\_rep. This gives us reference counting for free. We give all
implementations which are trivial directly in the header file and
postpone all others to the next section. Aside from this the header
file is in one-to-one correspondence to the manual page.


@(rat_vector.h@>=

#ifndef LEDA_RAT_VECTOR_H
#define LEDA_RAT_VECTOR_H @#

#include <LEDA/array.h>
#include <LEDA/integer.h>
#include <LEDA/rational.h>
#include <LEDA/integer_vector.h>
#include <LEDA/integer_matrix.h>
#include <LEDA/geo_rep.h> @#

class rat_point;

/*{\Manpage {rat_vector} {} {Rational Vectors} {v}}*/

class rat_vector: public handle_base { @#

/*{\Mdefinition

An instance of data type |rat_vector| is a vector of rational
numbers. A $d$-dimensional vector $r = (r_0,\ldots,r_{d-1})$ is
represented in homogeneous coordinates $(h_0,\ldots,h_d)$, where $r_i
= h_i/h_d$ and the $h_i$'s are of type |integer|. We call the $r_i$'s
the cartesian coordinates of the vector. The homogenizing coordinate
$h_d$ is positive.

This data type is meant for use in computational geometry. It realizes
free vectors as opposed to position vectors (type |rat_point|). The
main difference between position vectors and free vectors is their
behavior under affine transformations, e.g., free vectors are
invariant under translations.

|rat_vector| is an item type.}*/

geo_rep* ptr() const @+ { @+ return (geo_rep*)PTR; @+} 

rat_vector(const handle_base& p)@;: handle_base(p) @+ {}

friend class rat_point;

void print(ostream&) const;
void read(istream&);
 
public: @#

/*{\Mcreation 6}*/

rat_vector(int d = 2) @+ 
{@+ PTR = new geo_rep(d);@+ ptr()->v[d] = 1; @+}

/*{\Mcreate introduces a variable |\Mvar| of type |rat_vector| initialized  
            to the zero vector of dimension $d$.}*/

rat_vector(integer a, integer b, integer D = 1)
{ 
  if (D == 0) 
    error_handler(1,"rat_vector::constructor: D must be nonzero.");
  if (D < 0 ) 
    PTR = new geo_rep(-a,-b,-D);
  else 
    PTR = new geo_rep(a,b,D);
}
/*{\Mcreate introduces a variable  |\Mvar| of type |rat_vector|
            initialized to the two-dimensional vector with homogeneous
            representation $(a,b,D)$ if $D$ is positive and representation
            $(-a,-b,-D)$ if $D$ is negative.\\ 
            \precond $D$ is non-zero.}*/

rat_vector(const integer_vector & c, integer D) 
{
  if (D == 0) 
    error_handler(1,"rat_vector::constructor: D must be nonzero.");
  if (D < 0) 
    PTR = new geo_rep(-c,-D); 
  else 
    PTR = new geo_rep(c,D); 
}
/*{\Mcreate introduces a variable |\Mvar| of type |rat_vector|
            initialized to the vector with homogeneous 
            coordinates $(\pm c_0,\ldots,\pm c_{d-1},\pm D)$, 
            where $d$ is the dimension of $c$ and the sign 
            chosen is the sign of $D$.\\ 
            \precond $D$ is non-zero.}*/

rat_vector(const integer_vector& c)
{ 
  int d = c.dim();
  integer D = c[d-1];
  if (D == 0) 
    error_handler(1,"rat_point::constructor: D must be nonzero");
  if (D < 0) 
    PTR = new geo_rep(-c);
  else 
    PTR = new geo_rep(c);
}
/*{\Mcreate introduces a variable |\Mvar| of type |rat_vector|
            initialized to the direction with homogeneous coordinate  
            vector $\pm c$, where the sign chosen is the sign of the  
            last component of $c$. \\
            \precond The last component of $c$ is non-zero.}*/


rat_vector(const rat_vector& p)@; : handle_base(p) @+ {}
~rat_vector()       @+ {}     

rat_vector& operator=(const rat_vector& p) 
{ @+ handle_base::operator=(p); @+ return *this; @+ }


/*{\Moperations 3 4.5}*/
/*{\Mtext \headerline{3.1 Initialization, Access and Conversions}}*/

static rat_vector d2(integer a, integer b, 
       integer D = 1);
/*{\Mstatic returns a |rat_vector| of dimension $2$ initialized to a 
            vector with homogeneous representation $(a,b,D)$ if $D$ 
            is positive and representation $(-a,-b,-D)$ if $D$ is 
            negative.\\ 
            \precond $D$ is non-zero.}*/

static rat_vector d3(integer a, integer b, integer c, 
                  integer D = 1);
/*{\Mstatic returns a |rat_vector| of dimension $3$ initialized to a 
            vector with homogeneous representation $(a,b,c,D)$ if $D$ 
            is positive and representation $(-a,-b,-c,-D)$ if $D$ is 
            negative.\\ 
            \precond $D$ is non-zero.}*/

static rat_vector unit(int i, int d = 2);
/*{\Mstatic  returns a |rat_vector| of dimension $d$ initialized  
            to the $i$-th unit vector.\\
            \precond $0 \le i < d$.}*/

static rat_vector zero(int d = 2);
/*{\Mstatic  returns the zero vector in d-dimensional space.}*/


int    dim()    const  @+ 
{@+ return ptr()->dim; @+}
/*{\Mop       returns the dimension of |\Mvar|.}*/ 

rational coord(int i) const @+
{@+  return rational(ptr()->v[i], ptr()->v[ptr()->dim]); @+}
/*{\Mop     returns the $i$-th cartesian coordinate of |\Mvar|.}*/

rational operator[](int i) const @+ {@+ return coord(i); @+}
/*{\Marrop  returns the $i$-th cartesian coordinate of |\Mvar|.}*/

integer hcoord(int i) const @+
{@+ return ptr()->v[i]; @+}
/*{\Mop     returns the $i$-th homogeneous coordinate of |\Mvar|.}*/

/*{\Mtext \headerline{Additional Operations for vectors in 
two-dimensional space}}*/

rational xcoord() const@+
{@+ return rational(hcoord(0),hcoord(ptr()->dim)); @+}
/*{\Mop     returns the zeroth cartesian coordinate of |\Mvar|.}*/

rational ycoord() const@+
{@+ return rational(hcoord(1),hcoord(ptr()->dim)); @+}
/*{\Mop     returns the first cartesian coordinate of |\Mvar|.}*/


integer X() const @+ {@+ return hcoord(0); @+}
/*{\Mop     returns the zeroth homogeneous coordinate of |\Mvar|.}*/

integer Y() const @+ {@+ return hcoord(1); @+}
/*{\Mop     returns the first homogeneous coordinate of |\Mvar|.}*/

integer W() const @+ {@+ return hcoord(ptr()->dim); @+}
/*{\Mop     returns the homogenizing coordinate of |\Mvar|.}*/


/*{\Mtext \headerline{3.2 Tests}}*/

friend bool    identical(const rat_vector & v,const rat_vector& w) 
{ @+ return v.ptr() == w.ptr();@+ }
/*{\Mfunc     Test for identity.}*/

int cmp(const rat_vector& x, const rat_vector& y) const
{@+  return ( identical(x,y) ?  0 :  @|
              x.ptr()->cmp_rat_coords(x.ptr(),y.ptr())); @+}
friend int compare(const rat_vector& p, const rat_vector& q)
{@+ return p.cmp(p,q); @+}
  
bool     operator==(const rat_vector& w) const
{@+ return cmp(*this,w) == 0; @+}
/*{\Mbinop     Test for equality.}*/

bool     operator!=(const rat_vector& w)  const
{@+ return cmp(*this,w) != 0; @+}
/*{\Mbinop     Test for inequality.}*/

/*{\Mtext \headerline{3.3 Arithmetical Operators}}*/

rat_vector rat_vector::scale(integer m, integer n) const;
void       rat_vector::self_scale(integer m, integer n);
  
friend rat_vector operator*(int n, const rat_vector& v);

friend rat_vector operator*(integer n, const rat_vector& v);
/*{\Mbinopfunc   multiplies all cartesian coordinates by |n|.   }*/

friend rat_vector operator*(rational r, const rat_vector& v);
/*{\Mbinopfunc    multiplies all cartesian coordinates by |r|.   }*/

rat_vector& operator*=(integer n);
/*{\Mbinop  multiplies all cartesian coordinates by |n|.   }*/
rat_vector& operator*=(int n);

rat_vector& operator*=(rational r);
/*{\Mbinop  multiplies all cartesian coordinates by |r|.   }*/

friend rat_vector operator/(const rat_vector& v, int n);

friend rat_vector operator/(const rat_vector& v, integer n);
/*{\Mbinopfunc    divides all cartesian coordinates by $n$.}*/

friend rat_vector operator/(const rat_vector& v, rational r);
/*{\Mbinopfunc    divides all cartesian coordinates by $n$.}*/

rat_vector& operator/=(integer n);
/*{\Mbinop  divides all cartesian coordinates by |n|.   }*/
rat_vector& operator/=(int n);

rat_vector& operator/=(rational r);
/*{\Mbinop  divides all cartesian coordinates by |r|.   }*/

friend rational operator*(const rat_vector v, const rat_vector& w);
/*{\Mbinopfunc  scalar product, i.e., $\sum_{0 \le i < d} v_i w_i$, 
                where $v_i$ and $w_i$ are the cartesian coordinates 
                of $v$ and $w$ respectively. }*/

friend rat_vector operator+(const rat_vector& v, const rat_vector& w);
/*{\Mbinopfunc   adds cartesian coordinates.   }*/

rat_vector&      operator+=(const rat_vector& w);
/*{\Mbinop       addition plus assignment.   }*/

friend rat_vector operator-(const rat_vector& v, const rat_vector& w);
/*{\Mbinopfunc    subtracts cartesian coordinates   }*/

rat_vector&      operator-=(const rat_vector& w);
/*{\Mbinop       subtraction plus assignment.   }*/

rat_vector       operator-() const;
/*{\Munopfunc    returns -|\Mvar|.}*/


/*{\Mtext \headerline{3.4 Input and Output}}*/

friend ostream& operator<<(ostream& O, const rat_vector& v)
{ v.print(O); return O; }
/*{\Mbinopfunc  writes $v$'s homogeneous coordinates 
                componentwise to the output stream $O$.}*/

friend istream& operator>>(istream& I, rat_vector& v)
{ v.read(I); return I; }
/*{\Mbinopfunc  reads $v$'s homogeneous coordinates
                componentwise from the input stream $I$. 
                The operator uses the current dimension of |v|.}*/

@#
}; @#

inline void Print(const rat_vector& v, ostream& out) @+
{@+ out << v; @+}

inline void Read(rat_vector& v, istream& in) @+        
{@+ in >> v;  @+}


/*{\Mtext \headerline{3.5 Linear Hull, Dependence and Rank}}*/

bool contained_in_linear_hull(const array<rat_vector>& A,
     const rat_vector& x);
/*{\Mfuncl determines whether $x$ is contained in the linear hull
           of the vectors in $A$.}*/

int  linear_rank(const array<rat_vector>& A);
/*{\Mfuncl computes the linear rank of the vectors in $A$.}*/

bool linearly_independent(const array<rat_vector>& A);
/*{\Mfuncl decides whether the vectors in $A$ are linearly independent.}*/

array<rat_vector> linear_base(const array<rat_vector>& A);
/*{\Mfuncl computes a basis of the linear space spanned by the vectors 
           in $A$.}*/


/*{\Mimplementation 
Vectors are implemented by arrays of integers as an item type.  All
operations like creation, initialization, tests, vector arithmetic,
input and output on an vector $v$ take time $O(|v.dim()|)$. |dim()|,
coordinate access and conversions take constant time.  The operations
for linear hull, rank and independence have the cubic costs of the
used matrix operations.  The space requirement is $O(|v.dim()|)$.  }*/

#endif


@* The Implementation of class \rvec.

@(rat_vector.c@>=

#include <LEDA/rat_vector.h>

@<initialization@>@;
@<input and output@>@;
@<arithmetic operations@>@;
@<linear operations@>@;

@*1 Initialization.

We provide some easy initialization operations for the two and
three dimensional case.

@<initialization@>=

rat_vector rat_vector::d2(integer a, integer b, integer D)
{
  if (D == 0)
    error_handler(1,"rat_vector::d2: denominator must not be zero.");
  return rat_vector(a,b,D);
}

rat_vector rat_vector::d3(integer a, integer b, integer c, integer D)
{
  rat_vector d3v(3);
  if (D == 0)
    error_handler(1,"rat_vector::d3: denominator must not be zero.");
  if (D < 0)
    d3v.ptr()->init4(-a,-b,-c,-D);
  else
    d3v.ptr()->init4(a,b,c,D);
  return d3v;
}

rat_vector rat_vector::unit(int i, int d)
{
  if (i<0 || i>=d)
    error_handler(1,"rat_vector::unit: i out of range.");
  rat_vector uv(d);
  uv.ptr()->v[i] = 1; 
  return uv;
}

rat_vector rat_vector::zero(int d) @+
{ @+ return rat_vector(d); @+ }


@*1 Input and Output.
We just take care that in case of input we don't overwrite a
common representation. Apart from that we use the I/O-operations
of \grep.

@<input and output@>=

void rat_vector::print(ostream& out) const { out << ptr(); }

void rat_vector::read(istream& in)
{ int d = dim();
  if (refs() > 1) operator=(rat_vector(d));
  in >> ptr();
  if (hcoord(d) == 0)
    error_handler(1,"operator>>: denominator of vector must be nonzero.");
  if (hcoord(d) < 0) ptr()->negate(d+1);
}


@*1 Multiplication and Division.
We start with the operators |*| and |/|. We define a function
|scale(m,n)| that multiplies the homogenizing coordinate by $n$ and
all other coordinates by $m$ and reduce the operators to this
function.

@<arithmetic...@>=


rat_vector rat_vector::scale(integer m, integer n) const
{ 
  int d = dim();
  rat_vector result(d);
  result.ptr()->v[d] = ptr()->v[d] * n;
  integer g = gcd(ptr()->v[d],m);
  result.ptr()->v[d] /= g;
  m /= g;
  for (int i = 0; i < d; i++) 
    result.ptr()->v[i] = ptr()->v[i] * m; 
  return result;
}

void rat_vector::self_scale(integer m, integer n)
{ 
  int d = dim();
  ptr()->v[d] *= n;   
  integer g = gcd(ptr()->v[d],m);
  ptr()->v[d] /= g;
  m /= g;
  for (int i = 0; i < d; i++) 
    ptr()->v[i] *= m; 
}

rat_vector operator*(int n, const rat_vector& p)
{@+ return p.scale(n,1); @+}

rat_vector operator*(integer n, const rat_vector& p)
{@+ return p.scale(n,1); @+}

rat_vector operator*(rational r, const rat_vector& p)
{@+ return p.scale(r.numerator(),r.denominator()); @+}


rat_vector operator/(const rat_vector& p, int n)
{@+ return p.scale(1,n); @+}

rat_vector operator/(const rat_vector& p, integer n)
{@+ return p.scale(1,n); @+}

rat_vector operator/(const rat_vector& p, rational r)
{@+ return p.scale(r.denominator(),r.numerator()); @+}

rat_vector& rat_vector::operator*=(integer n)
{@+ self_scale(n,1); @+ return *this; @+}

rat_vector& rat_vector::operator*=(int n) 
{@+ self_scale(n,1); @+ return *this; @+}

rat_vector& rat_vector::operator*=(rational r)
{@+ self_scale(r.numerator(),r.denominator()); @+ return *this; @+}

rat_vector& rat_vector::operator/=(integer n)
{@+ self_scale(1,n); @+ return *this; @+}

rat_vector& rat_vector::operator/=(int n) 
{@+ self_scale(1,n); @+ return *this; @+}

rat_vector& rat_vector::operator/=(rational r)
{@+ self_scale(r.denominator(),r.numerator()); @+ return *this; @+}


rational operator*(const rat_vector v, const rat_vector& w)
{
  int d = v.dim();   
  if (d != w.dim()) 
    error_handler(1,"inner product: dimensions disagree.");
  integer nom = 0;
  for (int i = 0; i < d; i++) 
    nom += v.hcoord(i) * w.hcoord(i);
  integer denom = v.hcoord(d) * w.hcoord(d);
  return rational(nom,denom);  
}


@*1 Addition and Subtraction.
To implement addition and substraction of the cartesien coordinates we
use common code in our class \grep. As these operations have the same
functionality which are required for \rpnts.

@<arithmetic...@>=

rat_vector  operator+(const rat_vector& v, const rat_vector& w)
{
  rat_vector res(v.dim());
  c_add(res.ptr(), v.ptr(), w.ptr());
  return res;
}

rat_vector& rat_vector::operator+=(const rat_vector& w)
{ 
  int d = dim();
  rat_vector old(*this);
  if (ptr()->count > 2)
    *this = rat_vector(d);
  c_add(ptr(), old.ptr(), w.ptr());
  return *this;
}

rat_vector  operator-(const rat_vector& v, const rat_vector& w)
{
  rat_vector res(v.dim());
  c_sub(res.ptr(), v.ptr(), w.ptr());
  return res;
}
 
rat_vector& rat_vector::operator-=(const rat_vector& w)
{
  int d = dim();
  rat_vector old(*this);
  if (ptr()->count > 2)
    *this = rat_vector(d);
  c_sub(ptr(), old.ptr(), w.ptr());
  return *this;
}

rat_vector  rat_vector::operator-() const
{
  int d = dim();
  rat_vector result(d);
  result.ptr()->copy(ptr());
  result.ptr()->negate(d);
  return result;
}


@*1 Linear Algebra.

@ Containment in Linear Hull.
A vector |x| is contained in the linear hull of a set |A| of vectors if
|x| is a linear combination of the vectors in |A|, i.e., if the system
$\sum \lambda_i A_i = x$ has a solution.  We may scale each vector by
its homogenizing coordinate and hence forget about the homogenizing
coordinates.

@<linear...@>+=

bool contained_in_linear_hull(const array<rat_vector>& A,
     const rat_vector& x)
{
  int al = A.low();
  int k = A.high()-al+1; // |A| contains |k| vectors
  int d = A[al].dim();
  integer_matrix M(d,k);
  integer_vector b(d);
  for (int i = 0; i < d; i++)
  {
     b[i] = x.hcoord(i);
     for (int j = 0; j < k; j++) 
       M(i,j) = A[al+j].hcoord(i); 
  }
  return solvable(M,b);
}

@ Linear Rank.

We set up a matrix having the cartesian coordinates of the vectors in A
as its columns. The linear rank is the rank of this matrix.  Since the
rank of a matrix does not change under multiplication of a row with a
constant we may also use the homogeneous coordinates $0$ to $d-1$ of
the vectors in A.

@<linear...@>+=

int  linear_rank(const array<rat_vector>& A)
{
  int al = A.low();
  int k = A.high()-al+1; // |A| contains |k| vectors
  int d = A[al].dim();
  integer_matrix M(d,k);
  for (int i = 0; i < d  ; i++)
     for (int j = 0; j < k; j++)  
       M(i,j) = A[al+j].hcoord(i);
  return rank(M);
}


@ Linear Independence.

A set of vectors is linearly independent if their linear rank is
equal to the number of vectors in the set.

@<linear...@>+=

bool linearly_independent(const array<rat_vector>& A)
{@+  return (linear_rank(A) == A.high()-A.low()+1); @+}


@ Linear Base.

@<linear...@>+=

array<rat_vector> linear_base(const array<rat_vector>& A)
{
  int al = A.low();
  int k = A.high()-al+1;      // |A| contains |k| vectors
  int d = A[al].dim();
  integer denom;
  integer_matrix M(d,k);
  for (int j = 0; j < k; j++) 
    for (int i = 0; i < d; i++)
      M(i,j) = A[al+j].hcoord(i); 

  array<int> indcols;
  independent_columns(M,indcols);

  int indcolsdim = indcols.high()+1;
  array<rat_vector> L(indcolsdim);
  for (int i=0; i<indcolsdim; i++)
    L[i] = rat_vector(M.col(indcols[i]),1);
  return L;
}


@* A Test of class \rvec.
We test the construction, access to the components and all the
arithmetical operators in 3-Space.

@(rat_vector-test.c@>=

#include <LEDA/rat_vector.h>

main()
{
  /* construction and access */
  integer_vector vi1(4), vi2(3);
  vi1[0] = 1;@+ vi1[1] = 2;@+ vi1[2] = 3;@+ vi1[3] = 1;
  vi2[0] = 3;@+ vi2[1] = 2;@+ vi2[2] = 1;
  rat_vector a0(3), a1(vi1), a2(vi2,1), a3 = rat_vector::d3(4,-4,0,1), a4(3);
  cout << "four vectors a0,a1,a2,a3:\n" << a0 << a1 << a2 << a3;
  cout << "\nenter fifth vector a4: ";
  cin >> a4;
  cout << "access operations on a4:\n";
  cout << "cartesian: ";
  for (int i=0; i < a4.dim(); i++)     cout << a4.coord(i) << " ";
  cout << "\nhomogenous: ";
  for (int i=0; i <= a4.dim(); i++)    cout << a4.hcoord(i) << " ";

  /* arithmetical operations and compare */
  cout << "\nquadratic length of vector a4 = " << a4*a4;
  cout << "\ncompare(a0,a4) = " << compare(a0,a4);

  vi1[0] = 1;@+ vi1[1] = 1;@+ vi1[2] = 1;@+ vi1[3] = 1;
  rat_vector eins(vi1);

  cout << "\na0 = a1 + a2: " << (a0 = a1+a2);
  cout << "\na0 += eins: " << (a0 += eins);
  cout << "\na0 -= eins: " << (a0 -= eins);
  cout << "\na0 = a0 - a4: " << (a0 = a0 - a4);

  cout << "\na0 = 2 * a0: " << (a0 = 2 * a0);
  cout << "\na0 = 1/2 * a0: " << (a0 = rational(1,2)*a0);
  cout << "\na0 *= 3: " << (a0 *= 3);
  cout << "\na0 *= 1/3: " << (a0 *= rational(1,3));

  cout << "\na0 = a0 / 4: " << (a0 = a0 / 4);
  cout << "\na0 = a0 / 1/4: " << (a0 = a0 / rational(1,4));
  cout << "\na0 /= 5: " << (a0 /= 5);
  cout << "\na0 /= 1/5: " << (a0 /= rational(1,5));

  if ((a1+a2-a4) == a0)
    cout << "\n(a1+a2-a4) == a0\n";
  else
    cout << "\nsomething wrong: (a1+a2-a4) != a0\n";
  cout << "\na0 = -a4: " << (a0 = -a4);

  /* some linear algebra */
  array<rat_vector> P12(2);
  P12[0] = a1;@+ P12[1] = a2;
  cout << "\ncontained_in_linear_hull(<a1,a2>,a4) = ";
  cout << contained_in_linear_hull(P12,a4);
  cout << "\ncontained_in_linear_hull(<a1,a2>,a1+a2) = ";
  cout << contained_in_linear_hull(P12,a1+a2);
  array<rat_vector> B(3),A(4);
  B[0] = a0;@+ B[1] = a1;@+ B[2] = a2; 
  A[0] = a0;@+ A[1] = a1;@+ A[2] = a2;@+ A[3] = a1+a2;
  cout << "\nlinearly_independent(<a0,a1,a2>) = ";
  cout << linearly_independent(B);
  cout << "\nlinearly_independent(<a0,a1,a2,a1+a2>) = ";
  cout << linearly_independent(A);
  cout << "\nlinear_rank(<a0,a1,a2,a1+a2>) = ";
  cout << linear_rank(A);
  cout << "\nlinear_base(<a0,a1,a2,a1+a2>) = ";
  array<rat_vector> res = linear_base(A);
  for (int i=0; i < res.high()+1; i++)
    cout << res[i];
  cout << "\n\n";
}

@
\end{document}
