#include <LEDA/delaunay.h>
#include <LEDA/plane_alg.h>
#include <LEDA/window.h>


static window W("DYNAMIC DELAUNAY AND VORONOI DIAGRAMS OF POINTS");

static char* label[] = {"POINT SET",
                        "DELAUNAY TRIANG",
                        "VORONOI DIAGRAM",
                        "VORONOI + DELAUNAY",
                        "CONVEX HULL",
                        "MINIMUM SPANNING TREE"};

enum { InsertMode, LocateMode, DeleteMode, RangeMode};
enum { Point, Segment, Circle, Polygon };
enum { CircleQuery, TriangleQuery, RectangleQuery };


static points_on_seg    = 15;
static points_on_circle = 15;
static int display      = 1;
static int range        = CircleQuery;

static delaunay_triang T;



void draw_node(const point& p) 
{ //W.draw_filled_node(p,black);
  W.draw_filled_node(p,black); 
 }

void highlight_node(const point& p) 
{ int node_width = W.get_node_width();
  W.set_node_width(node_width+1);
  W.draw_filled_node(p,red); 
  W.set_node_width(node_width);}

void unhighlight_node(const point& p) 
{ W.draw_filled_node(p,black); }

void draw_voro_edge(const point& p, const point& q) 
{ W.draw_segment(p,q,blue); }

void draw_voro_ray(const point& p, const point& q) 
{ W.draw_ray(p,q,blue); }

void draw_triang_edge(const point& p, const point& q) 
{ W.draw_segment(p,q,yellow); }

void draw_hull_edge(const point& p, const point& q) 
{ W.draw_segment(p,q,red); }

void draw_diagram_edge(const point& p, const point& q) 
{ W.draw_segment(p,q,green2); }

void draw_poly(const list<point>& P)
{ W.draw_filled_polygon(P,yellow);
  W.draw_polygon(P,black); 
}


void show_nodes(const list<node>& L)
{ node v;
  forall(v,L)
  { point q = T.pos(v).topoint();
    highlight_node(q);
   }
  int val; double x; double y;
  while (W.read_event(val,x,y) != button_release_event);
/*
  forall(v,L)
  { point q = T.pos(v).topoint();
    unhighlight_node(q);
   }
*/
}


void redraw_nodes(const list<node>& L)
{ node v;
  forall(v,L)
  { point q = T.pos(v).topoint();
    draw_node(q);
    if (display == 1)
    { edge e;
      forall_adj_edges(e,v)
        T.draw_edge(e,draw_diagram_edge,draw_triang_edge,draw_hull_edge);
      }
   }
}


// Input

void random_points(int n)
{ list<rat_point> L;
  int dx = int((W.xmax() - W.xmin())/20);
  int dy = int((W.ymax() - W.ymin())/20);

  int xmin = int(W.xmin()+dx);
  int xmax = int(W.xmax()-dx);
  int ymin = int(W.ymin()+dy);
  int ymax = int(W.ymax()-dy);

  while (n--)
  { integer x = rand_int(xmin,xmax);
    integer y = rand_int(ymin,ymax);
    rat_point p(x,y);
    draw_node(p.topoint());
    L.append(p);
  }
  T.init(L);
}
   


void segment_points(segment s, int n)
{ 
  integer x  = n*s.xcoord1();
  integer y  = n*s.ycoord1();
  integer dx = s.xcoord2() - s.xcoord1();
  integer dy = s.ycoord2() - s.ycoord1();

  for(int i=0; i < n; i++)
  { rat_point p(x,y,n);
    T.insert(p);
    draw_node(p.topoint());
    x += dx;
    y += dy;
   }
}


void circle_points(circle C, int n)
{ 
  point c = C.center();

  int x = int(c.xcoord());
  int y = int(c.ycoord());
  int r = int(C.radius());

  double d = (2*LEDA_PI)/n;
  double eps = 0.001;
  
  double a = 0;
  for(int i=0; i < n; i++)
  { rat_point p = POINT_ON_CIRCLE(a, eps);
    rat_point q(r*p.xcoord(),r*p.ycoord());
    q = q.translate(x,y);
    T.insert(q);
    draw_node(q.topoint());
    a += d;
  }
}



void get_input(window& W, int inp)
{
   switch (inp) {

    case Point: 
         { point p;
           if (W >> p) draw_node(p);
           integer x = p.xcoord();
           integer y = p.ycoord();
           T.insert(rat_point(x,y,1));
           break;
          }
  
   case Segment: 
        { segment s;
          if (W >> s) segment_points(s,points_on_seg); 
          break;
         }

   case Circle:
        { circle c;
          if (W >> c) circle_points(c,points_on_circle);
          break;
         }

   case Polygon:
        { polygon P;
          if (W >> P) 
          { list<segment> L = P.edges();
            segment s;
            forall(s,L) segment_points(s,int(s.length()/4));
           }
          break;
         }
  }
}


void range_search(int range)
{
   switch(range) {

   case CircleQuery:
        { point p;
          W >> p; 
          W << p;
          point q;
  	  W.read_mouse_circle(p,q); 

          POINT pr(p.xcoord(),p.ycoord(),1);
          POINT qr(q.xcoord(),q.ycoord(),1);
          POINT qr90 = qr.rotate90(pr);
          POINT qr180 = qr90.rotate90(pr);
          CIRCLE C(qr,qr90,qr180);
  
          circle c(qr.topoint(),qr90.topoint(),qr180.topoint());
          W.draw_circle(c);
 
          show_nodes(T.range_search(C));
          W.draw_disc(c,white);
          redraw_nodes(T.range_search(C));

          break;
         }


   case TriangleQuery:
        { segment s;  
          W >> s; 
          W << s;
          point p = s.source();
          point q = s.target();
          point r; 
          W.read_mouse_seg(q,r);
          W.draw_segment(p,r); 
          W.draw_segment(q,r);

          POINT pr(p.xcoord(),p.ycoord(),1);
          POINT qr(q.xcoord(),q.ycoord(),1);
          POINT rr(r.xcoord(),r.ycoord(),1);
          list<node> L = T.range_search(pr,qr,rr);
          show_nodes(L);
          W.draw_filled_triangle(p,q,r,white);
          redraw_nodes(L);
          break;
       }

   case RectangleQuery:
        { point p;  
          W >> p; 
          W << p;
          point q; 
          W.read_mouse_rect(p,q);
          W.draw_rectangle(p,q);

          POINT pr(p.xcoord(),p.ycoord(),1);
          POINT qr(q.xcoord(),q.ycoord(),1);

          list<node> L = T.range_search(pr,qr);
          show_nodes(L);
          W.draw_filled_rectangle(p,q,white);
          redraw_nodes(L);
          break;
       }

   }
}


void draw_nodes()
{ T.draw_nodes(draw_node); }

void draw_edges()
{ 
  switch(display) {

  case 0: // nothing
          break;


  case 1: { T.draw_edges(draw_diagram_edge,draw_triang_edge,draw_hull_edge);
            edge ce = T.get_cur_edge();
            if (ce) W.draw_arrow(T.seg(ce).tosegment(), violet);
            break;
           }

  case 2: T.draw_voro_edges(draw_voro_edge,draw_voro_ray);
          break;

  case 3: T.draw_edges(draw_diagram_edge,draw_triang_edge,draw_hull_edge);
          T.draw_voro_edges(draw_voro_edge,draw_voro_ray);
          break;

  case 4: T.draw_hull(draw_poly);
          break;

  case 5: T.draw_edges(T.minimum_spanning_tree(),draw_diagram_edge);  
          break;

  }
}


void draw()
{ W.clear();
  draw_nodes();
  draw_edges();
}




void set_display(int i) 
{ display = i; 
  W.set_frame_label(label[i]);
  draw();
 }

void set_range(int i) { range = i; }


void setgrid(int d)
{ W.set_grid_mode(d);
  draw();
 }


int main()
{
  int input = Point;
  int mode  = InsertMode;

  menu disp_menu;
  disp_menu.button("Set of Points",         0, set_display);
  disp_menu.button("Delaunay Triangulation",1, set_display);
  disp_menu.button("Voronoi Diagram",       2, set_display);
  disp_menu.button("Voronoi and Delaunay",  3, set_display);
  disp_menu.button("Convex Hull",           4, set_display);
  disp_menu.button("Minimum Spanning Tree", 5, set_display);


  menu grid_menu;
  grid_menu.button("no grid", 0,setgrid);
  grid_menu.button("dist  5", 5,setgrid);
  grid_menu.button("dist 10",10,setgrid);
  grid_menu.button("dist 15",15,setgrid);
  grid_menu.button("dist 20",20,setgrid);
  grid_menu.button("dist 25",25,setgrid);

  menu rand_menu;
  rand_menu.button(" 500", 500,random_points);
  rand_menu.button("1000",1000,random_points);
  rand_menu.button("1500",1500,random_points);
  rand_menu.button("2000",2000,random_points);
  rand_menu.button("2500",2500,random_points);
  rand_menu.button("3000",3000,random_points);

  menu range_menu;
  range_menu.button("circle",   CircleQuery,   set_range);
  range_menu.button("triangle", TriangleQuery, set_range);
  range_menu.button("rectangle",RectangleQuery,set_range);
 
 
  W.choice_item("mode",  mode, "insert", "locate", "delete", "range query");
  W.choice_item("input", input,"point", "segment", "circle", "polygon");
  //W.choice_item("range", range,"circle", "triangle", "rectangle");

  W.button("display",17,disp_menu);
  W.button("range",  18,range_menu);
  W.button("grid",   19,grid_menu);
  W.button("random",  1,rand_menu);
  W.button("clear",   2);
  W.button("quit",    0);

  W.display();

  W.init(0,512,0);
  W.set_redraw(draw);
  W.set_node_width(2);
  W.set_frame_label(label[display]);


  for(;;)
  {
    double x,y;
    int but = W.read_mouse(x,y);
  
    if (but == 0) break;
  
    switch (but) {

    case MOUSE_BUTTON(1):
            { 
              switch (mode) {

              case InsertMode: put_back_event();
                               get_input(W,input);
                               draw();
                               break;
              case LocateMode:
              case DeleteMode:
                            { rat_point p = rat_point(integer(x),integer(y));
                              node v = T.nearest_neighbor(p);
                              if (v)
                              { show_nodes(v);
                                redraw_nodes(v);
                                if (mode == DeleteMode) 
                                { T.del(v);
                                  draw();
                                 }
                               }
                              break;
                             }

              case RangeMode: put_back_event();
                              range_search(range);
                              break;
              }
              break;
             }

    case 1: { // random points
              draw_edges();
              break;
             }
   

    case 2: { // clear
              T.clear();
              W.clear();
              break;
             }

    }

  }

  return 0;
}
