/*******************************************************************************
+
+  LEDA 3.4
+
+  graph_objects.h
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#ifndef LEDA_GRAPH_OBJECTS_H
#define LEDA_GRAPH_OBJECTS_H

#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 340046
#include <LEDA/REDEFINE_NAMES.h>
#endif


//------------------------------------------------------------------------------
// basic objects for graphs (nodes and edges0
//
// S. Naeher (1995)
//------------------------------------------------------------------------------


#include <LEDA/impl/olist.h>

class node_struct;
typedef node_struct* node;

class edge_struct;
typedef edge_struct* edge;


class face_struct;
typedef face_struct* face;



//------------------------------------------------------------------------------
// class node_struct: internal representation of nodes
//------------------------------------------------------------------------------

class node_struct : public obj_link  // used by node_list
{  
   friend class graph;
   friend class ugraph;
   friend class planar_map;
   friend class node_list;
   //friend template<int n> class b_node_pq;
   //friend template<class I> class node_pq;

   
   graph*    owner;              // pointer to graph of node 
   int       name;               // internal name (index)  
   edge      first_adj_edge[2];  // head of adj/in list
   edge      last_adj_edge[2];   // tail of adj/in list
   int       adj_length[2];       // outdeg/indeg
   edge      adj_iterator;       // for backward compatibility (old iteration)
   list_item pos_in_V;           // position of node in list of all nodes V

public:

   GenPtr data[3];      // data[0]: GRAPH
                        // data[1]: node_pq
                        // data[2]: node_partition

void append_adj_edge(edge e,int i, int chain_e);
void insert_adj_edge(edge e, edge e1, int i, int chain_e, int dir);
void del_adj_edge(edge e, int i, int chain_e);

node_struct(GenPtr i=0);

LEDA_MEMORY(node_struct)

friend inline graph* graph_of(node);
friend inline graph* graph_of(edge);
friend inline int    indeg(node);
friend inline int    outdeg(node);
friend inline int    degree(node);
friend inline int    index(node);

friend inline edge   First_Adj_Edge(node,int);
friend inline edge   Last_Adj_Edge(node,int);


friend void init_node_data(const graph&,int,GenPtr);

};



//------------------------------------------------------------------------------
// class edge_struct: internal representation of edges
//------------------------------------------------------------------------------

class edge_struct
{ 
   friend class node_struct;
   friend class graph;
   friend class ugraph;
   friend class planar_map;

   edge succ_adj_edge[2];   // chaining of adjacent out/in-edges
   edge pred_adj_edge[2];   // chaining of adjacent out/in-edges

   int  name;          // internal name (index)  
   node term[2];       // term[0] = source and term[1] = target
   edge rev;           // space for reverse edge (used by planar_map)
   face fa;            // associated face (used by planar_map);

public:

   GenPtr data[1];     // space for data (data[0] used by GRAPH)

   list_item pos_in_E;


edge_struct(node v=0, node w=0, GenPtr i=0);


public:

LEDA_MEMORY(edge_struct)

friend inline graph* graph_of(edge);
friend inline node   terminal(edge,int);
friend inline node   source(edge);
friend inline node   target(edge);
friend inline face   face_of(edge);
friend inline node   opposite(node,edge);
friend inline int    index(edge);
friend inline edge   Succ_Adj_Edge(edge,int);
friend inline edge   Succ_Adj_Edge(edge,node);
friend inline edge   Pred_Adj_Edge(edge,int);
friend inline edge   Pred_Adj_Edge(edge,node);

};


//------------------------------------------------------------------------------
// class face_struct: internal representation of faces
//------------------------------------------------------------------------------

class face_struct {

friend class graph;
friend class planar_map;

 int       name;     // internal name (index)  
 edge      head;     // first edge of face
 int       sz;       // size of face (# edges)
 list_item pos_in_F; // location in F_list
 GenPtr    data[1];  // space for data (data[0] used by PLANAR_MAP)

 graph*    owner;

 face_struct(GenPtr);


  LEDA_MEMORY(face_struct)

  friend index(face f)                { return f->name;  }
  friend graph*      graph_of(face f) { return f->owner; }
  friend planar_map* map_of(face f)   { return (planar_map*)f->owner; }

};



inline int    outdeg(node v) { return v->adj_length[0]; }
inline int    indeg(node v)  { return v->adj_length[1];  }
inline int    degree(node v) { return indeg(v) + outdeg(v); }

inline int    index(node v)    { return v->name;  }
inline graph* graph_of(node v) { return v->owner; }


inline node   terminal(edge e, int i) { return e->term[i]; }
inline node   source(edge e)   { return e->term[0]; }
inline node   target(edge e)   { return e->term[1]; }
inline face   face_of(edge e)  { return e->fa; }

inline node   opposite(node v, edge e) 
{ return (v==source(e)) ? target(e) : source(e); }

inline int    index(edge e)    { return e->name; }
inline graph* graph_of(edge e) { return source(e)->owner; }


// parameterized access of adjacent edges
// outgoing (i=0) or incoming (i=1) edges

inline edge First_Adj_Edge(node v, int i) { return v->first_adj_edge[i]; }
inline edge Last_Adj_Edge(node v, int i)  { return v->last_adj_edge[i]; }
inline edge Succ_Adj_Edge(edge e, int i) { return e->succ_adj_edge[i]; }
inline edge Pred_Adj_Edge(edge e, int i) { return e->pred_adj_edge[i]; }


// successor or predecessor of e in adjacency list of node v
// precond: v = source(e) or v = target(e)

inline edge Succ_Adj_Edge(edge e, node v) 
{ return Succ_Adj_Edge(e,(v==source(e)) ? 0:1); }

inline edge Pred_Adj_Edge(edge e, node v) 
{ return Pred_Adj_Edge(e,(v==source(e)) ? 0:1); }


inline int compare(const node& v, const node& w) 
{ return  int(long(v) - long(w)); }

inline int compare(const edge& x, const edge& y) 
{ return  int(long(x) - long(y)); }

inline int compare(const face& x, const face& y) 
{ return  int(long(x) - long(y)); }


#if LEDA_ROOT_INCL_ID == 340046
#undef LEDA_ROOT_INCL_ID
#include <LEDA/UNDEFINE_NAMES.h>
#endif

#endif
