/*******************************************************************************
+
+  LEDA 3.4
+
+  graphwin.h
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#ifndef LEDA_GRAPHWIN_H
#define LEDA_GRAPHWIN_H

#include <LEDA/window.h>
#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/map.h>
#include <LEDA/queue.h>
#include <LEDA/menu.h>

/****************************************************************************/
/*                                                                          */
/****************************************************************************/

class GraphWin;

/****************************************************************************/
/*             Internal handling of menu-function calls                     */
/****************************************************************************/

template<class function_t> void gw_call(GraphWin& gw, function_t f);

class base_function { public : virtual void call(GraphWin&)=0; };

template<class function_t>
class function : public base_function {
  function_t f;
 public :
  function(function_t g) : f(g) {}
  void call(GraphWin& gw) { gw_call(gw,f); } 
};

/****************************************************************************/
/*                           Constants                                      */
/****************************************************************************/

enum node_st { simple_node, filled_node }; 
typedef line_style edge_st;

enum node_t       { circle_node , ellipse_node , rectangle_node };
enum node_embed_t { random_embed , grid_embed , circle_embed };
enum generation_t { g_complete, ug_complete, g_random, bi_complete, bi_random,
                    pl_random, tri_planar, g_grid };
enum node_move_t  { move_single_node , move_all_nodes };
enum position_t   { central, left_above, above, right_above, 
                    right, right_below, below, left_below, left };

/****************************************************************************/
/*          Types for encoding/decoding Node_matrix & Menu-calls            */
/****************************************************************************/

typedef short ID_t;		// for identifying nodes & GraphWins

typedef long ecode_t;		// edge_code_type (!! double size of ID_t !!)
				// LO : source(e), HI : target(e)

typedef long mcode_t;		// menu_code_type (!! double size of ID_t !!)
				// LO : menu-item-index, HI : gw_id

/****************************************************************************/
/*               node_info & edge_info - structs                            */
/****************************************************************************/

struct node_info {
  ID_t id;
  color clr;
  node_t type;
  node_st style;
  point pos;
  double r1,r2;
  bool show_label;
  string label;
  color label_clr;
  position_t label_pos;  
  bool marked;
};

struct edge_info {
  color clr;
  edge_st style;
  int width;
  bool show_label;
  string label;
  color label_clr;
  position_t label_pos;
  list<point> p;
  bool auto_edge;
  bool marked;  
};

/****************************************************************************/
/*                         class  GraphWin                                  */
/****************************************************************************/

/*{\Manpage {GraphWin} {} {Graph-Windows}}*/

class GraphWin {

/*{\Mdefinition

An instance of ...

\medskip
With every node is assigned a \ \ldots
\begin{enumerate}
\item position ($point$)
\item color ($color$)
\item type ($node\_t$)
\item style ($node\_st$)
\item first radius ($double$)
\item second radius ($double$)
\item label ($string$)
\item label position ($position\_t$)
\item label color ($color$)
\item show\_label ($bool$)
\end{enumerate}

\medskip
With every edge is assigned a \ \ldots
\begin{enumerate}
\item list of points ($list<point>$)
\item color ($color$)
\item style ($edge\_st$)
\item width ($int$) 
\item label ($string$)
\item label position ($position\_t$)
\item label color ($color$)
\item show\_label ($bool$)

\end{enumerate}

The new types are:

$node\_t = \{\ circle\_node,\ ellipse\_node,\ rectangle\_node\ \}$ \\
$node\_st = \{\ simple\_node,\ filled\_node\ \}$ \\
$position\_t = \{\ central, left\_above, above, right\_above, right, right\_below, below, left\_below, left\ \}$ \\
$edge\_st = line\_style$

}*/

  private :
  
    graph *gr_p;			// graph
    window *win_p;			// window
    
    static map<window*,GraphWin*> win_vec; // find a gw-object from its window
 
    node_map<node_info> n_info;
    edge_map<edge_info> e_info;
    
    bool nodes_embedded;		// nodes already assigned to coordinates
    bool edges_embedded;		// edges typified?
    
    double x_min,x_max,y_min,y_max;	// window-size
    
    node_move_t node_move;
    node_embed_t node_embed;

    node_info n_model;
    edge_info e_model;

    color hi_node_color;		// highlighting color
    color hi_edge_color;
    
    double edge_dist;
    
    double units_per_pixel;    
    
    long create_flag;			// notice dynamic created memory objects

    string filename;
    string f_ext;
    string ps_ext;
    string gw_ext;

    void init_once();
    void create(long c_flag);
    void init_default();
    void init_node(node v);
    void init_edge(edge e); 
    void init_graph();
    void init_matrix();
    void init_window();
    void init_menu();
    void init_call_menu();
    
    bool directed;			// draw graph directed?

/****************************************************************************/
/*                          DRAWING                                         */
/****************************************************************************/

    void draw_node(node v);
    void draw_edge(edge e);
    void draw_node_with_edges(node v);
    void draw_nodes();
    void draw_edges();
    void draw_graph();

/****************************************************************************/
/*                                                                          */
/****************************************************************************/

    void update_win_bounds();

    static void s_redraw(window*);

    bool done;			// continue-button pressed? [edit finished, ..]
    bool done_cause;		// continue or quit-button pressed?
    
    node find_node(const point& p, node from=nil);
    edge find_edge(const point& p, edge from=nil);

    void save_embed(node,unsigned);
    void restore_embed(node,unsigned);
    void save_and_restore_embed(node,unsigned,unsigned);

    unsigned animation;

    void move_node(node v, vector trans, unsigned anim=0);
    void move_node(node v, point dest, unsigned anim=0);
    void move_nodes(list<node>& L, vector trans);
    void move_edges(vector trans);
    void move_edges(list<edge>& L, vector trans);
    void move_nodes(vector trans, unsigned anim=0);    
    void move_nodes(node_array<point>& dest, unsigned anim=0);
    void move_component(node v, double x, double y);
    void scroll_graph(double x, double y);

    bool flush;			// make changes on *gr_p visible
    bool window_is_mapped;

    bool constant_width;

    int buttons_per_line;

    int grid;
    bool show_crd;

    double zoom_factor;

    int gen_nodes,gen_edges;

    void embed_nodes(node_embed_t);
    void embed_edges();
    void embed_node_with_edges(node v);

    double max_edge_distance(node v, node w);
    point compute_leaving_point(node v, const point& p);
    void compute_leaving_points(node v, node w, point& p_v, point& p_w);
    point first_hit(node v, const line& l, bool from_left, bool from_above);

    void embed_edges(node v, node w);

/****************************************************************************/
/*          internal representation of the dynamic node_matrix              */
/****************************************************************************/

    ID_t max_node_id;

    map<ecode_t,list<edge>*> n_matrix;       // dynamic node-matrix

    static queue<ID_t> id_queue;	// re-use id's from deleted nodes
    
    ID_t get_id(node v) const { return n_info[v].id; }

    ecode_t ecode(node v, node w) const { 
      union { struct { ID_t lo,hi; } ii; ecode_t i; } c;
      c.ii.lo=get_id(v);
      c.ii.hi=get_id(w);
      return c.i;
    }
    
    ecode_t ecode(edge e) const {
      return ecode(source(e),target(e));
    }
    
    ecode_t ecode_rev(edge e) const {
      return ecode(target(e),source(e));
    }
    
    int number_of_incident_edges(node v, node w) {
      list<edge>* &Lp=n_matrix[ecode(v,w)];
      return Lp?Lp->size():0;
    }

    void new_incident_edge(edge e) {
      list<edge>* &Lp=n_matrix[ecode(e)];
      if (!Lp) {
        Lp=new list<edge>;
        n_matrix[ecode_rev(e)]=new list<edge>;
      }
      Lp->push(e);   
    }

    void del_incident_edge(edge e) {
      list<edge>* &Lp=n_matrix[ecode(e)];
      Lp->del_item(Lp->search(e)); 
    }   

    void del_incident_edges(node v, node w) {
      list<edge>* &Lp=n_matrix[ecode(v,w)];
      if (Lp) Lp->clear();
    }   

    list<edge>& GraphWin::incidence_list(node v, node w) {
      list<edge>* &Lp=n_matrix[ecode(v,w)];
      if (!Lp) {
        Lp=new list<edge>;
        n_matrix[ecode(w,v)]=new list<edge>;
      }
      return *Lp;	
    }

    #define forall_incident_edges(e,v,w) forall(e,*n_matrix[ecode(v,w)])
    //#define forall_incident_edges(e,v,w) forall(e,incidence_list(v,w))

/****************************************************************************/
/*       Handlers - called at a specified action                            */
/****************************************************************************/

    void (*new_node_handler)(GraphWin& gw, node v);
    void (*new_edge_handler)(GraphWin& gw, edge e);
    void (*move_node_handler)(GraphWin& gw, node v);
    void (*delete_node_handler)(GraphWin& gw, node v);
    void (*delete_edge_handler)(GraphWin& gw, edge e);

/****************************************************************************/
/*       MENUS - Representation and default-menus                           */
/****************************************************************************/

    ID_t gw_id;				// GraphWin - ID;
    static ID_t max_gw_id;
    static map<ID_t,GraphWin*> id_vec; // find a gw-object from its id 

    int call_entry_counter;
    base_function **menu_functions;
    
    int sub_menu_counter;
    menu **sub_menus;
    
    void continue_diag(int=0);
    void redraw_diag(int=0);
    void zoom_handler(int);
    void node_embed_handler(int);
    void call_handler(int n) { menu_functions[n]->call(*this); }
    void reset_handler(int);
    void settings_diag(int=0);
    void node_settings_diag(node);
    void edge_settings_diag(edge);
    void file_handler(int);
    void generation_handler(int n);

    typedef union { struct { ID_t lo,hi; } ii; mcode_t i; } m_union;

    mcode_t mcode(ID_t n) { m_union c; c.ii.lo=n; c.ii.hi=gw_id; return c.i; }

    friend ID_t hi_mcode(mcode_t n) { m_union c; c.i=n; return c.ii.hi; }
    friend ID_t lo_mcode(mcode_t n) { m_union c; c.i=n; return c.ii.lo; }

    friend void gw_continue_diag(int n) {
      GraphWin::id_vec[hi_mcode(n)]->continue_diag(lo_mcode(n));
    }
    friend void gw_redraw_diag(int n) {
      GraphWin::id_vec[hi_mcode(n)]->redraw_diag(lo_mcode(n));
    }
    friend void gw_zoom_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->zoom_handler(lo_mcode(n));
    }
    friend void gw_node_embed_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->node_embed_handler(lo_mcode(n));
    }
    friend void gw_call_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->call_handler(lo_mcode(n));
    }
    friend void gw_reset_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->reset_handler(lo_mcode(n));
    }
    friend void gw_settings_diag(int n) {
      GraphWin::id_vec[hi_mcode(n)]->settings_diag(lo_mcode(n));
    }
    friend void gw_file_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->file_handler(lo_mcode(n));
    }
    friend void gw_generation_handler(int n) {
      GraphWin::id_vec[hi_mcode(n)]->generation_handler(lo_mcode(n));
    }

/****************************************************************************/
/*                          PUBLIC - part                                   */
/****************************************************************************/

  public :

/*{\Mcreation GW }*/

    GraphWin(string win_label="LEDA - GraphWin"); 
    
/*{\Mcreate creates a graph-window with a new empty graph}*/
    
    GraphWin(graph& G, string win_label="LEDA - GraphWin"); 

/*{\Mcreate creates a graph-window using $G$}*/

    ~GraphWin();
    
/*{\Mtext
\bigskip
\headerline { Window operations} }*/

    void display();

/*{\Mop opens a window ... }*/

    bool edit();
    
/*{\Mop Change graph interactive until the Continue-button is pressed. }*/
    
    void open(); 

/*{\Mop $GW$.display() + $GW$.edit() }*/

    void close();

/*{\Mop closes the window ... }*/


/*{\Mtext
\bigskip       
\headerline { Update operations} }*/

    node new_node(point p);

/*{\Mopl        adds a new node at position $p$. Color, type, style and radii
                are taken from the values setted with functions below. }*/

    void del_node(node v);

/*{\Mop        deletes $v$ and all edges incident to $v$ from |\Mvar|. }*/

    edge new_edge(node v, node w);

/*{\Mopl       adds an edge $(v,w)$. Color and style are taken from values
               setted with functions below. }*/
                
    void del_edge(edge e);

/*{\Mop       deletes edge $e$ from |\Mvar|. }*/


/*{\Mtext
\bigskip
\headerline { Setting parameters } }*/
 

/*{\Mtext
\headerline { a) for new nodes and edges} }*/

    color get_node_color() const { return n_model.clr; }

    color set_node_color(color c);
    
/*{\Mopl       sets the color for a new node to $c$. }*/

    double get_node_radius() const { return n_model.r1; }

    double set_node_radius(double r);		

/*{\Mopl       sets the first radius for new nodes. }*/

    double get_node_radius2() const { return n_model.r2; }		

    double set_node_radius2(double r);		

/*{\Mopl       sets the second radius for new nodes. }*/

    double get_node_radius1() const { return get_node_radius(); }		

    double set_node_radius1(double r) { return set_node_radius(r); }		

/*{\Mopl       set\_node\_radius(r); }*/

    color get_edge_color() const { return e_model.clr; }

    color set_edge_color(color c);

/*{\Mopl       sets the color for a new edge to $c$. }*/

    node_t get_node_type() const { return n_model.type; }

    node_t set_node_type(node_t new_type);

/*{\Mopl       sets the style for a new node to $st$. }*/

    node_st get_node_style() const { return n_model.style; }

    node_st set_node_style(node_st st);

/*{\Mopl       sets the style for a new node to $st$. }*/

    edge_st get_edge_style() const { return e_model.style; }

    edge_st set_edge_style(edge_st st);

/*{\Mopl       sets the style for a new edge to $st$. }*/

    unsigned get_node_width() const { 
      return (unsigned) (get_node_radius()*2*win_p->scale());
    }		

    unsigned set_node_width(unsigned w) {
      return (unsigned) (set_node_radius(0.5*w/win_p->scale())*win_p->scale()*2);		
    }

/*{\Mopl       sets the width for a new node to $w$ pixels. }*/

    unsigned get_node_height() const { 
      return (unsigned) (get_node_radius2()*2*win_p->scale());
    } 

    unsigned set_node_height(unsigned w) {
      return (unsigned) (set_node_radius2(0.5*w/win_p->scale())*win_p->scale()*2);
    }

/*{\Mopl       sets the height for a new node to $w$. }*/

    int get_edge_width() const { return e_model.width; }		

    int set_edge_width(int w);		

/*{\Mopl       sets the width for a new edge to $w$. }*/

    position_t get_node_label_pos() const { return n_model.label_pos; }

    position_t set_node_label_pos(position_t lp);
    
/*{\Mopl       sets the label position of a new node to $lp$. }*/

    position_t get_edge_label_pos() const { return e_model.label_pos; }

    position_t set_edge_label_pos(position_t lp);
    
/*{\Mopl       sets the label position of a new edge to $lp$. }*/

    bool get_show_node_label() const { return n_model.show_label; }

    bool set_show_node_label(bool show);
    
/*{\Mopl       sets $show\_label$ of new nodes to $show$. }*/      

    bool get_show_edge_label() const { return e_model.show_label; }

    bool set_show_edge_label(bool show);

/*{\Mopl       sets $show\_label$ of new edges to $show$. }*/      

    bool show_node_label() { return set_show_node_label(true); }

/*{\Mopl       returns set_show_node_label(true). }*/      
    
    bool hide_node_label() { return set_show_node_label(false); }
    
/*{\Mopl       returns return set_show_node_label(false). }*/      
    
    bool show_edge_label() { return set_show_edge_label(true); }
    
/*{\Mopl       returns set_show_edge_label(true). }*/      
    
    bool hide_edge_label() { return set_show_edge_label(false); }

/*{\Mopl       returns set_show_edge_label(false). }*/      
    
/*{\Mtext
\medskip
\headerline { b) for existing nodes and edges} }*/
    
    ID_t get_index(node v) const { return n_info[v].id; }
        

    color get_color(node v) const { return n_info[v].clr; }

    void set_color(const list <node>& L, color c);

    color set_color(node v, color c);

/*{\Mopl       sets the color of $v$ to $c$. }*/

    color get_color(edge e) const { return e_info[e].clr; }

    void set_color(const list <edge>& L, color c);

    color set_color(edge e, color c);

/*{\Mopl       sets the color of $e$ to $c$. }*/

    node_t get_type(node v) const { return n_info[v].type; }

    void set_type(const list <node>& L, node_t new_type);

    node_t set_type(node v, node_t t);

/*{\Mopl       sets the type of $v$ to $t$. }*/

    node_st get_style(node v) const { return n_info[v].style; }

    void set_style(const list <node>& L, node_st st);

    node_st set_style(node v, node_st st);

/*{\Mopl       sets the style of $v$ to $st$. }*/

    edge_st get_style(edge e) const { return e_info[e].style; }

    void set_style(const list <edge>& L, edge_st st);

    edge_st set_style(edge e, edge_st st);
    
/*{\Mopl       sets the style of $e$ to $st$. }*/

    double get_radius(node v) const { return n_info[v].r1; }

    void set_radius(const list <node>& L, double r);

    double set_radius(node v, double r);
    
    double get_radius2(node v) const { return n_info[v].r2; }

    void set_radius2(const list <node>& L, double r);

    double set_radius2(node v, double r);
    
    double get_radius1(node v) const { return get_radius(v); }

    double set_radius1(node v, double r) { return set_radius(v,r); }

    int get_width(node v) const { 
      return int(get_radius(v)*2*win_p->scale());
    }

    void set_width(const list <node>& L, int w);

    int set_width(node v, int w) {
      return int(set_radius(v,0.5*w/win_p->scale())*2*win_p->scale());
    }

/*{\Mopl       sets the width of $v$ to $w$ pixels. }*/

    int get_height(node v) const {
      return int(get_radius2(v)*2*win_p->scale());
    }

    void set_height(const list <node>& L, int h);

    int set_height(node v, int h) {
      return int(set_radius2(v,0.5*h/win_p->scale())*2*win_p->scale());
    }

/*{\Mopl       sets the height of $v$ to $w$ pixels. }*/

    int get_width(edge e) const { return e_info[e].width; }
    
    void set_width(const list<edge>& L, int w);

    int set_width(edge e, int w);

/*{\Mopl       sets the width of $e$ to $w$. }*/

    point get_position(node v) const { return n_info[v].pos; }
    
    void set_position(const list<node>& L, point p);

    point set_position(node v, point p);

/*{\Mopl       sets the position of $v$ to $p$. }*/

    string get_label(edge e) const { return e_info[e].label; }

    void set_label(const list<edge>& L, string s);

    string set_label(edge e, string s);
    
    int set_label(edge e, int n) { return atoi(set_label(e,string("%d",n))); }

    string get_label(node v) const { return n_info[v].label; }

    void set_label(const list<node>& L, string s);

    string set_label(node v, string s);
    
/*{\Mopl       sets the label of $v$ to $s$. }*/

    int set_label(node v, int n) { return atoi(set_label(v,string("%d",n))); }

/*{\Mopl       sets the label of $v$ to $n$. }*/

    position_t get_label_pos(node v) const { return n_info[v].label_pos; }

    void set_label_pos(const list<node>& L, position_t lp);

    position_t set_label_pos(node v, position_t lp);
    
/*{\Mopl       sets the label position of $v$ to $lp$. }*/

    position_t get_label_pos(edge e) const { return e_info[e].label_pos; }

    void set_label_pos(const list<edge>& L, position_t lp);

    position_t set_label_pos(edge e, position_t lp);
    
/*{\Mopl       sets the label position of $e$ to $lp$. }*/

    list<point> get_polygon(edge e) const { return e_info[e].p; }

    bool get_show_label(node v) const { return n_info[v].show_label; }

    void set_show_label(const list<node>& L, bool show);

    bool set_show_label(node v, bool show);

/*{\Mopl       sets $show\_label$ of $v$ to $show$. }*/

    bool show_label(node v) { return set_show_label(v,true); }

/*{\Mopl       returns set\_show\_label(v,true). }*/

    bool hide_label(node v) { return set_show_label(v,false); }

/*{\Mopl       returns set\_show\_label(v,false). }*/

    bool get_show_label(edge e) const { return e_info[e].show_label; }

    void set_show_label(const list<edge>& L, bool show);

    bool set_show_label(edge e, bool show);
    
/*{\Mopl       sets $show\_label$ of $e$ to $show$. }*/

    bool show_label(edge e) { return set_show_label(e,true); }

/*{\Mopl       returns set\_show\_label(e,true). }*/

    bool hide_label(edge e) { return set_show_label(e,false); }

/*{\Mopl       returns set\_show\_label(e,false). }*/

    color get_label_color(node v) const { return n_info[v].label_clr; }
    
    void set_label_color(const list<node>& L, color c);
    
    color set_label_color(node v, color c);

/*{\Mopl       sets the label color of $v$ to $c$. }*/
    
    color get_label_color(edge e) const { return e_info[e].label_clr; }
    
    void set_label_color(const list<edge>& L, color c);
    
    color set_label_color(edge e, color c);
    
/*{\Mopl       sets the label color of $e$ to $c$. }*/
 
    int get_gen_nodes() const { return gen_nodes; }

    int set_gen_nodes(int n);
    
    int get_gen_edges() const { return gen_edges; }

    int set_gen_edges(int m);

/*{\Mtext
\medskip
\headerline { c) global} }*/

    double get_edge_dist() const { return edge_dist; }

    double set_edge_dist(double d);
    
/*{\Mopl       sets the maximum distance between multiple edges. }*/

    int get_grid_mode() const { return grid; }

    int set_grid_mode(int b);

/*{\Mopl       sets the window-grid ($b>0$) or removes it ($b=0$). }*/

    bool get_show_coordinates() const { return show_crd; }

    bool set_show_coordinates(bool b);

/*{\Mopl       show mouse-position ($b=true$) or not ($b=false$). }*/

    unsigned get_animation() const { return animation; }

    unsigned set_animation(unsigned s);

/*{\Mopl       move a node in $s$ steps to its new position. }*/

    bool get_flush() const { return flush & window_is_mapped; }

    bool set_flush(bool b);

/*{\Mopl       show operations on |\Mvar| instantly ($b=true$) 
               or not ($b=false$). }*/

    double get_zoom_factor() const { return zoom_factor; }

    double set_zoom_factor(double f);
    
/*{\Mopl       sets the zoom factor to f used when zooming from menu. }*/


    double get_x_min() const { return x_min; }

    double get_y_min() const { return y_min; }

    double get_x_max() const { return x_max; }

    double get_y_max() const { return y_max; }
               
    bool get_constant_width() const { return constant_width; }

    bool set_constant_width(bool b);           
               
/*{\Mtext
\bigskip
\headerline { Handler} }*/

    void set_new_node_handler(void (*f)(GraphWin& gw, node v)=NULL) {
      new_node_handler=f;
    }

/*{\Mopl	$f$ is called every time the user creates a new node.
		The new node is assigned to $v$. }*/

    void set_new_edge_handler(void (*f)(GraphWin& gw, edge e)=NULL) {
      new_edge_handler=f;
    }
   
/*{\Mopl	$f$ is called every time the user creates a new edge. 
		The new edge is assigned to $e$.}*/
    
    void set_move_node_handler(void (*f)(GraphWin& gw, node v)=NULL) {
      move_node_handler=f;
    }

/*{\Mopl	$f$ is called every time the user changes the position
		of a node. The moved node is assigned to $v$. }*/

    void set_delete_node_handler(void (*f)(GraphWin& gw, node v)=NULL) {
      delete_node_handler=f;
    }

/*{\Mopl        $f$ is called every time the user wants to delete a node.
		$f$ will be executed before the final removal of $v$. }*/

    void set_delete_edge_handler(void (*f)(GraphWin& gw, edge e)=NULL) {
      delete_edge_handler=f;
    }

/*{\Mopl        $f$ is called every time the user wants to delete an edge.
		$f$ will be executed before the final removal of $e$. }*/

/*{\Mtext
\bigskip
\headerline { Menus}

You can pass the graph created/edited in |\Mvar| through your own
functions per mouse-click if these functions are added to a menu.

To add your function $f$ of type {\it function\_t}, use (non-member-function) 

{\it void} gw\_add\_call({\it GraphWin}\& |\Mvar|, {\it function\_t} $f$, {\it string} $label$, {\it int} $sub\_menu=0$).

\medskip
To add a submenu, use (non-member-function)

{\it int} gw\_add\_menu({\it GraphWin}\& |\Mvar|, {\it string} $label$, {\it int} $sub\_menu=0$);

which returns an integer corresponding to a submenu you can use as parameter 
sub\_menu in gw\_add\_call and gw\_add\_menu. 

Since GraphWin don't know, how to handle with the arguments and the 
returning type of function $f$, you have to define 

{\it void} gw\_call({\it GraphWin}\& |\Mvar|, {\it function\_t} $f$);

for all your above used functions $f$. 
In gw\_call you create needed arguments, call $f$ with it and possibly show
the results (set\_color(..), set\_position(..) etc.).
}*/

    void add_call(base_function *f, string label, int sub_menu=0);
    int add_menu(string label, int sub_menu=0);
    int set_buttons_per_line(int n);

/*{\Mtext
\bigskip
\headerline { Some more} }*/
               
    void highlight(node v);

/*{\Mop ... }*/

    void highlight(edge e);

/*{\Mop ... }*/

    void unhighlight(node v);

/*{\Mop ... }*/

    void unhighlight(edge e);

/*{\Mop ... }*/

    color get_hi_node_color() const { return hi_node_color; }

    color set_hi_node_color(color c);

/*{\Mopl       sets the color used by highlight({\it node} $v$). }*/

    color get_hi_edge_color() const { return hi_edge_color; } 

    color set_hi_edge_color(color c);

/*{\Mopl       sets the color used by highlight({\it edge} $e$). }*/

    bool is_directed() const { return gr_p->is_directed(); }

/*{\Mop       returns true iff graph in |\Mvar| is directed. }*/

    void make_directed();

/*{\Mop       equivalent to the member-function of class graph.  }*/

    void make_undirected();

/*{\Mop       equivalent to the member-function of class graph.  }*/

    bool get_directed() const { return directed; }

    bool set_directed(bool b);    

/*{\Mop       show edges as arrows ($b=true$) or not ($b=false$). }*/

    void zoom_graph(double factor);

/*{\Mopl       zooms the graph by {\it factor}. }*/

    void zoom_up_area();

/*{\Mop       asks interactive for an rectangle to zoom in. }*/

    void fill_in_window();

/*{\Mop       zooms and translates the graph so it fits for the window. }*/

    void set_position(node_array<point>& dest);

/*{\Mopl       sets the node-positions to the given in $dest$. }*/

    void set_position(node_array<double>& destx, node_array<double>& desty);
    
/*{\Mopl       sets the node-positions to the given in ($destx$,$desty$). }*/

    node_move_t get_node_move() const { return node_move; }

    node_move_t set_node_move(node_move_t nm);
    
/*{\Mopl       When using set\_position({\it node\_array}\<..\> ..) the nodes 
               can be moved (see set\_animation) one after another 
               ($nm=${\it \ move\_single\_node}) 
               or all together ($nm=${\it \ move\_all\_nodes}). }*/

    bool wait(unsigned sec, string msg="");

/*{\Mopl       displays $msg$ and waits $sec$ seconds. 
               If the user presses the continue/quit-button from
               menu, wait terminates too. }*/

    bool wait(string msg="");

/*{\Mop		displays $msg$ and waits until continue/quit-button from
		menu is pressed. }*/
               
    void acknowledge(string s);

/*{\Mopl     displays string $s$ and asks for acknowledgement. }*/

    node get_node();
    
/*{\Mop     asks for node-determining and returns the chosen or nil, if
            there isn't one or the user canceled the action.}*/

    edge get_edge();
    
/*{\Mop     asks for edge-determining and returns the chosen or nil, if
            there isn't one or the user canceled the action.}*/

    const graph& get_graph() const { return *gr_p; }

/*{\Mop       returns a constant reference of the graph used in |\Mvar|. }*/

    void redraw();			

/*{\Mop		clears the window and draws the graph. }*/

    void clear_graph();

/*{\Mop  deletes all nodes and egdes. }*/

    void win_init(double xmin, double xmax, double ymin);

/*{\Mopl	reinitializes the window. }*/

    void set_frame_label(string label) {
      win_p->set_frame_label(label);
    }
    
/*{\Mopl        makes $label$ the window frame label. }*/

bool get_area(double& x0, double& y0, double& x1, double& y1, string msg="");
list<node> get_area_nodes(double x0, double y0, double x1, double y1);
void setup_nodes(list<node>& L);
unsigned long read_mouse_rect(double& x0, double& y0, double& x1, double& y1, bool pressed=false);

/*{\Mtext
\bigskip

There are some reset-functions, so values can set to their defaults. }*/

    void reset_node_types();

/*{\Mop  ...}*/

    void reset_node_colors(); 

/*{\Mop  ...}*/

    void reset_node_styles();

/*{\Mop  ...}*/

    void reset_node_radii();

/*{\Mop  ...}*/

    void reset_nodes();

/*{\Mop  reset\_node\_types() + reset\_node\_colors() + reset\_node\_styles() }*/

    void reset_edge_colors();

/*{\Mop  ...}*/

    void reset_edge_styles();

/*{\Mop  ...}*/

    void reset_edge_widths();

/*{\Mop  ...}*/

    void reset_edges();

/*{\Mop  reset\_edge\_colors/styles/widths() }*/

    void reset_nodes_and_edges() {
      reset_nodes(); 
      reset_edges();
    }    

/*{\Mopl  reset\_nodes() + reset\_edges() }*/

    void reset() { reset_nodes_and_edges(); }

/*{\Mopl  reset\_nodes\_and\_edges(); }*/

/*{\Mtext
\bigskip
\headerline { Input/Output} }*/

  int read_graph(string fname);

/*{\Mop read graph from file {\it fname} }*/

  bool save_graph(string fname);

/*{\Mop save graph to file {\it fname} }*/

  int read_gw(string name, string l_ext, string gw_ext);  

  int read_gw(const char* name="graph", const char* lext=".lgr", const char* gwext=".gw") {
    return read_gw(string(name),string(lext),string(gwext));
  }
  
/*{\Mopl ...}*/

  bool save_gw(string name, string l_ext, string gw_ext);  

  bool save_gw(const char* name="graph", const char* lext=".lgr", const char* gwext=".gw") {
    return save_gw(string(name),string(lext),string(gwext));
  }

/*{\Mopl ...}*/

  bool save_ps(string fname);

/*{\Mop ...}*/

};
 
//----------------------------------------------------------------------------

template<class function_t>
void gw_add_call(GraphWin& gw, function_t f, string label, int sub_menu) {
  base_function *bf_p=new function<function_t>(f);
  gw.add_call(bf_p,label,sub_menu);
} 

//----------------------------------------------------------------------------

template<class function_t>
void gw_add_call(GraphWin& gw, function_t f, string label) {
  gw_add_call(gw,f,label,0);
} 

//----------------------------------------------------------------------------

template<class function_t>
void gw_add_call(GraphWin& gw, function_t f, const char* label, int sub_menu) {
  gw_add_call(gw,f,string(label),sub_menu);
} 

//----------------------------------------------------------------------------

template<class function_t>
void gw_add_call(GraphWin& gw, function_t f, const char* label) {
  gw_add_call(gw,f,string(label));
} 

//----------------------------------------------------------------------------

inline int gw_add_menu(GraphWin& gw, string label, int sub_menu=0) {
  return gw.add_menu(label, sub_menu);
}

//----------------------------------------------------------------------------

class PostScript;		// defined elsewhere
PostScript& operator << (PostScript& ps, GraphWin& gw);

// ---------------------------------------------------------------------------

#endif /* LEDA_GRAPHWIN_H */
