/*******************************************************************************
+
+  LEDA 3.4
+
+  param_types.h
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#ifndef LEDA_PARAM_TYPE_H
#define LEDA_PARAM_TYPE_H

#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 340078
#include <LEDA/REDEFINE_NAMES.h>
#endif


//----------------------------------------------------------------------------
//
// Parameterized data types store copies of values of arbitrary types T
// in a single data field of type void* (GenPtr). There are two cases
// which have to be treated differently here.
//
// If T is a "big type" (sizeof(T) > sizeof(GenPtr)) we make a copy in the 
// dynamic memory using the new operator and store a pointer to this copy. 
// If T is a "small type" (sizeof(T) <= sizeof(GenPtr)) we avoid the 
// overhead of an extra level of indirection by copying the value directly 
// into the void* data field. This is done by using the "in-place" variant 
// of the new operator. A good description of this feature can be found 
// in  R.B. Murray: "C++ Strategies and Tactics" (chapter 9.5). We would
// like to thank all people contributing to this solution by sending us
// problem reports and helpful suggestions (see LEDA/ACKNOWLEDGEMENTS).
//
// Both cases are handled by the "leda_copy" function template defined in this 
// file. Similarly, the two cases (small or big type) have to be treated 
// differently,  when creating ("leda_create"), destroying ("leda_clear"),
// and accessing ("leda_access") data of a parameterized type. Note that 
// destruction of small types uses an an explicit destructor call which is 
// not supported by some compilers.
//
// More details on the implementation of parameterized data types in LEDA
// will be given in the text book "LEDA: A Platform ..."  by K. Mehlhorn
// and S. Naeher (to appear in 1996)
//
//
// S. Naeher (1995)
//
//
// MODIFICATIONS
//
// Mai  1996 : leda_const_access  (S. Naeher)
// July 1996 : removed default compare definition (S . Naeher)
//
//----------------------------------------------------------------------------



#if defined (__TEMPLATE_FUNCTIONS__)

// Most c++ implementations provide an in-place new operator in <new.h>.
// However, it is not always declared inline. So, for efficiency reasons,
// we define our own (inline) version. We use a third dummy argument to 
// avoid conflicts with user defined variants.
// For compilers not supporting explicit destructor calls or in-place new
// we use the old method based on casting and assignment. Then 0 must be
// legal value on the right side of assignments for all one-word types.


#if defined(__INPLACE_NEW__)

inline void Avoid_Unused_Warning(GenPtr) {}
inline void* operator new(size_t, void* where, int) { return where; }
#define COPY_INTO_WORD(T,x,p) Avoid_Unused_Warning(new((T*)&p,0) T(x))
#define CONSTRUCT_WORD(T,p)   Avoid_Unused_Warning(new((T*)&p,0) T)

#else

#define COPY_INTO_WORD(T,x,p) *(T*)&p = x;
#define CONSTRUCT_WORD(T,p)   { T x; COPY_INTO_WORD(T,x,p); }

#endif


#if defined(__EXPLICIT_DESTRUCTION__)

#if defined(__lucid)
#define DESTROY_WORD(T,p)  p->T::~T();
#else
#define DESTROY_WORD(T,p)  p->~T();
#endif

#else

inline void leda_explicit_destruction(GenPtr) {}
#define DESTROY_WORD(T,p)  leda_explicit_destruction((T*)p)

#endif


//----------------------------------------------------------------------------
// Function templates for leda_copy, leda_create, leda_clear, leda_cast, 
// leda_access, etc.  
// When using a compiler that does not support function templates
// these functions have to be created for class types by a call of the
// LEDA_TYPE_PARAMETER macro defined at end of this file. Most of the 
// templates distinguish the two cases that the actual type has size 
// greater than the size of a pointer or not, i.e., is a big or small 
// type. Note that this does not cause any overhead at run time since 
// the right case can be determined at compile time. 
//----------------------------------------------------------------------------


template<class T>
inline GenPtr leda_copy(const T& x)
{ GenPtr p=0;
  if (sizeof(T) <= sizeof(GenPtr)) COPY_INTO_WORD(T,x,p);
  if (sizeof(T) >  sizeof(GenPtr)) p = new T(x);
  return p;
 }

template <class T>
inline GenPtr leda_create(const T*) 
{ GenPtr p=0;
  if (sizeof(T) <= sizeof(GenPtr)) CONSTRUCT_WORD(T,p);
  if (sizeof(T) >  sizeof(GenPtr)) p = new T;
  return p;
 }

template <class T>
inline void leda_clear(T& x) 
{ T* p = &x;
  if (sizeof(T) <= sizeof(GenPtr)) DESTROY_WORD(T,p);
  if (sizeof(T) >  sizeof(GenPtr)) delete p;
}


template <class T>
inline GenPtr leda_cast(const T& x) 
{ GenPtr p=0;
  if (sizeof(T) == sizeof(GenPtr)) p = *(GenPtr*)&x;
  if (sizeof(T) <  sizeof(GenPtr)) *(T*)&p = x;
  if (sizeof(T) >  sizeof(GenPtr)) p = (GenPtr)&x;
  return p;
 }


#if !defined(_MSC_VER)

template <class T>
inline T& leda_access(const T*, const GenPtr& p)
{ if (sizeof(T) <= sizeof(GenPtr))
     return *(T*)&p;
  else
     return *(T*)p;
 }

template <class T>
inline const T& leda_const_access(const T*, const GenPtr& p)
{ if (sizeof(T) <= sizeof(GenPtr))
     return *(T*)&p;
  else
     return *(T*)p;
 }

#else

// Since Microsoft C++ cannot convert a GenPtr to a const GenPtr& in this 
// situation  we use a pointer instead of a reference as second argument in
// the leda_access template and a dummy function for converting the
// reference into a pointer (see also the definition of LEDA_ACCESS at the end
// of this file).

inline GenPtr* msc_hack(const GenPtr& p) { return (GenPtr*)&p; }

template <class T>
inline T& leda_access(const T*, GenPtr* p) 
{ if (sizeof(T) <= sizeof(GenPtr))
     return *(T*)p;
  else
     return *(T*)*p;
 }

template <class T>
inline const T& leda_const_access(const T*, GenPtr* p) 
{ if (sizeof(T) <= sizeof(GenPtr))
     return *(T*)p;
  else
     return *(T*)*p;
 }

#endif




template <class T> 
inline int   leda_itype(const T*) { return 0; }

template <class T>
inline char* leda_tname(const T*) { return "unknown"; }


//---------------------------------------------------------------------------
// default definitions for Print and Read
//---------------------------------------------------------------------------

inline istream& operator>>(istream& in, GenPtr) { return in; }

template <class T> 
inline void Print(const T& x, ostream& out) { out << x; }

template <class T>
inline void Read(T& x, istream& in) { in >> x; }


//---------------------------------------------------------------------------
// template declaration for compare
//---------------------------------------------------------------------------

#if defined(LEDA_COMPARE_TEMPLATE)
template <class T> int compare(const T&, const T&);
#endif


#else
//---------------------------------------------------------------------------
// Without function templates we have to provide functions for all
// pointer types (GenPtr versions) and Prind & Read for built-in types. 
//---------------------------------------------------------------------------

inline GenPtr& leda_access(const GenPtr, const GenPtr& p) 
                                          { return *(GenPtr*)&p; }
inline GenPtr  leda_copy(const GenPtr& x) { return x; }
inline GenPtr  leda_create(GenPtr)        { return 0; }
inline void    leda_clear(GenPtr)         {}
inline GenPtr  leda_cast(GenPtr x)        { return x; }
inline int     leda_itype(GenPtr)         { return 0; }
inline char*   leda_tname(GenPtr)         { return "unknown"; }

//inline int  compare(GenPtr x, GenPtr y) { return int(long(x)-long(y)); }
inline int  compare(const GenPtr& x, const GenPtr& y) 
{ return int(long(x)-long(y)); }
inline void Print(const GenPtr& x, ostream& out) { out << x; }
inline void Read(GenPtr, istream&) {}


inline void Print(const char& x, ostream& out  ) { out << x; }
inline void Print(const short& x, ostream& out ) { out << x; }
inline void Print(const int& x, ostream& out   ) { out << x; }
inline void Print(const long& x, ostream& out  ) { out << x; }
inline void Print(const float& x, ostream& out ) { out << x; }
inline void Print(const double& x, ostream& out) { out << x; }
#if defined(__BUILTIN_BOOL__)
inline void  Print(const bool& x, ostream& out) { out << (x ? "true":"false"); }
#endif

inline void Read(char& x,  istream& in) { in >> x; }
inline void Read(short& x, istream& in) { in >> x; }
inline void Read(int&  x,  istream& in) { in >> x; }
inline void Read(long& x,  istream& in) { in >> x; }
inline void Read(float& x, istream& in) { in >> x; }
inline void Read(double& x,istream& in) { in >> x; }
#if defined(__BUILTIN_BOOL__)
inline void  Read(bool&,  istream&) {}
#endif


#endif



//----------------------------------------------------------------------------
// Specializations for some built-in types (to be extended as needed) for
// compilers not supporting function templates at all or having particular 
// problems with function templates (watcomc, lucid) (g++ seems not to be
// able to instantiate function templates inline).
//----------------------------------------------------------------------------

#if !defined(__TEMPLATE_FUNCTIONS__) || defined(__WATCOMC__) || defined(__lucid) || defined(__GNUC__)

inline GenPtr leda_copy(const char& x)  { GenPtr p; *(char*)&p = x;  return p; }
inline GenPtr leda_copy(const short& x) { GenPtr p; *(short*)&p = x; return p; }
inline GenPtr leda_copy(const int& x)   { GenPtr p; *(int*)&p = x;   return p; }
inline GenPtr leda_copy(const long& x)  { GenPtr p; *(long*)&p = x;  return p; }
inline GenPtr leda_copy(const float& x) { GenPtr p; *(float*)&p = x; return p; }
 
inline GenPtr leda_cast(const char& x)  { return leda_copy(x); }
inline GenPtr leda_cast(const short& x) { return leda_copy(x); }
inline GenPtr leda_cast(const int& x)   { return leda_copy(x); }
inline GenPtr leda_cast(const long& x)  { return leda_copy(x); }
inline GenPtr leda_cast(const float& x) { return leda_copy(x); }
 
inline GenPtr leda_create(const char*)  { return 0; }
inline GenPtr leda_create(const short*) { return 0; }
inline GenPtr leda_create(const int*)   { return 0; }
inline GenPtr leda_create(const long*)  { return 0; }
inline GenPtr leda_create(const float*) { return 0; }
 
inline void   leda_clear(char&)  {}
inline void   leda_clear(short&) {}
inline void   leda_clear(int&)   {}
inline void   leda_clear(long&)  {}
inline void   leda_clear(float&) {}
 
inline char&  leda_access(const char*, const GenPtr& p) { return *(char*)&p;  }
inline short& leda_access(const short*,const GenPtr& p) { return *(short*)&p; }
inline int&   leda_access(const int*,  const GenPtr& p) { return *(int*)&p;   }
inline long&  leda_access(const long*, const GenPtr& p) { return *(long*)&p;  }
inline float& leda_access(const float*,const GenPtr& p) { return *(float*)&p; }

inline const char&  leda_const_access(const char*, const GenPtr& p) 
{ return *(char*)&p;  }
inline const short& leda_const_access(const short*,const GenPtr& p) 
{ return *(short*)&p; }
inline const int&   leda_const_access(const int*,  const GenPtr& p) 
{ return *(int*)&p;   }
inline const long&  leda_const_access(const long*, const GenPtr& p) 
{ return *(long*)&p;  }
inline const float& leda_const_access(const float*,const GenPtr& p) 
{ return *(float*)&p; }
 
#endif



//----------------------------------------------------------------------------
// A "special" specialization for "double" using the LEDA memory management 
// if sizeof(double) > sizeof(GenPtr). 
//----------------------------------------------------------------------------

inline GenPtr leda_copy(const double& x)
{ GenPtr p;
  if (sizeof(double) > sizeof(GenPtr))
    { p = std_memory.allocate_bytes(sizeof(double));
      *(double*)p = x; }
  else 
      *(double*)&p = x;
  return p;
 }

inline GenPtr leda_create(const double*)
{ GenPtr p;
  if (sizeof(double) > sizeof(GenPtr))
    { p = std_memory.allocate_bytes(sizeof(double));
      *(double*)p = 0; }
  else 
      *(double*)&p = 0;
  return p;
 }

inline void leda_clear(double& x) 
{ if (sizeof(double) > sizeof(GenPtr)) 
     std_memory.deallocate_bytes(&x,sizeof(double)); 
 }

inline GenPtr leda_cast(const double& x) 
{ return (sizeof(double) <= sizeof(GenPtr)) ? *(GenPtr*)&x : GenPtr(&x); }


inline double& leda_access(const double*, const GenPtr& p) 
{ return (sizeof(double) <= sizeof(GenPtr)) ? *(double*)&p : *(double*)p; }

inline const double& leda_const_access(const double*, const GenPtr& p) 
{ return (sizeof(double) <= sizeof(GenPtr)) ? *(double*)&p : *(double*)p; }




//----------------------------------------------------------------------------
// integer type or not ?
//----------------------------------------------------------------------------

inline int leda_itype(const int*)  { return 1; }
inline int leda_itype(const long*) { return sizeof(long) == sizeof(int); }


//----------------------------------------------------------------------------
// id numbers
//----------------------------------------------------------------------------

inline unsigned long ID_Number(GenPtr x) { return (unsigned long)x; }
inline unsigned long ID_Number(char   x) { return (unsigned long)x; }
inline unsigned long ID_Number(short  x) { return (unsigned long)x; }
inline unsigned long ID_Number(int    x) { return (unsigned long)x; }
inline unsigned long ID_Number(long   x) { return (unsigned long)x; }


//----------------------------------------------------------------------------
// type names
//----------------------------------------------------------------------------

inline char* leda_tname(const char*  ) { return "char";   }
inline char* leda_tname(const short* ) { return "short";  }
inline char* leda_tname(const int*   ) { return "int";    }
inline char* leda_tname(const long*  ) { return "long";   }
inline char* leda_tname(const float* ) { return "float";  }
inline char* leda_tname(const double*) { return "double"; }



//----------------------------------------------------------------------------
// pre-defined linear orders and hash functions
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// compare
//----------------------------------------------------------------------------

template <class T> class decl_compare_func
{ friend int compare(const T&, const T&); };


#define DEFINE_COMPARE(T)                 \
inline int compare(const T& x, const T& y)\
{ if (x < y) return -1;                   \
  else if (x > y) return  1;              \
  else return 0;                          \
}

DEFINE_COMPARE(char)
DEFINE_COMPARE(int)
DEFINE_COMPARE(unsigned int)
DEFINE_COMPARE(long)
DEFINE_COMPARE(unsigned long)
DEFINE_COMPARE(float)
DEFINE_COMPARE(double)

#if defined(__BUILTIN_BOOL__)
DEFINE_COMPARE(bool)
#endif


//----------------------------------------------------------------------------
// hashing
//----------------------------------------------------------------------------

inline int Hash(const GenPtr& x) { return int(long(x)); }
inline int Hash(const char&   x) { return int(x); } 
inline int Hash(const short&  x) { return int(x); }
inline int Hash(const int&    x) { return int(x); }
inline int Hash(const long&   x) { return int(x); }
inline int Hash(const float&  x) { return int(x); }
inline int Hash(const double& x) { return int(x); }
#if defined(__BUILTIN_BOOL__)
inline int Hash(const bool&   x) { return int(x); }
#endif



//----------------------------------------------------------------------------
// some useful macros 
// (used in data type templates)
//----------------------------------------------------------------------------

#if defined(__WATCOMC__) && __WATCOMC__ < 1050
#define MKNULLPTR(T) (const T*)0
#define ACCESSPTR(p) (const GenPtr&)p
#else
#if defined(_MSC_VER)
#define MKNULLPTR(T) (T*)0
#define ACCESSPTR(p) msc_hack(p)
#else
#define MKNULLPTR(T) (T*)0
#define ACCESSPTR(p) p
#endif
#endif


#if !defined(__TEMPLATE_FUNCTIONS__) || defined(__ATTCFRONT__)
#define LEDA_ACCESS(T,p)       (T&)(leda_access(MKNULLPTR(T),p))
#define LEDA_CONST_ACCESS(T,p) (const T&)(leda_access(MKNULLPTR(T),p))
#else
#define LEDA_ACCESS(T,p)       leda_access(MKNULLPTR(T),ACCESSPTR(p))
#define LEDA_CONST_ACCESS(T,p) leda_const_access(MKNULLPTR(T),ACCESSPTR(p))
#endif

#define LEDA_CREATE(T,x)    x=leda_create(MKNULLPTR(T))
#define LEDA_COPY(T,x)      x=leda_copy(LEDA_CONST_ACCESS(T,x))
#define LEDA_INT_TYPE(T)    leda_itype(MKNULLPTR(T))
#define LEDA_TYPE_NAME(T)   leda_tname(MKNULLPTR(T))
#define LEDA_CAST(x)        leda_cast(x)
#define LEDA_CLEAR(T,x)     leda_clear(LEDA_ACCESS(T,x))
#define LEDA_COMPARE(T,x,y) compare(LEDA_CONST_ACCESS(T,x),LEDA_CONST_ACCESS(T,y))
#define LEDA_HASH(T,x)      Hash(LEDA_CONST_ACCESS(T,x))
#define LEDA_PRINT(T,x,out) Print(LEDA_CONST_ACCESS(T,x),(ostream&)out)
#define LEDA_READ(T,x,in)   Read(LEDA_ACCESS(T,x),(istream&)in)

 

//----------------------------------------------------------------------------
// IF FUNCTION TEMPLATES DO NO WORK
//
// LEDA_TYPE_PARAMETER (T)
//
// has to be called for any "big type" T before using T as an actual type 
// argument in parameterized data types if function templates cannot be used.
//----------------------------------------------------------------------------
 
#define LEDA_TYPE_PARAMETER(T)\
inline GenPtr   leda_copy(const T& x)     { return new T(x); }\
inline GenPtr   leda_create(const T*)     { return new T; }\
inline void     leda_clear(T& x)          { delete &x; }\
inline GenPtr   leda_cast(const T& x)     { return (GenPtr)&x; }\
inline T&       leda_access(const T*, const GenPtr& p)  { return *(T*)p; }\
inline char*    leda_tname(const T*)      { return STRINGIZE(T); }\
inline void     Print(const T& x, ostream& out) { out << x; }\
inline void     Read(T& x, istream& in)         { in >> x; }
 

#if LEDA_ROOT_INCL_ID == 340078
#undef LEDA_ROOT_INCL_ID
#include <LEDA/UNDEFINE_NAMES.h>
#endif

#endif
