/*******************************************************************************
+
+  LEDA 3.4
+
+  real.h
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

/* This file has been automatically generated from "_real.w"
   by CTANGLE (Version 3.1 [km2c]) */


/*{\Manpage {real} {} {Real Numbers} }*/

#ifndef REAL_H
#define REAL_H

#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 340100
#include <LEDA/REDEFINE_NAMES.h>
#endif


/*{\Mdefinition

An instance $x$ of the data type |real| is an algebraic real number.
There are many ways to construct a |real|:
either by conversion from |double|, |bigfloat|, |integer| or |rational|
or by applying one of the arithmetic operators $+, -, *, /$ or $\sqrt{} $ to
|real| numbers.
One may test the sign of a |real| number or compare two |real| numbers by
any of the comparison relations $=, \neq, < , \leq, >$ and $\geq$.
The outcome of such a test is {\em exact}.
There is also a non--standard version of the sign function: the call
|x.sign(integer q)| computes the sign of |x| under the precondition
that $\Labs{x} \leq 2^{-q}$ implies $x = 0$.
This version of the sign function allows the user to assist the data type in the
computation of the sign of $x$, see the example below.

There are also several functions to compute approximations of |real|s.
The calls |x.to_bigfloat()| and |x.get_bigfloat_error()| return |bigfloat|s
$xnum$ and $xerr$ such that $\Labs{num - x} \leq xerr$.
The user may set a bound on $xerr$.
More precisely, after the call |x.reduce_error_in_approximation_to(integer q)| the
data type guarantees $xerr \leq 2^{-q}$.
One can also ask for |double| approximations of a real number $x$.
The calls |x.to_double()| and |x.get_double_error()| return |double|s $xnum$ and
$xerr$ such that $\Labs{xnum - x} \leq  \Labs{xnum} * xerr$.
Note that $xerr = \infty$ is possible.
}*/

#include <LEDA/bigfloat.h>
#include <LEDA/rational.h>
class real_rep;
class real {

  bool is_double;
  double value;
  real_rep *PTR;


  void adjust() const;


 public:


/*{\Mcreation x
}*/

/*{\Mtext |real|s may be constructed from data types |double|, |bigfloat|,
 |long|, |int| and |integer|.
 The default constructor |real()| initializes the |real| to zero.
}*/


   real();
   real(double);
   real(int);
   real(const integer &);
   real(const bigfloat &);
   real(const rational &);
   real(const real &);
   real(real_rep &);



   ~real();

   real & operator = (const real &);

/*{\Moperations 2.0 3.6
}*/

  double to_double() const;
/*{\Mop returns the current double approximation of |\Mvar|. }*/

  double todouble() const { return to_double(); }
/*{\Xop returns the current double approximation of |\Mvar|. }*/


  bigfloat to_bigfloat() const;
/*{\Mop returns the current bigfloat approximation of |\Mvar|.
}*/

  bigfloat tobigfloat() const { return to_bigfloat(); }
/*{\Xop returns the current bigfloat approximation of |\Mvar|.
}*/

  double get_double_error() const;
/*{\Mop returns the relative error of the current double approximation of |\Mvar|,
    i.e., $\Labs{\mbox{|x|} - \mbox{|x.to_double()|}} \leq
    \mbox{|x.get_double_error()|} \cdot \Labs{\mbox{|x.to_double()|}}$.
}*/

  bigfloat get_bigfloat_error() const;
/*{\Mop returns the absolute error of the current bigfloat approximation of
   |x|, i.e., $\Labs{\mbox{|x|} - \mbox{|x.to_bigfloat()|}} \leq
    \mbox{|x.get_bigfloat_error()|}$.
}*/

  int sign(const integer & q) const;
/*{\Mop as above. {\em Precondition:} if $\Labs{\mbox{|x|}}  \leq  2^{-q}$
   then $\mbox{|x|} = 0$.
}*/

  int sign(long) const;

  int sign() const;

  void reduce_error_in_approximation_to(const integer & q) const;
/*{\Mopl (re-)computes the approximation of |\Mvar| such that
  $\mbox{|x.get_bigfloat_error()|} \leq 2^{-q}$ after the call
  |x.reduce_error_in_approximation_to(q)|.
}*/

  void compute_with_mantissa_length(long k) const;
/*{\Mopl (re-)computes the approximation of |\Mvar|; each numerical operation
   is carried out with $k$ binary places.
}*/

  void guarantee_relative_error_two_to_minus(long k) const;
/*{\Mopl (re-)computes an approximation of |\Mvar| such that the relative error
    of the approximation is less than $2^{-k}$.
}*/

  friend ostream& operator << (ostream & O, const real & x);
/*{\Mbinopfunc writes the closest interval that is known to contain $x$
    to the output stream $O$.
}*/
  friend istream& operator >> (istream & I, real & x);
/*{\Mbinopfunc reads |real| number |x| from the output stream $I$
   (in |double| format).
}*/



  friend int operator < (const real &, const real &);
  friend int operator <= (const real &, const real &);
  friend int operator > (const real &, const real &);
  friend int operator >= (const real &, const real &);
  friend int operator == (const real &, const real &);
  friend int operator != (const real &, const real &);



  friend real operator + (const real &, const real &);
  friend real operator - (const real &, const real &);
  friend real operator *(const real &, const real &);
  friend real operator / (const real &, const real &);
  friend real operator - (const real &);
  friend real sqrt(const real &);


};

inline int compare(const real & x, const real & y)
{ return (x - y).sign(); }

inline char* leda_tname(const real *)
{ return "real"; }



inline real operator += (real & x, const real & y) {
  x = x + y;
  return x;
}
inline real operator -= (real & x, const real & y) {
  x = x - y;
  return x;
}
inline real operator *= (real & x, const real & y) {
  x = x * y;
  return x;
}


int sign(real x) { return x.sign(); }
/*{\Mfunc returns $-1$ if (the exact value of) $x$ is $< 0$, $1$ if $x$ 
   $> 0$, $0$ if $x$ is $= 0$.
}*/


real fabs(const real & x);

inline real abs(const real & x) {
  return fabs(x);
}
/*{\Mfunc returns the absolute value of |x|.
}*/
real sq(const real & x);
/*{\Mfunc returns the square of |x|.
}*/
real dist(const real & x, const real & y);
/*{\Mfunc returns the euclidean distance of vector (x,y) to the origin.
}*/
real powi(const real & x, int n);
/*{\Mfunc returns the |n|-th power of |x|.
}*/

/*{\Mimplementation

A |\Mtype| is represented by the expression which defines it and a |double|
approximation $\hat{x}$ together with a relative error bound
$\epsilon_{x}$.
The arithmetic operators $+, -, *, \, \sqrt{}$ take constant time.
When the sign of a |\Mtype| number needs to be determined, the data type first
computes a number $q$, if not already given as an argument to |sign|, such that
$\Labs{\mbox{\Mvar}} \leq 2^{-q}$ implies $x = 0$.
The bound $q$ is computed as described in \cite{Mignotte:book}.
The data type then computes a bigfloat approximation $xnum$ for |\Mvar| with
error bound $xerr \leq 2^{-q}$.
The sign of $xnum$ is then returned as the sign of |\Mvar|.

Two shortcuts are used to speed up the computation of the sign.
Firstly, if the |double| approximation already suffices to determine the sign,
then no |bigfloat| approximation is computed at all.
Secondly, the |bigfloat| approximation is first computed only with small precision.
The precision is then doubled until either the sign can be decided (i.e.,
if $xerr < \Labs{xnum}$) or the full precision $2^{-q}$ is reached.
}*/

/*{\Mexample
We give two examples of the use of the data type |\Mtype|.
The examples deal with the Voronoi diagram of line segments and the intersection
of line segments, respectively.

The following incircle test arises in the computation of Voronoi diagrams of
line segments \cite{BMS:ESA}.
For $i$, $1 \leq i \leq 3$, let $l_i : a_i x + b_i y + c_i = 0$ be a line
in two--dimensional space and let $p = (0,0)$ be the origin.
In general, there are two circles passing through $p$ and touching $l_1$ and
$l_2$.
The centers of these circles have coordinates
\begin{eqnarray*}
x_v = \frac{a_1 c_2 +a_2 c_1 \pm  \sqrt{2 c_1 c_2 (\sqrt{N} + C)}}
           {\sqrt{N} - a_1 a_2 - b_1 b_2}  ,
y_v = \frac{b_1 c_2 +b_2 c_1 \pm  \sqrt{2 c_1 c_2 (\sqrt{N} - C)}}
           {\sqrt{N} - a_1 a_2 - b_1 b_2}
\end{eqnarray*}
where $N = (a_1^{2} + b_1^{2}) (a_2^{2} + b_2^{2})$ and $C = a_1 a_2 - b_1 b_2$.
Let us concentrate on one of these (say, we take the plus sign in both cases).
The test whether $l_3$ intersects, touches or misses the circle amounts to
determining the sign of
\begin{eqnarray*}
D := dist^{2}(v,l_3) - dist^{2}(v,p) =
     \frac{(a_3 x_v + b_3 y_v + c_3)^2}{a_3^2 + b_3^2} - (x_v^2 + y_v^2)
\end{eqnarray*}
We show how to compute the sign of $\tilde{D} := (a_3^2 + b_3^2) \cdot D$ using
our data type |\Mtype|.
We assume that all coefficients $a_i, b_i$ and $c_i$, $1 \leq i \leq 3$,
are $k$ bit integers, i.e., integers whose absolute value is bounded
by $2^k -1$.
In \cite{BMS:ESA} we showed:
\begin{itemize}
\item if $\tilde{D} \neq 0$ then $\Labs{\tilde{D}} \geq 2^{-24 k - 50}$,
\item if $N$ is a square and $\tilde{D} \neq 0$ then $\tilde{D} \geq
          2^{-8 k - 18}
         $
\end{itemize}
where $k_1$ and $k_2$ are integer constants.
These observations suggest the following program.
\begin{verbatim}
int incircle(real k, a1, b1, c1, a2, b2, c2, a3, b3, c3)
{ real N = (a1*a1 + b1*b1) * (a2*a2 + b2*b2); 
  real C = a1 * a2 - b1 * b2;  
  real S = - (a1 * b2 + a2 * b1);  
  real D = sqrt(N) - (a1 * a2 + b1 * b2);  
  real x_v = (a1 * c2 + a2 * c1 + sqrt(2*c1*c2*(sqrt(N) + C)) / D;  
  real y_v = (b1 * c2 + b2 * c1 + S.sign() * sqrt(2*c1*c2*(sqrt(N) - C)) / D; 
  real E = (a3 * x_v + b3 * y_v + c3)*(a3 * x_v + b3 * y_v + c3)
                                       - (a3*a3 + b3*b3)*(x_v*x_v + y_v*y_v);  
  int max_prec = (is_square(N) ? 8*k + 18 : 24*k + 50); 
  return E.sign(max_prec);  
}  
\end{verbatim}

We turn to the line segment intersection problem next.
Assume that all endpoints have $k$--bit integer homogeneous coordinates.
This implies that the intersection points have homogeneous coordinates
$(X,Y,W)$ where $X,Y$ and $W$ are (4 k + 3) - bit integers.
The Bentley--Ottmann plane sweep algorithm for segment intersection
\cite{Me-Naeher:sweep} needs to sort points by their $x$--coordinates, i.e.,
to compare fractions $X_1 / W_1$ and $X_2 / W_2$ where $X_1, X_2, W_1, W_2$
are as above.
This is tantamount to determining the sign of the $8 k + 7$ bit integer
$X_1 * W_2 - X_2 * W_1$.
If all variables $X_i, W_i$ are declared {\bf real} then their sign
test will be performed quite efficiently.
First, a |double| approximation is computed and then, if necessary,
|bigfloat| approximations of increasing precision.
In many cases, the |double| approximation already determines the sign.
In this way, the user of the data type |real| gets the efficiency of a
floating point filter \cite{Fortune-vWijk,Me-Naeher:IFIP94} without
any work on his side.
This is in marked contrast to \cite{Fortune-vWijk,Me-Naeher:IFIP94} and
will be incorporated into \cite{Me-Naeher:sweep}.

}*/



#if LEDA_ROOT_INCL_ID == 340100
#undef LEDA_ROOT_INCL_ID
#include <LEDA/UNDEFINE_NAMES.h>
#endif


#endif
