#include <LEDA/plane_alg.h>
#include <LEDA/window.h>
#include <LEDA/map.h>

const double infty = 10000;

static point A(-infty,-infty);
static point B(+infty,-infty);
static point C(0,infty);

static edge e0;

static window W;
static window W_mess(180,65);

static int  flip_count = 0;
static int  point_count = 0;
static int  speed = 200;

static bool show_voronoi = false;
static bool show_delaunay = true;
static bool animate = false;

static point nil_point;

static map<edge,point> vnode(nil_point,0,2048);

void get_release_event()
{ double a,b;
  int v;
  while (W.read_event(v,a,b) != button_release_event);
 }



bool infi_edge(point a, point b)
{ return identical(a,A) || identical(a,B) || identical(a,C) ||
         identical(b,A) || identical(b,B) || identical(b,C);
}


void draw_triang_edge(const GRAPH<point,edge>& G, edge e)
{ point a = G[source(e)];
  point b = G[target(e)];
  if (show_delaunay && !infi_edge(a,b)) 
  W.draw_edge(a,b,blue);
}



void draw_voro_edge(const GRAPH<point,edge>& G, edge e)
{ if (show_voronoi)
  { edge r = G[e];
    if (vnode[e] != nil_point && vnode[r] != nil_point)
        W.draw_segment(vnode[e], vnode[r],violet); 
   }
}


void draw_delaunay(const GRAPH<point,edge>& G)
{ if (!show_delaunay) return;
  edge_array<bool> visited(G,false);
  edge e;
  forall_edges(e,G) 
     if (!visited[e])
       { draw_triang_edge(G,e);
         visited[G[e]] = visited[e] = true;
        }
 }


#define next_face_edge(x)  G.cyclic_adj_pred(G[x])
#define reverse_edge(x)    G[x]


static void compute_center(const GRAPH<point,edge>& G, edge e1)
{ edge e2 = next_face_edge(e1);
  edge e3 = next_face_edge(e2);
  point a = G[source(e1)];
  point b = G[source(e2)];
  point c = G[source(e3)];
  point cent = nil_point; 
  if (orientation(a,b,c) > 0) cent = circle(a,b,c).center();
  vnode[e1] = vnode[e2] = vnode[e3] = cent;
}


static void draw_circle(GRAPH<point,edge>& G, edge e)
{ edge x = next_face_edge(e);
  circle circ(G[source(e)],G[target(e)],G[target(x)]);
  W.draw_disc(circ,yellow);
  W.draw_circle(circ,black);
 }


static void build_voro(const GRAPH<point,edge>& G)
{ 
  edge e;
  forall_edges(e,G) vnode[e] = nil_point;
  forall_edges(e,G)
     if (identical(vnode[e],nil_point)) compute_center(G,e);
}


static void draw_voro(const GRAPH<point,edge>& G)
{ 
  edge_array<bool>  visited(G,false);
  edge e;

  if (show_voronoi)
  forall_edges(e,G)
     if (!visited[e])
     { edge r = G[e];
       visited[r] = visited[e] = true;
       draw_voro_edge(G,e);
     }
}



static void animate_flip(const GRAPH<point,edge>& G, edge e, int n)
{ 
  if (!show_delaunay) return;

  point a = G[source(e)];
  point b = G[target(G.cyclic_adj_pred(e))];
  point c = G[target(e)];
  point d = G[target(G.cyclic_adj_succ(e))];


  //W_mess.del_messages();
  //W_mess.message(string("%3d flips",flip_count++));

  W_mess.draw_text(3,15,string("%4d flips  (%.2f)",flip_count++,
                         float(flip_count)/point_count));

  segment s1(a,b);
  segment s2(c,d);
  double l1 = s1.length()/n;
  double l2 = s2.length()/n;
  double a1 = s1.angle(); 
  double a2 = s2.angle(); 

  if (animate && !infi_edge(a,c) && !infi_edge(b,d))
      { W.draw_edge(a,c,blue);
        W.draw_edge(a,c,red);
        while (n--)
        { point a_new = a.translate(a1,l1);
          point c_new = c.translate(a2,l2);
          W.draw_edge(a_new,c_new,red);
          W.draw_edge(a,c,red);
          a = a_new;
          c = c_new;
         }
        W.draw_edge(a,c,red);
        W.draw_edge(a,c,blue);
       }
   else
      { if (!infi_edge(a,c)) W.draw_edge(a,c,blue);
        if (!infi_edge(b,d)) W.draw_edge(b,d,blue);
       }

}




static edge locate_point(GRAPH<point,edge>& G, point p)
{ point a = G[source(e0)];
  point b = G[target(e0)];
  point m = a;

  int orient = orientation(a,b,p);

  if (orient == 0) 
  { e0 = next_face_edge(e0);
    a = b;
    b = G[target(e0)];
    orient = orientation(a,b,p);
   }

  if (orient >  0) e0 = G[e0]; 

  do { edge e1 = next_face_edge(G[e0]);
       e0 = (orientation(m,G[target(e1)],p) < 0) ? e1 : next_face_edge(e1);
  } while (orientation(G[source(e0)],G[target(e0)],p) <= 0);

  return e0;
}


static node nearest_neighbor(GRAPH<point,edge>& G, point p)
{
  edge e = locate_point(G,p);

  edge e1 = next_face_edge(e);

  node V[3];

  V[0] = source(e);
  V[1] = source(e1);
  V[2] = target(e1);

  double D_min =  MAXINT;
  int i_min = 0;

  for(int i=0; i<3; i++)
  { double dx = G[V[i]].xcoord() - p.xcoord();
    double dy = G[V[i]].ycoord() - p.ycoord();
    double D = dx*dx + dy*dy;
    if (D < D_min) 
    { D_min = D;
      i_min = i;
     }
   }

   return V[i_min];
}

  
   

   


inline bool flip_test(const GRAPH<point,edge>& G, edge e)
{ point a = G[source(e)];
  point b = G[target(G.cyclic_adj_pred(e))];
  point c = G[target(e)];
  point d = G[target(G.cyclic_adj_succ(e))];
  return right_turn(b,a,d) && right_turn(b,d,c) && incircle(a,b,c,d) > 0;
 }




static void delaunay_flipping(GRAPH<point,edge>& G, const list<edge>& el)
{
  edge_map<list_item> It(G,nil);
  list<edge> Q;
  edge E[4];
  edge e;

  forall(e,el) 
    if (It[e] == nil && flip_test(G,e)) 
       It[G[e]] = It[e] = Q.append(e);


  while ( !Q.empty() )
  { flip_count++;

    edge e = Q.pop();
    edge x = G.cyclic_adj_pred(e); 

    if (animate)
    { animate_flip(G,e,8000/speed);
      draw_voro_edge(G,e);              // delete old voronoi edge
     }

    // delete diagonal
    G.del_edge(G[e]);
    G.del_edge(e);

    // collect face edges of quadriliteral
    for(int i = 0; i < 4; i++) 
    { E[i] = x; 
      if (animate) draw_voro_edge(G,x); // delete old voronoi edges
      x = next_face_edge(x); 
     }


    // insert new diagonal
    edge diag = G.new_edge(E[1],source(E[3]));
    G[diag] = G.new_edge(E[3],source(E[1]),diag);

    e0 = diag;

    It[G[diag]] = It[diag] = nil;

    compute_center(G,E[0]);
    compute_center(G,E[1]);

    if (animate) draw_voro_edge(G,diag);  // draw new voronoi edge


    // test collected edges 
    for(int j=0; j<4; j++)
    { edge e = E[j];

      if (animate) draw_voro_edge(G,e); // draw new voronoi edges

      if (flip_test(G,e))
        { if (It[e] == nil) 
          It[G[e]] = It[e] = Q.push(e); 
         }
      else
          if (It[e] != nil)
          { Q.del_item(It[e]);
            It[G[e]] = It[e] = nil;
           }
     }

   }
 }


void initialize(GRAPH<point,edge>& G, list<point>& L)
{ // construct initial triangle (A,B,C)
  G.clear();
  L.clear();
  L.append(A);
  L.append(B);
  L.append(C);
  TRIANGULATE_POINTS(L,G);
  build_voro(G);
 }



void insert(GRAPH<point,edge>& G, list<point>& L, point p, bool inter=false)
{ 
    bool anim_save = animate; 
    node v = G.new_node(p);
    edge e = locate_point(G,p);

    L.append(p);
    point_count++;

    W_mess.draw_text(3,30, string("%4d points",point_count));

    animate = true;
    W.draw_filled_node(p,black);

    for(int i=0; i<3; i++)
    { draw_voro_edge(G,e); // delete old voronoi edge
      edge x = G.new_edge(e,v);
      G[x] = G.new_edge(v,source(e),x);
      draw_triang_edge(G,x);
      e = next_face_edge(e);
     }


    list<edge> E;

    forall_adj_edges(e,v)
    { compute_center(G,e);
      E.append(e);
      E.append(next_face_edge(e));
     }
    
    forall(e,E) draw_voro_edge(G,e);
 
    if (inter) get_release_event();

    delaunay_flipping(G,E);

    animate = anim_save;
}



void remove(GRAPH<point,edge>& G, list<point>& L, node v)
{ 
    bool anim_save = animate; 
    animate = true;
    point q = G[v];

    L.del_item(L.search(q));
    point_count--;

    W_mess.draw_text(3,40, string("%4d points",point_count));

    edge e_first = next_face_edge(G.first_adj_edge(v));

    // erase node and edges in W
    W.draw_filled_node(q,black);
    edge e;
    forall_adj_edges(e,v) 
    { draw_triang_edge(G,e);
      draw_voro_edge(G,e);
      draw_voro_edge(G,next_face_edge(e));
      if (e0 == e || e0 == G[e]) e0 = e_first;
     }

    G.del_node(v);

    // triangulate face, collect new edges in E

    v = source(e_first);

    list<edge> E;

    E.append(e_first);

    e = next_face_edge(e_first); 

    edge next = next_face_edge(e);
    edge x = e_first;

    while (target(next) != v)
    { x = G.new_edge(x,target(e));
      G[x] = G.new_edge(next,v,x);
      draw_triang_edge(G,x);
      compute_center(G,e);
      E.append(e);
      E.append(x);
      e = next;
      next = next_face_edge(e);
     }

    compute_center(G,e);
    E.append(e);
    E.append(next);

    forall(e,E) draw_voro_edge(G,e);

    W.draw_filled_node(q,red);
    get_release_event();
    W.draw_filled_node(q,red);

    delaunay_flipping(G,E);

    animate = anim_save;
}



static GRAPH<point,edge>* gp;

void  redraw()
{ node v;
  forall_nodes(v,*gp) W.draw_filled_node((*gp)[v],black);
  draw_delaunay(*gp);
  draw_voro(*gp);
 }


inline void add_voro_edge(GRAPH<circle,edge>& VD, node v, node w)
{ if (v != w) VD.new_edge(v,w,edge(nil)); }


node dfs(const GRAPH<point,edge>& DT, node v, edge e, GRAPH<circle,edge>& VD,
                                              edge_array<node>& vnode)
{ edge e1 = DT[e];
 
  node w = vnode[e1];
 
  if (w == nil)
   { edge e2 = DT.cyclic_adj_pred(DT[e1]);
     edge e3 = DT.cyclic_adj_pred(DT[e2]);
     point a = DT[source(e1)];
     point b = DT[source(e2)];
     point c = DT[source(e3)];

     if (v && VD[v].contains(c)) 
        w = v;
     else
        w = VD.new_node(circle(a,b,c));
 
     vnode[e1] = vnode[e2] = vnode[e3] = w;
 
     add_voro_edge(VD,w,dfs(DT,w,e1,VD,vnode));
     add_voro_edge(VD,w,dfs(DT,w,e2,VD,vnode));
     add_voro_edge(VD,w,dfs(DT,w,e3,VD,vnode));
    }
 
  return w;
}
 
 
 
void build_voro1(const GRAPH<point,edge>& DT, GRAPH<circle,edge>& VD)
{ edge_array<node>  vnode(DT,nil);
  edge e = DT.first_edge();
  dfs(DT,nil,e,VD,vnode);
 }
 



int main()
{
   int  N = 50;
   bool grid = false;

   //W.button("about",   7);
   W.button("random",  1);
   W.button("triang",  2);
   W.button("voro",    3);
   W.button("both",    4);
   W.button("clear",   5);
   W.button("quit",    0);

   //W.bool_item("animate",animate);
   //W.bool_item("grid",grid);
   W.int_item("rand points",N,0,500);

   W.set_frame_label("Delaunay Triangulation");
   W.set_redraw(redraw);
   W.init(-1100,1100,-1100);
   W.open();

   W_mess.set_frame_label("Flipping");
   W_mess.set_text_mode(opaque);
   W_mess.open(0,0);


   list<point> L;
   GRAPH<point,edge> G;

   gp = &G;

   //vnode.init(G,nil_point);

if (grid) W.set_grid_mode(30);
W.set_node_width(3);
W.set_mode(xor_mode);

initialize(G,L);

e0 = G.first_edge();

for(;;)
{ 
  point p;

  int but = W.read_mouse(p);

  if (but == 0) break;

  switch (but) {

/*
  case 1: { random_source x_ran(int(0.85*W.xmin()),int(0.85*W.xmax()));
            random_source y_ran(int(0.85*W.ymin()),int(0.85*W.ymax()));
            W.clear();
            G.clear();
            for(int i=0; i<N; i++) 
            { int x,y;
              x_ran >> x;
              y_ran >> y;
              L.append(point(x,y));
             }
            point p;
            forall(p,L) W.draw_filled_node(p,black);
            TRIANGULATE_POINTS(L,G);
            build_voro(G);
            delaunay_flipping(G,G.all_edges());
            draw_delaunay(G);
            draw_voro(G);
            break;
           }
*/

  case 1: { random_source x_ran(int(0.85*W.xmin()),int(0.85*W.xmax()));
            random_source y_ran(int(0.85*W.ymin()),int(0.85*W.ymax()));
            for(int i=0; i<N; i++) 
            { int x,y;
              x_ran >> x;
              y_ran >> y;
              point p(x,y);
              insert(G,L,p);
             }
            break;
           }


  case 2: { W.clear();
            W.set_frame_label("Delaunay Triangulation");
            show_voronoi = false;
            show_delaunay = true;
            node v;
            forall_nodes(v,G) W.draw_filled_node(G[v],black);
            draw_delaunay(G);
            break;
           }

  case 3: { W.clear();
            W.set_frame_label("Voronoi Diagram");
            show_voronoi = true;
            show_delaunay = false;
            node v;
            forall_nodes(v,G) W.draw_filled_node(G[v],black);
            draw_voro(G);
            break;
           }

  case 4: { W.clear();
            W.set_frame_label("Delaunay and Voronoi Diagram");
            show_voronoi = true;
            show_delaunay = true;
            node v;
            forall_nodes(v,G) W.draw_filled_node(G[v],black);
            draw_voro(G);
            draw_delaunay(G);
            break;
           }

  case 5: { W.clear();
            initialize(G,L);
            break;
           }

  case MOUSE_BUTTON(1):
  { if (W.shift_key_down())
       remove(G,L,nearest_neighbor(G,p));
    else
       insert(G,L,p,true);
    break;
   }


   case MOUSE_BUTTON(3):  // display circle 
   { edge e = locate_point(G,p);
     draw_circle(G,e);
     get_release_event();
     draw_circle(G,e);
     break;
    }


   }

}

return 0;
}
