#include <LEDA/plane_alg.h>
#include <LEDA/window.h>
#include <LEDA/rat_ray.h>
#include <LEDA/rat_line.h>


static window* Wp;
static color site_color   = black;
static color voro_color   = blue;
static color triang_color = red;
static color hull_color   = yellow;
static int   display      = 0;

static char* label[] = {"VORONOI DIAGRAM",
                        "DELAUNAY TRIANG",
                        "DELAUNAY DIAGRAM",
                        "CONVEX HULL"};

enum { POINT, SEGMENT, CIRCLE, POLYGON };



list<rat_point> p_list;

void random_points(int n)
{ 
  int ymax = int(Wp->ymax()-10);

  while (n--)
  { integer x = rand_int(10,500);
    integer y = rand_int(10,ymax);
    rat_point p(x,y);
    Wp->draw_point(p.to_point(),site_color);
    p_list.append(p);
  }
}
   


void segment_points(segment s, int n)
{ 
  integer x  = n*s.xcoord1();
  integer y  = n*s.ycoord1();
  integer dx = s.xcoord2() - s.xcoord1();
  integer dy = s.ycoord2() - s.ycoord1();

  for(int i=0; i < n; i++)
  { rat_point p(x,y,n);
    p_list.append(p);
    Wp->draw_point(p.to_point(),site_color);
    x += dx;
    y += dy;
   }
}



void circle_points(circle C, int n)
{ 
  point p = C.point1();
  point q = C.point2();
  point r = C.point3();

  int px = int(p.xcoord());
  int py = int(p.ycoord());
  int qx = int(q.xcoord());
  int qy = int(q.ycoord());
  int rx = int(r.xcoord());
  int ry = int(r.ycoord());

  rat_circle R(rat_point(px,py,1), rat_point(qx,qy,1), rat_point(rx,ry,1));

  double d = (2*LEDA_PI)/n;
  double eps = 0.001;
  
  double a = 0;
  for(int i=0; i < n; i++)
  { rat_point q = R.point_on_circle(a,eps);
    p_list.append(q);
    Wp->draw_point(q.to_point(),site_color);
    a += d;
  }
}


void draw_voro()
{ 
  GRAPH<rat_circle,rat_point>  VD;

  VORONOI(p_list,VD);
 
  edge_array<edge> rev(VD,nil);
 
  if (!Is_Bidirected(VD,rev))
    error_handler(1,"VD not bidirected.") ;
 
  edge e;
  forall_edges(e,VD)
  { node v = source(e);
    node w = target(e);
 
    if (rev[e])
    { rev[rev[e]] = nil;
 
      if (VD.outdeg(v) == 1 && VD.outdeg(w) == 1)
        { rat_line  l = p_bisector(VD[v].point1(),VD[v].point3());
          Wp->draw_line(l.to_line(),blue);
         }
      else
        if (VD.outdeg(v) == 1)
           { point cw = VD[w].center().to_point();
             point p1 = VD[v].point1().to_point();
             point p3 = VD[v].point3().to_point();
             vector vec = p3 - p1;
             point cv  = cw + vec.rotate90();
             Wp->draw_ray(cw,cv,blue);
            }
         else
           if (VD.outdeg(w) == 1)
           { point cv = VD[v].center().to_point();
             point p1 = VD[w].point1().to_point();
             point p3 = VD[w].point3().to_point();
             vector vec = p3 - p1;
             point cw  = cv + vec.rotate90();
             Wp->draw_ray(cv,cw,blue);
             }
           else
             { rat_point cv = VD[v].center();
               rat_point cw = VD[w].center();
               Wp->draw_segment(cv.to_point(),cw.to_point(),blue);
              }
 
     }
  }
}
 


void draw_edges(GRAPH<rat_point,edge>&  DT)
{ edge_array<edge> rev(DT,nil);
  if (!Is_Bidirected(DT,rev)) error_handler(1,"DT not bidirected.");
  edge e;
  forall_edges(e,DT) 
    if (DT[e])
    { DT[DT[e]] = nil;
      Wp->draw_edge(DT[source(e)].to_point(),DT[target(e)].to_point(),green2);
     }
 }



void draw_delau_triang()
{ GRAPH<rat_point,edge>  DT;
  TRIANGULATE_POINTS(p_list,DT);
  DELAUNAY_FLIPPING(DT);
  draw_edges(DT);
}


void draw_delau_diagram()
{ GRAPH<rat_point,edge>  DT;
  DELAUNAY_DIAGRAM(p_list,DT);
  draw_edges(DT);
 }



void draw_convex_hull()
{ GRAPH<rat_point,edge>  DT;
  TRIANGULATE_POINTS(p_list,DT);
  list<point> CH;
  edge e = DT.last_edge();
  if (e == nil) return;
  edge x = e;
  do { CH.append(DT[source(x)].to_point());
       x = DT.cyclic_adj_pred(DT[x]);
     } while (x != e);

  Wp->draw_filled_polygon(CH,hull_color);
  Wp->draw_polygon(CH,black);
 }



void get_input(window& W, int inp)
{
   switch (inp) {

    case POINT: 
         { point p;
           if (W >> p) W.draw_point(p);
           integer x = p.xcoord();
           integer y = p.ycoord();
           p_list.append(rat_point(x,y,1));
           break;
          }
  
   case SEGMENT: 
        { segment s;
          if (W >> s) segment_points(s,int(s.length()/5)); 
          break;
         }

   case CIRCLE:
        { circle c;
          if (W >> c) circle_points(c,int(c.radius()));
          break;
         }

   case POLYGON:
        { polygon P;
          if (W >> P) 
          { list<segment> L = P.edges();
            segment s;
            forall(s,L) segment_points(s,int(s.length()/4));
           }
          break;
         }
   }
}




void draw()
{ segment s;
  edge e;

  if (p_list.empty()) return;

  Wp->clear();

  rat_point p;
  forall(p,p_list) Wp->draw_point(p.to_point(),site_color);

  switch(display) {

  case 0: draw_voro();
          break;

  case 1: draw_delau_triang();
          break;

  case 2: draw_delau_diagram();
          break;

  case 3: draw_convex_hull();
          break;
  }
}





void set_display(int i) 
{ display = i; 
  Wp->set_frame_label(label[i]);
 }


int main()
{

  menu disp_menu;
  disp_menu.button("voronoi  diagram", 0, set_display);
  disp_menu.button("delaunay triang",  1, set_display);
  disp_menu.button("delaunay diagram", 2, set_display);
  disp_menu.button("convex   hull",    3, set_display);

  int input = POINT;

  window W("DYNAMIC VORONOI DIAGRAMS");

  Wp = &W;

  //W.buttons_per_line(4);

  W.choice_item("input",input,"point","segment","circle","polygon");

  W.button("display",  4,disp_menu);
  W.button("random",   1);
  W.button("settings", 2);
  W.button("clear",    3);
  W.button("quit",     0);


  W.set_frame_label(label[display]);
  W.display();

  W.init(0,512,0);
  W.set_redraw(draw);
  W.set_node_width(3);

  int N = 500;
  int grid_width = 0;
  int line_width = 0;

  panel SP("SETTINGS");

  SP.int_item("rand sites",N,1,1000);
  SP.int_item("grid width", grid_width,0,40);
  SP.int_item("line width", line_width,1,4);

  SP.color_item("sites ", site_color);
  SP.color_item("voro  ", voro_color);
  SP.color_item("triang", triang_color);
  SP.color_item("hull",   hull_color);
  SP.button("continue");


  for(;;)
  {
    point p;
    int but = W.read_mouse(p);
  
    if (but == 0) break;
  
    switch (but) {

    case MOUSE_BUTTON(1):
            { put_back_event(); 
              get_input(W,input);
              break;
             }

    case 1: { random_points(N);
              break;
             }
   

    case 2: { SP.open(W);
              W.set_grid_mode(grid_width);
              W.set_line_width(line_width);
              break;
             }

    case 3: { W.clear();
              p_list.clear();
              break;
             }


    } // switch

   draw();

  }

  return 0;
}
