#include <LEDA/plane_alg.h>
#include <LEDA/window.h>


window* Wp;

int ymax;


int segment_to_points(segment s, list<point>& out, double dist)
{ 
  int n = int(s.length()/dist) + 1;

  double x  = s.xcoord1();
  double y  = s.ycoord1();
  double dx = s.dx()/n;
  double dy = s.dy()/n;

  for(int i = 0; i<n; i++)
  { out.append(point(x,y));
    x += dx;
    y += dy;
   }

  return n;

}

int circle_to_points(circle C, list<point>& out, double dist)
{ 
  point c = C.center();
  double r = C.radius();
  int n = int(6.283 * r/dist);

  double alpha = 0;
  double d = 6.283/n;

  for(int i = 0; i<n; i++)
  { out.append(c.translate(alpha,r+double(rand_int(-100,100))/1000000));
    alpha += d;
   }

  out.permute();

 return n;

}



static color site_color   = black;
static color voro_color   = blue;
static color triang_color = red;
static color hull_color   = yellow;


list<point> p_list;

bool vor = false;
bool triang = false;

int display = 0;

char* label[] = {"VORONOI DIAGRAM",
                 "TRIANGULATION", 
                 "DELAUNAY TRIANGULATION",
                 "DELAUNAY DIAGRAM",
                 "CONVEX HULL"};


void insert_point(point p)
{ Wp->draw_point(p,site_color);
  p_list.append(p);
 }

void insert_segment(segment s)
{ list<point> pl;
  segment_to_points(s,pl,10/Wp->scale());
  point p;
  forall(p,pl) insert_point(p);
}

void insert_circle(circle c)
{ list<point> pl;
  circle_to_points(c,pl,10/Wp->scale());
  point p;
  forall(p,pl) insert_point(p);
}
            

void draw_sites()
{ point p;
  forall(p,p_list) Wp->draw_point(p,site_color);
}


void draw_voro_edges()
{ GRAPH<circle,point>  VD;

  VORONOI(p_list,VD);

  edge_array<edge> rev(VD,nil);
  if (!Is_Bidirected(VD,rev))  error_handler(1,"VD not bidirected.") ;

  edge e;
  forall_edges(e,VD) 
  { node v = source(e);
    node w = target(e);
    if (rev[e])
    { rev[rev[e]] = nil;

      if (VD.outdeg(v) == 1 && VD.outdeg(w) == 1)
        { line  l = p_bisector(VD[v].point1(),VD[v].point3());
          Wp->draw_line(l,voro_color);
         }
      else
        if (VD.outdeg(v) == 1) 
           { point  cw  = VD[w].center();
             vector vec = VD[v].point3() - VD[v].point1();
             point  cv  = cw + vec.rotate90();
             Wp->draw_ray(cw,cv,voro_color);
            }
         else
           if (VD.outdeg(w) == 1) 
            { point  cv  = VD[v].center();
              vector vec = VD[w].point3() - VD[w].point1();
              point  cw  = cv + vec.rotate90();
              Wp->draw_ray(cv,cw,voro_color);
             }
           else
              Wp->draw_segment(VD[v].center(),VD[w].center(),voro_color);
              
     }
  }
}

void draw_triang_edges()
{ GRAPH<point,edge>  DT;
  TRIANGULATE_POINTS(p_list,DT);

  edge_array<edge> rev(DT,nil);
  if (!Is_Bidirected(DT,rev)) error_handler(1,"DT not bidirected.");
  edge e;
  forall_edges(e,DT) 
    if (DT[e])
    { DT[DT[e]] = nil;
      Wp->draw_edge(DT[source(e)],DT[target(e)],green2);
     }
}


void draw_delaunay_triang()
{ GRAPH<point,edge>  DT;
  TRIANGULATE_POINTS(p_list,DT);
  DELAUNAY_FLIPPING(DT);

  edge_array<edge> rev(DT,nil);
  if (!Is_Bidirected(DT,rev)) error_handler(1,"DT not bidirected.");
  edge e;
  forall_edges(e,DT) 
    if (DT[e])
    { DT[DT[e]] = nil;
      Wp->draw_edge(DT[source(e)],DT[target(e)],triang_color);
     }
}


void draw_delaunay_diagram()
{ GRAPH<point,edge>  DT;
  DELAUNAY_DIAGRAM(p_list,DT);

  edge_array<edge> rev(DT,nil);
  if (!Is_Bidirected(DT,rev)) error_handler(1,"DT not bidirected.");
  edge e;
  forall_edges(e,DT) 
    if (DT[e])
    { DT[DT[e]] = nil;
      Wp->draw_edge(DT[source(e)],DT[target(e)],triang_color);
     }

 }



void draw_convex_hull()
{ GRAPH<point,edge>  DT;
  TRIANGULATE_POINTS(p_list,DT);
  list<point> CH;
  edge e = DT.last_edge();
  if (e == nil) return;
  edge x = e;
  do { CH.append(DT[source(x)]);
       x = DT.cyclic_adj_pred(DT[x]);
     } while (x != e);

  Wp->draw_filled_polygon(CH,hull_color);
  Wp->draw_polygon(CH,black);
 }



void draw()
{ segment s;
  edge e;

  if (p_list.empty()) return;

  draw_sites();

  switch(display) {

  case 0: draw_voro_edges();
          break;

  case 1: draw_delaunay_triang();
          break;

  case 2: draw_delaunay_diagram();
          break;

  case 3: draw_convex_hull();
          break;
  }
}





void set_display(int i) 
{ display = i; 
  Wp->set_frame_label(label[i]);
 }



int main()
{

  menu disp_menu;
  disp_menu.button("voronoi  diagram", 0, set_display);
  disp_menu.button("delaunay triang",  1, set_display);
  disp_menu.button("delaunay diagram", 2, set_display);
  disp_menu.button("convex   hull",    3, set_display);

  window W("DYNAMIC VORONOI DIAGRAMS");

  Wp = &W;

  W.buttons_per_line(5);

  //W.button("point",    0);
  W.button("segment",  1);
  W.button("circle",   2);
  W.button("random",   3);
  W.button("delete",   4);
  W.button("neighbor", 5);
  W.button("display",  8,disp_menu);
  W.button("settings", 6);
  W.button("clear",    7);
  W.button("quit",     9);


  W.set_frame_label(label[display]);
  W.display();

  W.init(0,512,0);
  //W.set_flush(false);
  W.set_redraw(draw);
  W.set_mode(xor_mode);
  W.set_node_width(3);

  ymax = int(W.ymax()-10);

  int N = 500;
  int grid_width = 0;
  int line_width = 0;

  panel SP("SETTINGS");
  SP.int_item("rand sites",N,1,1000);
  SP.int_item("grid width", grid_width,0,40);
  SP.int_item("line width", line_width,1,4);
  SP.color_item("sites ", site_color);
  SP.color_item("voro  ", voro_color);
  SP.color_item("triang", triang_color);
  SP.color_item("hull",   hull_color);
  SP.button("continue");

for(;;)
{
  point p;
  int but = W.read_mouse(p);

  if (but == MOUSE_BUTTON(1))
     insert_point(p);
  else
  { if (but < 0 || but > 9) continue;
    if (but == 9) break;

    switch (but) {

            
    case 1: { segment s;
              W >> s; 
              insert_segment(s);
              break;
             }
        
    case 2: { circle c;
              W >> c; 
              insert_circle(c);
              break;
             }

    case 3: { // random points
              for(int i=0; i<N; i++)
                 insert_point(point(rand_int(10,500),rand_int(10,ymax)));
              break;
             }
   

    case 6: { // settings
              SP.open(W);
              W.set_grid_mode(grid_width);
              W.set_line_width(line_width);
              W.set_mode(xor_mode);
              break;
             }

    case 7: { // clear
              W.clear();
              p_list.clear();
             }


   }  // switch
  }


  if (p_list.size() > 1) 
  { W.clear();
    draw();
   }

 }

 return 0;

}
