/*******************************************************************************
+
+  LEDA 3.4
+
+  _random.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <LEDA/random.h>
#include <time.h>
#include <stdio.h>

//------------------------------------------------------------------------------
// We use BSD Code for get_random (random), set_random (srandom), ...
// modified by S. Naeher  (1995 & 1996)
//------------------------------------------------------------------------------

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */



/*
 * random.c:
 *
 * An improved random number generation package.  In addition to the standard
 * rand()/srand() like interface, this package also has a special state info
 * interface.  The initstate() routine is called with a seed, an array of
 * bytes, and a count of how many bytes are being passed in; this array is
 * then initialized to contain information for random number generation with
 * that much state information.  Good sizes for the amount of state
 * information are 32, 64, 128, and 256 bytes.  The state can be switched by
 * calling the setstate() routine with the same array as was initiallized
 * with initstate().  By default, the package runs with 128 bytes of state
 * information and generates far better random numbers than a linear
 * congruential generator.  If the amount of state information is less than
 * 32 bytes, a simple linear congruential R.N.G. is used.
 *
 * Internally, the state information is treated as an array of longs; the
 * zeroeth element of the array is the type of R.N.G. being used (small
 * integer); the remainder of the array is the state information for the
 * R.N.G.  Thus, 32 bytes of state information will give 7 longs worth of
 * state information, which will allow a degree seven polynomial.  (Note:
 * the zeroeth word of state information also has some other information
 * stored in it -- see setstate() for details).
 * 
 * The random number generation technique is a linear feedback shift register
 * approach, employing trinomials (since there are fewer terms to sum up that
 * way).  In this approach, the least significant bit of all the numbers in
 * the state table will act as a linear feedback shift register, and will
 * have period 2^deg - 1 (where deg is the degree of the polynomial being
 * used, assuming that the polynomial is irreducible and primitive).  The
 * higher order bits will have longer periods, since their values are also
 * influenced by pseudo-random carries out of the lower bits.  The total
 * period of the generator is approximately deg*(2**deg - 1); thus doubling
 * the amount of state information has a vast influence on the period of the
 * generator.  Note: the deg*(2**deg - 1) is an approximation only good for
 * large deg, when the period of the shift register is the dominant factor.
 * With deg equal to seven, the period is actually much longer than the
 * 7*(2**7 - 1) predicted by this formula.
 */

/*
 * For each of the currently supported random number generators, we have a
 * break value on the amount of state information (you need at least this
 * many bytes of state info to support this random number generator), a degree
 * for the polynomial (actually a trinomial) that the R.N.G. is based on, and
 * the separation between the two lower order coefficients of the trinomial.
 */
#define	TYPE_0		0		/* linear congruential */
#define	BREAK_0		8
#define	DEG_0		0
#define	SEP_0		0

#define	TYPE_1		1		/* x**7 + x**3 + 1 */
#define	BREAK_1		32
#define	DEG_1		7
#define	SEP_1		3

#define	TYPE_2		2		/* x**15 + x + 1 */
#define	BREAK_2		64
#define	DEG_2		15
#define	SEP_2		1

#define	TYPE_3		3		/* x**31 + x**3 + 1 */
#define	BREAK_3		128
#define	DEG_3		31
#define	SEP_3		3

#define	TYPE_4		4		/* x**63 + x + 1 */
#define	BREAK_4		256
#define	DEG_4		63
#define	SEP_4		1

/*
 * Array versions of the above information to make code run faster --
 * relies on fact that TYPE_i == i.
 */
#define	MAX_TYPES	5		/* max number of types above */

//static int degrees[MAX_TYPES] = { DEG_0, DEG_1, DEG_2, DEG_3, DEG_4 };
//static int seps [MAX_TYPES] =	{ SEP_0, SEP_1, SEP_2, SEP_3, SEP_4 };

/*
 * Initially, everything is set up as if from:
 *
 *	initstate(1, &randtbl, 128);
 *
 * Note that this initialization takes advantage of the fact that set_random()
 * advances the front and rear pointers 10*rand_deg times, and hence the
 * rear pointer which starts at 0 will also end up at zero; thus the zeroeth
 * element of the state information, which contains info about the current
 * position of the rear pointer is just
 *
 *	MAX_TYPES * (rptr - state) + TYPE_3 == TYPE_3.
 */

//static long randtbl[DEG_3 + 1] = {
//	TYPE_3,
//	0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342, 0xde3b81e0, 0xdf0a6fb5,
//	0xf103bc02, 0x48f340fb, 0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
//	0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86, 0xda672e2a, 0x1588ca88,
//	0xe369735d, 0x904f35f7, 0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
//	0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b, 0xf5ad9d0e, 0x8999220b,
//	0x27fb47b9,
//};



/*
 * fptr and rptr are two pointers into the state info, a front and a rear
 * pointer.  These two pointers are always rand_sep places aparts, as they
 * cycle cyclically through the state information.  (Yes, this does mean we
 * could get away with just one pointer, but the code for get_random() is more
 * efficient this way).  The pointers are left positioned as they would be
 * from the call
 *
 *	initstate(1, randtbl, 128);
 *
 * (The position of the rear pointer, rptr, is really 0 (as explained above
 * in the initialization of randtbl) because the state table pointer is set
 * to point to randtbl[1] (as explained below).
 */

//long *fptr;
//long *rptr;

/*
 * The following things are the pointer to the state information table, the
 * type of the current generator, the degree of the current polynomial being
 * used, and the separation between the two pointers.  Note that for efficiency
 * of get_random(), we remember the first location of the state information, not
 * the zeroeth.  Hence it is valid to access state[-1], which is used to
 * store the type of the R.N.G.  Also, we remember the last location, since
 * this is more efficient than indexing every time to find the address of
 * the last element to see if the front and rear pointers have wrapped.
 */

//long* state;
//int   rand_type;
//int   rand_deg;
//int   rand_sep;
//long* end_ptr;


bsd_random::bsd_random()
{ 
  degrees = new int[MAX_TYPES];
  seps  = new int[MAX_TYPES];
  randtbl = new long[DEG_3 + 1];

  int i = 0;
  randtbl[i++] = TYPE_3;
  randtbl[i++] = 0x9a319039; 
  randtbl[i++] = 0x32d9c024; 
  randtbl[i++] = 0x9b663182; 
  randtbl[i++] = 0x5da1f342; 
  randtbl[i++] = 0xde3b81e0; 
  randtbl[i++] = 0xdf0a6fb5;
  randtbl[i++] = 0xf103bc02; 
  randtbl[i++] = 0x48f340fb; 
  randtbl[i++] = 0x7449e56b; 
  randtbl[i++] = 0xbeb1dbb0; 
  randtbl[i++] = 0xab5c5918; 
  randtbl[i++] = 0x946554fd;
  randtbl[i++] = 0x8c2e680f; 
  randtbl[i++] = 0xeb3d799f; 
  randtbl[i++] = 0xb11ee0b7; 
  randtbl[i++] = 0x2d436b86; 
  randtbl[i++] = 0xda672e2a; 
  randtbl[i++] = 0x1588ca88;
  randtbl[i++] = 0xe369735d; 
  randtbl[i++] = 0x904f35f7; 
  randtbl[i++] = 0xd7158fd6; 
  randtbl[i++] = 0x6fa6f051; 
  randtbl[i++] = 0x616e6b96; 
  randtbl[i++] = 0xac94efdc;
  randtbl[i++] = 0x36413f93; 
  randtbl[i++] = 0xc622c298; 
  randtbl[i++] = 0xf5a42ab8; 
  randtbl[i++] = 0x8a88d77b; 
  randtbl[i++] = 0xf5ad9d0e; 
  randtbl[i++] = 0x8999220b;
  randtbl[i++] = 0x27fb47b9;

  degrees[0] = DEG_0;
  degrees[1] = DEG_1;
  degrees[2] = DEG_2;
  degrees[3] = DEG_3;
  degrees[4] = DEG_4;
 
  seps[0] = SEP_0;
  seps[1] = SEP_1;
  seps[2] = SEP_2;
  seps[3] = SEP_3;
  seps[4] = SEP_4;
 
  fptr = &randtbl[SEP_3 + 1];
  rptr = &randtbl[1];
  state = &randtbl[1];
  rand_type = TYPE_3;
  rand_deg = DEG_3;
  rand_sep = SEP_3;
  end_ptr = &randtbl[DEG_3 + 1];
}

bsd_random::~bsd_random()
{ delete[] degrees;
  delete[] seps;
  delete[] randtbl;
}



/*
 * set_random:
 *
 * Initialize the random number generator based on the given seed.  If the
 * type is the trivial no-state-information type, just remember the seed.
 * Otherwise, initializes state[] based on the given "seed" via a linear
 * congruential generator.  Then, the pointers are set to known locations
 * that are exactly rand_sep places apart.  Lastly, it cycles the state
 * information a given number of times to get rid of any initial dependencies
 * introduced by the L.C.R.N.G.  Note that the initialization of randtbl[]
 * for default usage relies on values produced by this routine.
 */


void bsd_random::set_random(unsigned x)
{
	int i;
	state[0] = x;

	if (rand_type != TYPE_0)
	{  	state[0] = x;
		for (i = 1; i < rand_deg; i++)
			state[i] = 1103515245 * state[i - 1] + 12345;
		fptr = &state[rand_sep];
		rptr = &state[0];
		for (i = 0; i < 10 * rand_deg; i++) get_random();
	}
}



/*
 * initstate:
 *
 * Initialize the state information in the given array of n bytes for future
 * random number generation.  Based on the number of bytes we are given, and
 * the break values for the different R.N.G.'s, we choose the best (largest)
 * one we can and set things up for it.  set_random() is then called to
 * initialize the state information.
 * 
 * Note that on return from set_random(), we set state[-1] to be the type
 * multiplexed with the current value of the rear pointer; this is so
 * successive calls to initstate() won't lose this information and will be
 * able to restart with setstate().
 * 
 * Note: the first thing we do is save the current state, if any, just like
 * setstate() so that it doesn't matter when initstate is called.
 *
 * Returns a pointer to the old state.
 */

char* bsd_random::initstate(unsigned seed, char* arg_state, int n)
{
	register char *ostate = (char *)(&state[-1]);

	if (rand_type == TYPE_0)
		state[-1] = rand_type;
	else
		state[-1] = MAX_TYPES * (rptr - state) + rand_type;
	if (n < BREAK_0) {
		(void)fprintf(stderr,
		    "random: not enough state (%d bytes); ignored.\n", n);
		return 0;
	}
	if (n < BREAK_1) {
		rand_type = TYPE_0;
		rand_deg = DEG_0;
		rand_sep = SEP_0;
	} else if (n < BREAK_2) {
		rand_type = TYPE_1;
		rand_deg = DEG_1;
		rand_sep = SEP_1;
	} else if (n < BREAK_3) {
		rand_type = TYPE_2;
		rand_deg = DEG_2;
		rand_sep = SEP_2;
	} else if (n < BREAK_4) {
		rand_type = TYPE_3;
		rand_deg = DEG_3;
		rand_sep = SEP_3;
	} else {
		rand_type = TYPE_4;
		rand_deg = DEG_4;
		rand_sep = SEP_4;
	}
	state = &(((long *)arg_state)[1]);	/* first location */
	end_ptr = &state[rand_deg];	/* must set end_ptr before set_random */
	set_random(seed);
	if (rand_type == TYPE_0)
		state[-1] = rand_type;
	else
		state[-1] = MAX_TYPES*(rptr - state) + rand_type;
	return ostate;
}

/*
 * setstate:
 *
 * Restore the state from the given state array.
 *
 * Note: it is important that we also remember the locations of the pointers
 * in the current state information, and restore the locations of the pointers
 * from the old state information.  This is done by multiplexing the pointer
 * location into the zeroeth word of the state information.
 *
 * Note that due to the order in which things are done, it is OK to call
 * setstate() with the same state as the current state.
 *
 * Returns a pointer to the old state information.
 */

char* bsd_random::setstate(char* arg_state)
{
	register long *new_state = (long *)arg_state;
	register int type = int(new_state[0] % MAX_TYPES);
	register int rear = int(new_state[0] / MAX_TYPES);
	char *ostate = (char *)(&state[-1]);

	if (rand_type == TYPE_0)
		state[-1] = rand_type;
	else
		state[-1] = MAX_TYPES * (rptr - state) + rand_type;
	switch(type) {
	case TYPE_0:
	case TYPE_1:
	case TYPE_2:
	case TYPE_3:
	case TYPE_4:
		rand_type = type;
		rand_deg = degrees[type];
		rand_sep = seps[type];
		break;
	default:
		(void)fprintf(stderr,
		    "random: state info corrupted; not changed.\n");
	}
	state = &new_state[1];
	if (rand_type != TYPE_0) {
		rptr = &state[rear];
		fptr = &state[(rear + rand_sep) % rand_deg];
	}
	end_ptr = &state[rand_deg];		/* set end_ptr too */
	return ostate;
}


/*
 * random:
 *
 * If we are using the trivial TYPE_0 R.N.G., just do the old linear
 * congruential bit.  Otherwise, we do our fancy trinomial stuff, which is
 * the same in all the other cases due to all the global variables that have
 * been set up.  The basic operation is to add the number at the rear pointer
 * into the one at the front pointer.  Then both pointers are advanced to
 * the next location cyclically in the table.  The value returned is the sum
 * generated, reduced to 31 bits by throwing away the "least random" low bit.
 *
 * Note: the code takes advantage of the fact that both the front and
 * rear pointers can't wrap on the same call by not testing the rear
 * pointer if the front one has wrapped.
 *
 * Returns a 31-bit random number.
 */

unsigned long bsd_random::get_random()
{
	long i;

	if (rand_type == TYPE_0)
		i = state[0] = (state[0] * 1103515245 + 12345) & 0x7fffffff;
	else {
		*fptr += *rptr;
		i = (*fptr >> 1) & 0x7fffffff;	/* chucking least random bit */
		if (++fptr >= end_ptr) {
			fptr = state;
			++rptr;
		} else if (++rptr >= end_ptr)
			rptr = state;
	}
	return (unsigned long)i;
}




//------------------------------------------------------------------------------


//we assume that bsd_random::get_random() produces a 31 bit integer

#define RANDMAX 0x7FFFFFFF

mutex random_source::random_mutex_count;
int random_source::count = 0;

/* we have:
   the source is in bit mode iff bit_mode == true
   in bit mode: pat = 2^{p+1} - 1 where p is the precision.
   in integer mode: diff = high - low and 
                    if diff > 0 then pat = 2^p - 1 where 
                         p = 1 + \lfloor \log diff \rfloor
                    if diff = 0 then pat = 0
*/


random_source::random_source()
{ time_t seed;
  time(&seed);
  random_mutex_count.lock();
  count++;
  set_random(int(seed)*count);
  random_mutex_count.unlock();
  pat = 0xFFFFFFFF; 
  bit_mode = true;
  low = diff = 0;
 }

random_source::random_source(int bits) 
{ time_t seed;
  time(&seed);
  random_mutex_count.lock();
  count++;
  set_random(int(seed)*count);
  random_mutex_count.unlock();
  if (bits <= 0 || bits >= 32)
     pat = 0xFFFFFFFF; 
  else
     pat = (1 << bits) - 1;
  bit_mode = true;
  low = diff = 0;
 }

void random_source::set_range(int l, int h)
{ low  = l;
  diff = (h < l) ? 0 : h-l;
  bit_mode = false;
  pat = 1;
  while (pat <= unsigned(diff)) pat <<= 1;
  pat--;
 }


random_source::random_source(int l, int h) 
{ time_t seed;
  time(&seed);
  random_mutex_count.lock();
  count++;
  set_random(int(seed)*count);
  random_mutex_count.unlock();
  set_range(l,h);
   }


int random_source::set_precision(int bits) 
{ int oldprec = prec;
  prec = bits;
  if (0 >= bits || bits >= 32)
     pat = 0xFFFFFFFF; 
  else
     pat = (1 << bits) - 1;
  bit_mode = true;
  return oldprec;
 }


int random_source::operator()()
{ unsigned long x =  get_random() & pat;
  if (bit_mode)
     return int(x);
  else
    { while (x > unsigned(diff)) x = get_random() & pat;
      return (int)(low + x);
     }
 }

int  random_source::operator()(int bits)
{ return int(get_random() % ((1<<bits)-1)); }


int  random_source::operator()(int l, int h)
{ if (h < l) return 0;

  int  old_low = low;
  int  old_diff = diff;
  unsigned long  old_pat = pat;
  bool old_bit_mode = bit_mode;
  
  set_range(l,h);
  int x  = operator()();

  low = old_low; 
  diff = old_diff; 
  pat = old_pat; 
  bit_mode = old_bit_mode;

  return x; 
}



random_source& random_source::operator>>(char& x)
{ x = char(operator()());
  return *this;
 }

random_source& random_source::operator>>(int& x)      
{ x = int(operator()());
  return *this;
 }

random_source& random_source::operator>>(long& x)      
{ x = long(operator()());
  return *this;
 }


random_source& random_source::operator>>(unsigned char& x)
{ x = (unsigned char)(operator()());
  return *this;
 }

random_source& random_source::operator>>(unsigned int& x)
{ x = (unsigned int)(operator()());
  return *this;
 }

random_source& random_source::operator>>(unsigned long& x)
{ x = (unsigned long)(operator()());
  return *this;
 }


#if defined(__BUILTIN_BOOL__)
random_source& random_source::operator>>(bool& x)
{ x = get_random() & 1; 
  return *this;
 }
#endif


random_source& random_source::operator>>(float&  x)
{ x = float(get_random())/RANDMAX; 
  return *this;
 }

random_source& random_source::operator>>(double& x)
{ x = float(get_random())/RANDMAX; 
  return *this;
 }


// default random source

/*  moved to global.c
 *
 * random_source rand_int;
 *
 */

