/*******************************************************************************
+
+  LEDA 3.4
+
+  _g_map.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <LEDA/graph.h>

// reversal edges

edge graph::face_cycle_succ(edge e) const 
{ return cyclic_adj_pred(reversal(e)); }

edge graph::face_cycle_pred(edge e) const 
{ return reversal(cyclic_adj_succ(e)); }

edge graph::succ_face_edge(edge e) const 
{ return cyclic_adj_pred(reversal(e)); }

edge graph::pred_face_edge(edge e) const 
{ return reversal(cyclic_adj_succ(e)); }


void graph::set_reversal(edge e, edge r)
{ if ( source(e) != target(r) || target(e) != source(r) )
  error_handler(1,"graph::set_reversal: edges are not reversals of each other");
  e->rev = r;
  r->rev = e;
}


static int edge_ord1(const edge& e) { return index(source(e)); }
static int edge_ord2(const edge& e) { return index(target(e)); }

bool graph::make_map()     
{
 // computes for every edge e = (v,w) in G its reversal r = (w,v)
 // in G ( nil if not present). Returns true if every edge has a
 // reversal and false otherwise.

  int n     = max_node_index();
  int count = 0;

  edge e;
  forall(e,e_list) e->rev = 0;

  list<edge> El = e_list;
  El.bucket_sort(0,n,&edge_ord2);
  El.bucket_sort(0,n,&edge_ord1);
  
  list<edge> El1 = e_list;
  El1.bucket_sort(0,n,&edge_ord1);
  El1.bucket_sort(0,n,&edge_ord2);


  // merge El and El1 to find corresponding edges

  while (! El.empty() && ! El1.empty())
  { edge e = El.head();
    edge r = El1.head();
    if (target(r) == source(e))
      if (source(r) == target(e))
         { e->rev = r;
           El1.pop();
           El.pop();
           count++;
          }
      else
         if (index(source(r)) < index(target(e)))
             El1.pop();
         else
             El.pop();

    else
      if (index(target(r)) < index(source(e)))
          El1.pop();
      else
          El.pop();

   }

  return (count == e_list.length()) ? true : false;
}


extern bool PLANAR(graph&, bool=false);

void graph::make_planar_map()
{ if (!graph::make_map())
       error_handler(1,"planar_map: graph is not bidirected");
  if (!PLANAR(*this,true)) 
       error_handler(1,"planar_map: Graph is not planar."); 
  compute_faces();
}



face  graph::new_face(GenPtr i) 
{ copy_face_entry(i);
  return add_face(i);
 }

face  graph::new_face() 
{ GenPtr i = 0;
  init_node_entry(i);
  return add_face(i);
 }


list<edge> graph::adj_edges(face f) const
{ list<edge> result(f->head);
  edge e1 = face_cycle_succ(f->head);
  while (e1!=f->head)
  { result.append(e1);
    e1 = face_cycle_succ(e1);
   }
  return result;
 }

list<node> graph::adj_nodes(face f) const
{ list<node> result(source(f->head));
  edge e1 = face_cycle_succ(f->head);
  while (e1!=f->head)
  { result.append(source(e1));
    e1 = face_cycle_succ(e1);
   }
  return result;
 }

list<face> graph::adj_faces(node v) const
{ list<face> result;
  edge e;
  forall_out_edges(e,v) result.append(adj_face(e));
  return result;
 }


void graph::print_face(face f) const
{ cout << string("F[%2d]",index(f));
  cout << "(";
  print_face_entry(cout,f->data[0]);
  cout << "): ";
  edge e;
  forall_face_edges(e,f) 
   cout << string("[%2d]",index(target(e)));
 }




void graph::compute_faces()
{ 
  edge e;

  forall_edges(e,*this) 
  { if (e->rev == nil) 
      error_handler(1,"graph::compute_faces: no map (reversal edge missing)");
    e->fa = nil;
   }

  forall_edges(e,*this)
  { if (e->fa) continue;
    face f = new_face();
    f->head = e;
    edge e1 = e;
    do { e1->fa = f;
         e1 = face_cycle_succ(e1);
       } while (e1 != e);
   } 
} 




