/*******************************************************************************
+
+  LEDA 3.4
+
+  _g_misc.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <LEDA/graph.h>
#include <LEDA/ugraph.h>

//------------------------------------------------------------------------------
// S. Naeher  (1995)
//------------------------------------------------------------------------------

node_array<int>* num_ptr;
  
static int source_num(const edge& e) { return (*num_ptr)[source(e)]; }
static int target_num(const edge& e) { return (*num_ptr)[target(e)]; }


bool Is_Simple(graph& G)  
{ 
  // return true iff G is simple, i.e, has no parallel edges
 
  list<edge> el= G.all_edges();

  node v;
  
  edge e;
  int n= 0;
  
  node_array<int> num(G);
  forall_nodes(v,G) num[v]= n++;
  
  num_ptr= &num;
  
  el.bucket_sort(0,n-1,&source_num);
  el.bucket_sort(0,n-1,&target_num);
  
  edge e0 = el.pop();
  
  forall(e,el)
  { if ( source(e0) == source(e) && target(e0) == target(e) )
       return false;
    else
       e0 = e;
   }

  return true;
  
}
  
  
list<node> Delete_Loops(graph& G)
{ list<edge> loops;
  list<node> L;
  edge e;
  forall_edges(e,G)
  { node v = source(e);
    node w = target(e);
    if (v == w) 
    { L.append(v);
      loops.append(e);
     }
   }
  forall(e,loops) G.del_edge(e);
  return L;
 }


list<edge> Make_Simple(graph& G)
{ 
  list<edge> L;

  //use bucket sort to find and eliminate parallel edges
  
  list<edge> el = G.all_edges();

  node_array<int> num(G);
  int  n = 0;
  node v;
  forall_nodes(v,G) num[v] = n++;
  
  num_ptr = &num;

  el.bucket_sort(0,n-1,&source_num);
  el.bucket_sort(0,n-1,&target_num);
  
  bool deleted = false;
  edge e0 = el.pop();

  edge e;
  forall(e,el)  
    if (source(e0) == source(e) && target(e0) == target(e)) 
     { G.del_edge(e);
       if (!deleted) L.append(e0);
       deleted = true;
      }
    else 
     { deleted = false;
       e0 = e;
      }

  return L;
  
 }




static int edge_ord1(const edge& e) { return index(source(e)); }
static int edge_ord2(const edge& e) { return index(target(e)); }

bool Is_Bidirected(const graph& G, edge_array<edge>& reversal)     
{
 // computes for every edge e = (v,w) in G its reversal reversal[e] = (w,v)
 // in G ( nil if not present). Returns true if every edge has a
 // reversal and false otherwise.

  int n = G.max_node_index();
  int count = 0;

  edge e,r;

  forall_edges(e,G) reversal[e] = 0;

  list<edge> El = G.all_edges();
  El.bucket_sort(0,n,&edge_ord2);
  El.bucket_sort(0,n,&edge_ord1);
  
  list<edge> El1 = G.all_edges();
  El1.bucket_sort(0,n,&edge_ord1);
  El1.bucket_sort(0,n,&edge_ord2);


  // merge El and El1 to find corresponding edges

  while (! El.empty() && ! El1.empty())
  { e = El.head();
    r = El1.head();

    if (target(r) == source(e))
      if (source(r) == target(e))
         { reversal[e] = r;
           El1.pop();
           El.pop();
           count++;
          }
      else
         if (index(source(r)) < index(target(e)))
             El1.pop();
         else
             El.pop();

    else
      if (index(target(r)) < index(source(e)))
          El1.pop();
      else
          El.pop();

    }

  return (count == G.number_of_edges()) ? true : false;

}


void Make_Bidirected(graph& G, list<edge>& R)
{
  // make graph bi-directed by inserting reversal edges
  // appends new edges to R

  edge_array<edge> rev(G,nil);
  Is_Bidirected(G,rev);
 
  edge e = G.first_edge();
 
  while (e && rev[e]) e = G.succ_edge(e);
 
  edge last_edge = nil;
  if (e !=nil)
  { last_edge = G.new_edge(target(e),source(e));
    R.append(last_edge);
    e = G.succ_edge(e);
    while (e != last_edge)
    { if (rev[e] == nil)
      { edge e1 = G.new_edge(target(e),source(e));
        rev[e] = e1;
        R.append(e1);
        }
      e = G.succ_edge(e);
     }
   }
}


list<edge> Make_Bidirected(graph& G)
{ list<edge> R;
  Make_Bidirected(G,R);
  return R;
 }



static void dfs(graph& G, node v, node_array<bool>& reached)
{ reached[v] = true;
  edge e;
  forall_inout_edges(e,v) 
  { node w = G.opposite(v,e);
    if ( !reached[w] ) dfs(G,w,reached);
   }
 } 

bool Is_Connected(graph& G)
{
  node_array <bool>  reached(G, false);

  dfs(G,G.first_node(),reached);

  node v;
  forall_nodes(v, G) 
    if ( !reached[v] ) return false;

  return true;

 }


void Make_Connected(graph& G, list<edge>& L)
{
  node_array <bool>  reached(G, false);
  node u = G.first_node();

  node v;
  forall_nodes(v, G) 
    if ( !reached[v] ) 
    { dfs(G,v,reached);  // explore connected comp with root v 
      if (u != v)                     // link v's comp to the first comp
        L.append(G.new_edge(u, v));
     }
 }



list<edge> Make_Connected(graph& G)
{ list<edge> L;
  if (G.number_of_nodes() > 0) Make_Connected(G,L);
  return L;
 }





static void make_bicon_dfs(graph& G, node v, int& dfs_count, 
                                list<edge>& L,
                                node_array<int>&  dfsnum, 
                                node_array<int>&  lowpt,
                                node_array<node>& parent)
{ node u = nil;

  dfsnum[v] = dfs_count++;
  lowpt[v]  = dfsnum[v];

  edge e;
  forall_inout_edges(e,v) 
  { 
    node w = G.opposite(v,e);

    if (u == nil) u = w; //first child

    if ( dfsnum[w] == -1) // w not reached before; e is a tree edge
      { 
        parent[w] = v;
  
        make_bicon_dfs(G, w, dfs_count, L, dfsnum, lowpt, parent);
  
        if (lowpt[w] == dfsnum[v]) 	
        { // |v| is an articulation point. We now add an edge. If |w| is the
  	  // first child and |v| has a parent then we connect |w| and
  	  // |parent[v]|, if |w| is a first child and |v| has no parent then
  	  // we do nothing. If |w| is not the first child then we connect |w|
  	  // to the first child. The net effect of all of this is to link all
  	  // children of an articulation point to the first child and the first
  	  // child to the parent (if it exists)
  
  	  if (w == u && parent[v]) 
          { L.append(G.new_edge(w, parent[v]));
  	    //L.append(G.new_edge(parent[v], w)); (if bidirected)
  	   }
  
  	  if (w != u)
  	  { L.append(G.new_edge(u, w));
  	    //L.append(G.new_edge(w, u)); (if bidirected)
  	   }
         }
  
         lowpt[v] = Min(lowpt[v], lowpt[w]);
       } 
      else // non tree edge
       lowpt[v] = Min(lowpt[v], dfsnum[w]);

  }

}



static bool is_bicon_dfs(graph& G, node v, int& dfs_count, 
                                     node_array<int>&  dfsnum, 
                                     node_array<int>&  lowpt,
                                     node_array<node>& parent)
{ node u = nil;
  edge e;

  dfsnum[v] = dfs_count++;
  lowpt[v]  = dfsnum[v];

  forall_inout_edges(e,v) 
  { node w = G.opposite(v,e);
    if (u == nil) u = w;
    if ( dfsnum[w] == -1 )
      { parent[w] = v;
        is_bicon_dfs(G, w, dfs_count, dfsnum, lowpt, parent);
        if (lowpt[w] == dfsnum[v] && (w != u || parent[v]))  return false;
        lowpt[v] = Min(lowpt[v], lowpt[w]);
       } 
    else 
        lowpt[v] = Min(lowpt[v], dfsnum[w]);
  }

  return true;
}


bool Is_Biconnected(graph & G)
{ if ( ! Is_Connected(G) ) return false;
  node_array<int>  lowpt(G);
  node_array<int>  dfsnum(G,-1);
  node_array<node> parent(G,nil);
  int dfs_count = 0;
  return is_bicon_dfs(G, G.first_node(), dfs_count, dfsnum, lowpt, parent);
}				




void Make_Biconnected(graph& G, list<edge>& L)
{
  if (G.number_of_nodes() == 0) return;

  Make_Connected(G,L);

  node_array<int>  lowpt(G);
  node_array<int>  dfsnum(G,-1);  // dfsnum[v] == -1  <=>  v not reached 
  node_array<node> parent(G,nil);

  int dfs_count = 0;

  make_bicon_dfs(G, G.first_node(), dfs_count, L, dfsnum, lowpt, parent);
}				


list<edge> Make_Biconnected(graph & G)
{ list<edge> L;
  Make_Biconnected(G,L);
  return L;
 }






static bool bi_bfs(const graph& G, node s, node_array<int>& side)
{ 
  list<node> Q;

  Q.append(s);
  side[s] = 0;

  while ( ! Q.empty() )
  { node v = Q.head();
    edge e;
    forall_inout_edges(e,v)
    { node w = G.opposite(v,e);
      if (side[v] == side[w]) return false;
      if (side[w] == -1)
      { Q.append(w); 
        side[w] = 1 - side[v];
       }
     }
    Q.pop();
  }

  return true;
}


bool Is_Bipartite(const graph& G, list<node>& A, list<node>& B)
{
  node_array<int> side(G,-1);
  node v;

  forall_nodes(v,G)
    if (side[v] == -1) 
       if (! bi_bfs(G,v,side)) return false;

  forall_nodes(v,G)
  { if (side[v] == 0) A.append(v);
    if (side[v] == 1) B.append(v);
   }

  return true;

 }


int COMPONENTS(const graph&G, node_array<int>&);

int Genus(const graph& G, const edge_array<edge>& rev)
{ int n = G.number_of_nodes();
  int m = G.number_of_edges();
  node_array<int> cnum(G);
  int c = COMPONENTS(G,cnum);
  edge_array<bool> considered(G,false);
  int f = 0;
  edge e;
  forall_edges(e,G)
  { if (G.reversal(e) == nil)
       error_handler(1,"Genus: graph must be a map.");
    if ( !considered[e] )
    { // trace the face to the left of e
      f++;
      edge e1 = e;
      do { considered[e1] = true;
           e1 = G.cyclic_adj_pred(rev[e1]);
         }
      while (e1 != e);
    }
  }
  return (2 - n + m - f + (c-1))/2;
}




/*
void copy_graph(const graph& G, GRAPH<node,edge>& H, 
                node_array<node>& v_in_H, edge_array<edge>& e_in_H)
{

  node v;
  forall_nodes(v,G) v_in_H[v] = H.new_node(v);

  edge e;
  forall_edges(e,G) e_in_H[e] =
      H.new_edge(v_in_H[source(e)],v_in_H[target(e)],e);
}
*/


