/*******************************************************************************
+
+  LEDA 3.4
+
+  _graph.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <LEDA/graph.h>

//------------------------------------------------------------------------------
// basic graph operations
//
// by S. Naeher (1995,1996)
//------------------------------------------------------------------------------

graph::graph() 
{ max_n_index = -1;
  max_e_index = -1; 
  max_f_index = -1; 
  parent = 0; 
  undirected = false;
 }


void graph::copy_all_entries() const
{ node v;
  forall(v,v_list) copy_node_entry(v->data[0]);
  edge e;
  forall(e,e_list) copy_edge_entry(e->data[0]);
  face f;
  forall(f,f_list) copy_face_entry(f->data[0]);
 }


void graph::clear_all_entries() const
{ node v;
  forall(v,v_list) clear_node_entry(v->data[0]);
  edge e;
  forall(e,e_list) clear_edge_entry(e->data[0]);
  face f;
  forall(f,f_list) clear_face_entry(f->data[0]);
 }



void graph::copy_graph(const graph& G)  
{ 
  int n = G.number_of_nodes();
  //int m = G.number_of_edges();

  max_n_index = -1;
  max_e_index = -1;
  max_f_index = -1;
  parent = 0;

  if (n == 0) return;

  node* node_vec = new node[G.max_n_index+1];
  edge* edge_vec = new edge[G.max_e_index+1];

  if (node_vec == 0 || edge_vec == 0) 
    error_handler(1," copy_graph: out of memory");

  // allocate a single block of memory for all nodes
  // memory_allocate_block(sizeof(node_struct),n);

  node v;
  forall_nodes(v,G) 
     node_vec[v->name] = new_node(v->data[0]);

  // allocate a single block of memory for all edges
  // memory_allocate_block(sizeof(edge_struct),m);

  bool loops_deleted = false;

  forall_nodes(v,G) 
  { node a = node_vec[v->name];
    edge e;
    forall_adj_edges(e,v) 
    { if ( undirected && v == target(e)) //incoming edge 
       { if (v == source(e)) // loop 
           loops_deleted = true; 
        }
      else
       { node b = node_vec[target(e)->name];
         edge_vec[e->name] = new_edge(a,b,e->data[0]);
        }
     }
   }

  edge e;
  forall_edges(e,G) 
    if (e->rev) 
       edge_vec[e->name]->rev = edge_vec[e->rev->name];

  delete[] node_vec;
  delete[] edge_vec;

  if ( loops_deleted )
      error_handler(0,"selfloops deleted in ugraph copy constructor");
}


graph::graph(const graph& G)  
{ undirected = G.undirected; 
  copy_graph(G); 
 }


graph& graph::operator=(const graph& G)
{ if (&G != this)
  { graph::clear();
    undirected = G.undirected;
    copy_graph(G);
   }
  return *this;
}



// subgraph constructors  (do not work for undirected graphs)

graph::graph(graph& G, const list<node>& nl, const list<edge>& el)
{ // construct subgraph (nl,el) of graph G

  parent = &G;
  node v,w;
  edge e;

  node* N = new node[G.max_n_index+1];

  forall(v,nl)
   { if (graph_of(v) != parent) 
      error_handler(1,"graph: illegal node in subgraph constructor");
     N[v->name] = new_node((GenPtr)v);
    }

  forall(e,el)
   { v = source(e);
     w = target(e);
     if ( graph_of(e)!= parent || N[v->name]==0 || N[w->name]==0 ) 
      error_handler(1,"graph: illegal edge in subgraph constructor");
     new_edge(N[v->name],N[w->name],(GenPtr)e);
    }

  undirected = G.undirected;

  delete[] N;

 }

graph::graph(graph& G, const list<edge>& el)
{ /* construct subgraph of graph G with edge set el  */

  node  v,w;
  edge  e;
  node* N = new node[G.max_n_index+1];

  forall_nodes(v,G) N[v->name] = 0;

  parent = &G;

  forall(e,el)
   { v = source(e);
     w = target(e);
     if (N[v->name] == 0) N[v->name] = new_node((GenPtr)v);
     if (N[w->name] == 0) N[w->name] = new_node((GenPtr)w);
     if ( graph_of(e) != parent )
      error_handler(1,"graph: illegal edge in subgraph constructor");
     new_edge(N[v->name],N[w->name],(GenPtr)e);
    }

  undirected = G.undirected;

  delete[] N;

 }


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------

void graph::clear()
{ 
  node v;
  edge e;
  face f;

  clear_action();

  forall(v,v_list) delete v;
  v_list.clear();

  forall(v,v_free) delete v;
  v_free.clear();

  forall(e,e_list) delete e;
  e_list.clear();

  forall(e,e_free) delete e;
  e_free.clear();

  forall(f,f_list) delete f;
  f_list.clear();

  max_n_index = max_e_index = max_f_index = -1;
}



//------------------------------------------------------------------------------
// accessing node and edge lists
//------------------------------------------------------------------------------


const list<node>& graph::all_nodes() const { return v_list; }

const list<edge>& graph::all_edges() const { return e_list; }



list<edge> graph::out_edges(node v) const
{ list<edge> result;
  edge e;
  forall_out_edges(e,v) result.append(e);
  return result;
}


list<edge> graph::in_edges(node v) const
{ list<edge> result;
  edge e;
  forall_in_edges(e,v) result.append(e);
  return result;
}

list<edge> graph::adj_edges(node v) const
{ list<edge> result;
  edge e;
  forall_adj_edges(e,v) result.append(e);
  return result;
}


list<node> graph::adj_nodes(node v) const
{ list<node> result;
  edge e;
  forall_adj_edges(e,v) result.append(opposite(v,e));
  return result;
}



//------------------------------------------------------------------------------
// update operations
//------------------------------------------------------------------------------




list<edge> graph::insert_reverse_edges()
{ list<edge> L;
  edge e = first_edge();

  if (e != nil)
  { L.append(new_edge(target(e),source(e),e->data[0]));
    copy_edge_entry(e->data[0]);
    e = succ_edge(e);
   }

  edge stop = last_edge();

  while (e != stop)
  { L.append(new_edge(target(e),source(e),e->data[0]));
    copy_edge_entry(e->data[0]);
    e = succ_edge(e);
   } 

  return L;
}


face graph::add_face(GenPtr inf)
{ face f = new face_struct(inf);
  f->owner = this;
  f->name = ++max_f_index;
  f->pos_in_F = f_list.append(f);
  return f;
}

void graph::del_face(face f) 
{ f_list.del(f->pos_in_F); 
  delete f; 
 }


node graph::add_node(GenPtr inf)
{ node v;

  if ( v_free.empty() )
    { v = new node_struct(inf);
      v->owner = this;
      v->name = ++max_n_index;
      v->succ_link = nil;
     }
  else
    { v = v_free.pop();
      v->data[0] = inf;
     }

  v->pos_in_V = v_list.append(v);

  return v;
}



node graph::new_node()
{ GenPtr i = 0;
  init_node_entry(i);
  node v = add_node(i);
  new_node_action(v);
  return v;
}


node graph::new_node(GenPtr i)
{ node v = add_node(i);
  new_node_action(v);
  return v;
}


void graph::del_node(node v)
{ 
  if (v->owner != this) 
        error_handler(4,"del_node(v): v is not in G");


  // delete adjacent edges

  edge  e;
  while ((e=v->first_adj_edge[0]) != nil) del_edge(e);

  if (!undirected)
     while ((e=v->first_adj_edge[1]) != nil) del_edge(e);

  del_node_action(v);

  if (parent==0) clear_node_entry(v->data[0]);
 
  v_list.del(v->pos_in_V);
  v_free.append(v);
}







edge graph::add_edge(node v, node w, GenPtr inf)
{ edge e;

  if (v->owner != this)
     error_handler(6, "new_edge(v,w): v not in graph");
  if (w->owner != this) 
     error_handler(6, "new_edge(v,w): w not in graph");


  if ( e_free.empty() )
    { e = new edge_struct(v,w,inf);
      e->name = ++max_e_index;
     }
  else
    { e = e_free.pop();
      e->data[0] = inf;
      e->term[0] = v;
      e->term[1] = w;
      e->rev = nil;
      e->succ_adj_edge[0] = nil;
      e->succ_adj_edge[1] = nil;
      e->pred_adj_edge[0] = nil;
      e->pred_adj_edge[1] = nil;
    }

  e->pos_in_E = e_list.append(e);

  return e;
}


void graph::del_adj_edge(edge e, node v, node w)
{ if (undirected)
    { v->del_adj_edge(e,0,0);
      w->del_adj_edge(e,0,1);
     }
  else
    { v->del_adj_edge(e,0,0);
      w->del_adj_edge(e,1,1);
     }
 }



void graph::ins_adj_edge(edge e, node v, edge e1, node w, edge e2,int d1,int d2)
{ 
  // insert edge e
  // after(if d1=0)/before(if d1=1) e1 to adj_list of v
  // after(if d2=0)/before(if d2=1) e2 to in_list (adj_list) of w
  // (most general form of new_edge)

  if ( undirected ) 
   { if (v == w)  
        error_handler(1,"new_edge(v,e1,w,e2): selfloop in undirected graph.");
     if (e1 && v != source(e1) && v != target(e1)) 
        error_handler(1,"new_edge(v,e1,w,e2): v is not adjacent to e1.");
     if (e2 && w != source(e2) && w != target(e2)) 
        error_handler(1,"new_edge(v,e1,w,e2): w is not adjacent to e2.");

     v->insert_adj_edge(e,e1,0,0,d1);
     w->insert_adj_edge(e,e2,0,1,d2);
   }
  else
   { if (e1 && v != source(e1)) 
        error_handler(1,"new_edge(v,e1,w,e2): v is not source of e1.");
     if (e2 && w != source(e2) && w != target(e2)) 
        error_handler(1,"new_edge(v,e1,w,e2): w is not target of e2.");

     v->insert_adj_edge(e,e1,0,0,d1);
     w->insert_adj_edge(e,e2,1,1,d2);
    }
}



edge graph::new_edge(node v, edge e1, node w, edge e2, GenPtr i,int d1,int d2)
{ 
  // add edge (v,w,i)
  // after(if d1=0)/before(if d1=1) e1 to adj_list of v
  // after(if d2=0)/before(if d2=1) e2 to in_list (adj_list) of w
  // (most general form of new_edge)

  if ( undirected ) 
   { if (v == w)  
        error_handler(1,"new_edge(v,e1,w,e2): selfloop in undirected graph.");
     if (e1 && v != source(e1) && v != target(e1)) 
        error_handler(1,"new_edge(v,e1,w,e2): v is not adjacent to e1.");
     if (e2 && w != source(e2) && w != target(e2)) 
        error_handler(1,"new_edge(v,e1,w,e2): w is not adjacent to e2.");
   }
  else
   { if (e1 && v != source(e1)) 
        error_handler(1,"new_edge(v,e1,w,e2): v is not source of e1.");
     if (e2 && w != source(e2) && w != target(e2)) 
        error_handler(1,"new_edge(v,e1,w,e2): w is not target of e2.");
    }

  edge e = add_edge(v,w,i);
  ins_adj_edge(e,v,e1,w,e2,d1,d2);
  new_edge_action(e);
  return e ; 
}

edge graph::new_edge(node v, edge e1, node w, GenPtr i,int d)
{
  // add edge (v,w) after/before e1 to adj_list of v
  // append it to in_list (adj_list) of w

  return new_edge(v,e1,w,nil,i,d,0); 
}

edge graph::new_edge(node v, node w, edge e2, GenPtr i,int d)
{
  // append edge (v,w) to adj_list of v
  // insert it after/before e2 to in_list (adj_list) of w

  return new_edge(v,nil,w,e2,i,d,0); 
}

edge graph::new_edge(edge e1, node w, GenPtr i, int dir)  
{ 
  // add edge (source(e1),w) after/before e1 to adj_list of source(e1)
  // append it to in_list (adj_list) of w

  return new_edge(source(e1),e1,w,nil,i,dir,0);
 }

edge graph::new_edge(node v, edge e2, GenPtr i, int dir)  
{ 
  // append edge(v,target(e2))  to adj_list of v
  // insert it after/before e2 to in_list (adj_list) of target(e2)

  return new_edge(v,nil,target(e2),e2,i,0,dir);
 }

edge graph::new_edge(edge e1, edge e2, GenPtr i, int dir1, int dir2)  
{ 
  //add edge (source(e1),target(e2)) 
  //after(dir=0)/before(dir=1) e1 to adj_list of source(e1)
  //after(dir=1)/before(dir=1) e2 to in_list (adj_list) of target(e2)

  return new_edge(source(e1),e1,target(e2),e2,i,dir1,dir2);
 }


edge graph::new_edge(node v, node w, GenPtr i)
{ 
  // append (v,w) it to adj_list of v and to in_list (adj_list) of w

  return new_edge(v,nil,w,nil,i,0,0); 
 }






node graph::split_edge(edge e, GenPtr node_inf, edge& e1, edge& e2)
{
   // splits e into e1 and e2 by putting new node v on e 


  //node v = source(e);
  node w = target(e);
  node u = add_node(node_inf); 

  e1 = e;
  e2 = add_edge(u,w,e->data[0]);

  copy_edge_entry(e2->data[0]);

  if (undirected)
    { u->append_adj_edge(e2,0,0);
      w->insert_adj_edge(e2,e,0,1,0);
      w->del_adj_edge(e,0,1);
      e->term[1] = u;
      u->append_adj_edge(e,0,1);
     }
  else
    { u->append_adj_edge(e2,0,0);
      w->insert_adj_edge(e2,e,1,1,0);
      w->del_adj_edge(e,1,1);
      e->term[1] = u;
      u->append_adj_edge(e,1,1);
     }

 return u;
}




void graph::del_edge(edge e)
{ node v = source(e);
  node w = target(e);

  if (v->owner != this) error_handler(10,"del_edge(e): e is not in G");

  del_edge_action(e);

  e_list.del(e->pos_in_E);

  del_adj_edge(e,v,w);

  if (parent == 0)  // no subgraph
     clear_edge_entry(e->data[0]);

  e_free.append(e);
}

 
void graph::hide_edge(edge e)
{ if (e->succ_adj_edge[0] != e)  
  { node v = source(e);
    node w = target(e);
    del_adj_edge(e,v,w);
    e->succ_adj_edge[0] = e;
   }

 }

void graph::restore_edge(edge e)
{
  if (e->succ_adj_edge[0] == e) 
  { node v = source(e);
    node w = target(e);
    if (undirected)
      { v->append_adj_edge(e,0,0);
        w->append_adj_edge(e,0,1);
       }
    else
      { v->append_adj_edge(e,0,0);
        w->append_adj_edge(e,1,1);
       }
   }
 }


void graph::move_edge(edge e,edge e1,edge e2,int d1,int d2)
{ del_adj_edge(e,source(e),target(e));
  node v = source(e1);
  node w = target(e2);
  e->term[0] = v;
  e->term[1] = w;
  ins_adj_edge(e,v,e1,w,e2,d1,d2);
 }

void graph::move_edge(edge e,edge e1,node w,int dir)
{ del_adj_edge(e,source(e),target(e));
  e->term[0] = source(e1);
  e->term[1] = w;
  ins_adj_edge(e,source(e1),e1,w,nil,dir,0);
}

void graph::move_edge(edge e, node v, node w)
{ del_adj_edge(e,source(e),target(e));
  e->term[0] = v;
  e->term[1] = w;
  ins_adj_edge(e,v,nil,w,nil,0,0);
 }




edge graph::rev_edge(edge e)
{ node v = source(e);
  node w = target(e);

  rev_edge_action(e);

  if (e->succ_adj_edge[0] == e) // e hidden
  { e->term[0] = w;
    e->term[1] = v;
    return e;
   }

  if (undirected)
  { edge s = e->succ_adj_edge[0];
    edge p = e->pred_adj_edge[0];
    e->succ_adj_edge[0] = e->succ_adj_edge[1]; 
    e->pred_adj_edge[0] = e->pred_adj_edge[1]; 
    e->succ_adj_edge[1] = s;
    e->pred_adj_edge[1] = p;
    e->term[0] = w;
    e->term[1] = v;
   }
  else
  { del_adj_edge(e,v,w);
    e->term[0] = w;
    e->term[1] = v;
    ins_adj_edge(e,w,nil,v,nil,0,0);
   }

  return e; 
}


void graph::rev_all_edges()              
{ if (!undirected) 
  { edge e;
    forall(e,e_list) rev_edge(e);
   }
}

void graph::del_all_nodes() { clear(); }

void graph::del_all_edges()
{
  node v;
  forall(v,v_list) 
    for(int i=0; i<2; i++)
    { v->first_adj_edge[i] = nil;
      v->last_adj_edge[i] = nil;
      v->adj_length[i] = 0;
     }

  edge e;
  forall(e,e_list) delete e;
  e_list.clear();

  forall(e,e_free) delete e;
  e_free.clear();

  max_e_index = -1;
}


void graph::reset()  const   // reset all iterators
{ node v;
  forall(v,v_list) v->adj_iterator=nil;
}



void graph::make_undirected()
{
  if (undirected) return; 

  list<edge> loops;
  edge e;
 
  forall(e,e_list)
    if (source(e) == target(e))  loops.append(e);
 
  if ( ! loops.empty() )
      error_handler(0,"selfloops deleted in ugraph constructor");

  forall(e,loops) del_edge(e);


  /* adj_list(v) = out_list(v) + in_list(v) forall nodes v  */

  node v;
  forall(v,v_list)
  { 
    // append in_list to adj_list

    if (v->first_adj_edge[1] == nil) continue;


    if (v->first_adj_edge[0] == nil) // move in_list to adj_list
      { v->first_adj_edge[0] = v->first_adj_edge[1];
        v->last_adj_edge[0]  = v->last_adj_edge[1];
        v->adj_length[0]     = v->adj_length[1];
       }
    else // both lists are non-empty
      { v->last_adj_edge[0]->succ_adj_edge[0] = v->first_adj_edge[1];
        v->first_adj_edge[1]->pred_adj_edge[1] = v->last_adj_edge[0];
        v->last_adj_edge[0] = v->last_adj_edge[1];
        v->adj_length[0] += v->adj_length[1];
       }

    v->first_adj_edge[1] = nil;
    v->last_adj_edge[1] = nil;
    v->adj_length[1] = 0;

   }

  undirected = true; 
}


void graph::make_directed()
{ 
  if (!undirected) return;

  // for every node v delete entering edges from adj_list(v)
  // and put them back into in_list(v)

  node v;
  forall(v,v_list)
  { edge e = v->first_adj_edge[0];
    while (e)
      if (v == target(e))
         { edge e1 = e->succ_adj_edge[1];
           v->del_adj_edge(e,0,1);
           v->append_adj_edge(e,1,1); 
           e = e1;
          }
      else 
         e = e->succ_adj_edge[0];
   }

  undirected = false; 

}



void init_node_data(const graph& G,int i, GenPtr x)
{ node v;
  forall_nodes(v,G) v->data[i] = x;
 }



