/*******************************************************************************
+
+  LEDA 3.4
+
+  _planar_map.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

#include <LEDA/planar_map.h>

//------------------------------------------------------------------------------
// planar maps
//
// by S. Naeher  (1995,1996)
//------------------------------------------------------------------------------



edge planar_map::split_edge(edge e, GenPtr node_inf)
{ 
  /* splits edge e and its reversal by inserting a new node u (node_inf) 
              e                          e           rr
        ----------->                --------->   --------->
     (v)            (w)   ====>  (v)          (u)          (w)
        <-----------                <---------   <---------
              r                          er          r
     returns edge rr
  */

  edge r = e->rev;
  node v = source(e);
  node w = target(e);
  node u = graph::new_node();

  copy_node_entry(node_inf);

  u->data[0] = node_inf;

  // remove e and r from corresponding in-lists
  w->del_adj_edge(e,1,1);
  v->del_adj_edge(r,1,1);

  // insert e and r in in-list of u
  e->term[1] = u;
  r->term[1] = u;
  u->append_adj_edge(e,1,1);
  u->append_adj_edge(r,1,1);

  // create reverse edges rr and re
  edge rr = graph::new_edge(u,w);
  edge er = graph::new_edge(u,v);

  set_reversal(e,er);
  set_reversal(r,rr);

  rr->fa = e->fa;
  er->fa = r->fa;

  return rr;
}


edge planar_map::new_edge(edge e1, edge e2, GenPtr face_i)
{ 
  face f1 = adj_face(e1);
  face f2 = adj_face(e2);

  if (f1 != f2)
    error_handler(1,"planar_map::new_edge: new edge must lie in a face."); 

  f2 = new_face(face_i);

  edge x = graph::new_edge(e1,source(e2));
  edge y = graph::new_edge(e2,source(e1));

  f1->head = x;
  f2->head = y;

  graph::set_reversal(x,y);

  x->fa = f1;

  do { y->fa = f2;
       y = face_cycle_succ(y);
     } while (y != f2->head);

  return x;
}



node planar_map::new_node(const list<edge>& el, GenPtr node_inf)
{
  if (el.length() < 2)
      error_handler(1,"planar_map::new_node(el,i):  el.length() < 2."); 

  list_item it = el.first();

  edge e0 = el[it];

  it = el.succ(it);

  face f = adj_face(e0);

  edge e;
  forall(e,el)
  { if (adj_face(e) != f)
      error_handler(1,"planar_map::new_node: edges bound different faces."); 
   }

  e = el[it];

  it = el.succ(it);

  GenPtr face_inf = f->data[0];

  copy_face_entry(face_inf);
  copy_node_entry(node_inf);

  edge e1 = split_edge(new_edge(e0,e,face_inf),node_inf);

  node u = source(e1);

  while(it)
  { copy_face_entry(face_inf);
    e1 = new_edge(e1,el[it],face_inf);
    it = el.succ(it);
   }

  return u;
}

node planar_map::new_node(face f, GenPtr node_inf)
{ return new_node(adj_edges(f),node_inf);
 }


void planar_map::del_edge(edge x, GenPtr face_i)
{
  edge y  = reverse(x);

  face F1 = adj_face(x);
  face F2 = adj_face(y);

  edge e = face_cycle_succ(y);

  F1->data[0] = face_i;

  if (F1 != F2)
  { edge e = face_cycle_succ(y);
    F1->head = e;
    while ( e != y )
    { e->fa = F1;
      e = face_cycle_succ(e);
     }
    clear_face_entry(F2->data[0]);
    del_face(F2);
   }
  else
  { e = face_cycle_succ(e);

    if (e != y) // no isolated edge
      F1->head = e;   
    else 
      { clear_face_entry(F1->data[0]);
        del_face(F1);
       }

   }

  graph::del_edge(x);
  graph::del_edge(y);
}





list<edge> planar_map::triangulate()
{
  node v,w;
  edge e,e1,e2,e3;

  list<edge> L;

  node_array<bool> marked(*this,false);
 
  forall_nodes(v,*this)
  {
    list<edge> El = adj_edges(v);
 
    forall(e,El) marked[target(e)] = true;

    forall(e,El)
    { 
      e1 = e;
      e2 = face_cycle_succ(e1);
      e3 = face_cycle_succ(e2);

      while (target(e3) != v)
      { 
        // e1,e2 and e3 are the first three edges in a counter-clockwise 
        // traversal of a face incident to v and target(e3) != v

        if ( !marked[target(e2)] )
        { // we mark w and add the edge {v,w} inside F, i.e., after
          // dart e1 at v and after dart e3 at w.

          marked[target(e2)] = true;
  
          e1 = new_edge(e1,e3);
          e2 = e3;
          e3 = face_cycle_succ(e2);
          init_edge_entry(e1->data[0]);
          L.append(e1);
        }
        else
        { // we add the edge {source(e2),target(e3)} inside F, i.e.,
          // after dart e2 at source(e2) and after e4 at source(e4)

          e3 = face_cycle_succ(e3); 
          e2 = new_edge(e2,e3);
          init_edge_entry(e2->data[0]);
          L.append(e2);
        }
      }
    }

   forall_adj_nodes(w,v) marked[w] = false;

  }

 return L;

}


int planar_map::straight_line_embedding(node_array<int>& x, node_array<int>& y)
{ return STRAIGHT_LINE_EMBEDDING(*this,x,y); }

