/*******************************************************************************
+
+  LEDA 3.4
+
+  _pq_tree.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

//------------------------------------------------------------------------------
// PQ_TREES
//
// R. Hesse, E. Kalliwoda  (1995,1996)
//
// last modified 23.8.96
//
//------------------------------------------------------------------------------

#include <LEDA/queue.h>
#include <assert.h>
#include <LEDA/pq_tree.h>

#ifndef NDEBUG
#include "cons_pq.h"
#endif

enum { DPQ_RESET =  1,
       DPQ_P     =  2,
       DPQ_Q     =  3,
       DPQ_L     =  4,
       DPQ_unma  =  5,
       DPQ_queu  =  6,
       DPQ_bloc  =  7,
       DPQ_ublo  =  8,
       DPQ_empt  =  9,
       DPQ_part  = 10,
       DPQ_dpar  = 11,
       DPQ_full  = 12,
       DPQ_aful  = 13,
       DPQ_PKK   = 14,
       DPQ_nPKK  = 15,
       DPQ_pafu  = 16,
       DPQ_DIR   = 17,
       DPQ_nDIR  = 18,
       DPQ_fuDI  = 19,
       DPQ_NULL  = 20 };



pq_node_struct::pq_node_struct()
{ label = 0;
  leaf_index  = 0;
  child_count = 0;
  full_count  = 0;
  pert_child_count = 0;
  pert_leaf_count  = 0;
  parent = NULL;
  link1  = NULL;
  link2  = NULL;
  right_most  = NULL;
  left_most   = NULL;
  part_child1 = NULL;
  part_child2 = NULL;
}






void pq_node_struct::special_reset()
{ pert_leaf_count=0;            //
  pert_child_count=0;           //special reset
  full_count=0;                 //of this, because of the forbidden   
  mark(DPQ_RESET);              //Processed.push(x)
  part_child1=part_child2=0;    // 
}






pq_tree::pq_tree()
{ 
//  pq_mem_init();
  positiv = true;
  leaf_size = 0;
  leaves = NULL;
  root = NULL;
  too_many_part = 0;
  block_count = 0;
  blocked_nodes = 0;
  p_root = new pq_node_struct;
  p_root->leaf_index = 2;
 }





pq_tree::~pq_tree()
{ //cons_pq_tree(root,"a");
//  show_pq_tree_test(root);
//  cons_pq_tree(root,"b","destructor");
  if(!positiv)  del_subtree(root);
  delete  p_root;
  delete[] leaves;
//  pq_mem_off();
 }



void pq_node_struct::mark(char ttt)
{
 switch(ttt) {

 case DPQ_RESET:        //reset bubble & reduction information, 
                        //preserve type & PKK 
  label &= 0x83;        //10000011b
  break;

 //type:
 case DPQ_P:            //P-node
  label &= 0xfc;        //11111100b
  label |= 0x1;         //00000001b 
  break;
 case DPQ_Q:            //Q-node
  label &= 0xfc;        //11111100b
  label |= 0x2;         //00000010b 
  break;
 case DPQ_L:            //leaf
  label |= 0x3;         //00000011b
  break;

 //mark: 
 case DPQ_unma:         //unmarked
  label &= 0xf3;        //11110011b
  break;
 case DPQ_queu:         //queued
  label |= 0xc;         //00001100b 
  break;
 case DPQ_bloc:         //blocked node
  label &= 0xf3;        //11110011b
  label |= 0x8;         //00001000b 
  break;
 case DPQ_ublo:         //unblocked node
  label &= 0xf3;        //11110011b
  label |= 0x4;         //00000100b 
  break;

 //label:
 case DPQ_empt:         //empty, i.e. no pertinent leaf in subtree
  label &= 0xcf;        //11001111b
  break;
 case DPQ_part:         //pertinent and nonpertinent leaves in subtree
  label |= 0x30;        //00110000b 
  break;
 case DPQ_dpar:
  label &= 0xcf;        //11001111b
  label |= 0x10;        //00010000b 
  break;
 case DPQ_full:         //only pertinent leaves in subtree
  label &= 0xcf;        //11001111b
  label |= 0x20;        //00100000b 
  break;

 case DPQ_DIR:          //node becomes a direction pointer ...
  label |= 0x40;        //01000000b 
  break;
 case DPQ_nDIR:         //... and the contrary
  label &= 0xbf;        //10111111b

 case DPQ_PKK:          //"p-Knoten-Kind" - child of a P-node
  label |= 0x80;        //10000000b 
  break;
 case DPQ_nPKK:         //child of a Q-node
  label &= 0x7f;        //01111111b 
  break;
}

}


char pq_node_struct::check(char ttt)
{

 switch(ttt) {

 //type:
 case DPQ_P: return((label & 0x3) == 0x1);
 case DPQ_Q: return((label & 0x3) == 0x2);
 case DPQ_L: return((label & 0x3) == 0x3);

 //mark: 
 case DPQ_unma: return((label & 0xc) == 0x0);
 case DPQ_queu: return((label & 0xc) == 0xc);
 case DPQ_bloc: return((label & 0xc) == 0x8);
 case DPQ_ublo: return((label & 0xc) == 0x4);

 //label:
 case DPQ_empt: return((label & 0x30) == 0x00);
 case DPQ_part: return((label & 0x30) == 0x30);
 case DPQ_dpar: return((label & 0x30) == 0x10);
 case DPQ_full: return((label & 0x30) == 0x20);

//direction indicator
 case DPQ_DIR : return((label & 0x40));
 case DPQ_nDIR: return(!(label & 0x40));

//misspl.
 case DPQ_fuDI: return((label & 0x40) || ((label & 0x30) == 0x20));
                                        //"full" or "DIR"
 case DPQ_pafu: return((label & 0x30)); //mit "== 0x20" wird dpar nicht erfasst
 case DPQ_NULL: return(label == 0);

//P node child, or not
 case DPQ_PKK:  return(label & 0x80);    //"p-Knoten-Kind"
 case DPQ_nPKK: return(!(label & 0x80));

}

return 0;

}



void pq_tree::reset()
{ while (!Processed.empty())
  { pq_node k = Processed.pop();
    k->mark(DPQ_RESET);           // Knotentyp & PKK-Bit untouched
    k->pert_leaf_count = 0;
    k->pert_child_count = 0;
    k->full_count = 0;
    k->part_child1 = NULL;
    k->part_child2 = NULL;
   }
  p_root->mark(DPQ_P);
  p_root->special_reset();
}





bool pq_tree::bubble(list<int>& S)
{
  queue<pq_node> Q;
  pq_node       x,y,z,k,l;
  int           m,blocked_found;

  #ifndef NDEBUG
  cout << endl;
  #endif

  off_the_top = 0;
  block_count = blocked_nodes = 0;
  p_root->left_most = p_root->right_most = NULL;

 forall(m,S)
 { x = leaves[m];
   x->mark(DPQ_queu);  Q.append(x); 
  }

#ifndef NDEBUG
printf(" %d ",this->p_root->leaf_index);
cons_pq_tree(root,"a","bubble");
//show("in bubble",root);
#endif

 while((Q.size() + block_count + off_the_top) > 1)
 {
  if(!Q.size())  return(false);
 
  x = Q.pop(); 
 
  if(x->check(DPQ_PKK) || !x->link1 || !x->link2)  
     x->mark(DPQ_ublo);
  else
    { // try to make it valid in constant time 
      x->mark(DPQ_bloc);
      k = x->link1; l = x;
      skip_di(k,l);
      if(k->check(DPQ_ublo))
        { //the link1-sibling has a valid parent 
          x->mark(DPQ_ublo);
          x->parent = k->parent;
         } 
      else
        { k = x->link2;  l = x;
          skip_di(k,l);
          if(k->check(DPQ_ublo))
          { //the link2-sibling has a valid parent
            x->mark(DPQ_ublo);
            x->parent = k->parent;
           }
         }
      }
 
  if(x->check(DPQ_ublo))
  { //x has got a valid parent
    y = x->parent;                
    z = x->link1;
    if( z && (z->check(DPQ_bloc) || z->check(DPQ_DIR)) )
    { l = x;  
      blocked_found = 0;
      while( z && (z->check(DPQ_bloc) || z->check(DPQ_DIR)) )
      { if(z->check(DPQ_nDIR))
        { blocked_found = 1;
          z->parent = y;
          z->mark(DPQ_ublo);
          y->pert_child_count++;
          blocked_nodes--;
         }
        if (z->link1 == l) 
          { l = z; 
            z = z->link2; 
           }
        else
          { l = z; 
            z = z->link1; 
           }
       }
      if(blocked_found)  block_count--;
    } //to unblock a blocked chain in x->link1 direction
 
   z = x->link2;
 
   if ( z && (z->check(DPQ_bloc) || z->check(DPQ_DIR)) )
   { l = x;  
     blocked_found = 0;
     while( z && (z->check(DPQ_bloc) || z->check(DPQ_DIR)) )
     { if(z->check(DPQ_nDIR))
       { blocked_found = 1;
         z->mark(DPQ_ublo);
         z->parent = y;
         y->pert_child_count++;
         blocked_nodes--;
        }
        if (z->link1 == l) 
          { l = z; 
            z = z->link2; 
           }
        else
          { l = z; 
            z = z->link1; 
           }
       } //to unblock a blocked chain in x->link2 direction
      if(blocked_found)  block_count--;
     }
 
   if( !y )  
      off_the_top = 1; 
   else
    { y->pert_child_count++;
      if(y->check(DPQ_unma))
      { Q.append(y);
        y->mark(DPQ_queu);
       }
     }
 
  } 
  else
  { // x's parent is not valid
    k = x->link1;  
    l = x;  
    skip_di(k,l); 
    if(k->check(DPQ_bloc))  block_count--;
    k = x->link2;  
    l = x;  
    skip_di(k,l);
    if(k->check(DPQ_bloc))  block_count--;
    block_count++;
    blocked_nodes++;
   }
 
 } //end of "while(Q->size() + block_count + off_the_top > 1)"

 if(block_count)
 { 
   p_root->pert_child_count = blocked_nodes;
   p_root->mark(DPQ_Q);
   
   #ifndef NDEBUG
//   show("block_count am ende von bubble",root);
   #endif

  }

#ifndef NDEBUG
//if(p_root->leaf_index > 40)  show("Ende bubble", root); 
#endif

 return true;
}





void pq_tree::bubble_reset(pq_node x)
{ // in case bubble affects the nodes over the pertinent subtree root
  // we have to reset their pert_child_counts

  while(x && x->pert_child_count )
  { pq_node k = (x->check(DPQ_ublo)) ? x->parent : 0;
    x->pert_child_count = 0;
    x->mark(DPQ_RESET);
    x = k;
   }
}






bool pq_tree::reduce(list<int>& S)
{
  queue<pq_node> Q;
  pq_node x,y,k,l;
  
  int S_size = S.size();

//show_pq_tree_test(root);

  while(!S.empty())
  { x = leaves[S.pop()];
    x->pert_leaf_count = 1;
    Q.append(x);
   }

 while(!Q.empty())
 { x = Q.pop();

#ifndef NDEBUG
cons_pq_tree(root,"b","reduce");
int	aaa=0;
#endif

 if(x->check(DPQ_bloc))         
 { // a blocked chain exists and its members
   // get the "auxiliary" parent p_root

  x->parent = p_root;
  x->mark(DPQ_ublo);

  l = x;
  k = x->link1;
  skip_di(k,l);

  if (!k || k->check(DPQ_unma))
  { if(p_root->left_most)  
       p_root->right_most = l;
    else  
       p_root->left_most = l;
    }

  l = x;
  k = x->link2;
  skip_di(k,l);

  if (!k || k->check(DPQ_unma))
  { if(p_root->left_most)  
       p_root->right_most = l;
    else  
       p_root->left_most = l;
   }

  if (Q.empty()) p_root->mark(DPQ_Q);

 }

#ifndef NDEBUG
//if(!aaa && !x->check(DPQ_L)) show("in reduce, der momentane Unterbaum",x);
if(p_root->leaf_index > 40)  show("in rduce, der gesamte Baum", root);
cout << flush;
#endif

 if (x->pert_leaf_count < S_size)
 {      //"not pruned pert subtree root"?
   if(x != p_root)
   { y = x->parent; 
     y->pert_leaf_count += x->pert_leaf_count;
     if(!(--y->pert_child_count)) Q.append(y);
    }

   if(!template_P1(x))
    if(!template_P3(x))
     if(!template_P5(x))
      if(!template_Q1(x))
       if(!template_Q2(x)) return false;
   }
 else
 { //pruned pert subtree root (PRUNED(T,S))
   p_root->mark(DPQ_full);                      // ROOT(T,S) reached
   if(x->parent)  bubble_reset(x->parent);

   if(!template_P1(x))
    if(!template_P2(x))
     if(!template_P4(x))
      if(!template_P6(x))
       if(!template_Q1(x))
        if(!template_Q2(x))
         if(!template_Q3(x)) return false;
  }

 } // while Q not empty

#ifndef NDEBUG
cons_pq_tree(root,"b","reduce 2");
//show("in reduce, der gesamte Baum nach erfolgter reduction", root);
#endif

return true;

}





void pq_tree::frontier(list<int> &F)
{ F.clear(); 
  sequence(F,root); 
 }





void pq_tree::pert_sequence(list<int> &S) 
{
  int i;

  #ifndef NDEBUG
  if(!p_root->left_most){
   cout << "p_root->left_most gleich NULL in pert_sequebce" << endl;
   exit(1);
  }
  #endif

  pq_node l = p_root->left_most;
  pq_node k;

  S.clear();
  S.push(0);                   // preparation

  if (l == p_root->right_most) // if p_root has only one child life is easy.
  { sequence(S,l);  
    return; 
   }

        // Otherwise we have to find the direction to p_root->right_most.
        // Note that an endmost of p_root not necessary has any NULL-link.

  if(l->link1)
    { k = l->link1;
      skip_di(k,l);
      l = (k && k->check(DPQ_full)) ? p_root->left_most->link2 : 
                                      p_root->left_most->link1; 
     }
  else  
    l = NULL;
  
  k = p_root->left_most;
  
  // The direction is detected, we can scan the sequence S.

  do 
  { if(k->check(DPQ_DIR))       //do not call sequence(k)
     { i = S.pop() + 1;         //insert one more DIR-ptr in the sequence S
       if(k->link1 == l)        //then k is directed with the actual sequence
          S.push(k->leaf_index);
       else                     //then k is directed against the actual sequence
          S.push(-k->leaf_index);
       S.push(i);
       Processed.push(k);       //Now we have to delete the DIR-ptr in update()
     }
    else  
      sequence(S,k);
  
    if (k->link1 == l) 
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }

   } while(l != p_root->right_most);
  
}





bool pq_tree::array_copy()
{ 
  // if any leaf_index > leaf_size occurs we double the size of array
  // leaves, copy the content of the old array in the lower half of
  // the new and delete the old. Initially leaf_size is 16.
  
  int i;
  
  if (leaf_size)
    { pq_node* A = new pq_node[2*leaf_size];
      if (!A) error_handler(1,"pq_tree: out of memory");
      for (i=0; i < leaf_size; i++)  A[i] = leaves[i];
      leaf_size *= 2;
      while (i < leaf_size)  A[i++] = NULL;
      delete[] leaves;
      leaves = A;
     }
  else
    { leaf_size = 16;
      leaves = new pq_node[leaf_size];
      if (!leaves) error_handler(1,"pq_tree: out of memory");
      for(i = 0; i < 16; i++)  leaves[i] = NULL;
     }

  return true;
}





void pq_tree::del_pert_subtree()
{
  pq_node k = p_root->left_most;
  pq_node l = 0;

  while(k && k != p_root->right_most)
  { if (k->link1 == l)
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }

    del_subtree(l); // Note that also l is deleted in this function call.
                    // But to progress correctly in the chain we need l's
                    // value (see previous command).
  }
}






void pq_tree::del_subtree(pq_node x)
{
  pq_node k = x->left_most;
  pq_node l = 0;

  while(k)      
  { //Note that the value of l is essentially for the loop
    //but the pointer l is not valid 
    //(see also the comment in del_pert_subtree() ).

    if (k->link1 == l)
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }

    del_subtree(l);     //delete the subtree recursivly
  }

  delete x;

}






bool pq_tree::update(list<int>& S)
{
  int i=0;
  pq_node v,w,k=0,l;

  if (S.empty())
  { // delete the pert. subtree contained in processed...

    while(!Processed.empty())
    { w = Processed.pop();
      if(w->check(DPQ_L))  leaves[w->leaf_index] = NULL;
      delete w;
     } 

   // if there's anything else in the pq_tree the reduction has failed.

   for(i=0; i < leaf_size && !leaves[i]; i++);

   return i == leaf_size;
  }

  if (root && (p_root->left_most == root))
  { // A special case requires special treatment.
    // Note that p_root has only 1 child in this case.

    while (!Processed.empty())
    { w = Processed.pop();
      if (w->check(DPQ_L)) leaves[w->leaf_index] = NULL;
      delete w;
    } 

    // The pertinent subtree, here that means the whole pq_tree
    // is deleted. The application has failed.
    return false;
   }

  i = 0;
  v = new pq_node_struct;

  if(p_root->left_most && p_root->left_most->check(DPQ_PKK))
    { v->mark(DPQ_PKK); 
      v->parent = p_root->left_most->parent;
     }
  else  
    if(p_root->left_most)  v->mark(DPQ_nPKK);

  if (S.size() == 1)
    { // then create v as a leaf
      v->leaf_index = S.pop();

      while(v->leaf_index >= leaf_size) array_copy();

      if (leaves[v->leaf_index]) return false;

      leaves[v->leaf_index] = v;
      v->mark(DPQ_L);
     }
  else
    { //then create v as a P-node with leaves labelled with the elements of S
      //  v->leaf_index = p_root->leaf_index;  //only for testing
      v->mark(DPQ_P);
      v->child_count = S.size();
      v->left_most = l = new pq_node_struct;
      l->link1 = NULL;
      l->parent = v;
      l->leaf_index = S.pop();

      while(l->leaf_index >= leaf_size) array_copy();

      if(leaves[l->leaf_index])  return false;

      leaves[l->leaf_index]=l;
      l->mark(DPQ_PKK);
      l->mark(DPQ_L);

      while(!S.empty())
      { l->link2 = k = new pq_node_struct;
        k->leaf_index = S.pop();

        while(k->leaf_index >= leaf_size)  array_copy();
 
        if(leaves[k->leaf_index])  return false;
 
        leaves[k->leaf_index] = k;
        k->mark(DPQ_PKK);
        k->mark(DPQ_L);
        k->parent = v;
        k->link1 = l;
        l = k;
       }
      k->link2 = NULL;
      v->right_most = k;
    }
     
   if(!root)
   { //then the pq_tree is just constructed & still empty
     newroot = root = v;
     v->mark(DPQ_nPKK);
     return true;
    }

  //replace the full chain under p_root by v
  k = p_root->left_most;
  l = p_root->right_most;

  if (k == l)
    { //then p_root has only one child        
      v->link1 = k->link1;
      v->link2 = k->link2;
  
     if(v->link1)
     { if(v->link1->link1 == k)  
         v->link1->link1 = v;
       else 
         v->link1->link2 = v;
      }
  
     if(v->link2) 
     { if(v->link2->link1 == k) 
          v->link2->link1 = v;
       else 
          v->link2->link2 = v;
      }
  
     if(!k->link1 || !k->link2)
     { //then k is an endmost of his real father
       if(k->parent->right_most == k)
       { k->parent->right_most = v;
         v->parent = k->parent;
        }
       if(k->parent->left_most == k)
       { k->parent->left_most = v;
         v->parent = k->parent;
        }
      }
    }
  else
   { //then p_root has more then one child
 
     pq_node  k1,l1;

     if(k->link1 && k->link2)
     { // Then k has a full and an empty sibling.
       // The full is contained in the pertinent subtree and to be
       // deleted. The empty becomes a sibling of v.

      if(k->link1->check(DPQ_empt) && k->link1->check(DPQ_nDIR))
         v->link1 = k->link1;
      else
         if(k->link2->check(DPQ_empt) && k->link2->check(DPQ_nDIR))  
             v->link1 = k->link2;
         else
            if(k->link1->check(DPQ_full))  
               v->link1 = k->link2;
            else
               if(k->link2->check(DPQ_full)) 
                  v->link1 = k->link1;
               else
                { k1 = k->link1;  l1 = k;
                  skip_di(k1,l1);
                  if(k1->check(DPQ_empt))  
                     v->link1 = k->link1;
                  else  
                     v->link1 = k->link2;
                 }

       if (v->link1->link1 == k)
           v->link1->link1 = v;
       else
           v->link1->link2 = v;
      }
    else
      { //then k is an endmost of his real father
        v->link1 = NULL; 
        if (k->parent->right_most == k) 
           k->parent->right_most = v;
        else
           k->parent->left_most = v;
        v->parent = k->parent;
       }              

     // v is chained instead of k.

    if(l->link1 && l->link2)
    {  // analogous to treatment of k, as above
       if(l->link1->check(DPQ_empt) && l->link1->check(DPQ_nDIR))
          v->link2 = l->link1;
       else
          if(l->link2->check(DPQ_empt) && l->link2->check(DPQ_nDIR))
             v->link2 = l->link2;
          else
             if(l->link1->check(DPQ_full))
                v->link2 = l->link2;
             else
                if(l->link2->check(DPQ_full))
                   v->link2 = l->link1;
                else
                 { k1 = l->link1;  l1 = l;
                   skip_di(k1,l1);
                   if(k1->check(DPQ_empt))
                      v->link2 = l->link1;
                   else
                      v->link2 = l->link2;
                  }
       if (v->link2->link1 == l) 
           v->link2->link1 = v;
       else
           v->link2->link2 = v;
      }
   else
     { //then l is an endmost of his real father
       v->link2 = NULL;
       if (l->parent->right_most == l) 
          l->parent->right_most = v;
       else
          l->parent->left_most = v;

       v->parent = l->parent;
      }              
   // v is chained instead of l.

  }               

  //Now v replaces the chain of full nodes beyond 
  //the real father.

  //v is in the scanning direction of the pert subtree
  //and if necessary we can add a DIR-ptr

  if(p_root->left_most->check(DPQ_nPKK))
  { //then insert the DIR-ptr, but not as an end_most
    k = new pq_node_struct;

    k->leaf_index = p_root->leaf_index;
    k->mark(DPQ_DIR);
    if (v->link1)
      { k->link1 = v->link1;
        v->link1 = k;
        k->link2 = v;
        k->link1->link1 == v ? (k->link1->link1 = k) : (k->link1->link2 = k);
       }
    else
      { k->link2 = v->link2;
        v->link2 = k;
        k->link1 = v;
        k->link2->link1 == v ? (k->link2->link1 = k) : (k->link2->link2 = k);
       }
   }

   // Prepare p_root for the next possible DIR-ptr

   p_root->leaf_index++;
   p_root->mark(DPQ_RESET);

   //Now we delete the pertinent subtree.

   while(!Processed.empty())
   { w = Processed.pop();
     if(w->check(DPQ_L)) leaves[w->leaf_index] = NULL;
     delete w;
    }

#ifndef NDEBUG
cons_pq_tree(root,"b","update");
#endif
  return true;
} 






void pq_tree::sequence(list<int>& S, pq_node x)
{
 // Note that sequence(x) is only called for x=L,P,Q-node, but not for DIR-ptr. 

  if (x->check(DPQ_L))  
  { S.append(x->leaf_index);
    return;
   }

   pq_node k = x->left_most;
   pq_node l = 0;

   while(k != x->right_most)
   { if(k->check(DPQ_DIR)) // do not call sequence(k)
       { int i = S.pop() +1;   //insert one more DIR-ptr in the sequence S
                                //with respect to it's direction
         if (k->link1 == l) 
            S.push(k->leaf_index);
         else 
            S.push(-k->leaf_index);
         S.push(i);
         Processed.push(k);    // we have to delete the DIR-ptr in update()
       }
     else  
       sequence(S,k);

     if (k->link1 == l) 
        { l = k; k = k->link2; } 
     else
        { l = k; k = k->link1; }
    }

                //The while-loop doesn't process the right_most itselves.

   if(k->check(DPQ_DIR))
     { int i = S.pop() + 1;
       if(k->link1 == l)  
          S.push(k->leaf_index);
       else  
          S.push(-k->leaf_index);
       S.push(i);
       Processed.push(k);               
      }
   else 
      sequence(S,k);
}





//------------------------------------------------------------------------------
// TEMPLATES
//------------------------------------------------------------------------------



bool pq_tree::template_P1(pq_node x)
{

#ifndef NDEBUG
cout << "P1";
#endif

  // template for full P-nodes and leaves

  if (x->full_count != x->child_count) return false;      // empty children ex.
  
  if (!x->check(DPQ_L) && !x->check(DPQ_P)) return false; // no P-node, no leaf


  // Full children of P-nodes will be rearranged as right_most.
  // So when we match the P-node his full children are in a chain that
  // begins with its right_most

  if (x->check(DPQ_PKK) && x->parent && x->parent->right_most != x)
  { 
    if(x->parent->left_most == x)
       x->parent->left_most = (x->link1) ? x->link1 : x->link2;

    if (x->link1)
    { if (x->link1->link1 == x)  
        x->link1->link1 = x->link2;
      else  
        x->link1->link2 = x->link2;
     }

    if (x->link2)
    { if (x->link2->link1 == x)  
         x->link2->link1 = x->link1;
      else  
         x->link2->link2 = x->link1;
     }

    x->link1 = x->parent->right_most;
    x->link2 = NULL;

    if (x->parent->right_most->link1)  
       x->parent->right_most->link2 = x;
    else  
       x->parent->right_most->link1 = x;

    x->parent->right_most = x;
  }
  //full x is sorted to the right side

  x->mark(DPQ_full);
  Processed.push(x);

  if(p_root->check(DPQ_full)) 
    { // x is the root of the pertinent subtree
      p_root->left_most = x;
      p_root->right_most = x;
     }
  else
     if (x->parent) x->parent->full_count++;

  return true;
}






bool pq_tree::template_P2(pq_node x)
{

#ifndef NDEBUG
cout << "P2";
#endif

  // template for the root of the pertinent subtree, 
  // if it is a P-node without part-children

  pq_node y,k,l=NULL;

  if (!x->check(DPQ_P) || x->part_child1)  return false;

  if(x->full_count > 1)
  {  
    //collect x's full children under a new y

    y = new pq_node_struct;
    y->mark(DPQ_P);
    y->mark(DPQ_PKK);
    y->mark(DPQ_full);
    y->child_count = y->full_count = x->full_count;
    y->pert_leaf_count = x->pert_leaf_count;
    y->left_most = y->right_most = x->right_most;
    x->right_most = y;
    x->child_count -= y->child_count - 1;
    y->parent = x;  

    //now y is a son of x & father of the full's
   
    y->left_most->parent = y;
    l = NULL;
    for(int i = 1; i < y->full_count; i++)
    { if (y->left_most->link1 == l)
         { l = y->left_most;
           y->left_most = y->left_most->link2;
          }
      else
         { l = y->left_most;
           y->left_most = y->left_most->link1;
          }
     y->left_most->parent = y;
    }  

    // father pointer of the full's now are valid
   
    k = (y->left_most->link1 == l) ? y->left_most->link2 : y->left_most->link1;

    if (k->link1 == y->left_most)  
       k->link1 = y;
    else 
       k->link2 = y;

    if(y->left_most->link1 == l)  
       y->left_most->link2 = NULL; 
    else  
       y->left_most->link1 = NULL;

    y->link1 = k;
    Processed.push(y);
   }   

   // y-ch.is disconnected from the x-children 

  p_root->left_most = p_root->right_most = x->right_most;
  x->special_reset();

  return true;
}






bool pq_tree::template_P3(pq_node x)
{ 

#ifndef NDEBUG
cout << "P3";
#endif

  // template for a P-node that is not root of the pertinent
  // subtree, without part-child
  // x is not ROOT(T,S) <==> !p_root->check(DPQ_full)

  pq_node k,y,z,l;

  assert(x != root);

  if(!x->check(DPQ_P) || x->part_child1)  return false;

  if(x->full_count > 1)
  { 
    //then collect x's full children under a new y

    y = new pq_node_struct;
    y->mark(DPQ_P);
    y->mark(DPQ_full);
    y->child_count = y->full_count = x->full_count;
    x->child_count -= y->child_count - 1;
    y->pert_leaf_count = x->pert_leaf_count;
    x->full_count = 1;
    y->left_most = y->right_most = x->right_most;
    x->right_most = y;
    y->parent = x;                 

    // now y is a son of x & father of the full's
   
    y->left_most->parent = y;
    l = NULL;
    for(int i = 1; i < y->full_count; i++)
    { if (y->left_most->link1 == l)
        { l = y->left_most;
          y->left_most = y->left_most->link2;
         }
      else
        { l = y->left_most;
          y->left_most = y->left_most->link1;
         }
      y->left_most->parent = y;
     }

    //father pointer of the full's now are valid
   
    k = (y->left_most->link1 == l) ? y->left_most->link2 : y->left_most->link1;

    if (k->link1 == y->left_most)  
       k->link1 = y;  
    else 
       k->link2 = y;

    if (y->left_most->link1 == l)  
       y->left_most->link2 = NULL;
    else  
       y->left_most->link1 = NULL;

    y->link1 = k;
    Processed.push(y);
    //y-children are disconnected from the x-children 
   }
  else  
    y = x->right_most;
   

   //now y is the one and only full child of x
    
   if(x->child_count > 2)
   { 
     // is there more then one empty child?
     // then hang x with it's empty children beyond
     // a new pq_node z & arrange z on x's old place

    z = new pq_node_struct;

    if (x->check(DPQ_PKK)) 
       z->mark(DPQ_PKK);  
    else  
       z->mark(DPQ_nPKK);
    
    //for correct blocked chain handling in reduce():

    z->mark(DPQ_ublo);
    if(x == p_root->left_most)  p_root->left_most = z;
    if(x == p_root->right_most) p_root->right_most = z;
   
    z->pert_leaf_count = x->pert_leaf_count;
    x->pert_leaf_count = 0; 

    //now x becomes the father of the empties and a son of z

    z->parent = x->parent;
    z->link1 = x->link1;
    z->link2 = x->link2;

    if (z->link2)  
     if (z->link2->link1 == x)  
        z->link2->link1 = z; 
     else  
        z->link2->link2 = z;

    if (z->link1)  
     if (z->link1->link1 == x) 
        z->link1->link1 = z; 
     else  
        z->link1->link2 = z;

    if (z->parent->right_most == x) z->parent->right_most = z;
    if (z->parent->left_most == x)  z->parent->left_most = z;

    z->left_most = x;
    z->right_most = y;
    x->right_most = y->link1 ? y->link1 : y->link2;

    if (x->right_most->link1 == y) 
       x->right_most->link1 = NULL;
    else  
       x->right_most->link2 = NULL;

    x->parent = y->parent = z;
    x->link1 = NULL;
    x->link2 = y;

    if (y->link1) 
       y->link1 = x; 
    else  
       y->link2 = x;

    x->child_count--;
    x->special_reset(); 
   }
   else  z = x;
                                   //now z becomes partial
   z->mark(DPQ_Q);
   z->left_most->mark(DPQ_nPKK);
   z->right_most->mark(DPQ_nPKK);
   z->mark(DPQ_part);
   
   if(z->parent->part_child1)
     { if(z->parent->part_child2)
         { too_many_part = 1;  
           return false;
          }
       else  
          z->parent->part_child2 = z;
     }
   else  
      z->parent->part_child1 = z;  //z became one of the parents part_child's
   
   z->full_count = 1;
   
   return true;
}





bool pq_tree::template_P4(pq_node x)
{

#ifndef NDEBUG
cout << "P4";
if(p_root->leaf_index > 40) show("in P4",root);
#endif

  // template for the root of the pertinent subtree, if it is a P-node 
  // with exactly one part_child

  pq_node k,l,y;

  if (!x->check(DPQ_P) || x->part_child2)  return false;

  fix_part_child_direction(x);

  if(x->full_count > 1)
  { 
    // then collect x's full children under a new y,
    // y becomes an endmost of x's part_child

    y = new pq_node_struct;
    y->mark(DPQ_P);
    y->mark(DPQ_nPKK);
    y->mark(DPQ_full);
    y->child_count = y->full_count = x->full_count;
    y->pert_leaf_count = x->pert_leaf_count - x->part_child1->pert_leaf_count;
    x->full_count = 0;
    x->child_count -= y->child_count;
    y->left_most = y->right_most = x->right_most;
    y->parent = x->part_child1; 

   //now y is right_most son of part_ch.& father
   //of the full children of x

    y->left_most->parent = y;
    l = NULL;

    for(int i = 1; i < y->full_count; i++)
    { if (y->left_most->link1 == l)
        { l = y->left_most; 
          y->left_most = y->left_most->link2;
         }
      else
        { l = y->left_most, 
          y->left_most = y->left_most->link1;
         }

      y->left_most->parent = y;
    }  

    //father pointer of the full's now are valid

    k = (y->left_most->link1 == l) ? y->left_most->link2 : y->left_most->link1;

    if (k->link1 == y->left_most)  
       k->link1 = NULL;  
    else 
       k->link2 = NULL;

    x->right_most = k;

    if (y->left_most->link1 == l)  
       y->left_most->link2 = NULL; 
    else  
       y->left_most->link1 = NULL;

    y->link1 = x->part_child1->right_most;
    x->part_child1->right_most = y;

    if (y->link1->link1)
       y->link1->link2 = y;  
    else 
       y->link1->link1 = y;

    x->part_child1->child_count++;
    x->part_child1->full_count++;
    x->part_child1->pert_leaf_count += y->pert_leaf_count;
    Processed.push(y);
    //y-ch. is disconnected from the x-children 
    }
   else
     if (x->full_count == 1)
     { y = x->right_most;
       k = (y->link1) ? y->link1 : y->link2;
       x->right_most = k;
       if (k->link1 == y)  
          k->link1 = NULL;  
       else 
          k->link2 = NULL;

       k = x->part_child1->right_most;
       y->mark(DPQ_nPKK);

       if (k->link2)  
          k->link1 = y;  
       else 
          k->link2 = y;

       y->parent = x->part_child1;
       x->part_child1->right_most = y;

       if(y->link1)  
          y->link1 = k;  
       else 
          y->link2 = k;

       x->part_child1->child_count++;
       x->part_child1->full_count++;
       x->part_child1->pert_leaf_count = x->pert_leaf_count;
       x->full_count = 0;
       x->child_count--;
      } 

   // now x has no full child

   y = x->part_child1->right_most;
   p_root->right_most = y;
   k = y->link1 ? y->link1 : y->link2;

   while (k->check(DPQ_full) || k->check(DPQ_DIR))
   { if (k->link1 == y) 
        { y = k; k = k->link2; } 
     else
        { y = k; k = k->link1; }
    }

   p_root->left_most = y;
   y->parent = x->part_child1;
                                //to set p_root over the full chain 

   x->part_child1->special_reset();

   if(x->child_count == 1)
   {  // x became useless and can be deleted

      x->part_child1->parent = x->parent; 
      if(x->check(DPQ_nPKK)) x->part_child1->mark(DPQ_nPKK);

      x->part_child1->link1 = x->link1;
      x->part_child1->link2 = x->link2;
     
      if (x->link2)
        if (x->link2->link1 == x)  
           x->link2->link1 = x->part_child1;
        else  
           x->link2->link2 = x->part_child1;
     
      if (x->link1)
        if (x->link1->link1 == x)  
           x->link1->link1 = x->part_child1;
        else  
           x->link1->link2 = x->part_child1;
     
      if(x->parent)
       { if(x->parent->right_most == x) x->parent->right_most = x->part_child1;
         if(x->parent->left_most == x)  x->parent->left_most  = x->part_child1;
        }
      else  
        if(x == root)
        { root=x->part_child1;
          root->parent = root->link1 = root->link2 = NULL;
         }

     delete x;
   }
   else  
      x->special_reset();

return true;

}





bool pq_tree::template_P5(pq_node x)
{

#ifndef NDEBUG
cout << "P5";
#endif

  // template for a P-node that is not the root of the pertinent subtree,
  // with exactly one part_child

  pq_node y;
  pq_node k;
  pq_node l = NULL;
  
  #ifndef NDEBUG
  assert(x != root);
  cons_pq_tree(root,"b","P5"); 
  #endif

  if ( !x->check(DPQ_P) || x->part_child2 || too_many_part )  return false;
  
  fix_part_child_direction(x);
  
  if (x->full_count > 1)
  { // collect x's full children under a new y,
    // y becomes an endmost of x's part_child
    y = new pq_node_struct;
    y->mark(DPQ_P);
    y->mark(DPQ_nPKK);
    y->mark(DPQ_full);
    y->child_count = y->full_count = x->full_count;
    y->pert_leaf_count = x->pert_leaf_count - x->part_child1->pert_leaf_count;
    x->full_count = 0;
    x->child_count -= y->child_count;
    y->left_most = y->right_most = x->right_most;
    y->parent = x->part_child1;    

    // now y is right_most son of part_ch.& father
    // of the full children of x
  
    y->left_most->parent = y;
    l = NULL;
    for(int i = 1; i < y->full_count; i++)
    { if (y->left_most->link1 == l)
        { l = y->left_most;
          y->left_most = y->left_most->link2;
         }
      else
        { l = y->left_most;
          y->left_most = y->left_most->link1;
         }
      y->left_most->parent = y;
     } 

     // father pointer of the full's now are valid
  
    k = (y->left_most->link1 == l) ? y->left_most->link2 : y->left_most->link1;

    if (k->link1 == y->left_most)  
        k->link1 = NULL;
    else 
        k->link2 = NULL;

    x->right_most = k;

    if (y->left_most->link1 == l)  
        y->left_most->link2 = NULL; 
    else 
        y->left_most->link1 = NULL;

    y->link1 = x->part_child1->right_most;
   
   x->part_child1->right_most = y;

   if (y->link1->link1)
       y->link1->link2 = y;
   else 
       y->link1->link1 = y;

   x->part_child1->child_count++;
   x->part_child1->full_count++;
   x->part_child1->pert_leaf_count += y->pert_leaf_count;
   Processed.push(y);
  }                               // y-ch. is disconnected from the x-children 
  else
  if(x->full_count == 1)
  { y = x->right_most;
    k = y->link1 ? y->link1 : y->link2;
    x->right_most = k;

    if (k->link1 == y)
        k->link1 = NULL;  
    else
        k->link2 = NULL;

    k = x->part_child1->right_most;
    y->mark(DPQ_nPKK);

    if (k->link2)
        k->link1 = y;
    else
        k->link2 = y;

    y->parent = x->part_child1;
    x->part_child1->right_most = y;
    if (y->link1)
        y->link1 = k;
    else
        y->link2 = k;

    x->part_child1->child_count++;
    x->part_child1->full_count++;
    x->part_child1->pert_leaf_count = x->pert_leaf_count;
    x->full_count = 0;
    x->child_count--;
  } // now x has no full child
   
  if(x->child_count > 2)
  { // Is there more then one empty child?
    // Then hang x with it's empty children beyond the part_child

    if(!x->check(DPQ_PKK))  x->part_child1->mark(DPQ_nPKK);
    x->part_child1->pert_leaf_count = x->pert_leaf_count;
    x->child_count--;
    x->pert_leaf_count = 0;  

    // now x becomes the father of the empties & a son of it's part_child.
    // The part_child will replace x.

    l = x->part_child1;
    x->part_child1 = NULL;

    if(l->link1)
      if (l->link1->link1 == l)  
         l->link1->link1 = l->link2; 
      else  
         l->link1->link2 = l->link2;

    if(l->link2)
      if (l->link2->link1 == l)
         l->link2->link1 = l->link1;
      else 
         l->link2->link2 = l->link1;

    if (x->right_most == l) x->right_most = l->link1 ? l->link1 : l->link2;

    if (x->left_most == l)  x->left_most = l->link1 ? l->link1 : l->link2;

    // now part_child l is removed from the x-children's chain
  
    l->parent = x->parent; 
    l->link1 = x->link1;
    l->link2 = x->link2;

    if (l->link2)  
      if (l->link2->link1 == x)  
         l->link2->link1 = l;
      else  
         l->link2->link2 = l;

    if (l->link1)  
      if (l->link1->link1 == x)
         l->link1->link1=l;
      else 
         l->link1->link2 = l;

    if(l->parent->right_most == x)  l->parent->right_most = l;

    if(l->parent->left_most == x)  l->parent->left_most = l;

    if (l->parent->part_child1)
     { if(l->parent->part_child2) 
         { too_many_part = 1;  
//           return false;
          }
       else  
         l->parent->part_child2 = l;
      }
    else  
       l->parent->part_child1 = l;

    // l becomes one of the parents part_child's
    // now in the chain of it's brothers x is replaced by
    // it's former part_child
  
   if (l->left_most->link1)  
      l->left_most->link2 = x;
   else  
      l->left_most->link1 = x;

   x->link1 = NULL;
   x->link2 = l->left_most;
   l->left_most = x;
   x->parent = l;
   x->mark(DPQ_nPKK);
   // now x is a child of it's former part_child
   x->special_reset(); 

  } // if (x->child_count > 2) ...

  else 
    { // one or no empty child
      l = x->part_child1;      // to decrease pointer dereferencing
      if(x->child_count == 2)  // one empty child (the other is the part_child)
      {                        // then hang the empty beyond the partial l
        y = (x->right_most == l) ? x->left_most : x->right_most;
        y->mark(DPQ_nPKK);
        y->parent = l;
        y->link1 = NULL;
        y->link2 = l->left_most;
        if (l->left_most->link1)  
           l->left_most->link2 = y;
        else  
           l->left_most->link1 = y;
        l->left_most = y;
       } // now x has only one child and is useless
  
      l->parent = x->parent;
      if (x->check(DPQ_nPKK)) l->mark(DPQ_nPKK); 
     
      // for correct blocked chain handling in reduce():
      l->mark(DPQ_ublo);
      if (x == p_root->left_most)  p_root->left_most = l;
      if (x == p_root->right_most) p_root->right_most = l;
/*     
      if (l->parent->part_child1)
        { if (l->parent->part_child2) 
            { too_many_part = 1;  
              return false;
             }
          else 
             l->parent->part_child2 = l;
         }
      else  // l becomes one of the parents part_child's
        l->parent->part_child1 = l;   
*/    
      l->link1 = x->link1;
      l->link2 = x->link2;

      if (l->link1) 
        if (l->link1->link1 == x)  
           l->link1->link1 = l; 
        else  
           l->link1->link2 = l;

      if (l->link2) 
        if (l->link2->link1 == x)  
           l->link2->link1 = l; 
        else  
           l->link2->link2 = l;
 

     if (l->parent->part_child1)
        { if (l->parent->part_child2) 
            { too_many_part = 1;  
//              delete x;
//              return false;
             }
          else 
             l->parent->part_child2 = l;
         }
      else  // l becomes one of the parents part_child's
        l->parent->part_child1 = l;   

      if (l->parent->right_most == x) l->parent->right_most = l;
      if (l->parent->left_most == x)  l->parent->left_most = l;
      // x is replaced by it's former part_child & can be deleted
      delete x;
     }
if(too_many_part)  return false;  
 return true;
}






bool pq_tree::template_P6(pq_node x)
{

#ifndef NDEBUG
cout << "P6";
#endif

 // template for a P-node with 2 partial children
 
 pq_node k,l,y;
 
 if (!x->check(DPQ_P) || !x->part_child2 || too_many_part)  return false;
 
 if(x->part_child2->right_most->check(DPQ_full))
 { k = x->part_child2->right_most;
   x->part_child2->right_most = x->part_child2->left_most;
   x->part_child2->left_most = k;
  }
 
 // Now the full children of part_child2 are on it's lefthand side.

 if(x->part_child1->left_most->check(DPQ_full))
 { k = x->part_child1->left_most;
   x->part_child1->left_most = x->part_child1->right_most;
   x->part_child1->right_most = k;
 } 

 // Now the full children of part_child1 are on it's right hand side.
 fix_part_child_direction(x,1);

 if (x->full_count > 1)
    { // collect x's full children under a new y,
      // y becomes an endmost of x's part_child1
      y = new pq_node_struct;
      y->mark(DPQ_P);
      y->mark(DPQ_nPKK);
      y->mark(DPQ_full);
      y->child_count = y->full_count = x->full_count;
      y->pert_leaf_count = x->pert_leaf_count 
                         - x->part_child1->pert_leaf_count 
                         - x->part_child2->pert_leaf_count;
      x->full_count = 0;
      x->child_count -= y->child_count;
      y->left_most = y->right_most = x->right_most;
      y->parent = x->part_child1;
   
      // now y is right_most son of part_ch.& father
      // of the full children of x
     
      y->left_most->parent = y;
      l = NULL;
      for(int i = 1; i < y->full_count; i++)
      { if (y->left_most->link1 == l)
          { l = y->left_most;
            y->left_most = y->left_most->link2;
           }
        else
          { l = y->left_most;
            y->left_most = y->left_most->link1;
           }
        y->left_most->parent = y;
       } 
       // father pointer of the full's now are valid
     
      k = y->left_most->link1 == l ? y->left_most->link2 : y->left_most->link1;
   
      if (k->link1 == y->left_most)
         k->link1 = NULL;  
      else  
         k->link2 = NULL;
   
      x->right_most = k;
   
      if (y->left_most->link1 == l)
         y->left_most->link2 = NULL; 
      else
         y->left_most->link1 = NULL;
   
      y->link1 = x->part_child1->right_most;
      
      x->part_child1->right_most = y;
   
      if (y->link1->link1)
         y->link1->link2 = y;
      else
         y->link1->link1 = y;
   
      x->part_child1->child_count++;
      x->part_child1->full_count++;
      x->part_child1->pert_leaf_count += y->pert_leaf_count;
      Processed.push(y);
     } // y-ch. is disconnected from the x-children 
  else
     if (x->full_count == 1)
     { // hang the full child beyond part_child1
       y = x->right_most;
       k = y->link1 ? y->link1 : y->link2;
       x->right_most = k;

       if (k->link1 == y) 
          k->link1=NULL;  
       else  
          k->link2 = NULL;

       k = x->part_child1->right_most;
       y->mark(DPQ_nPKK);

       if (k->link2)  
          k->link1 = y;  
       else
          k->link2 = y;

       y->parent = x->part_child1;
       x->part_child1->right_most = y;

       if (y->link1) 
          y->link1 = k;
       else
          y->link2 = k;

       x->part_child1->child_count++;
       x->part_child1->full_count++;
       x->part_child1->pert_leaf_count = x->pert_leaf_count;
       x->pert_leaf_count = 0;
       x->full_count = 0;
       x->child_count--;
      }    

  // now x has no full child 
  // it's a good time to set p_root

  l = x->part_child1->right_most;
  k = l->link1 ? l->link1 : l->link2;

  while (k->check(DPQ_full) || k->check(DPQ_DIR))
     if (k->link1 == l)
       { l = k; k = k->link2; }
     else
       { l = k; k = k->link1; }

  p_root->left_most = l;
  l->parent = x->part_child1;
  l = x->part_child2->left_most;
  k = l->link1 ? l->link1 : l->link2;

  while (k->check(DPQ_full) || k->check(DPQ_DIR))
     if (k->link1 == l) 
       { l = k; k = k->link2; }
     else
       { l = k; k = l->link1; }

  p_root->right_most = l;

  // now meld together the part_children

  k = x->part_child1; 
  l = x->part_child2;
  x->part_child2 = NULL;

  if (x->left_most == l)  x->left_most  = l->link1 ? l->link1 : l->link2;
  if (x->right_most == l) x->right_most = l->link1 ? l->link1 : l->link2;

  if (l->link1) 
    if (l->link1->link1 == l)
       l->link1->link1 = l->link2;
    else
       l->link1->link2 = l->link2;

  if (l->link2)
    if (l->link2->link1 == l)
       l->link2->link1 = l->link1;
    else
       l->link2->link2 = l->link1;

  l->left_most->parent = l->right_most->parent = k;

  if (k->right_most->link1)
     k->right_most->link2 = l->left_most;
  else
     k->right_most->link1 = l->left_most;

  if (l->left_most->link1)
     l->left_most->link2 = k->right_most;
  else
     l->left_most->link1 = k->right_most;

  k->right_most = l->right_most;

  delete l;  
  l = NULL;
  x->child_count--;   // part_child1 got the children of part_child2
  
  x->part_child1->special_reset();
  
  if(x->child_count != 1)
     x->special_reset();
  else
    { // x is useless & can be deleted
      k->parent = x->parent;
      if(x->check(DPQ_nPKK))  k->mark(DPQ_nPKK);
      k->link1 = x->link1;
      k->link2 = x->link2;

      if(k->link1)
         if (k->link1->link1 == x)
            k->link1->link1 = k;
         else
            k->link1->link2 = k;

      if(k->link2)
         if (k->link2->link1 == x)
            k->link2->link1 = k;
         else
            k->link2->link2 = k;

      if (k->parent)
        { if (k->parent->right_most == x) k->parent->right_most = k;
          if (k->parent->left_most == x)  k->parent->left_most = k;
         }
      else  
        if(x == root)
        { root = k;
          root->parent = root->link1 = root->link2 = NULL;
         }
      delete x;
     }
  
  return true;
}





bool pq_tree::template_Q1(pq_node x)
{

#ifndef NDEBUG
cout << "Q1";
#endif

  // template for full Q-nodes

  pq_node k;      // to run through the children
  pq_node l;      // dito
  
  if ( !x->check(DPQ_Q) || x->part_child1 ) return false;
  
  if (x == p_root) return true;

  l = x->right_most;
  k = l->link1 ? l->link1 : l->link2;

  if (l->check(DPQ_DIR))
    { skip_di(k,l);
      if(!k->check(DPQ_full))  return false;
      k->link1 == l ? (l = k, k = k->link2) : (l = k, k = k->link1);
     }
  else  
    if (!l->check(DPQ_full))  return false;

  for (int i = 1; i < x->full_count; i++)
  { skip_di(k,l);
    if(!k->check(DPQ_full))  return false;
    k->link1 == l ? (l = k, k = k->link2) : (l = k,k = k->link1);
   }

  skip_di(k,l);

  if (k && k->check(DPQ_empt))  return false;
  
  //template is applicable, now the "replacement":
  
  if (p_root->check(DPQ_full))
    { p_root->left_most = p_root->right_most = x;
      Processed.push(x);
     }
  else
    { if (x->check(DPQ_PKK) && x->parent->right_most != x)
      { if (x->parent->left_most == x)
          x->parent->left_most = x->link1 ? x->link1 : x->link2;

        if (x->link1)
          if (x->link1->link1 == x)
             x->link1->link1 = x->link2;
          else
             x->link1->link2 = x->link2;

        if (x->link2)
          if (x->link2->link1 == x)
             x->link2->link1 = x->link1;
          else
             x->link2->link2 = x->link1;

        x->link1 = x->parent->right_most;
        x->link2 = NULL;

        if (x->parent->right_most->link1)
           x->parent->right_most->link2 = x;
        else
           x->parent->right_most->link1 = x;

        x->parent->right_most = x;
       } 
      // full x is sorted to the right side

      x->parent->full_count++;
      x->mark(DPQ_full);
      Processed.push(x);
    }
  
 return true;
}





bool pq_tree::template_Q2(pq_node x)
{

#ifndef NDEBUG
cout << "Q2" << endl;
if(p_root->leaf_index > 1)  show("in Q2",x);
//show("in Q2",x);
//show("in Q2",root);
#endif

  // template for a Q-node with empties and/or 1 partial child
  
  pq_node k;                      // to run through the children
  pq_node l;                      // dito
  pq_node l1,l2;                  // dito
  pq_node m,n;                    // dito, but see text
  pq_node p1=NULL;                // the part_child
  pq_node d1;                     // the side the full's will be turned to ...
  pq_node d2;                     // and the other side
  pq_node aux_leftm=NULL;
  pq_node aux_rightm=NULL;        // dummies for p_root->end_most's
  
  bool    full_found=false;       // turn the full's to the outside
  bool    a,b,c,d;                // for testing of blocked chain
  
  
  if(!x->check(DPQ_Q) || x->part_child2)  return false;   
  
  // only 0 or 1 part_child is allowed here
  
  if (p1 = x->part_child1)
  {
    l1 = l2 = x->part_child1;
  
    m = p1->link1;
    skip_di(m,l1);
  
    n = p1->link2;
    skip_di(n,l2);
  
   if (x->full_count)
   {
    a = m ? m->check(DPQ_full) : 0;
    b = n ? n->check(DPQ_full) : 0;   
  
    if (!(a ^ b))  return false;   //xor; (both (full)) or (both (NULL or empty))
  
    if(x->full_count == 1)
    {
      if (a) 
        { d1 = p1->link1; 
          d2 = p1->link2; }
      else
        { d2 = p1->link1; 
          d1 = p1->link2; }
    
//       if (!p_root->check(DPQ_Q) || x == p_root )
//       if (p_root->check(DPQ_full))
//       if (p_root->check(DPQ_full) ^ !p_root->check(DPQ_Q))
//       {
         if (a)
            { k = m; l = l1; }
         else
            { k = n; l = l2; }
     
         if  (k->link1 == l) 
            { l = k; k = k->link2; }
         else 
            { l = k; k = k->link1; }
     
         skip_di(k,l);
         aux_leftm = l;
//        }   
        // in case of a blocked chain the end_most's of p_root are already used
        // and valid
        // next reduce round Q2 will match p_root as the father of the blocked chain
    
       full_found = true;
     
      } //if(x->full_count == 1)
    else
      {
         if (a)
            { k = m; l = l1; }
         else
            { k = n; l = l2; }
  
         for(int i = 2; i <= x->full_count; i++)
         { if (k->link1 == l) 
             { l = k; k = k->link2; }
           else
             { l = k; k = k->link1; }
  
          skip_di(k,l);
          if( !k || !k->check(DPQ_full) )  return false;      
          //there is an empty between the full's
          //or not all full's are at one side
         }
  
         full_found = true;
  
//         if(!p_root->check(DPQ_Q) || x == p_root){
//         if(p_root->check(DPQ_full)){
          if (k->link1 == l) 
             { l = k; k = k->link2; }
           else
             { l = k; k = k->link1; }
  
           skip_di(k,l);
           aux_leftm = l;
//         }    //to take the DIR's
  
         if (a) 
           { d1 = p1->link1; d2 = p1->link2; } 
         else
           { d2 = p1->link1; d1 = p1->link2; }
      
        }
      
    } //if(x->full_count)
  else
    { //... no full's, only a part_child
      
      if(p_root->check(DPQ_full))
      { // ROOT(T,S) is reached
      
        d1 = p1->link1;  
        d2 = p1->link2;
  
        if(d1->check(DPQ_DIR))
        { l = p1;  
          k = d1;  
          skip_di(k,l);
          aux_leftm = l;
         }    
        else
          if (p1->right_most->check(DPQ_full)) 
             aux_leftm = p1->right_most;
          else  
             aux_leftm = p1->left_most;
       }
      else
      { if (m && n)  return false;        //-part_child is between empties
  
        d1 = NULL;                        //-part_child's full endmost will be
        d2 = m ? p1->link1 : p1->link2;   //turned outside and becomes an end_most
       }
   }
  
  } //end of "if(p1 = x->part_child1)"
  else
  { // there's no part_child
  
   if (x->left_most->check(DPQ_full))
     l = x->left_most;
   else
     { l = x->right_most;
       if (!l->check(DPQ_full))  return false;
      }
  
//   if (!p_root->check(DPQ_Q) || x == p_root)  aux_rightm = aux_leftm = l;
        aux_rightm = aux_leftm = l;
  
   if (x->full_count > 1)
   {
     k = l->link1 ? l->link1 : l->link2;
     skip_di(k,l);
  
     if(k->check(DPQ_full))
     { for(int i = 2; i <= x->full_count; i++)
       { skip_di(k,l);
         if (!k->check(DPQ_full))  return false;
         if (k->link1 == l) 
           { l = k; k = k->link2; }
         else
           { l = k; k = k->link1; }
        }
//       if(!p_root->check(DPQ_Q) || x == p_root )
//       if(p_root->check(DPQ_full))
//       { 
         skip_di(k,l);
         aux_leftm = l;
//        }
      }
     else return false;    // empties are intermingled with full's
   }
  } //else from "if (p1 = x->part_child1)"
                           //no empty's and no part_child between full's
                           //template applicable, now the replacement:
  if(p1)
  {
   if (p1->right_most->check(DPQ_full))  
     { m = p1->right_most; n = p1->left_most; }
   else 
     { m = p1->left_most, n = p1->right_most; }
  
   if (m->link1) 
      m->link2 = d1;
   else
      m->link1 = d1;
  
   if (d1) 
      if (d1->link1 == p1) 
         d1->link1 = m;
      else
         d1->link2 = m;
  
   if (n->link1) 
      n->link2 = d2;
   else
      n->link1 = d2;
  
   if (d2) 
      if (d2->link1 == p1) 
         d2->link1 = n;
      else
         d2->link2 = n;
  
   if(p1 == x->left_most)
   { x->left_most = full_found ? n : m;
     x->left_most->parent = x;
    }
  
   if(p1 == x->right_most) 
   { x->right_most = full_found ? n : m;
     x->right_most->parent = x;
    }
  
   x->part_child1 = NULL;

   //reversed (if necessary) and chained

   x->full_count += p1->full_count;
   delete p1;
  
   l = d1;  
   k = m;
  
   while (k && k->check(DPQ_full))
   { k->parent = x;
     if (k->link1 == l) 
       { l = k; k = k->link2; }
     else
       { l = k; k = k->link1; }
  
     skip_di(k,l);
    }
  
//   if(!p_root->check(DPQ_Q) || x == p_root )
//   if(p_root->check(DPQ_full))
     aux_rightm = l;
  
  }
   
  if (aux_leftm && aux_rightm)
  { c = aux_leftm->link1 && aux_leftm->link2;
    d = aux_rightm->link1 && aux_rightm->link2;
    if(c && d)  x->mark(DPQ_dpar);
   }


  if(p_root->check(DPQ_full) || x == p_root){
    p_root->left_most = aux_leftm;
    p_root->right_most = aux_rightm;
  }
  else
  if(x->check(DPQ_dpar))  return false;

/*
  if (x->check(DPQ_dpar))
     { 
      if (p_root->check(DPQ_Q))
         { if (x != p_root)  return false;
           p_root->left_most = aux_leftm;
           p_root->right_most = aux_rightm;
          }
      else
         { if(p_root->check(DPQ_full))
           { p_root->left_most = aux_leftm;
             p_root->right_most = aux_rightm;
            }
           else  return false;
          }
     }
  else
     { if(  !p_root->check(DPQ_Q) &&  p_root->check(DPQ_full))
       { p_root->left_most = aux_leftm;
         p_root->right_most = aux_rightm;
        }
      }
*/
     
  if(x != p_root)
  { // father pointer of x is valid, that means != NULL
   if(p_root->check(DPQ_full)) // ROOT(T,S) reached
       x->special_reset();
   else
     { x->mark(DPQ_part);
       if(x->parent->part_child1)
         { if(x->parent->part_child2)
             { too_many_part = 1;  
               return false;
              }
           else  
              x->parent->part_child2 = x;
          }
       else  
          x->parent->part_child1 = x;
                  //x becomes one of the x->parent->part_children
      }
   }
  
return true;

}






bool pq_tree::template_Q3(pq_node x)
{

#ifndef NDEBUG
cout << "Q3";
#endif

  // template for a Q-node with exactly 2 partial children

  pq_node  k;      //to run through the children
  pq_node  l;      //dito
  pq_node  p1;     //a part_child ...
  pq_node  d1;     //... and its neighbour in direction to the full's
  pq_node  p2;     //dito
  pq_node  d2;     //dito
  pq_node  mm,m;   //to run through the children
  pq_node  nn,n;   //dito
  pq_node  l1,l2;  //dito
  
  unsigned char   cc;     //for a check
  
  bool a,b;
  
  if ( !p_root->check(DPQ_full) ||
       !x->check(DPQ_Q) || 
       too_many_part || 
       !x->part_child2 ) return false;
  
  
   l1 = l2 = p1 = x->part_child1;
   m = p1->link1;
   skip_di(m,l1);
  
   n = p1->link2;
   skip_di(n,l2);
  
   l1 = l2 = p2 = x->part_child2;
   mm = p2->link1;
   skip_di(mm,l1);
   nn = p2->link2;
   skip_di(nn,l2);
   
   if(x->full_count)
   { // x has full children
    
    cc = m ? (m->check(DPQ_part) ? 1 : (m->check(DPQ_full) ? 2 : 0)) : 0;
    cc += n ? (n->check(DPQ_part) ? 1 : (n->check(DPQ_full) ? 2 : 0)) : 0;
    if(cc != 2)  return false;

    d1 = (m && m->check(DPQ_full)) ?  p1->link1 : p1->link2;
  
    cc = mm ? (mm->check(DPQ_part) ? 1 : (mm->check(DPQ_full) ? 2 : 0)) : 0;
    cc += nn ? (nn->check(DPQ_part) ? 1 : (nn->check(DPQ_full) ? 2 : 0)) : 0;
    if(cc != 2)  return false;

    d2 = (mm && mm->check(DPQ_full)) ? p2->link1 : p2->link2;
    
    /*
    explanation:
    check values for
    NULL empty part full
     0     0     1    2
  
    One sibling is "NULL" or "empty" and the other is "full" is a necessary 
    condition here for a valid Q3 situation (the rest of the test follows).
    */
  
    if(x->full_count > 1)
    {
      if (mm && mm->check(DPQ_full))
        { d2 = p2->link1;  k = mm;  l = l1;}
     else
        { d2 = p2->link2;  k = nn;  l = l2;}

     for(int i = 2; i <= x->full_count; i++)
     { if (k->link1 == l) 
          { l = k; k = k->link2; }
       else
          { l = k; k = k->link1; }
      skip_di(k,l);

      if(!k || !k->check(DPQ_full))  return false;  
         // because there is an empty or an part_child between the full's
     }
     
     if (k->link1 == l) 
       { l = k; k = k->link2; }
     else
       { l = k; k = k->link1; }
     skip_di(k,l);

     if (!k || !k->check(DPQ_part))  return false;  
    }
   }
   else
    { // x has no full child
  
      a = m ? m->check(DPQ_part) : 0;
      b = n ? n->check(DPQ_part) : 0;
      d1 = a ? p1->link1 : p1->link2;
      if (!(a ^ b))  return false;   
  
      a = mm ? mm->check(DPQ_part) : 0;
      b = nn ? nn->check(DPQ_part) : 0;
      d2 = a ? p2->link1 : p2->link2;
      if (!(a ^ b))  return false;   
    }
  
  //no empties and no part_child between full's
  //template is applicable, now the replacement:
  
  pq_node m1,m2;  //the full end_most of a part_child
  pq_node n1,n2;  //the empty end_most of a part_child
  pq_node o;      //a dummy
  
  if(d1 == p2)
  { // the partial children are neighbours
  
   if (p1->right_most->check(DPQ_full))
      { m1 = p1->right_most;
        n1 = p1->left_most;
       }
   else
      { m1 = p1->left_most;
        n1 = p1->right_most;
       }
  
   if (p2->right_most->check(DPQ_full))
      { m2 = p2->right_most;
        n2 = p2->left_most;
       }
   else 
      { m2 = p2->left_most;
        n2 = p2->right_most;
       }
  
   o = (p1->link1 == d1) ? p1->link2 : p1->link1;
  
   if (m1->link1) m1->link2 = m2; else m1->link1 = m2;
   if (m2->link1) m2->link2 = m1; else m2->link1 = m1;
   if (n1->link1) n1->link2 = o;  else n1->link1 = o;
  
   if (o)
     if (o->link1 == p1)
        o->link1 = n1;
     else
        o->link2 = n1;
  
   o = (p2->link1 == d2) ? p2->link2 : p2->link1;
  
   if (n2->link1) 
      n2->link2 = o;
   else
      n2->link1 = o;
  
   if (o)
     if (o->link1 == p2)
        o->link1 = n2;
     else
        o->link2 = n2;
  
   d1 = m2;
   d2 = m1;
                          //reversed and chained
  }
  else
  { // full children between the partial
  
   if (p1->right_most->check(DPQ_full))  
     { m1 = p1->right_most; 
       n1 = p1->left_most;
      }
   else 
     { m1 = p1->left_most; 
       n1 = p1->right_most;
      }
  
   o = (p1->link1 == d1) ? p1->link2 : p1->link1;
  
   if (m1->link1)       m1->link2 = d1; else m1->link1 = d1; 
   if (d1->link1 == p1) d1->link1 = m1; else d1->link2 = m1;
   if (n1->link1)       n1->link2 = o;  else n1->link1 = o;
  
   if(o)
      if(o->link1 == p1)  
         o->link1 = n1;
      else  
         o->link2 = n1;
  
   if(p2->right_most->check(DPQ_full))  
      { m2 = p2->right_most; 
        n2 = p2->left_most; 
       }
   else 
      { m2 = p2->left_most; 
        n2 = p2->right_most; 
       }
  
   o = (p2->link1 == d2) ? p2->link2 : p2->link1;
  
   if (m2->link1)       m2->link2 = d2; else m2->link1 = d2;
   if (d2->link1 == p2) d2->link1 = m2; else d2->link2 = m2;
   if (n2->link1)       n2->link2 = o;  else n2->link1 = o;
  
  
   if(o)
      if(o->link1 == p2)  
         o->link1 = n2;
      else  
         o->link2 = n2;
  
   // reversed and chained
  }
  
  l = d1;  
  k = m1;
  
  while(k->check(DPQ_full))
  { if (k->link1 == l) 
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }
    skip_di(k,l);
   }
  
  p_root->left_most = l;
  
  l = d2;  
  k = m2;
  
  while (k->check(DPQ_full))
  { if (k->link1 == l) 
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }
  
    skip_di(k,l);
   }
  
  p_root->right_most = l;         //end_most of p_root now is valid
  
  if (p1 == x->left_most)
    { x->left_most = n1;
      n1->parent = x;
     }
  else
    if (p2 == x->left_most)
    { x->left_most = n2;
      n2->parent = x;
     }
  
  if (p1 == x->right_most)
    { x->right_most = n1;
      n1->parent = x;
     }
  else
    if (p2 == x->right_most)
    { x->right_most = n2;
      n2->parent = x;
     }
  
  if (x != p_root)  x->special_reset();
  
  p_root->left_most->parent = x;
  
  delete p1;  
  delete p2;
  
  x->part_child1 = x->part_child2 = NULL;
  
  return true;
}






bool pq_tree::reduction(list<int>& S)
{ 
  reset(); 
  positiv = (bubble(S) && reduce(S)); 

  #ifndef NDEBUG
//  if(p_root->leaf_index > 32)  show("in pert_sequence",root);
  if(!positiv)  show("not positiv",root);
//  else  cout << "positiv" << endl;
  #endif

  return positiv;

 } 






void pq_tree::skip_di(pq_node& k, pq_node& l)  // skip direction indicators
{ 
  while (k && k->check(DPQ_DIR))
    if (k->link1 == l) 
      { l = k; k = k->link2; }
    else
      { l = k; k = k->link1; }
}

 



void pq_tree::fix_part_child_direction(pq_node x,char what){

 pq_node	k;

 if(x->part_child1->left_most->check(DPQ_full))
 { k = x->part_child1->left_most;
   x->part_child1->left_most = x->part_child1->right_most;
   x->part_child1->right_most = k;
 } 
 // Now the full children of part_child1 are on it's righthand side.

 if(what == 1)
 if(x->part_child2->right_most->check(DPQ_full))
 { k = x->part_child2->right_most;
   x->part_child2->right_most = x->part_child2->left_most;
   x->part_child2->left_most = k;
  }
 // Now the full children of part_child2 are on it's lefthand side.
 

}


/*

 list<pq_node>   node_del_list;
 int   exist_nodes;


void	pq_mem_init(void){

node_del_list.clear();
exist_nodes = 0;


}






void	pq_mem_off(){

while(node_del_list.size())  delete node_del_list.pop(); 

}





void*	pq_new(){

void*	p;

if(p = get_from_list())  return p;
else  return get_from_mem();

}






void*	get_from_list(){

return  node_del_list.pop();

}





void*  get_from_mem(){

#ifndef NDEBUG
exist_nodes++;
#endif

return new pq_node_struct;

}






void	pq_delete(void* p,size_t bytes){

#ifndef NDEBUG
pq_delete_cons();
#endif

node_del_list.push((pq_node) p);

}


*/
