/*******************************************************************************
+
+  LEDA 3.4
+
+  _planar.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

//------------------------------------------------------------------------------
//
// The Booth/Lueker Planarity Test  (based on PQ-Trees)
//
// bool PLANAR(graph& G, bool embed = false);
// bool PLANAR(graph& G, list<edge>& P, bool embed = false);
//
// R. Hesse, E. Kalliwoda, S. Naeher   (1996)
//
//------------------------------------------------------------------------------


#include <LEDA/graph_alg.h>
#include <LEDA/pq_tree.h>


#if !defined(__TEMPLATE_FUNCTIONS__)
LEDA_TYPE_PARAMETER(list<edge>)
#endif


enum { NORMAL = 0, REVERSE = -1 };


static void DFS(graph& G, node v, node_array<list<edge> >& upward,
                                  edge_array<edge>& reversal,
                                  node_array<int>& st_num, 
                                  list<edge>& embed_list)
{ 
  // I mark visited nodes by setting their st-number to -1

  if (st_num[v] < 0) return;
 
  st_num[v] = -1;

  edge e;
  forall_rev(e,upward[v])
  { embed_list.append(reversal[e]);
    DFS(G,target(e),upward,reversal,st_num,embed_list);
   }
}



static void embedding(graph& G, node t, node_array<int>& st_num,
                                        node_array<list<edge> >& upward,
                                        edge_array<edge>& reversal)
{
  // extend upward embedding to an entire embedding using DFS

  list<edge> embed_list;

  node v;
  forall_nodes(v,G)
  { edge e;
    forall(e,upward[v]) embed_list.append(e);
   }

  DFS(G,t,upward,reversal,st_num,embed_list);

  G.sort_edges(embed_list);
}



bool PLANAR(graph& G, bool do_embed)
{
  int n = G.number_of_nodes();
  
  if (n < 4)  return true;
  
  edge_array<edge> reversal(G,nil);
  list<edge> el;

  if ( do_embed )
    { if ( !Is_Bidirected(G,reversal) || !Is_Biconnected(G))
       error_handler(1,"PLANAR: can only embed biconnected bidirected graphs.");
     }
  else
    { Make_Biconnected(G,el); 
      Make_Bidirected(G,el);
     }


  int m = G.number_of_edges();
  
  node_array<int> st_num(G);     // st-numbering
  list<node>      st_list;       // nodes orderd by increaing st-numbers

  ST_NUMBERING(G,st_num,st_list);

  node s = st_list.head();
  node t = st_list.tail();
  

  // interface for pq_tree

  pq_tree T;

  list<int>* AU = new list<int>[n+1];
  
  edge_array<int> e_index(G);  // array of edge indices (--> pq_tree_items)
  edge* EDGE  = new edge[m+1]; // inverse array of e_index
  int   count = 0;
  node  v;
  edge  e;

  forall_edges(e,G)
  { e_index[e] = ++count;
    EDGE[count] = e;
   }
  
  
  // planarity test

  bool is_planar = true;

  forall(v,st_list)
  { 
    list<int> L;

    int stv = st_num[v];

    if ( v != s)
    { forall_adj_edges(e,v)
         if (st_num[target(e)] < stv) L.push(e_index[e]); 
      if ( !T.reduction(L) )  { is_planar = false; break; }
      T.pert_sequence(AU[stv]);
     }
    
    L.clear();

    forall_in_edges(e,v)
        if (st_num[source(e)] > stv) L.push(e_index[e]); 
    if ( !T.update(L) ) { is_planar = false; break; }
   }


  if ( is_planar && do_embed ) 
  {
    // construct adjacency lists of upward embedding

   node_array<list<edge> > upward(G); 

   list<int>* au = &AU[n];

   forall_rev(v,st_list)  
   { 
     if (v == s) break;   // for v = t down to s+1

     int stn = st_num[v];
     int l   = au->pop();
     int d   = 1;
     int i;

     if (l == REVERSE)
     { d = -1;
       l = au->pop();
      }

     for(i = 0; i < l; i++)
     { int h = -d * au->pop();
       if (h > 0) AU[h].push(REVERSE);
      }

     if (d > 0)
        forall(i,*au) upward[v].append(EDGE[i]);
     else
        forall(i,*au) upward[v].push(EDGE[i]);

     au--;
    }

   embedding(G,t,st_num,upward,reversal);
   G.sort_nodes(st_list);
  }
  
  forall(e,el) G.del_edge(e);
  
  delete[] EDGE;
  delete[] AU;
  
  return is_planar;
}


bool PLANAR(graph& G, list<edge>& P, bool embed)
{
  if (PLANAR(G, embed)) return true;

  GRAPH<node,edge> H;
  node_array<node> link(G);
  node v;

  forall_nodes(v, G) link[v] = H.new_node(v);

  int N = G.number_of_nodes();

  edge e;

  forall_edges(e,G)
  if (G.source(e) != G.target(e))
  { H.new_edge(link[G.source(e)], link[G.target(e)], e);
    if (--N < 1 && !PLANAR(H)) break;
   }

  list<edge> L = H.all_edges();

  edge eh;

  forall(eh, L) 
  { e = H[eh];
    node x = G.source(eh);
    node y = G.target(eh);
    H.del_edge(eh);
    if (PLANAR(H)) H.new_edge(x,y,e);
   }

  P.clear();

  forall_edges(eh, H) P.append(H[eh]);

  return false;
}

