/*******************************************************************************
+
+  LEDA 3.4
+
+  graphwin.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <time.h>
#include <math.h>

#include <LEDA/graphwin.h>
#include "local.h"

//------------------- static GraphWin-member --------------------------------

ID_t GraphWin::max_gw_id=0;
queue<ID_t> GraphWin::id_queue;
map<ID_t,GraphWin*> GraphWin::id_vec;
map<window*,GraphWin*> GraphWin::win_vec;

//----------------------------------------------------------------------------

GraphWin::GraphWin(string win_label) : n_matrix((list<edge>*)NULL) { 
  init_once();
  create(cr_win | cr_graph);
  init_graph();
  set_frame_label(win_label);
}
    
//----------------------------------------------------------------------------

GraphWin::GraphWin(graph& G, string win_label) : n_matrix((list<edge>*)NULL) { 
  init_once();
  gr_p=&G;
  create(cr_win);
  init_graph();
  set_frame_label(win_label);
}
    
//----------------------------------------------------------------------------

GraphWin::~GraphWin() {
  if (create_flag & cr_win) delete win_p;
  else win_p->set_redraw();
  if (create_flag & cr_graph) delete gr_p;
}

//----------------------------------------------------------------------------

void GraphWin::update_win_bounds() {

  x_min=win_p->xmin();
  x_max=win_p->xmax();
  y_min=win_p->ymin();
  y_max=win_p->ymax();

}

//----------------------------------------------------------------------------

void GraphWin::display() {
  init_window();
  bool old_flush=set_flush(false);
  if (!nodes_embedded) embed_nodes(node_embed);
  if (!edges_embedded) embed_edges();
  set_flush(old_flush);
  redraw();
}

//----------------------------------------------------------------------------

void GraphWin::open() {
  display();
  edit();
}
    
//----------------------------------------------------------------------------

void GraphWin::close() {
    win_p->close();
    window_is_mapped=false;
}

//----------------------------------------------------------------------------

node GraphWin::new_node(point p) {

  node v,w;

  v=gr_p->new_node();	// later : input_node_inf()
  
  init_node(v);
  
  n_info[v].pos=p;

  forall_nodes(w,*gr_p) { 
    del_incident_edges(v,w);
    del_incident_edges(w,v);
  }
  
  if (get_flush()) draw_node(v);
  
  return v;
}

//----------------------------------------------------------------------------

edge GraphWin::new_edge(node v, node w) {

  if (!is_directed() && v==w) return nil; 

  if (get_flush()) draw_node_with_edges(v);

  edge e=gr_p->new_edge(v,w);	// later : input_edge_inf()
  
  new_incident_edge(e);
  
  init_edge(e);

  if (get_flush()) {
    embed_node_with_edges(v);
    draw_node_with_edges(v);
  }
  else {
    edges_embedded=false;
  }
  
  return e;
}

//----------------------------------------------------------------------------

void GraphWin::del_node(node v) {

  if (get_flush()) draw_node_with_edges(v);

  bool old_flush=set_flush(false);		// don't follow del_edge 
  bool before=edges_embedded;			// edges_embedded not changed!!
  edge e;

  forall_inout_edges(e,v) del_edge(e);		// remove all assigned edges
  id_queue.append(get_id(v));
  
  gr_p->del_node(v);
  edges_embedded=before;
  set_flush(old_flush);
}

//----------------------------------------------------------------------------

void GraphWin::del_edge(edge e) {

  bool before=edges_embedded;
  node v=source(e),w=target(e);
  if (get_flush()) draw_node_with_edges(v);

  del_incident_edge(e);

  gr_p->del_edge(e);
  if (get_flush()) {
      embed_node_with_edges(v);
      draw_node_with_edges(v);
      edges_embedded=before;
  }
  else edges_embedded=false;
}

//----------------------------------------------------------------------------

bool GraphWin::wait(unsigned sec, string msg) {
  win_p->message(msg);
  bool old_done=done;
  done = false;
  time_t t=time(NULL);
  while (!done) {
    win_p->get_mouse();
    if (difftime(time(NULL),t) >= sec) break;
    wait(1);
  }   
  win_p->del_messages();
  done=old_done;
  return done_cause;
}    
    
//----------------------------------------------------------------------------

bool GraphWin::wait(string msg) {
  win_p->message(msg);
  bool old_done=done;
  done = false;
  while (!done) {  
    win_p->read_mouse();
  }
  win_p->del_messages();
  done=old_done;
  return done_cause;
}    
    
//----------------------------------------------------------------------------

int GraphWin::add_menu(string label, int sub_menu) {
  if (!sub_menus) init_call_menu();
  menu *m_sub=new menu;
  sub_menus[sub_menu_counter]=m_sub;
  if (sub_menu) sub_menus[sub_menu]->button(label,*m_sub);
  else win_p->button(label,*m_sub);
  return sub_menu_counter++;  
}

//----------------------------------------------------------------------------

void GraphWin::add_call(base_function* f, string label, int sub_menu) {
  if (!sub_menus) init_call_menu();
  menu_functions[call_entry_counter]=f;
  if (sub_menu) sub_menus[sub_menu]->button(label,mcode(call_entry_counter),gw_call_handler);
  else win_p->button(label,mcode(call_entry_counter),gw_call_handler);
  ++call_entry_counter;
}

//----------------------------------------------------------------------------

void GraphWin::clear_graph() {
  gr_p->clear();
  init_graph();
  if (get_flush()) redraw();
}

//----------------------------------------------------------------------------

