/*******************************************************************************
+
+  LEDA 3.4
+
+  io.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

#include <LEDA/graphwin.h>
#include "ps.h" 

//----------------------------------------------------------------------------

bool GraphWin::save_ps(string fname) {
  PostScript ps;
  ps << *this;
  return ps.save(fname);
}

//----------------------------------------------------------------------------

bool GraphWin::save_graph(string fname) {
  ofstream o(fname);
  if (o.fail()) return false;
  gr_p->write(o);
  if (o.fail()) return false;
  return true;
}

//----------------------------------------------------------------------------

int GraphWin::read_graph(string fname) {

  // need more information about graph::read	(1&3 means fail?)

  int result=gr_p->read(fname);
  if (result == 1) { // no such file ( !! no clear, graph is as it was before !! )
    return 1;    
  }
  if (result == 3) { // graph is not a LEDA-graph
    // future : try loading some other graph-types
    clear_graph();
    return 3;
  }

  // Sit.: graph was read from file

  init_graph();

  if (get_flush()) {
    set_flush(false);
    embed_nodes(node_embed);
    embed_edges();
    set_flush(true);
    redraw();
  }

  return 0;
}

//----------------------------------------------------------------------------

ostream& operator << (ostream& o, const node_info& inf) {
  o        << inf.pos.xcoord() 
    << ' ' << inf.pos.ycoord() 
    << ' ' << int(inf.type) 
    << ' ' << int(inf.style) 
    << ' ' << inf.r1 
    << ' ' << inf.r2
    << ' ' << int(inf.clr) 
    << ' ' << int(inf.marked)
    << ' ' << int(inf.label_clr)
    << ' ' << int(inf.label_pos)
    << ' ' << int(inf.show_label)
    << ' ' << inf.label;
  return o;
}

//----------------------------------------------------------------------------

istream& operator >> (istream& i, node_info& inf) {
  double x,y,r1,r2; 
  int clr,label_clr,type,style,label_pos,marked,show_label;
  i >> x 
    >> y 
    >> type 
    >> style 
    >> r1 
    >> r2
    >> clr 
    >> marked
    >> label_clr 
    >> label_pos 
    >> show_label;
  i.ignore(); // single space
  inf.label.read_line(i);
  inf.pos=point(x,y);
  inf.clr=clr;
  inf.r1=r1;
  inf.r2=r2;
  inf.type=(node_t)type;
  inf.style=(node_st)style;
  inf.label_pos=(position_t)label_pos;
  inf.label_clr=label_clr;
  inf.marked=marked;
  inf.show_label=show_label;
  return i;
}

//----------------------------------------------------------------------------

static void print_polygon(ostream& o, const list<point>& P) {
  point p;
  o << P.size() << ' ';
  forall(p,P) o << p.xcoord() << ' ' << p.ycoord() << ' ';
}

//----------------------------------------------------------------------------

static void read_polygon(istream& i, list<point>& P) {
  P.clear();
  int k,n;
  double x,y;
  i >> n;
  for(k=0;k<n;k++) {
    i >> x >> y;
    P.append(point(x,y));
  } 
}

//----------------------------------------------------------------------------

ostream& operator << (ostream& o, const edge_info& inf) {
  o        << inf.width
    << ' ' << int(inf.clr)
    << ' ' << int(inf.label_clr)
    << ' ' << int(inf.style)
    << ' ' << int(inf.label_pos)
    << ' ' << int(inf.show_label)
    << ' ' << int(inf.auto_edge)
    << ' ';
  print_polygon(o,inf.p);
  o << inf.label; 
  return o;
}

//----------------------------------------------------------------------------  
                                                                              
istream& operator >> (istream& i, edge_info& inf) {   
  int clr,label_clr,type,style,label_pos,show_label,auto_edge;
  i >> inf.width 
    >> clr
    >> label_clr 
    >> style 
    >> label_pos
    >> show_label
    >> auto_edge;
  read_polygon(i,inf.p);
  i.ignore();
  inf.label.read_line(i);
  inf.clr=clr;
  inf.style=(edge_st)style;
  inf.label_pos=(position_t)label_pos;
  inf.show_label=(bool)show_label;
  inf.auto_edge=(bool)auto_edge;
  return i;
}                                                                              
                                                                              
//----------------------------------------------------------------------------

bool GraphWin::save_gw(string fname, string g_extension, string gw_extension) {
  
  if (!save_graph(fname + g_extension)) return false;

  // My saving-procedure

  node v;
  edge e;

  ofstream out(fname + gw_extension);

  if (out.fail()) return false;

  forall_nodes(v,*gr_p) out << n_info[v] << endl;
  forall_edges(e,*gr_p) out << e_info[e] << endl;
  
  return !out.fail();
}  

//----------------------------------------------------------------------------

int GraphWin::read_gw(string fname, string g_extension, string gw_extension) {
 
  bool old_flush=set_flush(false);

  int err=read_graph(fname + g_extension);

  if (err == 1) {		// file couldn't be opened
    set_flush(old_flush);
    return 1;
  }

  if (err == 3) {		// no LEDA-graph
    if (old_flush) redraw();
    set_flush(old_flush);
    return 3;
  }
  
  node v;
  edge e;

  nodes_embedded=edges_embedded=false;

  ifstream in(fname + gw_extension);

  if (in.fail()) {
    if (old_flush) {
      embed_nodes(node_embed);
      embed_edges();
      redraw();
    }
    set_flush(old_flush);
    return 4;
  }

  forall_nodes(v,*gr_p) in >> n_info[v];

  if (in.fail()) {	// nodes may have undefined positions
    if (old_flush) {
      embed_nodes(node_embed);
      embed_edges();
      redraw();
    }
    set_flush(old_flush);
    return 5;
  }

  nodes_embedded=true;

  forall_edges(e,*gr_p) in >> e_info[e];

  if (in.fail()) {	// edges not embedded
    if (old_flush) {
      embed_edges();
      redraw();
    }
    set_flush(old_flush);
    return 6;
  }
  
  edges_embedded=true;

  if (old_flush) redraw();
 
  set_flush(old_flush);

  return 0;
}  

//----------------------------------------------------------------------------

PostScript& operator << (PostScript& ps, GraphWin& gw) {

  const graph &G=gw.get_graph();
  
  double gw_dx=gw.get_x_max() - gw.get_x_min();
  
  double factor=0.7*ps.get_x_width()/gw_dx;
  
  ps.scale(factor,factor);

  ps.set_origin(-gw.get_x_min()+0.2*gw_dx,-gw.get_y_min()+0.2*gw_dx);
  
  ps.set_line_width(1);

  double n_radius=gw.get_node_radius();
  
  node v,w;
  
  point p;
  
  double r1,r2,x,y;

  color c;
 
  node_st v_st;
 
  string s;
 
  ps.set_font("Helvetica",9);
    
  forall_nodes(v,G) {
  
    p=gw.get_position(v);
    x=p.xcoord();
    y=p.ycoord();
    r1=gw.get_radius(v);
    r2=gw.get_radius2(v);
    c=gw.get_color(v);
    v_st=gw.get_style(v);
    
    switch(gw.get_type(v)) {

      case circle_node:
        if (v_st == simple_node) ps.draw_circle(x,y,r1,c);
        else ps.draw_circle_node(x,y,r1,c);
        break;
        
      case ellipse_node:
        if (v_st == simple_node) ps.draw_ellipse(x,y,r1,r2,c);
        else ps.draw_ellipse_node(x,y,r1,r2,c);
        break;
        
      case rectangle_node:
        if (v_st == simple_node) ps.draw_rectangle(x-r1,y-r2,x+r1,y+r2,c);
        else ps.draw_rectangle_node(x-r1,y-r2,x+r1,y+r2,c);
        break;
    }

    if (!gw.get_show_label(v)) continue;

/*
    double dx=20;
    double dy=10;

    position_t label_pos=gw.get_label_pos(v);

    if (label_pos==right || label_pos==right_above || label_pos==right_below)
      x+=r1+factor;
    else {
      if (label_pos==left || label_pos==left_above || label_pos==left_below) 
        x-=r1+dx+factor;
      else x-=dx/2;
    }

    if (label_pos==above || label_pos==left_above || label_pos==right_above)
      y+=r2+dy+factor;
    else {
      if (label_pos==below || label_pos==left_below || label_pos==right_below)
        y-=r2+factor;
      else y+=dy/2;
    }
*/

    ps.draw_ctext(point(x,y),gw.get_label(v));
        
  }
 
  point e_p0,e_p1;

  edge e;

  bool directed=gw.get_directed();

  int width=gw.get_width(G.first_edge());

  ps.set_line_width(width);

  forall_edges(e,G) {
    if (width != gw.get_width(e)) {
      width=gw.get_width(e);
      ps.set_line_width(width);
    }
    ps.draw_polygon_edge(gw.get_polygon(e),gw.get_color(e));	// arrows?
  }    
  
  return ps;
}

// ---------------------------------------------------------------------------

