/*******************************************************************************
+
+  LEDA 3.4
+
+  _gx11.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

#include <LEDA/impl/x_window.h>
#include "gbase.h"

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const int disp_width  = 1000;
const int disp_height =  750;
const int MAX_COLORS  =  256; 

static Display*      display = NULL;
static Window        win;
static XGCValues     gc_val;
static GC            gc;
static int           screen;
static Colormap      color_map;
static unsigned long color_pix[MAX_COLORS];
static int           color_count= 0;

static int  mode[] = { GXcopy, GXand, GXor, GXxor };

static int alloc_color(const char* name)
{ XColor xcolor;
  if (DefaultDepth(display,screen)==1 && strcmp(name,"white")!=0) 
     XParseColor(display, color_map, "black", &xcolor);
  else
     XParseColor(display, color_map, name, &xcolor);
  XAllocColor(display,color_map,&xcolor);
  color_pix[color_count] = xcolor.pixel;
  return color_count++;
}


void vga_setpal(int, int, int, int) {}
void vga_getpal(int, int*, int*, int*) {}
 

int vga_width(void)  { return disp_width; }

int vga_height(void) { return disp_height; }

int vga_depth(void)  { return DefaultDepth(display,screen); }


void vga_init(int mode)
{
  if (mode == 0)
    { XCloseDisplay(display); 
      display = NULL;
     }
  else
  { if (display != NULL) return;
    if ((display = XOpenDisplay(0)) == NULL)	
    { fprintf(stderr, "Can\'t open display: \n");
      abort();
     }
    screen    = DefaultScreen(display);
    color_map = DefaultColormap(display,screen);
    color_count = 0;

    alloc_color("white");          // 0: white
    alloc_color("black");          // 1: black
    alloc_color("red");            // 2: red
    alloc_color("green2");         // 3: green
    alloc_color("blue");           // 4: blue
    alloc_color("yellow");         // 5: yellow
    alloc_color("purple");         // 6: violet
    alloc_color("darkorange");     // 7: orange
    alloc_color("cyan");           // 8: cyan
    alloc_color("sienna");         // 9: brown
    alloc_color("magenta");        //10: pink 
    alloc_color("#0cb3a0");        //11: green2
    alloc_color("cornflowerblue"); //12: blue2
    alloc_color("grey86");         //13: grey1
    alloc_color("grey70");         //14: grey2
    alloc_color("grey35");         //15: grey3

    gc_val.background = color_pix[0];
    gc_val.foreground = color_pix[1];
    gc = XCreateGC(display,RootWindow(display,screen),
                      GCBackground | GCForeground, &(gc_val));
  
    XSetLineAttributes(display,gc,1,LineSolid,CapButt,JoinMiter);

    // open window
  
    XSetWindowAttributes attrib;
    attrib.backing_store = Always;

    win = XCreateWindow(display, RootWindow(display,screen),
                        0, 0, disp_width,disp_height, 2,
                        DefaultDepth(display,screen), InputOutput, 
                        DefaultVisual(display,screen), 
                        CWBackingStore, &attrib);
  
    XSelectInput(display,win, EnterWindowMask | LeaveWindowMask    |
                              KeyPressMask    | PointerMotionMask  | 
                              ButtonPressMask | ButtonReleaseMask  |
                              ExposureMask    | StructureNotifyMask); 
    
    if (DefaultDepth(display,screen) == 1) 
       XSetWindowBackground(display,win,WhitePixel(display, screen)); 
    else
       XSetWindowBackground(display,win,color_pix[0]);
   
    XStoreName(display,win,"LEDA VGA Window");
    XSetIconName(display,win,"LEDA-VGA");
  
    XMapWindow(display,win);
  
    if (! XDoesBackingStore(XScreenOfDisplay(display,screen)))
    { XEvent e;
      do
         XNextEvent(display, &e);
      while (e.type != Expose);
     }
    
    //XEvent event;
    //
    //while (XCheckMaskEvent(display, KeyPressMask    | // PointerMotionMask  | 
    //                              ButtonPressMask | ButtonReleaseMask  |
    //                              ExposureMask    | StructureNotifyMask, 
    //                       &event)); 

   }
}



//------------------------------------------------------------------------------
// drawing parameters
//------------------------------------------------------------------------------


void vga_setcolor(int col)
{ gc_val.foreground = color_pix[col];
  XChangeGC(display,gc,GCForeground,&(gc_val));
 }

void vga_setmode(int m)
{ gc_val.function = mode[m];
  XChangeGC(display,gc,GCFunction,&gc_val);
 }



//------------------------------------------------------------------------------
// drawing functions
//------------------------------------------------------------------------------


void vga_pixel(int x, int y)          { XDrawPoint(display,win,gc,x,y); }
void vga_hline(int x0, int x1, int y) { XDrawLine(display,win,gc,x0,y,x1+1,y); }
void vga_vline(int x, int y0, int y1) { XDrawLine(display,win,gc,x,y0,x,y1+1); }

void vga_box(int x0, int y0, int x1, int y1)
{
  XFillRectangle(display,win,gc,x0,y0,x1-x0+1,y1-y0+1);
 }
 

struct Image {
 int w;
 int h;
 Pixmap P;
};

char* vga_getimage(int x1, int y1, int x2, int y2)
{ vga_setmode(0);
  Image* im = new Image;
  im->w = x2-x1+1;
  im->h = y2-y1+1;
  im->P = XCreatePixmap(display,win,im->w,im->h,DefaultDepth(display,screen));
  XCopyArea(display,win,im->P,gc,x1,y1,im->w,im->h,0,0);
  return (char*)im;
 }

void vga_putimage(int x, int y,  char* image)
{ vga_setmode(0);
  Image* im = (Image*)image;
  XCopyArea(display,im->P,win,gc,0,0,im->w,im->h,x,y);
 }


void vga_delimage(char* image)
{ Image* im = (Image*)image;
  XFreePixmap(display,im->P);
  delete im;
 }

void vga_copyimage(int x1,int y1,int x2,int y2,int x,int y)
{ char* I = vga_getimage(x1,y1,x2,y2);
  vga_putimage(x,y,I);
  vga_delimage(I);
 }


static void vga_bytes(int x, int y, unsigned char* data, int n)
{ 
  XPoint* points = new XPoint[8*n];
  XPoint* stop = points + 8*n;
  XPoint* p = points;

  while (p < stop) (p++)->y = (short)y;
  
  p = points;
  for (int i=0; i<n; i++)
  { unsigned char c = data[i];
    for (unsigned char pat = 0x80; pat; pat >>= 1, x++)
       if (c & pat) (p++)->x = x; 
   }

  XDrawPoints(display,win,gc,points,p-points,CoordModeOrigin);
  delete[] points;
}


void vga_bitmap(int x, int y, unsigned char* pm, int width, int height)
{ while (height--)
  { vga_bytes(x,y++,pm,width);
    pm += width;
   }
 }


/*
void vga_bitmap(int x, int y, unsigned char* data, int width, int height)
{ vga_setmode(0);
  width *= 8;
  Pixmap P=XCreatePixmapFromBitmapData(display,win,(char*)data,width,height,
                                       BlackPixel(display,screen),
                                       WhitePixel(display,screen),
                                       DefaultDepth(display,screen));
  XCopyArea(display,P,win,gc,0,0,width,height,x,y);
  XFreePixmap(display,P);
 }
*/
 

void vga_clear(int) { XClearWindow(display,win); }


//------------------------------------------------------------------------------
// event handling
//------------------------------------------------------------------------------

XEvent event;

int dos_next_event(int *val,int *x,int *y, int block)
{
  int kind = no_event;

  if (block)

      XNextEvent(display, &event);
  else
  { if (XCheckMaskEvent(display,
                      EnterWindowMask | LeaveWindowMask    |
                      KeyPressMask    | PointerMotionMask  |
                      ButtonPressMask | ButtonReleaseMask  |
                      ExposureMask    | StructureNotifyMask, &event) == 0)
    return no_event;
   }


  switch (event.type) {

  case ButtonPress: *val = event.xbutton.button;
                    *x = event.xbutton.x;
                    *y = event.xbutton.y;
                    kind = button_press_event;
                    if (event.xbutton.state & Mod1Mask)    *val = 2; 
                    if (event.xbutton.state & ShiftMask)   *val |= 256; 
                    if (event.xbutton.state & ControlMask) *val |= 512;
                    XUngrabPointer(display,CurrentTime);
                    break;

  case ButtonRelease: *val = event.xbutton.button;
                      *x = event.xbutton.x;
                      *y = event.xbutton.y;
                      if (event.xbutton.state & Mod1Mask)    *val = 2; 
                      if (event.xbutton.state & ShiftMask)   *val |= 256; 
                      if (event.xbutton.state & ControlMask) *val |= 512;
                      kind = button_release_event;
                      break;

  case MotionNotify: *x = event.xmotion.x;
                     *y = event.xmotion.y;
                     kind = motion_event;
                     break;

  case KeyPress: { char c = 0;
                   KeySym keysym;
                   XComposeStatus status;
                   XLookupString((XKeyEvent*)&event,&c,1, &keysym, &status);
                   *val = (c) ? c : keysym;
                   *x = event.xmotion.x;
                   *y = event.xmotion.y;
                   kind = key_press_event;
                   if (*val == 27)
                   { vga_init(0);
                     exit(0);
                    }
                   break;
                  }
   }

  return kind;
}

void dos_back_event() { XPutBackEvent(display,&event); }

void dos_mouse_init(int,int) {}

void dos_mouse_exit() {}

void dos_draw_pointer(int,int) {}
