/*******************************************************************************
+
+  LEDA 3.4
+
+  _rat_polygon.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/
#include <LEDA/rat_polygon.h>
#include <LEDA/graph.h>
#include <math.h>
#include <assert.h>

extern void SWEEP_SEGMENTS(const list<rat_segment>&,GRAPH<rat_point,rat_segment>&);


//------------------------------------------------------------------------------
// rat_polygons
//------------------------------------------------------------------------------


rat_polygon_rep::rat_polygon_rep(const list<rat_segment>& sl) : seg_list(sl)
{ rat_segment s;
  forall(s,sl) pt_list.append(s.start());
 }


ostream& operator<<(ostream& out, const rat_polygon& p) 
{ p.vertices().print(out);
  return out << endl;
 } 

istream& operator>>(istream& in,  rat_polygon& p) 
{ list<rat_point> L; 
  L.read(in,'\n'); 
  p = rat_polygon(L); 
  return in;
}


rational rat_polygon::compute_area(const list<rat_segment>& seg_list) const
{
  if (seg_list.length() < 3) return 0;

  list_item it = seg_list.item(1);
  rat_point p  = seg_list[it].source();

  it = seg_list.succ(it);

  rational  A  = 0;

  while (it)
  { rat_segment s = seg_list[it];
    A += ::area(p,s.source(),s.target());
    it = seg_list.succ(it);
   }

  return -A;
}



static void check_simplicity(const list<rat_segment>& seg_list)
{ GRAPH<rat_point,rat_segment> G;
  SWEEP_SEGMENTS(seg_list,G);
  node v;
  forall_nodes(v,G)
      if (G.degree(v) != 2)
         error_handler(1,"rat_polygon: rational polygon must be simple");
}


rat_polygon::rat_polygon(const list<rat_segment>& sl)
{ PTR = new rat_polygon_rep(sl); }
 

rat_polygon::rat_polygon(const list<rat_point>& pl)
{ 
  list<rat_segment> seglist;

  for(list_item it = pl.first(); it; it = pl.succ(it))
     seglist.append(rat_segment(pl[it],pl[pl.cyclic_succ(it)]));

  if (compute_area(seglist) < 0)
  { // reverse edge list
    seglist.clear();
    for(list_item it = pl.last(); it; it = pl.pred(it))
      seglist.append(rat_segment(pl[it],pl[pl.cyclic_pred(it)]));
   }

  check_simplicity(seglist);

  PTR = new rat_polygon_rep(seglist);
}


rat_polygon rat_polygon::rotate90(const rat_point& cent) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.rotate90(cent));
  return rat_polygon(L);
}


rat_polygon rat_polygon::rotate90() const
{ return rotate90(rat_point()); }


rat_polygon rat_polygon::translate(integer dx, integer dy, integer dw) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.translate(dx,dy,dw));
  return rat_polygon(L);
}


rat_polygon rat_polygon::translate(const rational& dx, const rational& dy) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.translate(dx,dy));
  return rat_polygon(L);
}

rat_polygon rat_polygon::translate(const rat_vector& v) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.translate(v));
  return rat_polygon(L);
}




rat_polygon rat_polygon::reflect(const rat_point& p, const rat_point& q) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.reflect(p,q));
  return rat_polygon(L);
}


rat_polygon rat_polygon::reflect(const rat_point& p) const
{ list<rat_segment> L;
  rat_segment s;
  forall(s,ptr()->seg_list) L.append(s.reflect(p));
  return rat_polygon(L);
}



polygon rat_polygon::topolygon() const
{ list<point> L;
  rat_segment s;
  forall(s,ptr()->seg_list) 
    L.append(point(s.xcoord1D(), s.ycoord1D()) );
  return polygon(L);
}



bool rat_polygon::inside(const rat_point& p) const
// straightforward transformation of polygon::inside
// Too many multiplications this must be done much cheaper !$#@!$#@
{
  list<rat_segment>& Ls = ptr()->seg_list;
 
  int count = 0;
 
  integer pX = p.X();
  integer pW = p.W();
 
  list_item it0 = Ls.first();
  list_item it1 = Ls.first();

  integer x0 = Ls[it0].X2(); 
  integer w0 = Ls[it0].W2(); 
  integer x1 = Ls[it1].X2(); 
  integer w1 = Ls[it1].W2(); 

  list_item it;
 
  forall_items(it,Ls)
  { rat_segment s = Ls[it];
      if (s.X2()*w0 < x0*s.W2())
    { it0 = it;
      x0 = s.X2(); w0 = s.W2();
    }
    if (s.X2()*w1 > x1*s.W2()) 
    { it1 = it1;
      x1 = s.X2(); w1 = s.W2();
    }
   }
 
  if ((pX*w0 <= x0*pW) || (x1*pW <= pX*w1)) return false;
 
  while ( Ls[it0].X2()*pW <= pX*Ls[it0].W2() ) it0 = Ls.cyclic_succ(it0);
 
  it = it0;
  do
  { while ( Ls[it].X2()*pW >= pX*Ls[it].W2() ) it = Ls.cyclic_succ(it);
    if (orientation(Ls[it],p) > 0) count++;
    while ( Ls[it].X2()*pW <= pX*Ls[it].W2() ) it = Ls.cyclic_succ(it);
    if (orientation(Ls[it],p) < 0) count++;
  } while (it != it0);
 
  return count % 2;
 
}


bool rat_polygon::outside(const rat_point& p) const { return !inside(p); }



// intersection or union with rational polygon

static bool test_edge(GRAPH<rat_point,rat_segment>& G, edge i1, int mode)
{
  // if mode = 0   test_edge is used for intersection
  // if mode = 1   test_edge is used for union

  node v  = G.target(i1);
  edge i2 = G.cyclic_in_succ(i1);

  if (i1 == i2) return false;

  edge o1 = G.first_adj_edge(v);
  edge o2 = G.last_adj_edge(v);

  rat_point p1 = G[o1].target();
  rat_point p2 = G[o2].target();

  rat_segment si1 = G[i1];
  rat_segment si2 = G[i2];

  if (mode == 0) // intersection
  { if (orientation(si1,si2.source()) > 0)
      return orientation(si1,p1) > 0 && orientation(si2,p1) < 0 && 
             orientation(si1,p2) > 0 && orientation(si2,p2) < 0;
   else
      return (orientation(si1,p1) > 0 || orientation(si2,p1) < 0) && 
             (orientation(si1,p2) > 0 || orientation(si2,p2) < 0);
  }
  else // union
  { if (orientation(si1,si2.source()) < 0)
      return orientation(si1,p1) < 0 && orientation(si2,p1) > 0 && 
             orientation(si1,p2) < 0 && orientation(si2,p2) > 0;
   else
      return (orientation(si1,p1) < 0 || orientation(si2,p1) > 0) && 
             (orientation(si1,p2) < 0 || orientation(si2,p2) > 0);
   }
}



static edge next_edge(GRAPH<rat_point,rat_segment>& G, edge i1, int dir)
{ 
  // if dir = +1 turn left
  // if dir = -1 turn right

  node v = G.target(i1);

  edge o1 = G.first_adj_edge(v);
  edge o2 = G.last_adj_edge(v);

  if (o1 == o2) return o1;

  rat_segment si1 = G[i1];
  rat_segment so1 = G[o1];
  rat_segment so2 = G[o2];


  int orient1 = orientation(si1,so1.target());
  int orient2 = orientation(si1,so2.target());

  if (orient1 == orient2)
     return (orientation(so1,so2.target()) == dir) ? o2 : o1;
  else
     return (orient1 - orient2 == dir) ? o1 : o2;

}



list_rat_polygon_ rat_polygon::intersection(const rat_polygon& P) const
{
  list<rat_polygon> result;
  list<rat_segment> seg_list;
  GRAPH<rat_point,rat_segment> G;

  rat_segment s;
  forall(s,ptr()->seg_list) seg_list.append(s);
  forall(s,P.ptr()->seg_list) seg_list.append(s);

  SWEEP_SEGMENTS(seg_list,G);

  bool borders_intersect = false;

  node v;
  forall_nodes(v,G) 
  { int deg = G.degree(v);
    // assert(deg == 2 || deg == 4);
    if (deg == 4) borders_intersect = true;
   }
    

  if ( ! borders_intersect )
  { // no intersections between edges of (*this) and P

    // check for equality

    if (G.number_of_nodes() == size()) 
    { result.append(*this);
      return result;
     }


    // check for inclusion

    rat_segment s1 = ptr()->seg_list.head();
    rat_segment s2 = P.ptr()->seg_list.head();

    if ( P.inside(s1.start()) ) result.append(*this);
    if ( inside(s2.start()) ) result.append(P);                   

    return result;

   }

  edge_array<bool> marked(G,false);

  edge e;
  forall_edges(e,G) 
  { if ( ! marked[e] && test_edge(G,e,0) )
    { list<rat_segment> pol;
      edge start = e;
      do { node v = source(e);
           node w = target(e);
           pol.append(rat_segment(G[v],G[w]));
           marked[e] = true;
           e = next_edge(G,e,+1);
          } while (e != start);
      result.append(rat_polygon(pol));
     }
   }

 return result;

}


list_rat_polygon_ rat_polygon::unite(const rat_polygon& P) const
{
  list<rat_polygon> result;
  list<rat_segment> seg_list;
  GRAPH<rat_point,rat_segment> G;

  rat_segment s;
  forall(s,ptr()->seg_list) seg_list.append(s);
  forall(s,P.ptr()->seg_list) seg_list.append(s);

  SWEEP_SEGMENTS(seg_list,G);

  if (G.number_of_nodes() == size() + P.size())
  { // no intersections between edges of (*this) and P

    // check for equality

    if (G.number_of_nodes() == size()) 
    { result.append(*this);
      return result;
     }


    // check for inclusion

    rat_segment s1 = ptr()->seg_list.head();
    rat_segment s2 = P.ptr()->seg_list.head();

    if ( ! P.inside(s1.start())) result.append(*this);
    if ( ! inside(s2.start())) result.append(P);                   

    return result;

   }

  edge_array<bool> marked(G,false);

  edge e;
  forall_edges(e,G) 
    if ( ! marked[e] && test_edge(G,e,1) )
    { list<rat_segment> pol;
      edge start = e;
      do { node v = source(e);
           node w = target(e);
           pol.append(rat_segment(G[v],G[w]));
           marked[e] = true;
           e = next_edge(G,e,-1);
          } while (e != start);
      result.append(rat_polygon(pol));
     }

 // now find outer polygon and move it to front of result

  list_item min_it = nil;
  double    min_x = MAXDOUBLE;

  list_item it;
  forall_items(it,result)
  { rat_segment s;
    forall(s,result[it].ptr()->seg_list) 
    { if (s.xcoord1D() < min_x)
      { min_x = s.xcoord1D(); 
        min_it = it;
       }
    }
   }

  assert(min_it != nil);

  result.move_to_front(min_it);


 return result;

}

