/*******************************************************************************
+
+  LEDA 3.4
+
+  delaunay.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

//------------------------------------------------------------------------------
// DELAUNAY TRIANGULATIONS & DIAGRAMS
//
// - DELAUNAY_FLIPPING
// - DELAUNAY_TRIANG
// - DELAUNAY_DIAGRAM
//------------------------------------------------------------------------------


// We use bidirected graphs of type GRAPH<POINT,edge> for representing planar 
// subdvivisons (G[v] = position of v and G[e] = reverse edge of e)


int DELAUNAY_FLIPPING(GRAPH<POINT,edge>& G, list<edge>& L)
{
  // Takes as input a planar triangulation $G$.
  // Tranforms it into a Delaunay triangulation by flipping edges.
  // Diagonals of possibly co-circular convex quadriliterals are collect in L
  // We maintain a stack $S$ of edges containing diagonals which might
  // have to be flipped (initially all edges). 

  if (G.number_of_nodes() <= 3) return 0;

  list<edge> S = G.all_edges();

  L.clear();

  int flip_count = 0;

  while ( !S.empty() )
  { edge e = S.pop();
    edge r = G.reversal(e);

    // e1,e2,e3,e4: edges of quadriliteral with diagonal e
    edge e1 = G.next_face_edge(r);
    edge e3 = G.next_face_edge(e);

    // flip test
    POINT a = G[source(e1)];
    POINT b = G[target(e1)];
    POINT c = G[source(e3)];
    POINT d = G[target(e3)];

    if (left_turn(b,d,a) && right_turn(b,d,c))
    {
      int soc = side_of_circle(a,b,c,d);

      if (soc == 0) 
      { L.append(e);
        L.append(r);
       }

      if (soc > 0)
      { flip_count++;

        edge e2 = G.next_face_edge(e1);
        edge e4 = G.next_face_edge(e3);
  
        S.push(e1); 
        S.push(e2); 
        S.push(e3); 
        S.push(e4); 
  
        // flip diagonal
        G.move_edge(e,e2,source(e4));
        G.move_edge(r,e4,source(e2));
      }
    }
  }

 return flip_count;
}

int DELAUNAY_FLIPPING(GRAPH<POINT,edge>& G)
{ list<edge> dummy;
  return DELAUNAY_FLIPPING(G,dummy);
}


void DELAUNAY_TRIANG(const list<POINT>& L, GRAPH<POINT,edge>& G)
{ TRIANGULATE_POINTS(L,G);
  DELAUNAY_FLIPPING(G);
}



void DELAUNAY_DIAGRAM(const list<POINT>& L, GRAPH<POINT,edge>& G)
{ 
  list<edge> d_list;

  TRIANGULATE_POINTS(L,G);
  DELAUNAY_FLIPPING(G,d_list);
  
  if ( ! d_list.empty() )
  { 
    edge_array<bool> seen(G,false);

    list_item it = d_list.first();

    while (it)
    { list_item  next = d_list.succ(it);

      edge e  = d_list[it];
  
      if (seen[e])
      { d_list.del_item(it);
        it = next;
        continue;
       }

      seen[e] = true;
     
      edge r  = G.reversal(e);
      edge e1 = G.next_face_edge(r);
      edge e3 = G.next_face_edge(e);
  
      POINT a = G[source(e1)];
      POINT b = G[target(e1)];
      POINT c = G[source(e3)];
      POINT d = G[target(e3)];
  
  
      if (! (left_turn(b,d,a) && right_turn(b,d,c) && cocircular(a,b,c,d)))
          d_list.del_item(it);
  
      it = next;
  
    }
  
    edge e;
    forall(e,d_list) G.del_edge(e);
  }
    
}


