/*******************************************************************************
+
+  LEDA 3.4
+
+  _base_panel.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/


// defines the BASE_WINDOW operations declared in <leda/BASE_WINDOW.h>
// using the basic graphics routines from <leda/impl/x_basic.h>

#include <LEDA/base_window.h>
#include <LEDA/impl/x_basic.h>
#include <LEDA/bool.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>


//------------------------------------------------------------------------------
// some auxiliary functions for string manipulation
//------------------------------------------------------------------------------

static char* duplicate_string(const char* p)
{ char* q = new char[strlen(p)+1];
  if (q==0) 
  { fprintf(stderr,"duplicate_string: out of memory");
    abort();
   }
  strcpy(q,p);
  return q;
}

static char* make_string(int x)
{ char str[256];
  sprintf(str,"%d",x);
  return duplicate_string(str);
 }

static char* make_string(double x)
{ char str[256];
  sprintf(str,"%f",x);
  return duplicate_string(str);
 }


//------------------------------------------------------------------------------
// construction and destruction of panel items
//------------------------------------------------------------------------------


Panel_Item::Panel_Item(int k, const char* s, void* addr, int h)
{ kind = k;
  label_str = duplicate_string(s);
  ref = addr;
  height  = h;
  data_str = 0;
  xcoord = 0;
  ycoord = 0;
  width = 0;
  dat1 = 0;
  dat2 = 0;
  offset = 0;
  choices = 0;
  action = 0;
  menu_win = 0;
}


Panel_Item::~Panel_Item()
{ 
  delete[] label_str;

  if (kind != String_Item && kind != String_Menu_Item && data_str) 
     delete[] data_str;

  if (kind == String_Menu_Item || kind == Choice_Item || kind == Bool_Item )
  { for(int i = 0; i < dat1; i++) delete[] choices[i];
    delete[] choices;
    if (menu_win) delete menu_win;
   }

  if (kind == Bitmap_Choice_Item) {
    for(int i = 0; i < dat1; i++)  
       x_delete_bitmap(choices[i]); 
   }

  if (kind == Button_Item && choices != 0) { 
    x_delete_bitmap(choices[0]);
    delete[] choices;
   }
}



//------------------------------------------------------------------------------
// panel member functions
//------------------------------------------------------------------------------

void BASE_WINDOW::panel_init()
{ 
  item_count = 0;
  but_count = 0;

  item_bg_color  = white;

  if ( mono() )
   { panel_bg_color = white;   
     shadow_color = black;   
     item_bg_color = white;   
    }
  else
   { panel_bg_color = grey1;
     //press_color    = blue;
     press_color    = grey2;
     shadow_color   = grey3;
    }

  panel_width = 0;
  panel_height = 0;


  tw = x_text_width(0,"H");
  th = x_text_height(0,"H");

  ytskip   = th + 2;
  yskip    = th + 11;
  yoff     = th/2;
  xoff     = tw + 1;

  slider_h = th - 2;
  string_h = th + 2;
  button_h = th + 4;
  choice_h = th + 4;

  label_w   = 50;          // minimal label length
  number_w  = 5*tw;        // slider num field width
  slider_w  = 16*choice_h; // minimal slider length
  string_w  = 16*choice_h; // minimal string/int/float item length
  choice_w  = 35;          // minimal choice field width

  //button_w  = 40;
  button_w  = 20;          // minimal button width

  buts_per_line = 0;  
  center_buttons = 1;

  act_str_item = 0;
  last_sel_button = 0;
  panel_menu_mode = 0;
 }


void BASE_WINDOW::item_error()
{ fprintf(stderr,"sorry, too many items."); 
  exit(1);
 }


//------------------------------------------------------------------------------
// adding items
//------------------------------------------------------------------------------


void BASE_WINDOW::bitmap_choice_item(const char* label, int *address, int argc,
                                     int width, int height, char **argv)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  int byskip=choice_h-th+(height>th?height:th);
  panel_item p = new Panel_Item(Bitmap_Choice_Item,label,address,byskip);  

  p->dat1=argc;		// number of bitmaps
  p->dat2=width+5;	// width in dat2
  p->offset=height+4;	// save height in offset
  p->choices=new char*[argc]; // space for bitmaps
  for(int i=0; i < argc; i++) 
     p->choices[i] = x_create_bitmap(draw_win,width,height,argv[i]);

  int w;
  if ( (w=x_text_width(draw_win,label)+number_w) > label_w) label_w=w;
  Item[item_count++]=p;
}


void BASE_WINDOW::text_item(const char* s)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Text_Item,s,0,ytskip);
  Item[item_count++] = p;
}


void BASE_WINDOW::string_item(const char* s, void* x)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(String_Item,s,x,yskip);
  p->data_str = *access_str(x);
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }



static BASE_WINDOW* str_menu;
static panel_item str_menu_it;

void str_menu_selection(int sel)
{ if (str_menu_it && sel >= 0)
  { delete[] *str_menu->access_str(str_menu_it->ref);
    *str_menu->access_str(str_menu_it->ref) = 
              duplicate_string(str_menu_it->choices[sel]);
    str_menu_it->data_str = *str_menu->access_str(str_menu_it->ref);
    str_menu->clipping(1);

    int draw_w = str_menu->draw_win;

    int          save_lw = x_set_line_width(draw_w,1);
    line_style   save_ls = x_set_line_style(draw_w,solid);
    text_mode    save_tm = x_set_text_mode(draw_w,transparent);
    drawing_mode save_mo = x_set_mode(draw_w,src_mode);
    
    str_menu->draw_string_item(str_menu_it);

    int xr = str_menu_it->xcoord + str_menu->label_w - 10;
    int yr = str_menu_it->ycoord + str_menu->yskip/2;

    str_menu->draw_down_menu_button(xr,yr,0);

    x_set_line_width(draw_w,save_lw);
    x_set_line_style(draw_w,save_ls);
    x_set_text_mode(draw_w,save_tm);
    x_set_mode(draw_w,save_mo);
   }
}


void BASE_WINDOW::string_menu_item(const char* s, void* x, const char* ,
                                  int argc,const char** argv)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(String_Menu_Item,s,x,yskip);
  p->data_str = *access_str(x);
  p->dat1 = argc;
  p->choices = new char*[argc];

  int max_choice_len = 0;
  int i;
  for(i = 0; i < argc; i++) 
  { p->choices[i] = duplicate_string(argv[i]);
    int l = strlen(argv[i]);
    if (l > max_choice_len) max_choice_len = l;
   }

  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;

  // build menu panel

  int bpl = string_w/(max_choice_len*tw);
  if (bpl >  2) bpl = 2;
  if (bpl == 0 || argc < 6) bpl = 1;

  p->menu_win = new BASE_WINDOW(-1,-1);
  p->menu_win->buttons_per_line(bpl);
  p->menu_win->center_buttons = 0;
  for(i=0; i<argc; i++) 
       p->menu_win->button(p->choices[i],i,str_menu_selection);

  Item[item_count++] = p;
 }


void BASE_WINDOW::int_item(const char* s, int* x)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Int_Item,s,x,yskip);
  p->data_str = make_string(*x);
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }

void BASE_WINDOW::float_item(const char* s, double* x)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Float_Item,s,x,yskip);
  p->data_str = make_string(*x);
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }


void  BASE_WINDOW::slider_item(const char* s, int* x, int min, int max, 
                                                      panel_action_func F)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Slider_Item,s,x,yskip);
  p->data_str = make_string(*x);
  p->dat1 = min;
  p->dat2 = max;
  p->action = F;
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }

void BASE_WINDOW::choice_item(const char* s, int* address, int argc,
                             const char** argv, int step, int off)
{ 
  if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Choice_Item,s,address,yskip);
  p->data_str = 0;
  p->dat1 = argc;
  p->dat2 = step;
  p->offset = off;
  p->choices = new char*[argc];
  for(int i=0; i<argc; i++) 
  { int w;
    p->choices[i] = duplicate_string(argv[i]);
    if ((w = x_text_width(draw_win,argv[i])+10) > choice_w) choice_w = w;
    }
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }


void BASE_WINDOW::bool_item(const char* s, char* address)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Bool_Item,s,address,yskip);
  p->dat1 = 2;
  p->dat2 = 1;
  p->offset = 0;
  p->choices = new char*[2];
  p->choices[0] = duplicate_string("off");
  p->choices[1] = duplicate_string("on");
  int w;
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  if ((w = x_text_width(draw_win,"off")+10) > choice_w) choice_w = w;
  Item[item_count++] = p;
 }


void BASE_WINDOW::color_item(const char* s, int* x)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  int w;
  panel_item p = new Panel_Item(Color_Item,s,x,yskip);
  if ( mono() )
   { // choice item [white][black]
     p->kind = Choice_Item;
     p->dat1 = 2;
     p->dat2 = 1;
     p->offset = 0;
     p->choices = new char*[2];
     p->choices[0] = duplicate_string("white");
     p->choices[1] = duplicate_string("black");
     if ((w = x_text_width(draw_win,"white")+10) > choice_w) choice_w = w;
    }
  if ((w = x_text_width(draw_win,s)+number_w) > label_w) label_w = w;
  Item[item_count++] = p;
 }



int BASE_WINDOW::button(const char* s, int val, panel_action_func F)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Button_Item,s,0,button_h);
  p->dat1 = (val == -1) ? but_count : val;
  p->action = F;
  Item[item_count++] = p;
  int w = x_text_width(draw_win,s) + 20;
  if (w  > button_w) button_w = w;
  but_count++;
  return p->dat1;
 }


int BASE_WINDOW::button(const char* s, int val, BASE_WINDOW* wp)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  panel_item p = new Panel_Item(Button_Item,s,0,button_h);
  p->dat1 = (val == -1) ? but_count : val;
  p->ref = wp;
  Item[item_count++] = p;
  int w = x_text_width(draw_win,s) + 30;
  if (w  > button_w) button_w = w;
  but_count++;
  return p->dat1;
 }


int BASE_WINDOW::button(int w, int h, char* bits, const char* s, int val, 
                                                        panel_action_func F)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  if (h+2 > button_h) button_h = h+2;
  if (w+3 > button_w) button_w = w+3;
  panel_item p = new Panel_Item(Button_Item,s,0,button_h);
  p->choices = new char*[1];
  p->choices[0] = x_create_bitmap(draw_win,w,h,bits);
  p->dat1 = (val == -1) ? but_count : val;
  p->action = F;
  Item[item_count++] = p;
  but_count++;
  return p->dat1;
 }


int BASE_WINDOW::button(int w, int h, char* bits, const char* s, int val, 
                                                               BASE_WINDOW* wp)
{ if (item_count >= MAX_ITEM_NUM) item_error();
  if (h+2  > button_h) button_h = h+2;
  if (w+3 > button_w) button_w = w+3;
  //if (w+20 > button_w) button_w = w+20;
  panel_item p = new Panel_Item(Button_Item,s,0,button_h);
  p->choices = new char*[1];
  p->choices[0] = x_create_bitmap(draw_win,w,h,bits);
  p->dat1 = (val == -1) ? but_count : val;
  p->ref = wp;
  Item[item_count++] = p;
  but_count++;
  return p->dat1;
 }


//------------------------------------------------------------------------
// drawing buttons & items
//------------------------------------------------------------------------

void BASE_WINDOW::draw_label(int x, int y, const char* str)
{ x_text(draw_win,x,y,str);
/*
  if (!mono()) // boldface
  { x_text(draw_win,x+1,y,  str);
    x_text(draw_win,x,  y+1,str);
    x_text(draw_win,x+1,y+1,str);
   }
*/
}

void BASE_WINDOW::draw_box_with_shadow(int x1,int y1,int x2,int y2,int c,int w)
{ // below and right of box(x1,y1,x2,y2)
  x_set_color(draw_win,shadow_color);
  x_box(draw_win,x1+3,y1+3,x2+w,y2+w);
  x_set_color(draw_win,c);
  x_box(draw_win,x1,y1,x2,y2);
  x_set_color(draw_win,black);
  x_rect(draw_win,x1,y1,x2,y2);
 }


void BASE_WINDOW::draw_d3_box(int x1,int y1,int x2,int y2, int pressed)
{
  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,x1,y1,x2,y2);

  int c1 = (pressed) ? black : white;
  int c2 = (pressed) ? shadow_color : panel_bg_color;
  int c3 = (pressed) ? panel_bg_color : shadow_color;
  int c4 = (pressed) ? white : black;
  int c5 = (pressed) ? press_color  : panel_bg_color;

  x_set_color(draw_win,c5);
  x_box(draw_win,x1,y1,x2,y2);

  x_set_color(draw_win,c1);
  x_line(draw_win,x1,y1,x2,y1);
  x_line(draw_win,x1,y1,x1,y2);

  x_set_color(draw_win,c2);
  x_line(draw_win,x1+1,y1+1,x2-1,y1+1);
  x_line(draw_win,x1+1,y1+1,x1+1,y2-1);
  
  x_set_color(draw_win,c3);
  x_line(draw_win,x1+1,y2,x2,y2);
  x_line(draw_win,x2,y1+1,x2,y2);

  x_set_color(draw_win,c4);
  x_line(draw_win,x1,y2+1,x2+1,y2+1);
  x_line(draw_win,x2+1,y1,x2+1,y2+1);
}



void BASE_WINDOW::draw_text_item(panel_item it)
{ x_text(draw_win,it->xcoord,it->ycoord,it->label_str); }


void BASE_WINDOW::draw_string_item(panel_item it, const char* s)
{ 
  if (it->kind == String_Item || it->kind == String_Menu_Item)  
      it->data_str = *access_str(it->ref);

  int active = it == act_str_item;

  int  yt = it->ycoord + (yskip-th)/2;
  draw_label(it->xcoord,yt,it->label_str);

  int x1 = it->xcoord + label_w;
  int x2 = x1 + string_w;
  int y1 = yt - (string_h-th)/2 - 1;
  int y2 = yt + string_h + (string_h-th)/2;
  
  if (s == 0)  s = it->data_str;

  int max_c = string_w/tw - 1;
  int d = strlen(s) - max_c;

  if (active && d > 0) s += d;

  if (panel_bg_color == white)
    { x_set_color(draw_win,panel_bg_color);
      x_box(draw_win,x1,y1,x2,y2);
      x_set_color(draw_win,black);
      x_rect(draw_win,x1,y1,x2,y2);
     }
  else
    draw_d3_box(x1,y1,x2,y2,active);

  // write text into box
  x_set_color(draw_win,black);
  x_text(draw_win,x1+4,yt,s,max_c);


  if (active) x_text(draw_win,x1+4+strlen(s)*tw,yt,"|");
}


void BASE_WINDOW::activate_string_item(panel_item it)
{ if (act_str_item == it) return;
  panel_item it_old = act_str_item;
  act_str_item = it;
  draw_string_item(it_old);
  draw_string_item(it);
 }
   

void BASE_WINDOW::panel_text_edit(panel_item it)
{ 
  char  str[128];
  strcpy(str,it->data_str);

  if (it->kind != String_Item && it->kind != String_Menu_Item) 
     delete[] it->data_str;

  int j = strlen(str);

  for(;;)
  { int  k,val;
    char c = 13;

    while(1)
    { int w;
      int xc,yc;
      unsigned long t;
      k = x_get_next_event(&w,&val,&xc,&yc,&t);
      if (w != draw_win) continue;
      if (k == key_press_event || k == button_press_event) break;
     }

     if (k == key_press_event) c = val;
     if (c == 13) 
     { x_put_back_event();
       break;
      }

     if (isprint(c))
     { str[j++]=c;
       str[j] = '\0';
      }

     if(c==8 && j>0) str[--j] = '\0';

     draw_string_item(it,str);
   }

  it->data_str = duplicate_string(str);
}


void BASE_WINDOW::draw_choice_bool_item(panel_item it,int x,int y, int c, int n)
{
  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,x,y,x+n*choice_w,y+choice_h);

  int j;

  if (panel_bg_color != white)
     for(j=0; j<n; j++)
     { draw_d3_box(x,y,x+choice_w-1,y+choice_h,j==c);
       x_set_color(draw_win,black);
       x_ctext(draw_win,x+choice_w/2,y+choice_h/2,it->choices[j]);
       x += choice_w;
      }
  else
    for(j=0; j<n; j++)
    { x_set_color(draw_win,item_bg_color);
      x_box(draw_win,x+1,y+1,x+choice_w-1,y+choice_h-1);
      x_set_color(draw_win,black);
      x_ctext(draw_win,x+choice_w/2,y+choice_h/2,it->choices[j]);
      if (panel_bg_color == item_bg_color) 
         x_rect(draw_win,x,y,x+choice_w,y+choice_h);

      if (j == c) 
      { x_rect(draw_win,x,y,x+choice_w,y+choice_h);
        x_rect(draw_win,x+1,y+1,x+choice_w-1,y+choice_h-1);
       }
      x += choice_w;
     }

}



void BASE_WINDOW::draw_choice_item(panel_item it)
{ 
  int yt = it->ycoord + (yskip-th)/2;

  draw_label(it->xcoord,yt,it->label_str);

  int y = yt - (choice_h-th)/2 - 1;
  int c = (*(int*)it->ref - it->offset)/it->dat2;

  draw_choice_bool_item(it,it->xcoord+label_w,y,c,it->dat1);
}



void BASE_WINDOW::draw_bool_item(panel_item it)
{ 
  int yt = it->ycoord + (yskip-th)/2 + 1;
  draw_label(it->xcoord,yt,it->label_str);

  int y = yt - (choice_h-th)/2;
  int c = (*(bool*)it->ref - it->offset)/it->dat2;

  draw_choice_bool_item(it,it->xcoord+label_w,y,c,it->dat1);
}


void BASE_WINDOW::change_bitmap_choice_item(panel_item it, int j)
{
  int yt  = it->ycoord + (it->height-th)/2 + 1; 
  int y   = yt - (it->offset-th)/2 - 1;
  int old = *(int*)it->ref;

  if (j == old) return;

  *(int*)it->ref = j;

  int xoff1 = it->xcoord + label_w;
  int x     = xoff1 + j*it->dat2;

  draw_d3_box(x,y+1,x+it->dat2-2,y+it->offset,1);
  x_set_color(draw_win,black);
  x_insert_bitmap(draw_win,x+3,y+it->offset-1,it->choices[j]);

  j = old;
  x = xoff1 + j*it->dat2;
  draw_d3_box(x,y+1,x+it->dat2-2,y+it->offset,0);
  x_set_color(draw_win,black);
  x_insert_bitmap(draw_win,x+2,y+it->offset-2,it->choices[j]);
}


void BASE_WINDOW::draw_bitmap_choice_item(panel_item it) 
{
  int yt = it->ycoord + (it->height-th)/2 + 1;
  int x  = it->xcoord + label_w;
  int y  = yt - (it->offset-th)/2 - 1;
  int c  = *(int*)it->ref;

  draw_label(it->xcoord,yt,it->label_str);

  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,x,y,x+it->dat1*it->dat2,y+it->offset);

  for(int j=0;j<it->dat1;j++) 
  { draw_d3_box(x,y+1,x+it->dat2-2,y+it->offset,j==c);
    x_set_color(draw_win,black);
    x_insert_bitmap(draw_win,x+2,y+it->offset-2,it->choices[j]);
    x += it->dat2;
   }
}


void BASE_WINDOW::change_color_item(panel_item it, int j)
{ 
  int yt = it->ycoord + (yskip-th)/2 + 1;
  int y  = yt - (choice_h-th)/2 - 1;

  int old = *(int*)it->ref;

  if (j == old) return; 

  *(int*)it->ref = j;

  char col_str[8];

  int xoff1 = it->xcoord + label_w;

  int x = xoff1 + j*choice_h;
  sprintf(col_str,"%d",j);
  press_color = j;
  draw_d3_box(x,y,x+choice_h-1,y+choice_h-1,1);
  x_set_color(draw_win,text_color(j));
  x_ctext(draw_win,x+choice_h/2,y+choice_h/2,col_str);

  j = old;
  x = xoff1 + j*choice_h;
  sprintf(col_str,"%d",j);
  press_color = grey2;
  draw_d3_box(x,y,x+choice_h-1,y+choice_h-1,0);
  x_set_color(draw_win,(j==panel_bg_color) ? black : j);
  x_ctext(draw_win,x+choice_h/2,y+choice_h/2,col_str);

  x_set_color(draw_win,black);
}


void BASE_WINDOW::draw_color_item(panel_item it)
{ 
  int yt = it->ycoord + (yskip-th)/2 + 1;
  int x  = it->xcoord + label_w;
  int y  = yt - (choice_h-th)/2 - 1;

  draw_label(it->xcoord,yt,it->label_str);

  char c = *(int*)it->ref;

  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,x,y,x+16*choice_h,y+choice_h);

  for(int j=0; j < 16; j++)
  { char col_str[8];
    sprintf(col_str,"%d",j);
    press_color = j;
    draw_d3_box(x,y,x+choice_h-1,y+choice_h-1,j==c);
    if (j != c)
       x_set_color(draw_win,(j==panel_bg_color) ? black : j);
    else
       x_set_color(draw_win,text_color(j));
    x_ctext(draw_win,x+choice_h/2,y+choice_h/2,col_str);
    x += choice_h;
   }
  press_color = grey2;
  x_set_color(draw_win,black);

  //end_group(draw_win);
}



void BASE_WINDOW::draw_slider_item(panel_item it, float x)
{ 
  int x0 = it->xcoord + label_w;
  int y0 = it->ycoord + (yskip-slider_h)/2;

  int x1 = x0 + slider_w;
  int y1 = y0 + slider_h;

  int yt = it->ycoord + (yskip-th)/2;

  int first_time = 0;

  float mi = it->dat1;
  float ma = it->dat2;

  if (x == 0) // first time
  { first_time = 1;
    float d = (ma > mi) ? float(slider_w)/(ma-mi) : 1;
    x = x0  + d * (*(int*)it->ref - it->dat1);
    draw_label(it->xcoord,  yt,  it->label_str);
   }

  if (x < x0) x = x0;
  if (x > x1) x = x1;

  int val  = int(mi + (ma-mi)*(x-x0)/slider_w + 0.5);
  int val0 = *(int*)it->ref;

  *(int*)it->ref = val;

  // text
  char text[16];
  sprintf(text,"%4d",val);
  //int t_w = x_text_width(draw_win,text) + 3;
  //int t_h = x_text_height(draw_win,text);
  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,x0-number_w,y0,x1,y1);
  x_set_color(draw_win,black);
  draw_label(x0-number_w,yt,text);

  // slider
  if (panel_bg_color == white)
    { x_set_color(draw_win,black);
      x_box(draw_win,x0,y0,int(x+0.5),y1);
      x_set_color(draw_win,white);
      x_box(draw_win,int(x+0.5),y0,x1,y1);
      x_set_color(draw_win,black);
      x_rect(draw_win,x0,y0,x1,y1);
     }
  else
    { draw_d3_box(int(x+0.5),y0,x1,y1,1);
      draw_d3_box(x0,y0,int(x+0.5)+1,y1);
     }
  x_set_color(draw_win,black);

  if (first_time == 0)
  { clipping(2);
    if (it->action && val != val0) it->action(val);
    clipping(1);
   }

}


void BASE_WINDOW::draw_button(panel_item it, int pressed)
{ 
   char* s = it->label_str;

   int x1 = it->xcoord;
   int y1 = it->ycoord;
   int x2 = x1 + it->width;
   int y2 = y1 + it->height+1;
   int yt = y1 + it->height/2;
   int xt = x1 + it->width/2;

   if (panel_bg_color == white)
     { draw_box_with_shadow(x1,y1,x2,y2,white,3);
       if (pressed)
       { x_box(draw_win,x1+1,y1+1,x2-1,y2-1);
         x_set_color(draw_win,white);
        }
      }
   else
     { draw_d3_box(x1,y1,x2,y2,pressed);
       if (it->choices == 0)
          x_set_color(draw_win,black);
       else   
          x_set_color(draw_win,blue);
      }


     if (it->choices == 0)
        if (center_buttons)
           if (pressed)
              x_ctext(draw_win,xt+1,yt+1,s);
           else
              x_ctext(draw_win,xt,yt,s);
        else
           x_text(draw_win,x1+10,yt-th/2,s);
     else
        if (pressed)
           x_insert_bitmap(draw_win,x1+3,y2-1,it->choices[0]);
        else
           x_insert_bitmap(draw_win,x1+2,y2-2,it->choices[0]);


   x_set_color(draw_win,black);

   if (it->ref && it->choices == 0) // text button
     if (buts_per_line == 1 && item_count == but_count) // menu
        draw_right_menu_button(x2-8,(y1+y2)/2,-1);
     else
        draw_down_menu_button(x1+8,(y1+y2)/2,-1);

   x_flush_display();
}


void BASE_WINDOW::draw_down_menu_button(int x, int y, int pressed)
{ 
  if (pressed > -1) 
    draw_d3_box(x-9,y-10,x+9,y+10,pressed);

  int x1 = x-4;
  int y1 = y-3;
  int x2 = x+4;
  int y2 = y-3;
  int x3 = x;
  int y3 = y+4;

  if ( !mono() )
    x_set_color(draw_win,grey2);
  else
    x_set_color(draw_win,white);

  for(int i=1; i<7; i++)
     x_line(draw_win,x1+i,y1+i,x2-i,y2+i);

  x_set_color(draw_win,black);
  x_line(draw_win,x1,y1,x2,y2);
  x_line(draw_win,x1+1,y1+1,x3,y3);

  x_set_color(draw_win,white);
  x_line(draw_win,x2,y2,x3,y3);

  x_set_color(draw_win,black);
  x_flush_display();
}


void BASE_WINDOW::draw_right_menu_button(int x, int y, int)
{ 
  int y1 = y-4;
  int x1 = x-3;
  int y2 = y+4;
  int x2 = x-3;
  int y3 = y;
  int x3 = x+4;

  if ( !mono() )
    x_set_color(draw_win,grey2);
  else
    x_set_color(draw_win,white);

  for(int i=1; i<7; i++)
     x_line(draw_win,x1+i,y1+i,x2+i,y2-i);

  x_set_color(draw_win,black);
  x_line(draw_win,x1,y1,x2,y2);
  x_line(draw_win,x1+1,y1+1,x3,y3);

  x_set_color(draw_win,white);
  x_line(draw_win,x2+1,y2-1,x3+1,y3);

  x_set_color(draw_win,black);
  x_flush_display();
}


void BASE_WINDOW::place_panel_items()
{
  int i;
  int w;

  int bxoff = 0;     // left and right button boundary space
  int bskip = th/2;  // horizontal space between buttons

  if (item_count == but_count) bskip = 1 + button_h/8;

  bool is_menu = (p_win && item_count == but_count);

  // adjust panel width

  if (yskip < button_h+2) yskip = button_h+2;

  if (is_menu)  // menu
    { yoff  = 1;
      xoff  = 1;
      bxoff = 1;
      bskip = 1;
      yskip = button_h+3;
      panel_width = buts_per_line*(button_w + bskip) - bskip + 4;
     }
  else
  { int w = x_text_width(draw_win,default_frame_label);
    if (panel_width == -1 && item_count == but_count) // only buttons
    { if (w > 300) w = 300;
      if (w > panel_width) panel_width = w;
     }
   }

  if (buts_per_line > but_count) buts_per_line = but_count;


  // adjust width of all items

  int last_ycoord = yoff;

  //int last_ycoord = (item_count > but_count) ? yoff : bskip;

  for(i = 0; i<item_count; i++)
  { panel_item it = Item[i];

    it->xcoord = xoff;
    it->ycoord = last_ycoord;

    switch (it->kind)
    {
      case  Text_Item:
             it->width = x_text_width(draw_win,it->label_str);
             last_ycoord += ytskip;
             break;

      case Choice_Item:
             it->width = label_w + choice_w*it->dat1;
             last_ycoord += yskip;
             break;

      case Color_Item:
             it->width = label_w + 16*choice_h;
             last_ycoord += yskip;
             break;

      case Bitmap_Choice_Item:
             it->width = label_w + it->dat1 * it->dat2;
             last_ycoord += choice_h-th+(it->height>th?it->height:th);
             break;

      case Slider_Item:
             it->width = label_w + slider_w;
             last_ycoord += yskip;
             break;

      case String_Menu_Item:
             it->width = label_w + string_w + yskip;
             last_ycoord += yskip;
             break;

      case Button_Item:
             break;

      default:   
             it->width = label_w + string_w;
             last_ycoord += yskip;
             break;
     }

     if ((w = 2*xoff + it->width) > panel_width) panel_width = w;

   }


  // slider & string boxes

  w = panel_width - label_w - 4*xoff;

  //if (w > slider_w) slider_w = w;
  //if (w > string_w) string_w = w;

  for(i = 0; i<item_count; i++)
  { panel_item it = Item[i];
    
    switch (it->kind)
    {
      case  Text_Item:
             break;
      case  Button_Item:
             break;
      case Choice_Item:
             break;
      case Color_Item:
             break;
      case Bitmap_Choice_Item:
             break;
      case Slider_Item:
             it->width = label_w + slider_w;
             break;
      case String_Menu_Item:
             it->width = label_w + string_w + yskip;
             break;
      default:  
             it->width = label_w + string_w;
             break;
     }

   }


  //  button layout

  int bw = button_w + bskip;



  if ( !is_menu)
  {
    if (buts_per_line == 0) 
    { buts_per_line = (panel_width-2*xoff+bskip)/bw;
      if (buts_per_line == 0) buts_per_line = 1;
     }
  
    if ((w = buts_per_line*bw+2*xoff-bskip) > panel_width) panel_width = w;
  
    if (item_count > but_count) // center buttons
    { if (but_count < buts_per_line)
         bxoff = (panel_width - (but_count * bw - bskip))/2;
      else
         bxoff = (panel_width - (buts_per_line * bw - bskip))/2;
     }
    else
      bxoff = xoff;
 
   }

  // place buttons

  int x = bxoff;
  int y = last_ycoord;
  int j = 0;

  if (item_count > but_count) y += yoff;

  for(i = 0; i<item_count; i++)
  { panel_item it = Item[i];
    if (it->kind == Button_Item)
    { it->xcoord = x;
      it->ycoord = y;
      it->width  = button_w;
      j++;
      x += bw;
      if (j >= buts_per_line)
      { y += yskip;
        x = bxoff;
        j = 0;
       }
    }
   }

  if (!is_menu && j != 0) y += yskip;

  //y += bskip;
  y += yoff;

  if (panel_height < y) panel_height = y;
}


void BASE_WINDOW::draw_panel_items()
{
  int          save_lw = x_set_line_width(draw_win,1);
  line_style   save_ls = x_set_line_style(draw_win,solid);
  text_mode    save_tm = x_set_text_mode(draw_win,transparent);
  drawing_mode save_mo = x_set_mode(draw_win,src_mode);

  x_set_color(draw_win,panel_bg_color);
  x_box(draw_win,0,0,window_width-1,panel_height-1);
  x_set_color(draw_win,black);

  for(int i=0; i<item_count; i++)
  { panel_item it=Item[i];

    switch (it->kind) {

    case Button_Item:
        { draw_button(it,0);
          break;
         }

    case Text_Item:
        { draw_text_item(it);
          break;
         }

    case Choice_Item:
        { draw_choice_item(it);
          break;
         }

    case Bool_Item:
        { draw_bool_item(it);
          break;
         }

    case Color_Item:
        { draw_color_item(it);
          break;
         }

    case Bitmap_Choice_Item:
        { draw_bitmap_choice_item(it);
          break;
         }

    case Slider_Item:
        { draw_slider_item(it,0);
          break;
         }

    case Int_Item:
        { if (act_str_item == 0) act_str_item = it;
          draw_string_item(it);
          break;
         }
 
    case Float_Item:
        { if (act_str_item == 0) act_str_item = it;
          draw_string_item(it);
          break;
         }
  
    case String_Item:
        { if (act_str_item == 0) act_str_item = it;
          draw_string_item(it);
          break;
         }

    case String_Menu_Item:
        { if (act_str_item == 0) act_str_item = it;
          draw_string_item(it);
          int xr = it->xcoord + label_w - 10;
          draw_down_menu_button(xr,it->ycoord+yskip/2,0);
          break;
         }
   }

 }

 x_set_line_width(draw_win,save_lw);
 x_set_line_style(draw_win,save_ls);
 x_set_mode(draw_win,save_mo);
 x_set_text_mode(draw_win,save_tm);

}


    
void BASE_WINDOW::open_sub_panel(panel_item it)
{ 
  BASE_WINDOW* wp = (BASE_WINDOW*)it->ref;

  int x = window_xpos + window_width + 2;
  int y = window_ypos + it->ycoord;
    
  if (this == p_root) 
  { x = it->xcoord; 
    y = it->ycoord + button_h + 5;
   }

  wp->p_win = this;
  wp->p_root = p_root;
  wp->display(x,y,p_root);

  wp->last_sel_button = 0;

  wp->panel_menu_mode = 1;
}

    
void BASE_WINDOW::close_sub_panel(BASE_WINDOW* wp)
{ wp->last_sel_button = 0;
  wp->panel_menu_mode = 0;

  if (wp->is_open())
  { wp->close();
    if (!x_display_bits_saved())
       while (event_handler(wp,1) != exposure_event);
    else
       while (event_handler(wp,0) != no_event || wp !=0 );
   }
}


void  BASE_WINDOW::trace_button(panel_item it, unsigned long T)
{ 
  if (last_traced_button == it) return;

  panel_item it0 = last_traced_button;

  if (it0)
    button_tracer(0,this,it0->xcoord,it0->ycoord,it0->dat1,it0->label_str);
  last_traced_button = 0;
 
  if (it) 
  { //if (T - last_traced_button_time  >  1000)
    { button_tracer(1,this,it->xcoord,it->ycoord,it->dat1,it->label_str);
      last_traced_button = it;
      last_traced_button_time = T;
     }
   }
}



int BASE_WINDOW::panel_event_handler(int w, int k, int b, int x, int y, 
                                                         unsigned long t)
{
  int          save_lw = x_set_line_width(draw_win,1);
  line_style   save_ls = x_set_line_style(draw_win,solid);
  text_mode    save_tm = x_set_text_mode(draw_win,transparent);
  drawing_mode save_mo = x_set_mode(draw_win,src_mode);
  clipping(1);

  if (w != draw_win) 
  { x = -1;
    y = -1;
   }

  int but = -1;


  panel_item it = 0;

  if (k == key_press_event && act_str_item != 0)  
    { // inp_ut for active string item
        if (b == 13)   // return
           { int j = 0;
             while (Item[j] != act_str_item) j++;
             j = (j + 1) % item_count;
             k =  Item[j]->kind; 
             while (k != String_Item && k != String_Menu_Item && 
                    k != Int_Item    && k != Float_Item)
             { j = (j+1) % item_count;
               k =  Item[j]->kind; 
              }
             activate_string_item(Item[j]);
            }
         else // character input for active string item
            { it = act_str_item;
              x = xoff + label_w;
              y = it->ycoord;
              x_put_back_event(); // event will be handled by panel_text_edit
             }
       }
    else
     { int i;
       for(i=0; i<item_count; i++) // search for selected item
       { it = Item[i];
         if (x <= it->xcoord || x >= it->xcoord+it->width + 4) continue;
         if (y <= it->ycoord || y >= it->ycoord+it->height   ) continue;
         break;
        }

       if (i==item_count) it = 0;

       if (button_tracer && k == motion_event) trace_button(it,t);

       if (panel_menu_mode == 0 && k != button_press_event) it = 0;
  
       if (k == button_release_event) 
         if (last_sel_button == 0) 
           { p_root->active_window = this;
             panel_menu_mode = 0;
            }
         else
           { it = last_sel_button;
             if (it->ref)
               { BASE_WINDOW* wp =  ((BASE_WINDOW*)it->ref);
                 wp->p_root->active_window = wp;
                 wp->panel_menu_mode = 0;
                }
             else
               { draw_button(it,0);
                 but = it->dat1;
                }
            }
  
      }


    if (but == -1 && it) // an item was selected
    {
      switch (it->kind) {
  
      case Text_Item: break;
  
      case Slider_Item:
      { int xoff1 = it->xcoord + label_w;
        int w = draw_win;
        int xx = 0;
        if (x < xoff1 || x > xoff1+slider_w) break;
        while(w == draw_win)
        { if (xx != x)
          { xx = x;
            draw_slider_item(it,xx);
           }
          if (x_get_next_event(&w,&b,&x,&b,&t) == button_release_event) break;
         }
        break;
       }
  
      case Int_Item:
      { activate_string_item(it);
        panel_text_edit(it);
        int x = atoi(it->data_str);
        delete[] it->data_str;
        it->data_str = make_string(x);
        *(int*)it->ref = x;
        break;
       }
  
      case Float_Item:
      { activate_string_item(it);
        panel_text_edit(it);
        double x = (double)atof(it->data_str);
        delete[] it->data_str;
        it->data_str = make_string(x);
        *(double*)it->ref = x;
        break;
       }
  
      case String_Item:
      { activate_string_item(it);
        panel_text_edit(it);
        delete[] *access_str(it->ref);
        *access_str(it->ref) = it->data_str;
        break;
       }
  
      case String_Menu_Item:
      { int xr = it->xcoord + label_w - 10;
        int yr = it->ycoord + yskip/2;

        activate_string_item(it);

        if (x > xr+8)
         { panel_text_edit(it);
           delete[] *access_str(it->ref);
           *access_str(it->ref) = it->data_str;
          }
        else
          if (x > xr-8)
          { draw_down_menu_button(xr,yr,1);

            BASE_WINDOW* wp = it->menu_win;

            wp->p_win = this;
            wp->p_root = p_root;
            wp->last_sel_button = 0;
            wp->panel_menu_mode = 1;
            wp->panel_bg_color = white;
            wp->shadow_color = white;

            int px = xr-9;

            int maxlen = 0;
            for(int i = 0; i < it->dat1; i++) 
            { int l = strlen(it->choices[i]);
              if (l > maxlen) maxlen = l;
             }
            maxlen *= tw;
            maxlen += 20;

            int bpl  = it->dat1;
            int lines = 1;  

            while (bpl*maxlen > lines*yskip) 
            { bpl--;
              lines = it->dat1/bpl;
             }

            wp->buts_per_line = bpl; //(window_width - px)/maxlen;

            wp->place_panel_items();

            int py = it->ycoord + yskip;

            str_menu = this;
            str_menu_it = it;
            last_sel_button = 0;

            //if (py < window_height-wp->window_height)
            //  wp->display(px,py,this);
            //else
              wp->display(px+window_xpos,py+window_ypos);

            //p_root->active_window = wp;
           }

        break;
      }
  
  
     case Choice_Item:
     { if (x > it->xcoord + label_w)
       { int d = (x-it->xcoord-label_w)/choice_w;
         if (d < it->dat1)
         { *(int*)it->ref = it->offset + d * it->dat2;
           draw_choice_item(it);
          }
        }
       break;
      }
  
     case Bool_Item:
     { if (x > it->xcoord + label_w)
       { int d = (x-it->xcoord-label_w)/choice_w;
         if (d < it->dat1)
         { *(bool*)it->ref = bool(it->offset + d * it->dat2);
           draw_bool_item(it);
          }
        }
       break;
      }
  
     case Color_Item:
     { if (x > it->xcoord + label_w)
       { int d = (x - it->xcoord - label_w)/choice_h;
         if (d < 16) change_color_item(it,d);
        }
       break;
      }
  
     case Bitmap_Choice_Item:
     { if (x > it->xcoord + label_w)
       { int d = (x - it->xcoord - label_w)/it->dat2;
         if (d < it->dat1) change_bitmap_choice_item(it,d);
        }
       break;
     }

     case Button_Item:
     { 
       if (last_traced_button)
       { panel_item it = last_traced_button;
         button_tracer(0,this,it->xcoord,it->ycoord,it->dat1,it->label_str);
         last_traced_button = 0;
        }
    
       if (last_sel_button != it)
       { 
         if (last_sel_button) 
         { draw_button(last_sel_button,0);
           if (last_sel_button->ref)
                     close_sub_panel((BASE_WINDOW*)last_sel_button->ref);
          }
         if (it) 
         { draw_button(it,1);
           if (it->ref) open_sub_panel(it);
           last_sel_button = it;
          }
       }

       it = 0;
       break;
     }

    }
  }

  int but1 = but;

  if (it  && it->kind == Button_Item  && it->ref == 0 ) 
  { BASE_WINDOW* wp = this;
    while(wp->p_win) 
    { BASE_WINDOW* wpp = wp->p_win;
      close_sub_panel(wp);
      wp->p_win = 0;
      wp = wpp;
     }
    wp->active_window = wp;

    if (wp->last_sel_button)
    { drawing_mode save_mo = x_set_mode(wp->draw_win,src_mode);
      wp->clipping(1);
      wp->draw_button(wp->last_sel_button,0);
      x_set_mode(wp->draw_win,save_mo);
      wp->clipping(2);
      if (wp->last_sel_button->dat1 > -1)
              but1 = wp->last_sel_button->dat1;
      wp->last_sel_button = 0;
     }
      else but1 = -1;
   }

  x_set_line_width(draw_win,save_lw);
  x_set_line_style(draw_win,save_ls);
  x_set_mode(draw_win,save_mo);
  x_set_text_mode(draw_win,save_tm);
  clipping(2);

 if (it && it->kind == Button_Item && it->action) it->action(but);

 return but1;

}

