/*******************************************************************************
+
+  LEDA 3.4
+
+  _graph_edit.c
+
+  This file is part of the LEDA research version (LEDA-R) that can be 
+  used free of charge in academic research and teaching. Any commercial
+  use of this software requires a license which is distributed by the
+  LEDA Software GmbH, Postfach 151101, 66041 Saarbruecken, FRG
+  (fax +49 681 31104).
+
+  Copyright (c) 1991-1996  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 66123 Saarbruecken, Germany     
+  All rights reserved.
+ 
*******************************************************************************/

#include <LEDA/graph_edit.h>
#include <LEDA/node_matrix.h>
#include <math.h>


static window*           Wp;
static GRAPH<point,int>* Gp;
static node_matrix<int>  M;

static color node_color = yellow;
static color edge_color = black;

static int  node_width;
static int  edge_width;
static bool directed;

static int x_min, x_max, y_min, y_max;
static double node_radius;


inline void draw_node(window& W, GRAPH<point,int>& G, node v)
{ //W.draw_text_node(G[v],string("%d",index(v)),node_color); 
  W.draw_filled_node(G[v],node_color); 
 }



static void draw_edge(window& W, GRAPH<point,int>& G, node v, node w)
{ 
  // M(v,w) = number of edges from v to w (defines thickness of drawn edge)

  if (v == w) return; // still cannot draw self-loops 

  if (M(v,w) == 0) return;  // there is no edge (v,w)

  int lw_save = W.set_line_width(M(v,w)*edge_width);

  segment s(G[v],G[w]);

  if (directed) 
    if (M(w,v) == 0)
       W.draw_edge_arrow(s,edge_color);
    else
       W.draw_arc_edge_arrow(s,0.7*s.length(),edge_color);
  else 
    if (M(w,v) == 0)
       W.draw_edge(s,edge_color); 
    else
       W.draw_arc_edge(s,0.7*s.length(),edge_color);

  W.set_line_width(lw_save);
}

 
static void draw_node_with_edges(window& W, GRAPH<point,int>& G, node v)
{ edge e;
  forall_in_edges(e,v)  draw_edge(W,G,source(e),v);
  forall_out_edges(e,v) draw_edge(W,G,v,target(e));
  draw_node(W,G,v);
}



inline void message(window& W, string s)
{ //W.del_messages();
  //W.message(s + string("                              "));
  W.acknowledge(s);
 }


static edge find_edge(GRAPH<point,int>& G, point p)  
{ 
  edge   e_min = nil;
  double e_len = 0;
  double a_min = MAXDOUBLE;
  edge e;
  forall_edges(e,G)
  { point  s = G[source(e)];
    point  t = G[target(e)];
    double a = fabs(area(s,t,p));
    if (a < a_min)
    { a_min = a;
      e_min = e;
      e_len = s.distance(t);
     }
   }

  if ((a_min/e_len) < 0.25 * node_radius) return e_min;

  return nil;
 }


static node find_node(GRAPH<point,int>& G, point p) 
{ node v;
  forall_nodes(v,G)
  { point q = G[v];
    if (p.distance(q) < node_radius) return v;
   }
  return nil;
 }


void init_matrix(const graph& g)
{ edge e;
  int n = 100;
  if (2*g.number_of_nodes() > n) n = 2*g.number_of_nodes();
  M.init(g,n,0);
  forall_edges(e,g) M(source(e),target(e))++;
 }


static void read_graph(window& W, GRAPH<point,int>& G,string s,bool clear=false)
{
  //message(W,string("Reading file %s.",~s));

  GRAPH<point,int> X;
  node v;
  node w;

  int x = X.read(s);

  if (x == 1) 
  { message(W,string("Cannot open file %s.",~s));
    return;
   }

  if (x == 2)
  { message(W,"File is not written by graph_edit (random embedding).");
    int max = int(W.xmax())-3;
    int min = 2;
    forall_nodes(v,X)
      X[v] = point(rand_int(min,max), rand_int(min,max));
   }

  if (clear) 
  { G.clear(); 
    W.clear(); 
   }

  node_array<node> corr(X);

  forall_nodes(v,X) 
  { node u = G.new_node(X[v]);
    corr[v] = u;
    draw_node(W,G,u);
   }

  edge e;
  forall_edges(e,X) 
  { v = corr[source(e)];
    w = corr[target(e)];
    G.new_edge(v,w);
    M(v,w)++;
   }

  if (clear) init_matrix(G);

  forall_nodes(v,G)
     forall_nodes(w,G) 
        draw_edge(W,G,v,w);

}



static void save_graph(window& W, GRAPH<point,int>& G,string s)
{ if (s=="") 
  { message(W,"Cannot open file.");
    return;
   }
  //message(W,string("writing to file %s",~s));
  G.write(s);
 }



void circular_embed(GRAPH<point,int>& G , point m, double r)
{ double alpha = 0;
  double step  = 6.2832/G.number_of_nodes();
  node v;
  forall_nodes(v,G)  
  { G[v] = m.translate(alpha,r);
    alpha+=step;
  }
}


void bi_embed(GRAPH<point,int>& G, list<node>& A, list<node>& B,
              double x_min, double x_max, double y_min, double y_max)
{
  double dy = (y_max-y_min)/A.length();
  double y = y_min + dy/2;
  node v;

  forall(v,A) 
  { G[v] = point(x_min,y);
    y += dy;
  }

  dy = (y_max-y_min)/B.length();
  y  = y_min + dy/2;
  forall(v,B) 
  { G[v] = point(x_max,y);
    y += dy;
  }
}

void embed(GRAPH<point,int>& G, node_array<double>& x, 
                                node_array<double>& y, double w, double h)
{ node v;
  forall_nodes(v,G)  
      G[v] = point(x_min + w*x[v], y_min + h*y[v]);
 }


static void draw_graph(window& W, GRAPH<point,int>& G)
{ node v,w;
  forall_nodes(v,G) 
  { draw_node(W,G,v);
    forall_nodes(w,G) draw_edge(W,G,v,w);
   }
}


static void redraw_func()  { draw_graph(*Wp, *Gp); }

static void window_init(window& W)
{ W.set_redraw(redraw_func);
  W.set_node_width(node_width);
  W.init(x_min,x_max,y_min,0);
  node_radius = node_width/W.scale();
  W.set_mode(xor_mode);
  W.set_show_coordinates(true);
}


static void gen_diag(int)
{
  panel P("GRAPH GENERATORS");

  int n = 10;
  int m = 20;

  P.int_item("# nodes", n,0,500);
  P.int_item("# edges", m,0,500);

  P.button("random");
  P.button("complete");
  P.button("bi_random");
  P.button("bi_complete");
  P.button("planar");
  P.button("grid");
  P.button("triang");
  P.button("exit");

  list<node> A,B;

  node_array<double> xcoord(*Gp);
  node_array<double> ycoord(*Gp);

  double w  = x_max - x_min;
  double h  = y_max - y_min;
  double xl = x_min + w/4;
  double xr = x_max - w/4;
  double R  = w/2.5;
  point  M  = point(x_min + w/2, y_min + h/2);

  Gp->clear();

  switch(P.open(*Wp)) {

    case 0: random_graph(*Gp,n,m);
            circular_embed(*Gp,M,R);
            break;
    case 1: complete_graph(*Gp,n);
            circular_embed(*Gp,M,R);
            break;

    case 2: random_bigraph(*Gp,n,n,m,A,B);
            bi_embed(*Gp,A,B,xl,xr,y_min,y_max);
            break;

    case 3: complete_bigraph(*Gp,n,n,A,B);
            bi_embed(*Gp,A,B,xl,xr,y_min,y_max);
            break;

    case 4: random_planar_graph(*Gp,xcoord,ycoord,n);
            embed(*Gp,xcoord,ycoord,w,h);
            break;

    case 5: grid_graph(*Gp,xcoord,ycoord,n);
            embed(*Gp,xcoord,ycoord,w,h);
            break;

    case 6: triangulated_planar_graph(*Gp,xcoord,ycoord,n);
            embed(*Gp,xcoord,ycoord,w,h);
            break;
   }

  init_matrix(*Gp);
  Wp->clear();
  draw_graph(*Wp,*Gp);
}



static void file_diag(int)
{
  string filename = "graph.ggg";

  panel P("FILE PANEL");
  P.string_item(string("File:"),filename);
  P.button("load");
  P.button("read");
  P.button("save");

  switch (P.open(*Wp)) {

      case 0 : // load 
               read_graph(*Wp,*Gp,filename,true);
               break;

      case 1 : // read
               read_graph(*Wp,*Gp,filename,false);
               break;
  
      case 2 : // save
               save_graph(*Wp,*Gp,filename);
               break;
   }
}


static void settings_diag(int)
{ panel P("GRAPH EDIT SETTINGS");
  P.int_item("x_min",x_min);
  P.int_item("x_max",x_max);
  P.int_item("y_min",y_min);
  P.int_item("node width",node_width,1,20);
  P.int_item("edge width",edge_width,1,5);
  P.bool_item("directed",directed);
  P.color_item("node color",node_color);
  P.color_item("edge color",edge_color);
  P.button("continue");
  P.open(*Wp);
  if (Wp->mono()) node_color = white;
  window_init(*Wp);
  draw_graph(*Wp,*Gp);
}


static void help_diag(int)
{ panel P("MOUSE BUTTONS");
  P.text_item("");
  P.text_item("    LEFT           MIDDLE           RIGHT    ");
  P.text_item("ins/move node   insert edge     del node/edge");
  P.button("continue");
  P.open(*Wp);
}


static void clear_all(int)
{ Wp->clear();
  Gp->clear();
  init_matrix(*Gp);
}

static bool done;

static void exit_diag(int) { done = true; }

void graph_edit(window& W, GRAPH<point,int>& G, bool dir, bool redraw)
{
  drawing_mode save_mode = W.get_mode();

  int save_but[8];
  W.std_buttons(save_but);
  
  directed = dir;

  Wp = &W;
  Gp = &G;

  init_matrix(G);

  panel main_panel("Poor Man's Graph Editor");

  main_panel.button("FILE", file_diag);
  main_panel.button("GEN",  gen_diag);
  main_panel.button("CLEAR",clear_all);
  main_panel.button("PROPS",settings_diag);
  main_panel.button("INFO", help_diag);
  main_panel.button("DONE", exit_diag);
  main_panel.buttons_per_line(3);

  //W.init(0,100,0);

  x_min = (int)W.xmin();
  x_max = (int)W.xmax();
  y_min = (int)W.ymin();
  y_max = (int)W.ymax();

  node_width  = W.get_node_width();
  edge_width  = W.get_line_width();
  node_radius = node_width/W.scale();

  if (W.mono()) node_color = white;

  window_init(W);
  if (redraw) draw_graph(W,G);

  main_panel.display(0,0);



  done = false;

  while ( ! done )
  {
    double x,y;

    window* wp;

    int k = read_mouse(wp,x,y);

    if (wp != &W) continue;

    unsigned long msec = W.button_press_time();

    point p = point(x,y);

    switch (k) {

      case MOUSE_BUTTON(1):  { 

                node v = find_node(G,p);

                if (v == nil)        // new node
                { v  = G.new_node(p);
                  draw_node(W,G,v);
                  break;
                 }
                else                 // move node or new edge
                { point  p0 = G[v];
                  double dx = p0.xcoord() - x;
                  double dy = p0.ycoord() - y;
                  int val;

                  while ( W.read_event(val,x,y) != button_release_event)
                  { point q = point(x+dx,y+dy);
                    if (q != G[v])
                    { p = G[v];
                      G[v] = q;
                      draw_node_with_edges(W,G,v);
                      G[v] = p;
                      draw_node_with_edges(W,G,v);
                      G[v] = q;
                     }
                   }

                  msec = W.button_release_time() - msec;

                  p = G[v];
                  G[v] = p0;

                  node u = find_node(G,p); // node at new position

                  if (msec > 200  || p.distance(p0) >  node_radius/2) 
                  { // move node certain distance or for certain time 
                    // assign new position if free
                    if (u == nil)
                    { G[v] = p; 
                      break;
                     }
                   }
                  // move node back
                  draw_node_with_edges(W,G,v);
                  G[v] = p;
                  draw_node_with_edges(W,G,v);
                  G[v] = p0;

                  // stop if new position was occupied by some other node
                  if (u != nil && u != v) break;  

                  // start new edge at p = p0 if time or distance was to small
                  p = p0;  
                  // do not break; continue with BUTTON(2)-case
                }
              }


      case MOUSE_BUTTON(2):  { // new edge
                
                node v = find_node(G,p);   // start node

                W.message("new edge");
 
                if (v != nil)            
                { p = G[v];

                  W.draw_filled_node(p);   // highlight start node

                  if (W.read_mouse_seg(p.xcoord(),p.ycoord(),x,y) != 3)
                  {
                    point q(x,y);
                    node w = find_node(G,q);
  
                    if (w == nil)      // no hit: create a new node w at q
                    { w  = G.new_node(q);
                      draw_node(W,G,w);
                     }
  
                    draw_edge(W,G,v,w);  // delete existing edges between v and w
                    draw_edge(W,G,w,v);
  
                    G.new_edge(v,w);  // insert new edge
                    M(v,w)++;
  
                    draw_edge(W,G,w,v);   // reinsert old edges
                    draw_edge(W,G,v,w);
                   }

                  W.draw_filled_node(p);  // unhighlight start node
                 }

                W.del_messages();
                break;
               }




      case MOUSE_BUTTON(3):  { // delete node or edge

               node v = find_node(G,p);
               if (v != nil) 
               { draw_node_with_edges(W,G,v);
                 edge e;
                 forall_inout_edges(e,v) M(source(e),target(e)) = 0;
                 G.del_node(v);
                 break;
                }

               edge e = find_edge(G,p);
               if (e != nil)
               { node v = source(e);
                 node w = target(e);
                 draw_edge(W,G,v,w);
                 draw_edge(W,G,w,v);
                 G.del_edge(e);
                 M(v,w)--;
                 draw_edge(W,G,w,v);
                 draw_edge(W,G,v,w);
                }
               break;
              }

     } // switch

  } // while ! done



  W.reset_frame_label();
  W.set_redraw();
  W.set_mode(save_mode);
  W.set_buttons(save_but);
}
